/*-------------------------------------------------------------------*/
/* ct.c , based on ACF_new.C
/*
/*  grown up by Akimov Alexander
	was born by Pavel Kopylov                                          */
/*-------------------------------------------------------------------*/

#define  ProgName       "HZ111"
#define  VersionNumber  "Version 2.22"
#define  LastUpdated    "22.22.2222"

#include <assert.h>
#include <math.h>
#include <time.h>
#include <stdlib.h>
#include <stdio.h>

#include "ct.h"
#include "..\Modules\memctrl.h"
#include "..\Modules\interfc.h"
#include "..\Modules\cb.h"
#include "..\Modules\mapseg.h"
#include "..\Modules\bintree.h"
#include "..\Modules\color.h"
 

int       QuietLevel, NColors;
ULONG     NodesAdded, ContextsAdded;
double Delta;
double COEFF;
BYTE LocalPath[31];
MULTITREE *Root;

CLPR Props;

int PrintCounter;

#define N_LOG_2 10000
#define OPT 0//vsegda

#define REDUCED 0

double LOG_2[N_LOG_2];
double LOG_2_Delta;


ULONG DescentCounter;
int Counter;

//For FREE TREE
int FreeTreeDepth;
int BitLength;
TEMPLATEFT *RootFT;

double *ConfigLut;

FILE *flog;

char **TmpTrees;
int TmpTreesCI;

int AlreadyBuiltDepth;
int NonDeleteLevel;

int QuantizedNChilds;
int ChildsLut[256];



void ConstructLevel(MULTITREE *CT, int depth, int Pos);
void CalculateLevel(MULTITREE *CT, IMAGE* Image, int depth, TRAININGSET* TS, MULTITREEm* MMT);
void FillRoot(MULTITREE *CT);
void PostProcessLevel(MULTITREE *CT, int depth);
void DeallocateTree(MULTITREE *CT);
void PrintNodes_acf(MULTITREE *CT, int *Path, int idx, int Color);
void PrintTree(MULTITREE *CT);
void ReadMTree1(MULTITREE *MT, FLOW *inFlow, ULONG* NModels, int depth);
void WriteMTree1(FLOW *outFlow, MULTITREE *CT);
//void CreateChildren(MULTITREE *CT);
void DeallocateChildren(MULTITREE *CT);


/* ----------------------------------------------------------------- */
/* -- New node                                                    -- */
/* ----------------------------------------------------------------- */

void OpenLogFile()
{
	flog = fopen("zPruneLog.txt", "a");
}

void CloseLogFile()
{
	fclose(flog);
}

void InitLog()
{
	int i = 0;
	for(i = 1; i<= N_LOG_2; i++) LOG_2[i-1] = log((double)i/(double)(N_LOG_2))/LOG2;
}

void SetNColors(int N)
{
	int i;

	NColors = N;
	QuantizedNChilds = N;
	for(i=0;i<256;i++) ChildsLut[i] = i;

	QuietLevel = 3;
	Counter = 0;
	TmpTrees = NULL;
	TmpTreesCI = 0;
	BitLength = 0;
	FreeTreeDepth = 0;

	Props.SizeX = 0;
	Props.SizeY = 0;
}
 
void SetDelta(double N)
{
//	IS_CONFIG_LUT = 0;
	Delta = N;
	LOG_2_Delta = log(Delta)/LOG2;
//	InitLog();
	DescentCounter = 0;
}

void SetCoeff(double N)
{
	COEFF = N;
}

void SetRoot(MULTITREE* r)
{
	Root = r;
}

void SetFreeTreeParam(int FTD, int BL)
{
	FreeTreeDepth = FTD;
	BitLength = BL;
}


void SetAlreadyBuiltDepth(int num1, int num2)
{
	AlreadyBuiltDepth = num1;
	NonDeleteLevel = num2;
}

void SetQuantizedNChilds(int N, int *CL)
{
	int i;
	QuantizedNChilds = N;
	for(i=0;i<NColors;i++)
		ChildsLut[i] = CL[i];
}

void SetClusterSizes(int SizeX, int SizeY)
{
	Props.SizeX = SizeX;
	Props.SizeY = SizeY;
}


MULTITREE *NewTreeNode()
{
  MULTITREE *CT=0;
//  int i;

  CT=(MULTITREE*)malloc(sizeof(MULTITREE));
  CT->pMostProbablePixelsItem=0;
//  CT->Children = NULL;
  CT->pChildren = NULL;
  CT->pPixels = NULL;
  CT->ChildPosition = 0;
  CT->Total = 0;
  CT->Context = -1;
  
  
  //DEBUG
 // CT->Pixels = NULL;

//  CT->Pixels=(int*)allocate(sizeof(int)*NColors);
//  for(i=0;i<NColors;i++) CT->Pixels[i]=0;

  NodesAdded++;
  return(CT);
}

ChildListItem *FindChildItem(MULTITREE *CT, int PixelID)
{
	ChildListItem *pCur=CT->pChildren;

	for(;pCur;pCur=pCur->pNext)
		if(pCur->PixelID == PixelID)
			break;
		
	return pCur;
}

PixelsListItem *FindPixelsItem(MULTITREE *CT, int PixelID)
{
	PixelsListItem *pCur=CT->pPixels;

	for(;pCur;pCur=pCur->pNext)
		if(pCur->PixelID == PixelID)
			break;
		
	return pCur;
}

MULTITREE *GetVarContext(MULTITREE *CT, BYTE *Data, int x)
{
  int       i;
  int       pixelN;
  MULTITREE *CCT;
  ChildListItem *pCurItem=0;

  CCT=CT;
  i=0;
  while (i!=-1)
  {
 
	if(x - CCT->ChildPosition >=0)
		pixelN = Data[x - CCT->ChildPosition];
	else
		pixelN = 0;

//    if (CCT->Children!=NULL && CCT->Children[pixelN]!=NULL)
	if(CCT->pChildren!=NULL)
    {
		pCurItem = FindChildItem(CCT,pixelN);
		if(pCurItem)
			CCT=pCurItem->pNode;
		//CCT->Children[pixelN];
    }
    else i=-1;
  }   
  return (CCT);
}


void PrintVarContext(MULTITREE *CT, BYTE *Data, int x)
{
  int       i, k;
  //int j;
  int       pixelN;
  MULTITREE *CCT;
  ChildListItem *pCurListItem =0;
  PixelsListItem *pCurPixItem =0;


  CCT=CT;
  i=0; k=1;
  while (i!=-1)
  {
//    if (CCT->Children!=NULL)
	  if(CCT->pChildren)
    {

	  if(x-CCT->ChildPosition>=0)
		pixelN=Data[x-CCT->ChildPosition];

      PrintMessage("\n%d:\n", k);
//      for (j=0;j<NColors;j++) PrintMessage("%d\t",CCT->Pixels[j]); PrintMessage("\n");
	  for(pCurPixItem=CCT->pPixels; pCurPixItem; pCurPixItem=pCurPixItem->pNext)
		  PrintMessage("%d\t",pCurPixItem->Counter); PrintMessage("\n");

      PrintMessage("%d\n", pixelN);

      PrintMessage("%d\n", pixelN);
//      CCT=CCT->Children[pixelN];
	  pCurListItem = FindChildItem(CCT,pixelN);
	  CCT=pCurListItem->pNode;
    }
    else i=-1;
    k++;
  }   
  PrintMessage("\n");
}


/* ----------------------------------------------------------------- */
/* -- Alternative tree construction                               -- */
/* ----------------------------------------------------------------- */
void ConstructTree(MULTITREE *CT, BYTE* Data, int DataLength, int depth)
{
  MULTITREE *CCT;
  ULONG     x;
  int       pixelN;
	//DEBUG
	int counter = 0;


  if(QuietLevel>2)
  {
    SetProgressVisibility(1);
    OpenProgressWindow();
  }

  for(x=0;x<(unsigned)DataLength;x++)
  {
    pixelN = Data[x];
    CCT = SetContext(CT, Data, x, depth, pixelN);

   
	if (CCT!=NULL) 
		CCT->Total++;

    if(QuietLevel>2) 
		UpdateProgressBar(x, DataLength);
  }
  if(QuietLevel>2) CloseProgressWindow();
}

void GetClusterBounds(IMAGE *GrayImage, int index, int *StartX, int *StartY, int *EndX, int *EndY)
{
	int h_num, v_num;
	int ShiftX, ShiftY;
	
	if(Props.SizeX == 0 || Props.SizeY == 0)
	{
		(*StartX) = 1;
		(*StartY) = 1;

		(*EndX) = GrayImage->ImageSizeX;
		(*EndY) = GrayImage->ImageSizeY;
	}
	else
	{
		h_num = (int)ceil((double)GrayImage->ImageSizeX/(double)Props.SizeX);
		v_num = (int)ceil((double)GrayImage->ImageSizeY/(double)Props.SizeY);

		ShiftY = index/h_num;
		ShiftX = index - ShiftY*h_num;

		(*StartX) = Props.SizeX * ShiftX + 1;
		(*StartY) = Props.SizeY * ShiftY + 1;
		
		(*EndX) = min(Props.SizeX * (ShiftX+1), (int)GrayImage->ImageSizeX);
		(*EndY) = min(Props.SizeY * (ShiftY+1), (int)GrayImage->ImageSizeY);
	}
}

void ConstructTreeImage(MULTITREE *CT, IMAGE *GrayImage, int *lut, int depth, int index)
{
  MULTITREE *CCT;
  ULONG     x,y;
  int       pixelN;
  int StartX, StartY;
  int EndX, EndY;


  if(QuietLevel>2)
  {
    SetProgressVisibility(1);
    OpenProgressWindow();
  }

  GetClusterBounds(GrayImage, index, &StartX, &StartY, &EndX, &EndY);

//  for(y=1;y <= GrayImage->ImageSizeY;y++)
	for(y = (unsigned)StartY; y <= (unsigned)EndY; y++)
	{
		for(x = (unsigned)StartX; x <= (unsigned)EndX; x++)
		{
			pixelN = GetGrayPixel(GrayImage,x,y);
			CCT = SetContextImage(CT,GrayImage,lut,x,y,depth,pixelN, StartX, StartY, EndX, EndY);

			if (CCT!=NULL) 
				CCT->Total++;
		}
		if(QuietLevel>2) 
			UpdateProgressBar(y, GrayImage->ImageSizeY);
  }
  if(QuietLevel>2) CloseProgressWindow();
  ImageRewind(GrayImage);
}


ChildListItem* AddChildListItem(MULTITREE *CT, int PixelID, int Position)
{
	ChildListItem *pNewItem=0, *pCurItem=0,*pPrevItem=0;

	pNewItem = (ChildListItem*) malloc(sizeof(ChildListItem));

	pNewItem->PixelID = PixelID;
	pNewItem->pNode = NewTreeNode();
	if(Position!=-1)  // otherwise just ignore the position parameter
		pNewItem->pNode->ChildPosition = Position;

/*
	if(CT->pChildren==NULL || CT->pChildren->PixelID>PixelID)
	{
		if(CT->pChildren)
			pNewItem->pNext=CT->pChildren;
		else
			pNewItem->pNext=NULL;
		CT->pChildren=pNewItem;
		return pNewItem;
	}

	for(pCurItem=CT->pChildren;pCurItem;pCurItem=pCurItem->pNext)
	{
		if(pCurItem->PixelID>PixelID)
			break;
		pPrevItem=pCurItem;
	}

	pPrevItem->pNext = pNewItem;
	if(pCurItem)
		pNewItem->pNext = pCurItem;
	else
		pNewItem->pNext = NULL;
*/
	pNewItem->pNext = CT->pChildren;
	CT->pChildren = pNewItem;
	return pNewItem;
}

IncreasePixelsListItem(MULTITREE *CT, int PixelID)
{
	PixelsListItem *pItem=0;
	pItem = FindPixelsItem(CT,PixelID);
	if(pItem)
		pItem->Counter++;
	else
		AddPixelsListItem(CT,PixelID,1);
}

PixelsListItem* AddPixelsListItem(MULTITREE *CT, int PixelID, int Counter)
{
	PixelsListItem *pNewItem=0;

	if(Counter==0)		// do not add zero pixel couters;
		return 0;

	pNewItem = (PixelsListItem*) malloc(sizeof(PixelsListItem));
	pNewItem->PixelID = PixelID;
	pNewItem->Counter = Counter;

	pNewItem->pNext = CT->pPixels;
	CT->pPixels = pNewItem;
	return pNewItem;
}

MULTITREE *SetContext(MULTITREE *CT, BYTE* Data, int x, int depth, int refPixelN)
{
  int       i;
  int       pixelN;
  MULTITREE *CCT;
  ChildListItem *pCurChildItem=0;
  PixelsListItem *pCurPixItem=0;

  CCT=CT;
  for (i=1;i<=depth;i++)
  {
	if(x - i >= 0)
		pixelN = Data[x-i];
	else
		pixelN = 0;
    
 
    CCT->ChildPosition = i;
	CCT->Total += 1;

//	CCT->Pixels[refPixelN]++;
	IncreasePixelsListItem(CCT,refPixelN);
	
//    if (CCT->Children==NULL) CreateChildren(CCT);

	pCurChildItem=FindChildItem(CCT,pixelN);
//	if (CCT->Children[pixelN]==NULL)
	if(pCurChildItem==NULL)
    {
      if (i==depth) 
		  ContextsAdded++;

//      CCT->Children[pixelN]=NewTreeNode();
//	  CCT->Children[pixelN]->ChildPosition = i;
	  pCurChildItem = AddChildListItem(CCT,pixelN,i);

//      assert(CCT->Children[pixelN]!=NULL);
    }
//    CCT=CCT->Children[pixelN];
	CCT=pCurChildItem->pNode;	// go to the child
  }   

//  CCT->Pixels[refPixelN]++;
  IncreasePixelsListItem(CCT,refPixelN);
  
  return (CCT);
}



MULTITREE *SetContextImage(MULTITREE *CT, IMAGE* GrayImage, int *lut, 
						   int x, int y, int depth, int refPixelN, int StartX, int StartY, int EndX, int EndY)
{
	ChildListItem *pCurChildItem=0;
	PixelsListItem *pCurPixItem=0;

  int       i;
  int       pixelN;
  MULTITREE *CCT;

  int refx, refy;

  int ChildIndex;

#ifdef TEMPLATE_2D
//    int  tJBIGX[31] = {30,-1,0,-1,+1,-2,0,-2,+2,-1,+1,-2,+2,-3,0,-3,+3,-1,+1,-3,+3,-2,+2,-4, 0,-4,+4,-1,+1,-3,+3 };
//    int  tJBIGY[31] = {30,0,-1,-1,-1,0,-2,-1,-1,-2,-2,-2,-2,0,-3,-1,-1,-3,-3,-2,-2,-3,-3, 0,-4,-1,-1,-4,-4,-3,-3 };
int  tJBIGX[36+1] = {36, 1, 0,-1, 0, 1,-1,-1, 1, 2, 0,-2, 0, 2,-1,-2, 1, 1,-2,-1, 2, 3, 0,-3, 0, 2,-2,-2, 2, 3,-1,-3, 1, 1,-3,-1, 3};
int  tJBIGY[36+1] = {36, 0, 1, 0,-1, 1, 1,-1,-1, 0, 2, 0,-2, 1, 2,-1,-2, 2, 1,-2,-1, 0, 3, 0,-3, 2, 2,-2,-2, 1, 3,-1,-3, 3, 1,-3,-1};

// Second template              //
//                              //
//            41 - 77           //
//                              //
//                              //
//           76 65 73           //
//        68 60 53 57 69        //
//     72 56 49 45 46 61 77     //
//     64 52 44 41 42 50 62     //
//     75 59 48 43 47 54 70     //
//        67 55 51 58 66        //
//           71 63 74           //
//                              //
#endif

  refPixelN = lut[refPixelN];
  CCT=CT;

  for (i=1;i<=depth;i++)
  {

	refx = x + tJBIGX[i];
	refy = y + tJBIGY[i];
	
	if((refx>=StartX && refx <= EndX) && (refy>=StartY && refy<=EndY))
		pixelN = lut[GetGrayPixel(GrayImage,refx,refy)];
	else
		pixelN = 0;
    
    CCT->ChildPosition = i;

//	if(CCT->Pixels)
//		CCT->Pixels[refPixelN]++;
	IncreasePixelsListItem(CCT,refPixelN);

	CCT->Total += 1;

	ChildIndex = pixelN;
	
//	if(QUANTIZE_CHILDS)
//		pixelN = ChildsLut[pixelN];

//    if (CCT->Children==NULL) CreateChildren(CCT);
 
	pCurChildItem=FindChildItem(CCT,pixelN);
//	if (CCT->Children[pixelN]==NULL)
    if(pCurChildItem==NULL)
	{
      if (i==depth) 
		  ContextsAdded++;

//     CCT->Children[pixelN] = NewTreeNode();
//	  CCT->Children[pixelN]->ChildPosition = i;

	  pCurChildItem = AddChildListItem(CCT,pixelN,i);
//	  pCurItem=CCT->pChildren;	// new Elements are added to the head
//      assert(CCT->Children[pixelN]!=NULL);
    }
//    CCT=CCT->Children[pixelN];
	CCT=pCurChildItem->pNode;	// go to the child
  }   


  CCT->ChildPosition = i;

//  if(CCT->Pixels)
//    CCT->Pixels[refPixelN]++;
  IncreasePixelsListItem(CCT,refPixelN);

  return (CCT);
}



/* ----------------------------------------------------------------- */
/* -- Delete unused nodes at level                                -- */
/* ----------------------------------------------------------------- */
/*
EXCLUDED SINCE NOT USED
void PostProcessLevel(MULTITREE *CT, int depth)
{
  int i;

  if (depth==1)
  {
    if (CT->Children!=NULL)
    {
      // Pruning !HERE!
      for(i=0;i<NColors;i++)
      {
        // Empty child
        if (CT->Children[i]->Total==0)
        {
          deallocate(CT->Children[i]->Pixels);
          deallocate(CT->Children[i]);
          CT->Children[i]=NULL;
        }
        else
        {
          // To do something when positive decision
        }
      }
    }
  }
  else
  for(i=0;i<NColors;i++)
  {
    if(CT->Children[i]!=NULL) PostProcessLevel(CT->Children[i], depth-1);
  }
}
*/
/*
EXCLUDED SINCE NOT USED
void FillRoot(MULTITREE *CT)
{
  int   i;
  ULONG k;
  
  k=0;

  for(i=0;i<NColors;i++)
  {
        k+=CT->Pixels[i];
  }
  CT->Total=k;
}*/

/* ----------------------------------------------------------------- */
/* -- Deallocate whole tree                                      -- */
/* ----------------------------------------------------------------- */
void DeallocateChildList(MULTITREE *CT)
{
	ChildListItem *pCurItem,*ptodie;
	for(pCurItem=CT->pChildren; pCurItem; )
	{
		ptodie = pCurItem;
		pCurItem=pCurItem->pNext;
		free(ptodie);
	}
	CT->pChildren=NULL;
}

void DeallocatePixelsList(MULTITREE *CT)
{
	PixelsListItem *pCurItem,*ptodie;
	for(pCurItem=CT->pPixels; pCurItem; )
	{
		ptodie = pCurItem;
		pCurItem=pCurItem->pNext;
		free(ptodie);
	}
	CT->pPixels = NULL;
}

void DeallocateTree(MULTITREE *CT)
{
  ChildListItem *pCurItem=0;
  
//  if (CT->Children!=NULL)
  if(CT->pChildren)
  {
//    for(i=0;i<QuantizedNChilds;i++)
	for(pCurItem=CT->pChildren; pCurItem; pCurItem=pCurItem->pNext)
    {
//      if (CT->Children[i]!=NULL)
      {
        //DeallocateTree(CT->Children[i]);
		  DeallocateTree(pCurItem->pNode);
		  pCurItem->pNode=NULL;
      }
    }
  }

//  deallocate(CT->Pixels);
//  deallocate(CT->Children);
  DeallocateChildList(CT);
  DeallocatePixelsList(CT);

  deallocate(CT);
}


void DeallocateChildren(MULTITREE *CT)
{
//  int i;
  ChildListItem *pCurItem=0;

  if (CT->pChildren!=NULL)
  {
    for(pCurItem=CT->pChildren; pCurItem; pCurItem=pCurItem->pNext)
    {
      //if (CT->Children[i]!=NULL)
      //{
        DeallocateTree(pCurItem->pNode/*CT->Children[i]*/);
		pCurItem->pNode=NULL;
      //}
    }
	  //deallocate(CT->Children);
	DeallocateChildList(CT);
	//  CT->Children = NULL;
  }
}


/*
EXCLUDED
void CreateChildren(MULTITREE *CT)
{
  int i;
  
  CT->Children=(MULTITREE**)allocate(sizeof(MULTITREE*)*QuantizedNChilds);
  assert(CT->Children!=NULL);
  for (i=0;i<QuantizedNChilds;i++) 
	  CT->Children[i]=NULL;
}
*/


/* ----------------------------------------------------------------- */
/* -- Print tree                                                  -- */
/* ----------------------------------------------------------------- */
void PrintNodes_acf(MULTITREE *CT, int *Path, int idx, int Color)
{
	ChildListItem *pCurItem;
  int i;//,j;
  int flag = 1;
  int counter=0;


  PrintCounter += 1;

  if(PrintCounter == 100)
  {
         PrintCounter = 1;
  }

  if (idx>0) Path[idx-1]=Color;

//  if (CT->Children == NULL)
  if(CT->pChildren==NULL)
  {
//	if(CT->Total<50 && CT->Total>40 )
//	{

    PrintMessage("%d\t", CT->Context);
    for(i=0;i<idx;i++) PrintMessage("%d ", Path[i]); PrintMessage("\t");
    
    if(CT->pPixels)
	{
		PixelsListItem *pPixItem=0;
		PrintMessage("\n");
		//for(i=0;i<NColors;i++) 
		for(pPixItem=CT->pPixels; pPixItem; pPixItem=pPixItem->pNext)
		{
			//PrintMessage("%d ", CT->Pixels[i]); PrintMessage("\t");
			PrintMessage("%i->%i ", pPixItem->PixelID, pPixItem->Counter); PrintMessage("\t");
		}
		PrintMessage("\n");
		
	}
	    PrintMessage("%d\t", CT->Total);
		PrintMessage("\n");

 //   }
//	if(CT->Pixels)
//	{
//		entropy = GetEntropySemi(CT->Pixels, CT->Total);
//		PrintMessage("%.4f\n", entropy);
//	}
//	else
		
  }
  else
  {
	  counter = 0;
	  //	for(i=0;i<QuantizedNChilds;i++)
	  for(pCurItem=CT->pChildren;pCurItem;pCurItem=pCurItem->pNext)  // for ALL children
	  {
		  //	  if (CT->Children[i] != NULL)
		  {
			  //PrintNodes_acf(CT->Children[i], Path, idx+1, i);
			  PrintNodes_acf(pCurItem->pNode, Path, idx+1, pCurItem->PixelID);
		  }
		  counter++;
	  }
//	  else
//	  {
/*	    if(counter<QuantizedNChilds)
		{

			PrintMessage("%d\t", CT->Context);
			for(i=0;i<idx;i++) PrintMessage("%d ", Path[i]); PrintMessage("\t");
    
//			if(CT->Pixels)
//			{
//				for(j=0;j<NColors;j++) PrintMessage("%d ", CT->Pixels[j]); PrintMessage("\t");
//			}

			PrintMessage("%d\n", CT->Total);
			flag = 0;
		}
//	  }
*/
  }
}

void PrintTree(MULTITREE *CT)
{
  int Path[256];

  PrintCounter = 0;

  PrintMessage("Context-N\tPath\tColors\tTotal\n");
  PrintNodes_acf(CT, Path, 0, 0);
  PrintMessage("\n");
}


/* ----------------------------------------------------------------- */
/* -- Tree I/O                                                    -- */
/* ----------------------------------------------------------------- */
/*
void WriteMTree(FLOW *outFlow, MULTITREE *CT)
{
  WriteULongToFlow(outFlow, NColors, 4);

  WriteMTree1(outFlow, CT);
  outFlow->FlushOutput(outFlow);
}

void WriteMTree1(FLOW *outFlow, MULTITREE *CT)
{
  int         i;

  for(i=0;i<NColors;i++) WriteULongToFlow(outFlow, CT->Pixels[i], 4);

  if (CT->Children != NULL)
  {
    for(i=0;i<NColors;i++)
    {
      if (CT->Children[i]!=NULL && CT->Pixels[i]!=0)
      {
        WriteMTree1(outFlow, CT->Children[i]);
      }
    }
  }
}
*/
/*
void ReadMTree(MULTITREE *MT, FLOW *inFlow, ULONG* NModels)
{
  int depth=0;

  (*NModels) = 0;
  ReadULongFromFlow(inFlow, &NColors, 4);
  ReadMTree1(MT, inFlow, NModels, depth);
}


void ReadMTree1(MULTITREE *MT, FLOW *inFlow, ULONG* NModels, int depth)
{
  int i;
  
  MT = NewTreeNode();

  for(i=0;i<NColors;i++) ReadULongFromFlow(inFlow, &(MT->Pixels[i]), 4);

  MT->Children=(MULTITREE**)allocate(sizeof(MULTITREE*)*NColors);
  assert(MT->Children!=NULL);
 
  MT->ChildPosition = depth;

  for(i=0;i<NColors;i++)
  {
    if (MT->Pixels[i]!=0)
    {
      NModels++;
      ReadMTree1(MT->Children[i], inFlow, NModels, depth+1);
    }
  }
}
*/


/*--------------------------------------*/
// Pruning and seting of the order numbers to contexts
/**---------------------------------------*/

void SetNumbers(MULTITREE *CT, int *ContextNumber)
{
//	int i;
	int flag = 1;
	ChildListItem *pCurItem=0;
	int counter=0;
		
	if(CT->pChildren == NULL)
	{
		CT->Context = (*ContextNumber);
		(*ContextNumber) += 1;
	}
	else 
	{
//		children = (char*)malloc(QuantizedNChilds*sizeof(char));
//		memset(children,0,QuantizedNChilds*sizeof(char));
		//for(i=0;i<QuantizedNChilds;i++)
		counter=0;
		for(pCurItem=CT->pChildren;pCurItem;pCurItem=pCurItem->pNext)
		{
			//if(CT->Children[i] != NULL)
			{
//				CT->Total = CT->Total-CT->Children[i]->Total;
//				for(j=0;j<NColors;j++)
//					CT->Pixels[j] =CT->Pixels[j] - CT->Children[i]->Pixels[j];


//				CT->Total -= CT->Children[i]->Total;
//				for(j=0;j<QuantizedNChilds;j++)
//					CT->Pixels[j] -= CT->Children[i]->Pixels[j];

				SetNumbers(pCurItem->pNode/*CT->Children[i]*/, ContextNumber);
			}
			counter++;
		}
//			else
			if(counter<QuantizedNChilds)
			{
				if(/*flag && */!TREE_COMPLETE)
				{
					CT->Context = (*ContextNumber);
					(*ContextNumber) += 1;
				}
//				flag = 0;
			}

	}
}

double GetCodeSizeSemi(MULTITREE *CT)
{
//	int i;
	double total = 0;
	double result = 0;
	double prob;
//	ChildListItem *pCurChildItem;
	PixelsListItem *pCurPixItem;

//	for(i=0;i<NColors;i++)
	for(pCurPixItem=CT->pPixels;pCurPixItem;pCurPixItem=pCurPixItem->pNext)
		total += pCurPixItem->Counter; //CT->Pixels[i];

	if(!total)
		return 0;
	
//	for(i=0;i<NColors;i++)
	for(pCurPixItem=CT->pPixels;pCurPixItem;pCurPixItem=pCurPixItem->pNext)
	{
//		prob = (double)(CT->Pixels[i])/(double)(total);
		prob = (double)(pCurPixItem->Counter)/(double)(total);
//		if(CT->Pixels[i])
		//result += prob*(log((total) /(double)(CT->Pixels[i]))/LOG2);
		result += prob*(log((total) /(double)(pCurPixItem->Counter))/LOG2);
	}

	return result*total + CONST;

}


double GetCodeSizeAdaptS(MULTITREE *CT)
{
	ULONG j;
	double total = 0;
	double result = 0;

	ChildListItem *pCurChildItem=0;
	PixelsListItem *pCurPixItem=0;

//	for(i=0;i<(unsigned)NColors;i++)
	for(pCurPixItem=CT->pPixels;pCurPixItem;pCurPixItem=pCurPixItem->pNext)
		total += pCurPixItem->Counter;//total += CT->Pixels[i];

	if(!total)
		return 0;
	
	//for(i=0;i<(unsigned)NColors;i++)
	for(pCurPixItem=CT->pPixels;pCurPixItem;pCurPixItem=pCurPixItem->pNext)
	{
		for(j=0;j < pCurPixItem->Counter/*CT->Pixels[i]*/;j++)
			result += log(Delta + j);
	}

	for(j = 0;j < (ULONG)total;j++)
		result -= log(j + NColors*Delta);

	return (-1)*result/LOG2;

}

//FAST
double GetCodeSizeAdapt(MULTITREE *CT)
{
	ChildListItem *pCurChildItem=0;
	PixelsListItem *pCurPixItem=0;
	ULONG j;
	double total = 0;
	double result = 0;


	double TempProduct;
	double Thr;

	
	double _2In52;
	double _2In32;
	double _2In24;
	double _2In16;
	double _2In8;

	int StartPower;

	_2In52 = 0x8000000000000;
	_2In32 = 0x100000000;
	_2In24 = 0x1000000;
	_2In16 = 0x10000;
	_2In8  = 0x100;

//	for(i=0;i<(unsigned)NColors;i++)
	for(pCurPixItem=CT->pPixels;pCurPixItem;pCurPixItem=pCurPixItem->pNext)
		total += pCurPixItem->Counter;//total += CT->Pixels[i];

	if(!total)
		return 0;

//	for(i=0;i<(unsigned)NColors;i++)
	for(pCurPixItem=CT->pPixels;pCurPixItem;pCurPixItem=pCurPixItem->pNext)
		result += LOG_2_Delta*pCurPixItem->Counter;//result += LOG_2_Delta*CT->Pixels[i];

//	for(i=0;i<(unsigned)NColors;i++)
	for(pCurPixItem=CT->pPixels;pCurPixItem;pCurPixItem=pCurPixItem->pNext)
	{

		TempProduct = 1;

		for(j=0;j<pCurPixItem->Counter/*CT->Pixels[i]*/;j++)
		{
			TempProduct = TempProduct*(NColors*j + 1);
			
			if(TempProduct >= _2In32)
			{
				TempProduct = TempProduct/(double)_2In32;
				result += 32;
			}
			
/*			if(TempProduct >= _2In24)
			{
				TempProduct = TempProduct/(double)_2In24;
				result += 24;
			}
			
			if(TempProduct >= _2In16)
			{
				TempProduct = TempProduct/(double)_2In16;
				result += 16;
			}
			
			if(TempProduct >= _2In8)
			{
				TempProduct = TempProduct/(double)_2In8;
				result += 8;
			}*/
		}

		StartPower = -1;

		if(TempProduct >= _2In52)
		{
			TempProduct = TempProduct/(double)_2In52;
			result += 52;
		}
		
		if(TempProduct >= _2In32)
		{
			TempProduct = TempProduct/(double)_2In32;
			result += 32;
		}
		
		if(TempProduct >= _2In24)
		{
			TempProduct = TempProduct/(double)_2In24;
			result += 24;
		}
		
		if(TempProduct >= _2In16)
		{
			TempProduct = TempProduct/(double)_2In16;
			result += 16;
		}
		
		if(TempProduct >= _2In8)
		{
			TempProduct = TempProduct/(double)_2In8;
			result += 8;
		}

		do
		{
			StartPower++;
			Thr = (1<<StartPower);
		}while(Thr<TempProduct);

		result += StartPower;
		TempProduct = TempProduct/(double)Thr;
		

		if(TempProduct<1)
//			result += LOG_2[max((int)(TempProduct*N_LOG_2 + 0.5) - 1,0)];
			result += log(TempProduct)/LOG2;

	}


	TempProduct = 1;
	for(j = 0;j<(ULONG)total;j++)
	{
		TempProduct = TempProduct*(j + 1);
		if(TempProduct >= _2In32)
		{
			TempProduct = TempProduct/(double)_2In32;
			result -= 32;
		}

/*		if(TempProduct >= _2In24)
		{
			TempProduct = TempProduct/(double)_2In24;
			result -= 24;
		}
	
		if(TempProduct >= _2In16)
		{
			TempProduct = TempProduct/(double)_2In16;
			result -= 16;
		}
	
		if(TempProduct >= _2In8)
		{
			TempProduct = TempProduct/(double)_2In8;
			result -= 8;
		}*/
	}

	StartPower = -1;

	if(TempProduct >= _2In52)
	{
		TempProduct = TempProduct/(double)_2In52;
		result -= 52;
	}

	if(TempProduct >= _2In32)
	{
		TempProduct = TempProduct/(double)_2In32;
		result -= 32;
	}
	
	if(TempProduct >= _2In24)
	{
		TempProduct = TempProduct/(double)_2In24;
		result -= 24;
	}
	
	if(TempProduct >= _2In16)
	{
		TempProduct = TempProduct/(double)_2In16;
		result -= 16;
	}
	
	if(TempProduct >= _2In8)
	{
		TempProduct = TempProduct/(double)_2In8;
		result -= 8;
	}

	do
	{
		StartPower++;
		Thr = (1<<StartPower);
	}while(Thr<TempProduct);

	result -= StartPower;
	TempProduct = TempProduct/(double)Thr;


	if(TempProduct<1)
//		result -= LOG_2[max((int)(TempProduct*N_LOG_2 + 0.5) - 1,0)];
		result -= log(TempProduct)/LOG2;

	result = (-1)*result;

	result = (double)((llong)(result*10000 + 0.5))/(double)10000;

	return result;

}



double GetChildCodeSize(MULTITREE *CT, int index)
{
	ChildListItem *pCurChildItem=0;
	PixelsListItem *pCurPixItem=0;

	double Amount=0;
	double total = 0;
	double result = 0;

//	for(i=0;i<(unsigned)NColors;i++)
	for(pCurPixItem=CT->pPixels;pCurPixItem;pCurPixItem=pCurPixItem->pNext)
		total += pCurPixItem->Counter;//total += CT->Pixels[i];

	if(!total)
		return 0;

	pCurPixItem = FindPixelsItem(CT,index);
	Amount = pCurPixItem ? (double)pCurPixItem->Counter : 0;
	
//	for(i=0;i<NColors;i++)
//	result += (CT->Pixels[i]+Delta)*(log((total+NColors*Delta) /(double)(CT->Pixels[i]+Delta))/LOG2);
//	result += (CT->Pixels[index])*(log((total+NColors*Delta) /(double)(CT->Pixels[index]+Delta))/LOG2);
	result += (Amount)*(log((total+NColors*Delta) /(double)(Amount+Delta))/LOG2);

	return result;

}


double GetSplitCosts()
{
	return (double)(NColors+1);
}


double GetEntropySemi(ULONG *Pixels, double total)
{
	double result = 0;

	int i;
	double a,b;
	double c;

	for(i=0;i<NColors;i++)
	{
		if(Pixels[i])
		{
			a = total + NColors*Delta;
			b = Pixels[i] + Delta;
			c = LOG2;
			result += Pixels[i]*log(a/b)/LOG2;
		}
	}

	return result;
}



double GetEntropyAdaptS(ULONG *Pixels, double total)
{
	double result = 0;

	ULONG i,j;

	if(!total)
		return 0;

	for(i=0;i<(unsigned)NColors;i++)
	{
		if(Pixels[i])
		{
			for(j=0;j<Pixels[i];j++)
				result += log(Delta + j);
		}
	}

	for(j=0;j<(ULONG)total;j++)
		result -= log(j+NColors*Delta);

	return (-1)*result/LOG2;
}



//FAST
double GetEntropyAdapt(ULONG *Pixels, double total)
{
	ULONG i,j;
	double result = 0;


	double TempProduct;
	double Thr;

	
	double _2In52;
	double _2In32;
	double _2In24;
	double _2In16;
	double _2In8;

	int StartPower;

	_2In52 = 0x8000000000000;
	_2In32 = 0x100000000;
	_2In24 = 0x1000000;
	_2In16 = 0x10000;
	_2In8  = 0x100;


	if(!total)
		return 0;
	

	for(i=0;i<(unsigned)NColors;i++)
		result += LOG_2_Delta*Pixels[i];


	for(i=0;i<(unsigned)NColors;i++)
	{

		TempProduct = 1;

		for(j=0;j<Pixels[i];j++)
		{
			TempProduct = TempProduct*(NColors*j + 1);
			
			if(TempProduct >= _2In32)
			{
				TempProduct = TempProduct/(double)_2In32;
				result += 32;
			}

/*			if(TempProduct >= _2In24)
			{
				TempProduct = TempProduct/(double)_2In24;
				result += 24;
			}
			
			if(TempProduct >= _2In16)
			{
				TempProduct = TempProduct/(double)_2In16;
				result += 16;
			}
			
			if(TempProduct >= _2In8)
			{
				TempProduct = TempProduct/(double)_2In8;
				result += 8;
			}*/
		}

		StartPower = -1;

		if(TempProduct >= _2In52)
		{
			TempProduct = TempProduct/(double)_2In52;
			result += 52;
		}
		
		if(TempProduct >= _2In32)
		{
			TempProduct = TempProduct/(double)_2In32;
			result += 32;
		}
		
		if(TempProduct >= _2In24)
		{
			TempProduct = TempProduct/(double)_2In24;
			result += 24;
		}
		
		if(TempProduct >= _2In16)
		{
			TempProduct = TempProduct/(double)_2In16;
			result += 16;
		}
		
		if(TempProduct >= _2In8)
		{
			TempProduct = TempProduct/(double)_2In8;
			result += 8;
		}

		do
		{
			StartPower++;
			Thr = (1<<StartPower);
		}while(Thr<TempProduct);

		result += StartPower;
		TempProduct = TempProduct/(double)Thr;
		

		if(TempProduct<1)
		//	result += LOG_2[max((int)(TempProduct*N_LOG_2 + 0.5) - 1,0)];
			result += log(TempProduct)/LOG2;

	}

//	for(j = 0;j<(ULONG)total;j++)
//		result -= log(j + 1);


	TempProduct = 1;
	for(j = 0;j<(ULONG)total;j++)
	{
		TempProduct = TempProduct*(j + 1);
		if(TempProduct >= _2In32)
		{
			TempProduct = TempProduct/(double)_2In32;
			result -= 32;
		}

/*		if(TempProduct >= _2In24)
		{
			TempProduct = TempProduct/(double)_2In24;
			result -= 24;
		}
		
		if(TempProduct >= _2In16)
		{
			TempProduct = TempProduct/(double)_2In16;
			result -= 16;
		}
		
		if(TempProduct >= _2In8)
		{
			TempProduct = TempProduct/(double)_2In8;
			result -= 8;
		}*/
	}

	StartPower = -1;

	if(TempProduct >= _2In52)
	{
		TempProduct = TempProduct/(double)_2In52;
		result -= 52;
	}

	if(TempProduct >= _2In32)
	{
		TempProduct = TempProduct/(double)_2In32;
		result -= 32;
	}
	
	if(TempProduct >= _2In24)
	{
		TempProduct = TempProduct/(double)_2In24;
		result -= 24;
	}
	
	if(TempProduct >= _2In16)
	{
		TempProduct = TempProduct/(double)_2In16;
		result -= 16;
	}
	
	if(TempProduct >= _2In8)
	{
		TempProduct = TempProduct/(double)_2In8;
		result -= 8;
	}

	do
	{
		StartPower++;
		Thr = (1<<StartPower);
	}while(Thr<TempProduct);

	result -= StartPower;
	TempProduct = TempProduct/(double)Thr;


	if(TempProduct<1)
	//	result -= LOG_2[max((int)(TempProduct*N_LOG_2 + 0.5) - 1,0)];
		result -= log(TempProduct)/LOG2;


	result = (-1)*result;

	result = (double)((llong)(result*10000 + 0.5))/(double)10000;

	return result;

}

double GetConfigurationCodeLength(BYTE *Config, double *ConfBitSize, MULTITREE *CT)
{
	double value;
	PixelsListItem *pCurPixItem=0;
	ChildListItem *pCurChildItem=0;

	int i;

	int flag;

	int ReducedNChilds = QuantizedNChilds;

	ULONG *Pixels;
	ULONG total;

	Pixels = (ULONG*)malloc(sizeof(ULONG)*NColors);
	total = 0;

	if(REDUCED)
		ReducedNChilds = GetReducedNChilds(CT);

	
	if(OPT)
		value = ReducedNChilds*COEFF;
	else
	{
		value = 1;
		for(i=0;i<ReducedNChilds;i++) 
			if(Config[i])
			{
				value = ReducedNChilds*COEFF + 1;
				break;
			}
	}

	memset(Pixels,0,NColors*sizeof(ULONG));
	//for(i=0;i<NColors;i++)
	for(pCurPixItem=CT->pPixels;pCurPixItem;pCurPixItem=pCurPixItem->pNext)
		Pixels[pCurPixItem->PixelID] = pCurPixItem->Counter;//CT->Pixels[i];

	//for(i=0;i<ReducedNChilds;i++)
	for(pCurChildItem=CT->pChildren;pCurChildItem;pCurChildItem=pCurChildItem->pNext)
	{
		if(Config[pCurChildItem->PixelID/*i*/]/* && CT->Children[i]*/)
		{
			//for(j=0;j<NColors;j++)
			for(pCurPixItem=pCurChildItem->pNode->pPixels;pCurPixItem;pCurPixItem=pCurPixItem->pNext)
			{
				//Pixels[j] = Pixels[j] - CT->Children[i]->Pixels[j];
				Pixels[pCurPixItem->PixelID] -= pCurPixItem->Counter;
			}

			value += ConfBitSize[pCurChildItem->PixelID/*i*/];
		}
	}

	for(i=0;i<NColors;i++)
		total += Pixels[i];

	if(total)
	{
		if(!SEMI)
			value += GetEntropyAdapt(Pixels, total);
		else
			exit(-1);
	}



	free(Pixels);
//	if(IS_CONFIG_LUT)
///		Put2Lut(Config,value, CT);

	if(FREE_TREE == 1 && CT->ChildPosition <= FreeTreeDepth)
	{
		flag = 0;
		
		//for(i = 0; i < QuantizedNChilds; i++)
		for(pCurChildItem=CT->pChildren;pCurChildItem;pCurChildItem=pCurChildItem->pNext)
		{
			if(Config[pCurChildItem->PixelID/*i*/] /*&& CT->Children[i]*/)
			{
				flag = 1;
				break;
			}
		}

		if(flag)
			value += BitLength;
	}

	return value;
}


void ZeroLut(MULTITREE* CT)
{
	int LocalNColors;

	LocalNColors = GetPaletteNumber(CT);

	ConfigLut = (double*)malloc(sizeof(double)*(1<<LocalNColors));

	memset(ConfigLut,0, (1<<LocalNColors)*sizeof(double));
	
//	for(i=0;i<(1<<NColors);i++) ConfigLut[i] = -1;

}

double GetFromLut(BYTE *Config, MULTITREE* CT)
{
	ChildListItem *pCurItem=0;
	ULONG index = 0;
//	int i;

//	for(i = 0; i<NColors-1; i++)
	for(pCurItem=CT->pChildren;pCurItem;pCurItem=pCurItem->pNext)
	{
//		if(CT->Children[i])
		{
//			index |= Config[i];
			index |= Config[pCurItem->PixelID];
			index = index<<1;
		}
	}
	
//	index |= Config[i];
	index |= Config[NColors];

	return ConfigLut[index];

}

void Put2Lut(BYTE* Config, double Val, MULTITREE* CT)
{
	ChildListItem *pCurItem=0;
	ULONG index = 0;
//	int i;

	//for(i = 0; i<NColors-1; i++)
	for(pCurItem=CT->pChildren;pCurItem;pCurItem=pCurItem->pNext)
	{
//		if(CT->Children[i])
		{
			//index |= Config[i];
			index |= Config[pCurItem->PixelID];
			index = index<<1;
		}
	}
	
//	index |= Config[i];
	index |= Config[NColors];

	ConfigLut[index] = Val;
}


int GetPaletteNumber(MULTITREE* CT)
{
	ChildListItem *pCurItem=0;
	int result = 0;
//	int i;

//	if(!CT->Children)
	if(!CT->pChildren)
		return result;

//	for(i = 0;i<NColors;i++)
	for(pCurItem=CT->pChildren;pCurItem;pCurItem=pCurItem->pNext)
	{
//		if(CT->Children[i])
			result++;
	}
	return result;
}

double DescentOptimum(double *ConfBitSize, MULTITREE *CT, BYTE *Config, double *StartValue, int Direction, int Start, int LowBound)
{
	ChildListItem *pCurItem=0;
	MULTITREE **pChildren=0;
	int i;
	double ResultValue = (*StartValue);
	double MinValue;
	double *MinValues;
	double temp;
	int LocalLowBound;

	BYTE **LocalMinConfig;


	pChildren = (MULTITREE**)malloc(sizeof(MULTITREE*)*QuantizedNChilds);
	memset(pChildren,0,sizeof(MULTITREE*)*QuantizedNChilds);
	for(pCurItem=CT->pChildren;pCurItem;pCurItem=pCurItem->pNext)
		pChildren[pCurItem->PixelID]=pCurItem->pNode;

	DescentCounter += 1;

	//printf("\nDescentCounter %d", DescentCounter);

	if(Start)
		LocalLowBound = LowBound;
	else
		LocalLowBound = 0;

	ResultValue = (*StartValue);

	MinValue = ResultValue;
	if(ResultValue > (*StartValue))
		return -1;

	MinValues = (double*)malloc(sizeof(double)*QuantizedNChilds);
	LocalMinConfig = (BYTE**)malloc(sizeof(BYTE*)*QuantizedNChilds);

	for(i=0;i<QuantizedNChilds;i++) LocalMinConfig[i] = (BYTE*)malloc(sizeof(BYTE)*QuantizedNChilds);

	
	for(i=LocalLowBound;i<QuantizedNChilds;i++) 
	{
		memcpy(LocalMinConfig[i], Config, QuantizedNChilds*sizeof(BYTE));
		MinValues[i] = -1;
		if((LocalMinConfig[i][i] + Direction)>=0 && (LocalMinConfig[i][i] + Direction)<=1)
		{
			LocalMinConfig[i][i] += Direction;
			MinValues[i] = GetConfigurationCodeLength(LocalMinConfig[i], ConfBitSize,CT);

			if(MinValues[i] < MinValue)
			{
				MinValue = MinValues[i];
			}
		}
	}


	if(*StartValue > MinValue && CT->ChildPosition<5)
		i = i;

	for(i=LocalLowBound;i<QuantizedNChilds;i++)
	{
		if(pChildren[i]/*CT->Children[i]*/)
		{
			if(MinValues[i]!=(double)(-1) && (double)fabs(MinValues[i] - MinValue)<(double)0.01)
			{
				temp = DescentOptimum(ConfBitSize, CT, LocalMinConfig[i], &MinValue, Direction,Start,i);

				if(temp!=-1)
				{
					if(temp < ResultValue)
					{
						ResultValue = temp;
						memcpy(Config,LocalMinConfig[i], QuantizedNChilds*sizeof(BYTE));
					}
				}
			}
		}
	}


	for(i=0;i<QuantizedNChilds;i++) free(LocalMinConfig[i]);

	free(pChildren);
	free(LocalMinConfig);
	free(MinValues);
	return ResultValue;
}


int DescentConfig(double *ConfBitSize, MULTITREE *CT, double *Value, BYTE *ResConfig)
{
	ChildListItem *pCurItem=0;
	int conf = 0;

	int i;

	int Direction;
	BYTE *Config;
	double Val1, Val2,  MinVal, TempVal;

	Config = (BYTE*)malloc(sizeof(BYTE)*QuantizedNChilds);
	DescentCounter = 0;

	memset(Config, 0, sizeof(BYTE)*QuantizedNChilds);
	Val1 = GetConfigurationCodeLength(Config, ConfBitSize, CT);

	//for(i=0;i<QuantizedNChilds; i++)
	for(pCurItem=CT->pChildren;pCurItem;pCurItem=pCurItem->pNext)
		//if(CT->Children[i])
			//Config[i] = 1;
			Config[pCurItem->PixelID] = 1;
		

	Val2 = GetConfigurationCodeLength(Config, ConfBitSize, CT);

	MinVal = Val2;

	if(Val1 < Val2)
	{
		Direction = 1;
		memset(Config,0, sizeof(BYTE)*QuantizedNChilds);
		MinVal = Val1;
	}
	else
		Direction = -1;
	
	TempVal = DescentOptimum(ConfBitSize, CT, Config, &MinVal, Direction, 1, 0);

	if(TempVal != -1)
	{
		if(MinVal>TempVal)
		{
			MinVal = TempVal;
			TempVal = DescentOptimum(ConfBitSize,CT,Config,&MinVal, Direction,1,0);
			if(TempVal != -1)
				MinVal = TempVal;
		}
	}

	for(i=0;i<QuantizedNChilds;i++)
		ResConfig[i] = Config[i];

	free(Config);

	(*Value) = MinVal;
	
	
/*	if(IS_CONFIG_LUT)
		free(ConfigLut);*/

	return conf;
}

/*
EXCLUDED
void CheckTree(MULTITREE *CT)
{
	int i;
	
	if(!CT)
		return;
	
	if(!CT->Children)
		return;

	CheckChildren(CT);

	for(i=0;i<NColors;i++)
	{
		CheckTree(CT->Children[i]);
	}
}
*/
/*
EXCLUDED
void CheckChildren(MULTITREE *CT)
{
	int i,j;

	if(!CT->Children)
		return;

	for(i=0;i<NColors;i++)
	{
		if(CT->Children[i] == NULL)
		{
			for(j=0;j<NColors;j++)
			{
				if(CT->Children[j])
				{
					if(CT->Children[j]->Children)
					if(CT->Children[j]->Children[i] != NULL)
						printf("\n Wrong assumption");
				}
			}
		}
	}
}
*/

void PrintNodeInfo(MULTITREE *CT)
{
	//int i;
	PixelsListItem *pCurItem=0;
	
	fprintf(flog,"\nTotal = %d\t", CT->Total);
//	for(i = 0; i< NColors; i++)
	for(pCurItem=CT->pPixels;pCurItem;pCurItem=pCurItem->pNext)
		//fprintf(flog, "Pixels[%d]=%d\t", i, CT->Pixels[i]);
		fprintf(flog, "Pixels[%d]=%d\t", pCurItem->PixelID, pCurItem->Counter);
}

void RemoveChildItem(MULTITREE *CT, int PixelID)
{
	ChildListItem *pCurItem=CT->pChildren,*pPrevItem=CT->pChildren;

	for(; pCurItem && pCurItem->PixelID!=PixelID; )
	{
		pPrevItem=pCurItem;
		pCurItem=pCurItem->pNext;
	}

	if(pCurItem==NULL)
		return;		// No item found

	if(pCurItem==CT->pChildren)
	{
		CT->pChildren=pCurItem->pNext;
		DeallocateTree(pCurItem->pNode);
		pCurItem->pNode=NULL;
		free(pCurItem);
		pCurItem=NULL;
		return;
	}

	pPrevItem->pNext=pCurItem->pNext;
	DeallocateTree(pCurItem->pNode);
	pCurItem->pNode=NULL;
	free(pCurItem);
	pCurItem=NULL;
}

double PruneTreeOptimal(MULTITREE *CT)
{
	ChildListItem *pCurItem=0,*pTempItem=0;

	int i;//,j;
	double *ConfBitSize;
	BYTE *Config;
	double Min = (double)MAXLLONG;
	ULONG MinConfig;

//	double total;
//	double CurCodeLength;
	ULONG *Pixels;
	int flag;

	int ReducedNChilds;

	double DescentValue;
	
	if(CT->pChildren==NULL/*CT->Children == NULL*/ )
	{
		if(CT->ChildPosition>AlreadyBuiltDepth+1)
		{
			if(REDUCED)
			{
				ReducedNChilds = GetReducedNChilds(CT);
				printf("\n%d",ReducedNChilds);
			}
			
			
			if(OPT)
			{
				ReducedNChilds = NColors;
				
				if(QUANTIZE_CHILDS)
					ReducedNChilds = QuantizedNChilds;

			}
			else
				ReducedNChilds = 1;

			
			if(SEMI)
				return CT->Total*GetCodeSizeSemi(CT) + NColors + NColors*NUMBER_M;
			else
			{
				return GetCodeSizeAdapt(CT) + ReducedNChilds*COEFF/* + ((FREE_TREE == 1 && CT->ChildPosition <= FreeTreeDepth) ? BitLength : 0)*/;
			}
		}
		else
			return 0;
	}
	else
	{

		ConfBitSize = (double*)malloc(sizeof(double)*QuantizedNChilds);
		Pixels = (ULONG*)malloc(sizeof(ULONG)*NColors);

		for(i=0;i<QuantizedNChilds;i++)
			ConfBitSize[i] = (TREE_COMPLETE ? 1:0);

		for(pCurItem=CT->pChildren;pCurItem;pCurItem=pCurItem->pNext)
//			if(CT->Children[i] != NULL)
			{
				LocalPath[CT->ChildPosition-1] = pCurItem->PixelID;//i;
				ConfBitSize[pCurItem->PixelID] = PruneTreeOptimal(pCurItem->pNode/*CT->Children[i]*/);
			}
//			else
//				ConfBitSize[i] = (TREE_COMPLETE ? 1:0);

		Config = (BYTE*)malloc(sizeof(BYTE)*(QuantizedNChilds));
	
		if(AlreadyBuiltDepth)
		{
			if(CT->ChildPosition <= AlreadyBuiltDepth)
				goto EXIT;
		}
		
		if(!FULL_SEARCH)
		{
			DescentConfig(ConfBitSize,CT,&DescentValue, Config);
			Min = DescentValue;
		}
		
		if(FULL_SEARCH)
		{
			exit(0);
/* NOT SUPPORTED YET
			ULONG step = 1;
			
			if(TREE_COMPLETE)
				step = ((1<<QuantizedNChilds)-1);
			
			for(i=0;i<(1<<QuantizedNChilds);i+=step)
			{
				
				if(TREE_COMPLETE)
				{
					if(i!=0 && i!=((1<<QuantizedNChilds)-1))
						continue;
				}
				
				flag = 0;
				for(j=0;j<NColors;j++)
				{
					Config[j] = (BYTE)((i>>j)&0x01);
					if(Config[j] && !CT->Children[j])
					{
						flag = 1;
						break;
					}
				}

				if(flag && !TREE_COMPLETE)
					continue;


				CurCodeLength = QuantizedNChilds*COEFF;
				

				if(!OPT && i==0)
					CurCodeLength=1;

											 //The space, necessary 
											 // to store the current node
				if(SEMI)
				if(i<(1<<NColors)-1)
					CurCodeLength += QuantizedNChilds*NUMBER_M;

				total = CT->Total;

				for(j=0;j<NColors;j++)
					Pixels[j] = CT->Pixels[j];


				for(j=0;j<QuantizedNChilds;j++)
				{
					if(Config[j])
					{
						if(CT->Children[j] != NULL)
						{
							CurCodeLength += ConfBitSize[j];
							total -= CT->Children[j]->Total;

							for(k=0;k<NColors;k++)
							{
								Pixels[k] -= CT->Children[j]->Pixels[k];
								if(Pixels[k] == -1)
									j = j;
							}
						}
					}
				}

				if(total)
				{
					if(SEMI)
						CurCodeLength += GetEntropySemi(Pixels,total);
					else
						CurCodeLength += GetEntropyAdapt(Pixels,total);
				}

				if(CurCodeLength < Min)
				{
					Min = CurCodeLength;
					MinConfig = i;
				}
			}
		*/}

		for(i=0;i<QuantizedNChilds;i++)
//		for(pCurItem=CT->pChildren;pCurItem;)
		{
			if(FULL_SEARCH)
			{ exit(0);
			// NOT SUPPORTED
/*				if(!((MinConfig>>i)&0x01))
				{
					if(CT->Children[i])
					{
						DeallocateTree(CT->Children[i]);
						CT->Children[i] = NULL;
					}
				}
				else if(TREE_COMPLETE)
				{
					if(!CT->Children[i])
						CT->Children[i] = NewTreeNode();
				}
*/			}
			else
			{
				if(Config[i]==0)
				{
//					if(CT->Children[i])
					{
						//DeallocateTree(pCurItem->pNode/*CT->Children[i]*/);
//						pTempItem = pCurItem->pNext; // go to the next
						RemoveChildItem(CT,i);//CT->Children[i] = NULL;
//						pCurItem = pTempItem;
					}
				}
				else
				{
					if(TREE_COMPLETE)
					{
						exit(0);	// ALARM!!!! OPTION IS NOT SUPPORTED YET
	//					if(!CT->Children[i])
	//						CT->Children[i] = NewTreeNode();
					}
//					pCurItem=pCurItem->pNext;
				}
			}

		}

		
		if(FULL_SEARCH)
		{
			if(MinConfig == 0)
			{
				DeallocateChildList(CT);
//				deallocate(CT->Children);
//				CT->Children = NULL;
			}
		}
		else
		{
			flag = 0;

			for(i=0;i<QuantizedNChilds;i++)
				if(Config[i])
				{
					flag = 1;
					break;
				}

			if(!flag)
			{
				DeallocateChildList(CT);
//				deallocate(CT->Children);
//				CT->Children = NULL;
			}
		}

EXIT:	
		
		if(CT->pChildren/*CT->Children*/)
		{
			//for(i=0;i<QuantizedNChilds;i++)
			for(pCurItem=CT->pChildren;pCurItem;pCurItem=pCurItem->pNext)
			{
//				if(CT->ChildPosition < NonDeleteLevel || AlreadyBuiltDepth>0)
				if(CT->ChildPosition < NonDeleteLevel /*|| AlreadyBuiltDepth>0*/)
				{
//					if(CT->Children[i])
					{
						DeallocatePixelsList(pCurItem->pNode);
//						deallocate(CT->Children[i]->Pixels);
//						CT->Children[i]->Pixels = NULL;
					}
				}
			}
		}


		free(Pixels);
		free(Config);
		free(ConfBitSize);
		return Min;
	}
}


void WriteMT(MULTITREE *CT, FLOW *flow, int *counter)
{
	int i;
	MULTITREE **pchildren = 0;
	ChildListItem *pCurItem=0;

	if(!TREE_COMPLETE)
	{
		if(OPT)
			(*counter) += QuantizedNChilds;
		else
		{
			if(CT->pChildren)
				(*counter) +=(QuantizedNChilds+1);
			else
				(*counter) += 1;
		}

/*		if(CT == NULL)
		{
			for(i=0; i<QuantizedNChilds;i++) 
			flow->OutputBit(flow,0);
			return;
		}*/

		if(CT->pChildren == NULL)
		{
//			for(i=0; i<NColors;i++) 
			flow->OutputBit(flow,0);
			return;
		}
		else
		{
			
			//new
			flow->OutputBit(flow,1);

			pchildren = (MULTITREE**)malloc(QuantizedNChilds*sizeof(MULTITREE*));
			memset(pchildren,0,QuantizedNChilds*sizeof(MULTITREE*));

			for(pCurItem=CT->pChildren;pCurItem;pCurItem=pCurItem->pNext)
				pchildren[pCurItem->PixelID]=pCurItem->pNode;

			for(i=0; i<QuantizedNChilds;i++) 
//				if(CT->Children[i])
				if(pchildren[i])
					flow->OutputBit(flow,1);
				else
					flow->OutputBit(flow,0);

			for(i=0; i<QuantizedNChilds;i++) 
				if(pchildren[i])
					WriteMT(pchildren[i], flow, counter);

			free(pchildren);

			return;
		}
	}
	else
	{
		exit(0);
		// NOT SUPPORTED
/*		(*counter) +=1;
		
		if(CT->Children == NULL)
		{
			flow->OutputBit(flow,0);
			return;
		}
		else
		{
			flow->OutputBit(flow,1);
			for(i=0; i<QuantizedNChilds;i++)
				if(CT->Children[i])
				WriteMT(CT->Children[i], flow, counter);
		}
*/	}
}

void ReadMT(MULTITREE *CT, FLOW *flow, int *counter)
{
	int i,r;
	ChildListItem *pCurItem=0;
	MULTITREE **pchildren = 0;

	if(TREE_COMPLETE)
	{
		// NOT SUPPORTED
/*		(*counter)++;
		
		if(flow->InputBit(flow) == 0)
		{
			
		}
		else
		{
//			CreateChildren(CT);
			for(i=0;i<QuantizedNChilds;i++)
			{
				AddChildListItem(CT,i,CT->ChildPosition+1);
				//CT->Children[i] = NewTreeNode();
				//CT->Children[i]->ChildPosition = CT->ChildPosition+1;
				//ReadMT(CT->Children[i], flow, counter);
				ReadMT(CT->pChildren->pNode, flow, counter);  // since new elements are added to the head of the list
			}
		}
*/	}
	else
	{
		if(flow->InputBit(flow))
		{
			pchildren = (MULTITREE**)malloc(QuantizedNChilds*sizeof(MULTITREE*));
			memset(pchildren,0,QuantizedNChilds*sizeof(MULTITREE*));
		
			for(i=0;i<QuantizedNChilds;i++)
			{
				r = flow->InputBit(flow);
				(*counter)++;

				if(r)
				{
//					if(CT->Children == NULL)
//					{
//						CreateChildren(CT);
//					}

					pCurItem = AddChildListItem(CT,i,CT->ChildPosition + 1);
					pchildren[i] = pCurItem->pNode;

					//CT->Children[i] = NewTreeNode();
					//CT->Children[i]->ChildPosition = CT->ChildPosition + 1;
				}
			}

			if(CT->pChildren)
			{
				for(i=0;i<QuantizedNChilds;i++)
				//for(pCurItem=CT->pChildren;pCurItem;pCurItem=pCurItem->pNext)
				{
					if(pchildren[i])
						ReadMT(pchildren[i], flow, counter);

				}
			}
			free(pchildren);

//			PrintTree(CT);
		}
	}
}


/************************************************************/


MULTITREE *SetContextImage2(MULTITREE *CT, IMAGE* GrayImage, int *lut, int x, int y, 
							int StartDepth, int EndDepth, int refPixelN, int StartX, int StartY, int EndX, int EndY)
{
	ChildListItem *pCurChildItem=0;
	PixelsListItem *pCurPixItem=0;

  int       i;
  int       pixelN;
  MULTITREE *CCT;

  int refx, refy;
  

#ifdef TEMPLATE_2D
    int  tJBIGX[31] = {30,-1,0,-1,+1,-2,0,-2,+2,-1,+1,-2,+2,-3,0,-3,+3,-1,+1,-3,+3,-2,+2,-4, 0,-4,+4,-1,+1,-3,+3 };
    int  tJBIGY[31] = {30,0,-1,-1,-1,0,-2,-1,-1,-2,-2,-2,-2,0,-3,-1,-1,-3,-3,-2,-2,-3,-3, 0,-4,-1,-1,-4,-4,-3,-3 };

//	int  tJBIGX[31] = {30,-1,  0, -2, +1, -1, 0, -3, +2, -2, +1, -1, 0, -4, +3, -3, +2, -2, +1, -1, 0, -5, +4, -4, +3, -3, +2, -2, +1, -1,  0};
//    int  tJBIGY[31] = {30, 0, -1,  0, -1, -1, -2, 0, -1, -1, -2, -2, -3, 0, -1, -1, -2, -2, -3, -3, -4, 0, -1, -1, -2, -2, -3, -3, -4, -4, -5};

#else 
    int  tJBIGX[31] = {30,-1,-2,-3,-4,-5, -6,-7,-8,-9,-10,-11,-12,-13, -14,-15, -16,-17, -18,-19,-20,-21,-22, -23, -24,-25,-26,-27,-28,-29,-30};
    int  tJBIGY[31] = {30,0,0,0,0,0,00,00,00,00,00,00,00,0,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00 };
#endif


  refPixelN = lut[refPixelN];


	for(i=1;i<=StartDepth;i++)
	{
		refx = x + tJBIGX[i];
		refy = y + tJBIGY[i];
		
	if((refx>=StartX && refx <= EndX) && (refy>=StartY && refy<=EndY))
		pixelN = lut[GetGrayPixel(GrayImage,refx,refy)];
	else
		pixelN = 0;
		

		if(!QUANTIZE_CHILDS)
		{
			if(LocalPath[i-1] != pixelN)
				return NULL;
		}
		else
		{
			if(LocalPath[i-1] != ChildsLut[pixelN])
				return NULL;
		}
	}

	CCT=CT;

	for (i=StartDepth+1;i<=EndDepth;i++)
	{

		refx = x + tJBIGX[i];
		refy = y + tJBIGY[i];
		
		if((refx>=StartX && refx <= EndX) && (refy>=StartY && refy<=EndY))
			pixelN = lut[GetGrayPixel(GrayImage,refx,refy)];
		else
			pixelN = 0;

		
		if(i>StartDepth+1)
		{
			IncreasePixelsListItem(CCT,refPixelN);
			//CCT->Pixels[refPixelN]++;
			CCT->Total += 1;
			CCT->ChildPosition = i;
		}


		if(QUANTIZE_CHILDS)
			pixelN = ChildsLut[pixelN];

//		if (CCT->Children==NULL) CreateChildren(CCT);
 
		pCurChildItem = FindChildItem(CCT,pixelN);
//		if (CCT->Children[pixelN]==NULL)
		if(pCurChildItem==NULL)
		{
		  if (i==EndDepth) 
			  ContextsAdded++;

//		  CCT->Children[pixelN] = NewTreeNode();
		  pCurChildItem = AddChildListItem(CCT,pixelN,-1);
		  
//		  assert(CCT->Children[pixelN]!=NULL);
		}
//		CCT=CCT->Children[pixelN];
		CCT=pCurChildItem->pNode;
	}   
  
  CCT->ChildPosition = i;
  IncreasePixelsListItem(CCT,refPixelN);
  
//  CCT->Pixels[refPixelN]++;
  return (CCT);
}
/************************************************************/

void ConstructTreeImage2(MULTITREE *CT, IMAGE *GrayImage, int *lut, int StartDepth, int EndDepth, int index)
{
  MULTITREE *CCT;
  ULONG     x,y;
  int       pixelN;
  int StartX, StartY;
  int EndX, EndY;


  if(QuietLevel>2)
  {
    SetProgressVisibility(1);
    OpenProgressWindow();
  }

  GetClusterBounds(GrayImage, index, &StartX, &StartY, &EndX, &EndY);
  ImageRewind(GrayImage);


  for(y=(unsigned)StartY;y<=(unsigned)EndY;y++)
  {
	for(x = (unsigned)StartX; x <= (unsigned)EndX; x++)
	{
		pixelN = GetGrayPixel(GrayImage,x,y);
		CCT = SetContextImage2(CT,GrayImage,lut,x,y,StartDepth,EndDepth,pixelN, StartX, StartY, EndX, EndY);

		if (CCT!=NULL) 
			CCT->Total++;
	}
	
	if(QuietLevel>2) 
		UpdateProgressBar(y, GrayImage->ImageSizeY);

//	if(y>=186)	
//		PrintTree(CT);
  }
  
  if(QuietLevel>2) CloseProgressWindow();
	ImageRewind(GrayImage);
}
/************************************************************/
// nadstroika CT nad already constructed CT fron StartDepth to EndDepth
MULTITREE *SetContextImage3(MULTITREE *CT, IMAGE* GrayImage, int *lut, int x, int y, 
							int StartDepth, int EndDepth, int refPixelN, int StartX, int StartY, int EndX, int EndY)
{
  int       i;
  int       pixelN;
  MULTITREE *CCT;

  ChildListItem *pCurChildItem=0;
  PixelsListItem *pCurPixItem=0;

  int refx, refy;
  

#ifdef TEMPLATE_2D
    int  tJBIGX[31] = {30,-1,0,-1,+1,-2,0,-2,+2,-1,+1,-2,+2,-3,0,-3,+3,-1,+1,-3,+3,-2,+2,-4, 0,-4,+4,-1,+1,-3,+3 };
    int  tJBIGY[31] = {30,0,-1,-1,-1,0,-2,-1,-1,-2,-2,-2,-2,0,-3,-1,-1,-3,-3,-2,-2,-3,-3, 0,-4,-1,-1,-4,-4,-3,-3 };

//	int  tJBIGX[31] = {30,-1,  0, -2, +1, -1, 0, -3, +2, -2, +1, -1, 0, -4, +3, -3, +2, -2, +1, -1, 0, -5, +4, -4, +3, -3, +2, -2, +1, -1, 0};
//  int  tJBIGY[31] = {30, 0, -1,  0, -1, -1, -2, 0, -1, -1, -2, -2, -3, 0, -1, -1, -2, -2, -3, -3, -4, 0, -1, -1, -2, -2, -3, -3, -4, -4, -5};

#else 
    int  tJBIGX[31] = {30,-1,-2,-3,-4,-5, -6,-7,-8,-9,-10,-11,-12,-13, -14,-15, -16,-17, -18,-19,-20,-21,-22, -23, -24,-25,-26,-27,-28,-29,-30};
    int  tJBIGY[31] = {30,0,0,0,0,0,00,00,00,00,00,00,00,0,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00 };
#endif

  refPixelN = lut[refPixelN];


	CCT=CT;
	for (i=1;i<=EndDepth;i++)
	{

		refx = x + tJBIGX[i];
		refy = y + tJBIGY[i];
		
		if((refx>=StartX && refx <= EndX) && (refy>=StartY && refy<=EndY))
			pixelN = lut[GetGrayPixel(GrayImage,refx,refy)];
		else
			pixelN = 0;

		
		if(CCT->ChildPosition > StartDepth+1)
		{
			pCurPixItem = FindPixelsItem(CCT,refPixelN);
			pCurPixItem->Counter++;
			//CCT->Pixels[refPixelN]++;
			CCT->Total += 1;
			CCT->ChildPosition = i;
		}

		if(CCT->pChildren==NULL) 
		{
			if(CCT->ChildPosition > StartDepth)
				;//CreateChildren(CCT);
			else
				return (CCT);
		}
 
		if(QUANTIZE_CHILDS)
			pixelN = ChildsLut[pixelN];

		pCurChildItem = FindChildItem(CCT,pixelN);
		//if (CCT->Children[pixelN]==NULL)
		if (pCurChildItem==NULL)
		{
		  if(CCT->ChildPosition > StartDepth)
		  {
			  if (i==EndDepth) 
				  ContextsAdded++;

			  pCurChildItem = AddChildListItem(CCT,pixelN,CCT->ChildPosition+1);

			  //CCT->Children[pixelN] = NewTreeNode();
			  //CCT->Children[pixelN]->ChildPosition = CCT->ChildPosition+1;

			  //assert(CCT->Children[pixelN]!=NULL);
		  }
		  else
			  return (CCT);
		}
		//CCT=CCT->Children[pixelN];
		CCT = pCurChildItem->pNode;
	}   
  
  CCT->ChildPosition = i;
  pCurPixItem = FindPixelsItem(CCT,refPixelN);
  pCurPixItem->Counter++;
  //CCT->Pixels[refPixelN]++;
  return (CCT);
}




/************************************************************/
/*
void ConstructTreeImage3(MULTITREE *CT, IMAGE *GrayImage, int *lut, int StartDepth, int EndDepth, int index)
{
  MULTITREE *CCT;
  ULONG     x,y;
  int       pixelN;
  int StartX, StartY;
  int EndX, EndY;


  if(QuietLevel>2)
  {
    SetProgressVisibility(1);
    OpenProgressWindow();
  }

  ImageRewind(GrayImage);

  GetClusterBounds(GrayImage, index, &StartX, &StartY, &EndX, &EndY);

  for(y = (unsigned)StartY; y <= (unsigned)EndY; y++)
  {

	for(x = (unsigned)StartX; x <= (unsigned)EndX; x++)
	{
		pixelN = GetGrayPixel(GrayImage,x,y);
		CCT = SetContextImage3(CT,GrayImage,lut,x,y,StartDepth,EndDepth,pixelN, StartX, StartY, EndX, EndY);

		if (CCT!=NULL && CCT->ChildPosition>StartDepth+1) 
			CCT->Total++;
	}
	
	if(QuietLevel>2) 
		UpdateProgressBar(y, GrayImage->ImageSizeY);
  }
  
  if(QuietLevel>2) CloseProgressWindow();
  
  ImageRewind(GrayImage);
}
*/
/************************************************************/

void AddOnSubtree(MULTITREE* CT, ULONG* counter, int StartDepth)
{
	int tmp = 0;
	FLOW flow;
//	int i;
	FILE *f;
//	char *temp;

	ChildListItem *pCurItem=0;


	if(CT->pChildren == NULL)
	{
		
		if(CT->ChildPosition == StartDepth+1)
		{
			
//			if(CT->Total == 0)
//			{
				f = fopen(TmpTrees[(*counter)], "rb");
				if(f == NULL)
					f = f;
				InitFlow(&flow, F_FILE, f,1);
		
				ReadMT(CT, &flow, &tmp);
				flow.FlushInput(&flow);
				fclose(f);
//			}

//			(*counter) += 1;
		}
		else
		{
//			(*counter)+=(int)pow(QuantizedNChilds, StartDepth - (CT->ChildPosition-1));
		}
	}
	else
	{
		
//		temp = (char*)malloc(QuantizedNChilds*sizeof(char));
//		memset(temp,0,QuantizedNChilds*sizeof(char));

		//for(i=0;i<QuantizedNChilds;i++)
		for(pCurItem=CT->pChildren;pCurItem;pCurItem=pCurItem->pNext)
		{
			*counter = pCurItem->PixelID;
			//if(CT->Children[i])
			AddOnSubtree(pCurItem->pNode/*CT->Children[i]*/,counter, StartDepth);
//			temp[pCurItem->PixelID]=1;
//			else
//				(*counter)+=(int)pow(QuantizedNChilds, StartDepth - (CT->ChildPosition));
		}
/*
		for(i=0;i<QuantizedNChilds;i++)
		{
			if(temp[i]==0)
				(*counter)+=(int)pow(QuantizedNChilds, StartDepth - (CT->ChildPosition));
		}
*/
//		free(temp);
	}
}

void ReconstructTree(int StartDepth)
{
	int i;
	ULONG N = 0;

	AddOnSubtree(Root, &N, StartDepth);

	for(i=0;i<(int)pow(QuantizedNChilds, StartDepth);i++)
	{
		remove(TmpTrees[i]);
		free(TmpTrees[i]);
	}
	
	
	free(TmpTrees);

	TmpTrees = NULL;

}

void StoreSubtree(MULTITREE *CT)
{
	FLOW flow;
	int counter = 0;
	FILE *f;

	f = fopen(TmpTrees[Counter],"wb");

	InitFlow(&flow, F_FILE, f,1);

	WriteMT(CT, &flow, &counter);
	flow.FlushOutput(&flow);
	fclose(f);
}

void InitTmpTrees(int StartDepth)
{
	char *name;
	int i;

	TmpTrees = (char**)malloc(sizeof(char*)*((int)pow(QuantizedNChilds,StartDepth)));
	
	for(i=0;i<(int)pow(QuantizedNChilds,StartDepth);i++)
	{
		name = tmpnam(NULL);
		if(name == NULL)
		{
			printf("\nUnable to create tmp name\n");
			exit(-1);
		}

		TmpTrees[i] = (char*)malloc(sizeof(char)*256);
		TmpTrees[i][0] = 0;
		strcpy(TmpTrees[i], name);

//		if(name)
//			free(name);
	}

}


int GetNumberOfClustersMT(IMAGE *GrayImage)
{
	int h_num, v_num;

	if(Props.SizeX == 0 || Props.SizeY == 0)
		return 1;

	h_num = (int)ceil((double)GrayImage->ImageSizeX/(double)Props.SizeX);
	v_num = (int)ceil((double)GrayImage->ImageSizeY/(double)Props.SizeY);


	return h_num*v_num;
}



double PruneTreeOptimalWithConstructing(MULTITREE *CT, IMAGE *GrayImage, int *lut, int StartDepth, int EndDepth, int TemplateX, int TemplateY)
{
	ChildListItem *pCurItem=0;
	int i,k;
	double *ConfBitSize;
	BYTE *Config;
	double Min = (double)MAXLLONG;
//	ULONG MinConfig;

//	double total;
//	double CurCodeLength;
	ULONG *Pixels;
	int flag;

	ULONG Total = 0, Useless = 0;

	double DescentValue;
	
	int ReducedNChilds;
	int NumberOfClusters;
//	char* temp;
	
	if(CT == Root)
	{
		if(TmpTrees == NULL)
			InitTmpTrees(StartDepth);
	}
		
	if(CT->pChildren == NULL || (FREE_TREE == 1 && CT->ChildPosition == StartDepth+1))
	{
		if(CT->ChildPosition == StartDepth+1)
		{
			if(FREE_TREE == 1)
			{
				if(EndDepth > FreeTreeDepth)
					ConstructTreeImageFTreeHybrid2(CT, GrayImage, lut, StartDepth, EndDepth, TemplateX, TemplateY);


				printf("\n  Hybrid tree constructed");
				printf("\n Local path: ");
				for(k = 0; k<CT->ChildPosition; k++)
					printf("%d ", LocalPath[k]);
//				CheckTheTree(Root);
			}
			else
			{

				NumberOfClusters = GetNumberOfClustersMT(GrayImage);
				for(k = 0; k<NumberOfClusters; k++)
					ConstructTreeImage2(CT, GrayImage, lut, StartDepth, EndDepth, k);

				//DEBUG	
//				CheckTheTree(Root);
//				PrintTree(Root);

			}

			
//			GetMemoryUsage(Root,&Total, &Useless, QuantizedNChilds);
//			printf("Tree cost | Total : %i Useless : % i. %f%% wasted\n",Total,Useless,(double)Useless/(double)Total*100.0);

//			PrintTree(CT);
			Min = PruneTreeOptimal(CT);
//			PrintTree(CT);

			if(FREE_TREE == 1)
			{
				if(CT->ChildPosition <= FreeTreeDepth && CT->pChildren)
					Min += BitLength;
			}
			
			if(TmpTrees)
			{
				StoreSubtree(CT);
				DeallocateChildren(CT);
			}

//			Counter++;

			printf("\n%d\n", Counter);
				
			return Min;
		}
		else
		{
//			Counter += (int)(pow(QuantizedNChilds, StartDepth - CT->ChildPosition + 1));
			
			ReducedNChilds = NColors;

			if(QUANTIZE_CHILDS)
				ReducedNChilds = QuantizedNChilds;

			if(REDUCED)
				ReducedNChilds = GetReducedNChilds(CT);

			if(!OPT)
				ReducedNChilds = 1;


			if(SEMI)
				return CT->Total*GetCodeSizeSemi(CT) + NColors + NColors*NUMBER_M;
			else
				return GetCodeSizeAdapt(CT) + ReducedNChilds*COEFF /*+ ((FREE_TREE == 1 && CT->ChildPosition <= FreeTreeDepth) ? BitLength : 0)*/;
		}
	}
	else
	{
		
		ConfBitSize = (double*)malloc(sizeof(double)*(QuantizedNChilds));
		Pixels = (ULONG*)malloc(sizeof(ULONG)*NColors);

		//for(i=0;i<QuantizedNChilds;i++)
		for(pCurItem=CT->pChildren;pCurItem;pCurItem=pCurItem->pNext)
			//if(CT->Children[i] != NULL)
		{
			LocalPath[CT->ChildPosition-1] = pCurItem->PixelID;//i
			Counter = pCurItem->PixelID;
			ConfBitSize[pCurItem->PixelID/*i*/] = PruneTreeOptimalWithConstructing(pCurItem->pNode/*CT->Children[i]*/, GrayImage, lut, StartDepth, EndDepth, TemplateX, TemplateY);
		}

/*		temp = (char*)malloc(QuantizedNChilds*sizeof(char));
		memset(temp,0,QuantizedNChilds*sizeof(char));
		for(pCurItem=CT->pChildren;pCurItem;pCurItem=pCurItem->pNext)
			temp[pCurItem->PixelID]=1;

		for(i=0;i<QuantizedNChilds;i++)
			if(temp[i]==0)
			{
				Counter += (int)(pow(NColors, StartDepth - CT->ChildPosition));
				ConfBitSize[i] = (TREE_COMPLETE ? 1:0);
			}
		free(temp);
*/
//		PrintTree(CT);

		Config = (BYTE*)malloc(sizeof(BYTE)*(QuantizedNChilds));
	
		if(!FULL_SEARCH)
		{
			DescentConfig(ConfBitSize,CT,&DescentValue, Config);
			Min = DescentValue;
		}

		if(FULL_SEARCH)
		{
/* FULL SEARCH IS TEMPORARILY NOT SUPPORTED
			ULONG step = 1;
			
			if(TREE_COMPLETE)
				step = ((1<<QuantizedNChilds)-1);
			
			for(i=0;i<(1<<QuantizedNChilds);i+=step)
			{
				if(TREE_COMPLETE)
				{
					if(i!=0 && i!=((1<<QuantizedNChilds)-1))
						continue;
				}
				
				flag = 0;
				for(j=0;j<QuantizedNChilds;j++)
				{
					Config[j] = (BYTE)((i>>j)&0x01);
					if(Config[j] && !CT->Children[j])
					{
						flag = 1;
						break;
					}
				}

				if(flag && !TREE_COMPLETE)
					continue;


				CurCodeLength = QuantizedNChilds*COEFF;
				
				if(!OPT && i==0)
					CurCodeLength=1;							 //The space, necessary 
											 // to store the current node
				if(SEMI)
				if(i<(1<<NColors)-1)
					CurCodeLength += QuantizedNChilds*NUMBER_M;

				total = CT->Total;

				for(j=0;j<NColors;j++)
					Pixels[j] = CT->Pixels[j];


				for(j=0;j<NColors;j++)
				{
					if(Config[j])
					{
						if(CT->Children[j] != NULL)
						{
							CurCodeLength += ConfBitSize[j];
							total -= CT->Children[j]->Total;

							for(k=0;k<NColors;k++)
							{
								Pixels[k] -= CT->Children[j]->Pixels[k];
								if(Pixels[k] == -1)
									j = j;
							}
						}
					}
				}

				if(total)
				{
					if(SEMI)
						CurCodeLength += GetEntropySemi(Pixels,total);
					else
						CurCodeLength += GetEntropyAdapt(Pixels,total);
				}

				if(CurCodeLength < Min)
				{
					Min = CurCodeLength;
					MinConfig = i;
				}
			}
	*/	}


		for(i=0;i<QuantizedNChilds;i++)
		{
			if(FULL_SEARCH)
			{
/*				if(!((MinConfig>>i)&0x01))
				{
					if(CT->Children[i])
					{
						DeallocateTree(CT->Children[i]);
						CT->Children[i] = NULL;
					}
				}
				else if(TREE_COMPLETE)
				{
					if(!CT->Children[i])
						CT->Children[i] = NewTreeNode();
				}
*/			}
			else
			{
				if(!Config[i])
				{
					RemoveChildItem(CT,i);
					/*if(CT->Children[i])
					{
						DeallocateTree(CT->Children[i]);
						CT->Children[i] = NULL;
					}*/
				}
				else if(TREE_COMPLETE)
				{
					exit(0);
					// NOT SUPPORTED TEMPORARILY
//					if(!CT->Children[i])
//						CT->Children[i] = NewTreeNode();
				}
			}
		}

		
		if(FULL_SEARCH)
		{
/*			if(MinConfig == 0)
			{
				deallocate(CT->Children);
				CT->Children = NULL;
			}
*/		}
		else
		{
			flag = 0;

			for(i=0;i<QuantizedNChilds;i++)
				if(Config[i])
					flag = 1;

			if(!flag)
			{
				DeallocateChildList(CT);
				//deallocate(CT->Children);
				//CT->Children = NULL;
			}
		}

		free(Pixels);
		free(Config);
		free(ConfBitSize);


		if(CT->pChildren)
		{
			//for(i=0;i<QuantizedNChilds;i++)
			for(pCurItem=CT->pChildren;pCurItem;pCurItem=pCurItem->pNext)
			{
				//if(CT->Children[i])
				{
					//deallocate(CT->Children[i]->Pixels);
					DeallocatePixelsList(pCurItem->pNode);
					//CT->Children[i]->Pixels = NULL;
				}
			}
		}

		if(CT == Root && TmpTrees)
			ReconstructTree(StartDepth);

		return Min;
	}
}



/*************************************************************/
/*REDUCED TREE*/


int GetReducedNChilds(MULTITREE* CT)
{
	int i;
	int result = QuantizedNChilds;
	ChildListItem *pCurItem=0;

	MULTITREE *CCT;

	CCT = Root;
	
	for(i=1;i<CT->ChildPosition-1;i++)
	{
		if(CCT->pChildren)
		{
			pCurItem = FindChildItem(CT,LocalPath[i]);
			if(pCurItem)
				CCT = pCurItem->pNode;
//			if(CCT->Children[LocalPath[i]])
//				CCT = CCT->Children[LocalPath[i]];
			else
				break;
		}
		else
			break;
	}

	if(!CCT->pChildren)
		return 0;

	i=0;
	for(pCurItem=CT->pChildren;pCurItem;pCurItem=pCurItem->pNext)
		i++;
	result -= QuantizedNChilds-i;

//	for(i=0;i<QuantizedNChilds;i++)
//		if(!CCT->Children[i])
//			result = result - 1;
	
	return result;
}
//FREE Template - for best times
// Construction of free tree

TEMPLATEFT *NewFTNode()
{
  TEMPLATEFT *FT;

  FT = (TEMPLATEFT*)allocate(sizeof(TEMPLATEFT));
  FT->Children = NULL;
	
  FT->ShiftX = 0;
  FT->ShiftY = 0;
  
  return(FT);
}

void DeallocateFT(TEMPLATEFT *FT)
{
	int i;

	if(FT->Children)
	{
		for(i=0;i<QuantizedNChilds;i++)
		{
			if(FT->Children[i])
				DeallocateFT(FT->Children[i]);
		}

		deallocate(FT->Children);
	}

	deallocate(FT);
}

void CreateChildrenFT(TEMPLATEFT *FT)
{
  int i;
  
  FT->Children=(TEMPLATEFT**)allocate(sizeof(TEMPLATEFT*)*QuantizedNChilds);
  assert(FT->Children!=NULL);
  
  for (i=0;i<QuantizedNChilds;i++) FT->Children[i]=NULL;
}


void DefineRange(ULONG ChildsNumber, int x_size, int y_size, int *range)
{
	double UpperLimit;
	double MaxMem = 0x8000000;
//	double MaxMem = 0x80000;
	int i;
	int TemplateSize = ( x_size*y_size + (x_size+1)*(y_size-1));


//	return QuantizedNChilds;

	UpperLimit = (double)TemplateSize * (double)ChildsNumber * (double)QuantizedNChilds * (double)NColors * (double)(sizeof(ULONG));

	if(UpperLimit <= MaxMem)
		(*range) = ChildsNumber;
	else
	{
		i = 2;
		while(1)
		{
			if(UpperLimit/i <= MaxMem)
			{
				(*range) = ChildsNumber/i;
				break;
			}
			else
				i++;

			if((unsigned)i>ChildsNumber)
			{
				(*range) = 1;
				break;
			}
		}
	}
}

int Shifts2Index(int x_size, int y_size, int ShiftX, int ShiftY)
{
	int result;
	int row_length;

	ShiftY = (-1)*ShiftY;

	//check, whether the shifts belong to TemplateArea
	if(ShiftX>=0)
	{
		if(ShiftX > x_size || ShiftY >= y_size)
			return -1;
	}

	if(ShiftX < 0)
	{
		if(ShiftY >= y_size || (-1)*ShiftX > x_size)
		{
			return -1;
		}
	}
	

	

	if(!ShiftY)
		result = x_size + ShiftX;
	else
	{
		result = x_size;
		row_length = 2*x_size + 1;

		result += (ShiftY-1)*row_length;

		result += (x_size + ShiftX);
	}

	return result;
}

void Index2Shifts(int index, int x_size, int y_size, int *ShiftX, int *ShiftY)
{
	int row_length = 2*x_size + 1;
	
	(*ShiftX) = 0;
	(*ShiftY) = 0;

	if(index < x_size)
	{
		(*ShiftX) = index - x_size;
	}
	else
	{
		index -= x_size;
		(*ShiftY) = index/row_length + 1;

		(*ShiftX) = index - ((*ShiftY) - 1)*row_length - x_size;
	}

	(*ShiftY) = (-1)*(*ShiftY);
}



void InitLeavesIndexes(MULTITREE *CT, ULONG* index)
{
//	int i;
	ChildListItem *pCurItem;
	
	if(CT->pChildren == NULL)
	{
		CT->Context = (*index);
		(*index) += 1;
	}
	else
	{
		//for(i = 0;i < QuantizedNChilds; i++)
		for(pCurItem=CT->pChildren;pCurItem;pCurItem=pCurItem->pNext)
		{
			//if(CT->Children[i])
				InitLeavesIndexes(pCurItem->pNode/*CT->Children[i]*/, index);
		}
		CT->Context = (*index);
	}
}

void DefaultLeavesIndexes(MULTITREE *CT)
{
//	int i;
	ChildListItem *pCurItem;

	if(CT->pChildren == NULL)
		CT->Context = -1;
	else
	{
		CT->Context = -1;

		for(pCurItem=CT->pChildren;pCurItem;pCurItem=pCurItem->pNext)
		//for(i = 0; i < QuantizedNChilds; i++)
		{
			//if(CT->Children[i])
				DefaultLeavesIndexes(pCurItem->pNode/*CT->Children[i]*/);
		}
	}
}

void RestoreLeafByIndex(MULTITREE *CT, TEMPLATEFT *FT, ULONG SearchIndex, MULTITREE **resCT, TEMPLATEFT **resFT)
{
//	int i;
	ChildListItem *pCurItem;

	if(CT->pChildren == NULL)
	{
		if(CT->Context == SearchIndex)
		{
			(*resCT) = CT;
			(*resFT) = FT;
		}
		else
			(*resCT) = NULL;
	}
	else
	{
		
		if(SearchIndex <= CT->Context)
		{
			//for(i = 0; i < QuantizedNChilds; i++)
			for(pCurItem=CT->pChildren;pCurItem;pCurItem=pCurItem->pNext)
			{
				//if(CT->Children[i])
				{
					RestoreLeafByIndex(pCurItem->pNode/*CT->Children[i]*/, FT->Children[pCurItem->PixelID], SearchIndex, resCT, resFT);
					if((*resCT) != NULL)
						return;
				}
			}
		}
		else
			(*resCT) = NULL;
	}
}

void ConstructTreeImageFTree(MULTITREE *CT, TEMPLATEFT *FT, IMAGE *GrayImage, int *lut, int depth, int x_size, int y_size)
{
  ULONG     x,y;
  int       pixelN;

  int cd;
  int range;
  int coeff;
//  ULONG BigNumber, VeryBigNumber;
  ULONG i,j,step;

  ULONG ChildsNumber;

  ULONG ***PixelsCounts;

  int TemplateSize = ( x_size*y_size + (x_size+1)*(y_size-1));
  PixelsCounts = NULL;

  if(QuietLevel>2)
  {
    SetProgressVisibility(1);
    OpenProgressWindow();
  }


  for(cd = 1; cd <= depth; cd++)
  {

	//define range
	//init PixelsCounts
	//Find number of Childs



	ChildsNumber = 0;
	InitLeavesIndexes(CT,&ChildsNumber);
	

	DefineRange(ChildsNumber, x_size, y_size, &range);

	coeff = (int)ceil((double)ChildsNumber/(double)range);
	PixelsCounts = (ULONG***)malloc(sizeof(ULONG**)*TemplateSize);

	printf("%d %d ", cd, coeff);

	for(i = 0; i < (unsigned)TemplateSize; i++)
	{
		PixelsCounts[i] = (ULONG**)malloc(sizeof(ULONG*)*(range));
		for(j = 0; j < (unsigned)range; j++)
			PixelsCounts[i][j] = (ULONG*)malloc(sizeof(ULONG)*NColors*QuantizedNChilds);
	}

	for(step = 0; step < (unsigned)coeff; step++)
	{
		printf("%d\n", step);
		
		for(i = 0; i < (unsigned)TemplateSize; i++)
			for(j = 0; j < (unsigned)range; j++)
				memset(PixelsCounts[i][j],0,sizeof(ULONG)*NColors*QuantizedNChilds);

		for(y = 1; y <= GrayImage->ImageSizeY; y++)
		{
			for(x = 1; x <= GrayImage->ImageSizeX; x++)
			{
				pixelN = GetGrayPixel(GrayImage,x,y);
				UpdatePixelsCountsFT(CT, FT, GrayImage, lut, x, y, cd-1, cd, pixelN,
					TemplateSize, x_size, y_size, range, step, PixelsCounts);
			}

			if(QuietLevel>2) 
				UpdateProgressBar(y, GrayImage->ImageSizeY);
		}
		
		if(QuietLevel>2) CloseProgressWindow();
			ImageRewind(GrayImage);


		// find optimal template position for each child
		// and update CT and FT structures


		for(i = step*range; i<min((step+1)*range, ChildsNumber); i++)
		{
			FindOptimalTemplatePosition( i, (i - step*range), PixelsCounts, 
											CT, FT, x_size, y_size);
		}

		//DEBUG
//		CheckTheTree(CT);

	}

	for(i = 0; i < (unsigned)TemplateSize; i++)
	{
		for(j = 0; j < (unsigned)range; j++)
			free(PixelsCounts[i][j]);
		free(PixelsCounts[i]);
	}

	free(PixelsCounts);
	PixelsCounts = NULL;

	DefaultLeavesIndexes(CT);
  }
}

MULTITREE *SetContextImageHybrid(MULTITREE *CT, IMAGE* GrayImage, int *lut, int x, int y, int EndDepth, int refPixelN, 
								 TEMPLATEFT *FT, int x_size, int y_size)
{
  ChildListItem *pCurChildItem=0;
  PixelsListItem *pCurPixItem=0;

  int       i;
  int       pixelN;
  MULTITREE *CCT;
  TEMPLATEFT *FFT;

  int refx, refy;
  

#ifdef TEMPLATE_2D
    int  tJBIGX[31] = {30,-1,0, -1,+1,-2, 0,-2,+2,-1,+1,-2,+2,-3, 0,-3,+3,-1,+1,-3,+3,-2,+2,-4, 0,-4,+4,-1,+1,-3,+3 };
    int  tJBIGY[31] = {30,0, -1,-1,-1, 0,-2,-1,-1,-2,-2,-2,-2, 0,-3,-1,-1,-3,-3,-2,-2,-3,-3, 0,-4,-1,-1,-4,-4,-3,-3 };

//	int  tJBIGX[31] = {30,-1,  0, -2, +1, -1, 0, -3, +2, -2, +1, -1, 0, -4, +3, -3, +2, -2, +1, -1, 0, -5, +4, -4, +3, -3, +2, -2, +1, -1,  0};
//    int  tJBIGY[31] = {30, 0, -1,  0, -1, -1, -2, 0, -1, -1, -2, -2, -3, 0, -1, -1, -2, -2, -3, -3, -4, 0, -1, -1, -2, -2, -3, -3, -4, -4, -5};

#else 
    int  tJBIGX[31] = {30,-1,-2,-3,-4,-5, -6,-7,-8,-9,-10,-11,-12,-13, -14,-15, -16,-17, -18,-19,-20,-21,-22, -23, -24,-25,-26,-27,-28,-29,-30};
    int  tJBIGY[31] = {30,0,0,0,0,0,00,00,00,00,00,00,00,0,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00 };
#endif


	char *Template;
	int TemplateSize;
	int ShiftX, ShiftY;
	int index;
	char j_used[31];

	FFT = FT;
	CCT = CT;

	TemplateSize = (x_size*y_size) + (x_size+1)*(y_size-1);

	Template = malloc(TemplateSize);

	memset(Template,0,TemplateSize);
	memset(j_used, 0, 31);
	refPixelN = lut[refPixelN];

	for(i = 1; i <= FreeTreeDepth;i++)
	{
		refx = x + FFT->ShiftX;
		refy = y + FFT->ShiftY;
		
		index = Shifts2Index(x_size, y_size, FFT->ShiftX, FFT->ShiftY);

		Template[index] = 1;
		
		if((refx>0 && (unsigned)refx <= GrayImage->ImageSizeX) && (refy>0 && (unsigned)refy<=GrayImage->ImageSizeY))
			pixelN = lut[GetGrayPixel(GrayImage,refx,refy)];
		else
			pixelN = 0;

		if(QUANTIZE_CHILDS)
			pixelN = ChildsLut[pixelN];
		
		
		if(CCT->pChildren)
		{
			pCurChildItem = FindChildItem(CT,pixelN);
			if(pCurChildItem/*CCT->Children[pixelN]*/)
				CCT = pCurChildItem->pNode;//CCT->Children[pixelN];
			else
				return NULL;
		}
		else
			return NULL;

		if(i<FreeTreeDepth && FFT->Children)
			if(FFT->Children[pixelN])
				FFT = FFT->Children[pixelN];
	}

	for (i = FreeTreeDepth+1;i <= EndDepth; i++)
	{
		GetClosestUnusedTemplatePosition(Template, &ShiftX, &ShiftY, x_size, y_size, i, j_used);

		refx = x + ShiftX;
		refy = y + ShiftY;
		
		if((refx>0 && (unsigned)refx <= GrayImage->ImageSizeX) && (refy>0 && (unsigned)refy<=GrayImage->ImageSizeY))
			pixelN = lut[GetGrayPixel(GrayImage,refx,refy)];
		else
			pixelN = 0;

		pCurPixItem = FindPixelsItem(CT,refPixelN);
		if(i>FreeTreeDepth+1)
		{
			pCurPixItem->Counter++;
//			CCT->Pixels[refPixelN]++;
			CCT->Total += 1;
			CCT->ChildPosition = i;
		}

		if(QUANTIZE_CHILDS)
			pixelN = ChildsLut[pixelN];

//		if (CCT->Children==NULL) CreateChildren(CCT);
 
		pCurChildItem = FindChildItem(CT,pixelN);
		if (pCurChildItem/*CCT->Children[pixelN]==NULL*/)
		{
		  if (i==EndDepth) 
			  ContextsAdded++;

		  pCurChildItem = AddChildListItem(CCT,pixelN,-1);
		  //CCT->Children[pixelN] = NewTreeNode();

		  //assert(CCT->Children[pixelN]!=NULL);
		}
		//CCT=CCT->Children[pixelN];
		CCT = pCurChildItem->pNode;
	}   

	if(i>FreeTreeDepth+1)
	{
		CCT->ChildPosition = i;
		pCurPixItem = FindPixelsItem(CT,refPixelN);
		pCurPixItem->Counter++;
//		CCT->Pixels[refPixelN]++;
		CCT->Total++;
	}

	free(Template);

	return (CCT);
}


void ConstructTreeImageFTreeHybrid(MULTITREE *CT, TEMPLATEFT *FT, IMAGE *GrayImage, int *lut, int EndDepth, int x_size, int y_size)
{
	MULTITREE *CCT;
	ULONG     x,y;
	int       pixelN;

	if(QuietLevel>2)
	{
		SetProgressVisibility(1);
		OpenProgressWindow();
	}

	ImageRewind(GrayImage);

	for(y=1;y<=GrayImage->ImageSizeY;y++)
	{

		for(x = 1; x <= GrayImage->ImageSizeX; x++)
		{
			pixelN = GetGrayPixel(GrayImage,x,y);
			CCT = SetContextImageHybrid(CT,GrayImage,lut,x,y,EndDepth,pixelN, FT, x_size, y_size);

//			if(CCT != NULL)
//				CCT->Total++;
		}
		
		if(QuietLevel>2) 
			UpdateProgressBar(y, GrayImage->ImageSizeY);
	}

	if(QuietLevel>2) CloseProgressWindow();
		ImageRewind(GrayImage);
}

double GetNodeCodeLength(ULONG *PixelsCounts)
{
	//two options are possible  - non optimized and optimized (fast and slow)
	double result = (double)MAXLLONG;
	ULONG sum;
	ULONG *Pixels;
	int i,j;

	BYTE *Config;
	MULTITREE *CT;
	double *ConfBitSize;

	ChildListItem *pCurItem=0, *pCurItem2=0;
	PixelsListItem *pCurPixItem=0;

	CT = NULL;

	Pixels = (ULONG*)malloc(sizeof(ULONG)*NColors);
	Config = (BYTE*)malloc(QuantizedNChilds);
	ConfBitSize = (double*)malloc(sizeof(double)*QuantizedNChilds);

	CT = NewTreeNode();

	memset(ConfBitSize, 0, sizeof(double)*QuantizedNChilds);

	for(i = 0; i<QuantizedNChilds; i++)
	{
		sum = 0;
		for(j = i*NColors; j<(i+1)*NColors; j++)
		{
			Pixels[j-i*NColors] = PixelsCounts[j];
			sum += Pixels[j-i*NColors];
		}

		if(sum)
		{
//			if(!CT->Children)
//				CreateChildren(CT);

			pCurItem = AddChildListItem(CT,i,-1);
//			CT->Children[i] = NewTreeNode();

			for(j = 0; j<NColors;j++)
			{
				//CT->Children[i]->Pixels[j] = Pixels[j];
				AddPixelsListItem(pCurItem->pNode,j,Pixels[j]);
			}

			//CT->Children[i]->Total = sum;
			pCurItem->pNode->Total=sum;
		}
	}

	memset(Pixels,0,sizeof(ULONG)*NColors);

	//for(i = 0; i<QuantizedNChilds; i++)
	for(pCurItem=CT->pChildren;pCurItem;pCurItem=pCurItem->pNext)
	{
		//if(CT->Children[i])
			//for(j = 0; j<NColors;j++)
		for(pCurPixItem=pCurItem->pNode->pPixels;pCurPixItem;pCurPixItem=pCurPixItem->pNext)
				//Pixels[j] += CT->Children[i]->Pixels[j];
				Pixels[pCurPixItem->PixelID] += pCurPixItem->Counter;
	}

	sum = 0;
	//for(i = 0; i<NColors; i++)
	for(pCurPixItem=CT->pPixels;pCurPixItem;pCurPixItem=pCurPixItem->pNext)
	{
		//CT->Pixels[i] = Pixels[i];
		pCurPixItem->Counter = Pixels[pCurPixItem->PixelID];
		//sum += Pixels[i];
		sum += Pixels[pCurPixItem->PixelID];
	}

	CT->Total = sum;

	result = 0;
	//for(i=0;i<QuantizedNChilds;i++)
	for(pCurItem=CT->pChildren;pCurItem;pCurItem=pCurItem->pNext)
	{
//		if(CT->Children[i])
		{
			if(CT->Total < 500000)
				ConfBitSize[pCurItem->PixelID] = GetCodeSizeAdapt(pCurItem->pNode/*CT->Children[i]*/);
			else
				ConfBitSize[pCurItem->PixelID] = GetCodeSizeSemi(pCurItem->pNode/*CT->Children[i]*/); //super fast
		}
	}

	if(0) //slow calculation
	{
		DescentConfig(ConfBitSize, CT, &result, Config);
	}
	else //fast
		for(i=0;i<QuantizedNChilds;i++)
			result += ConfBitSize[i];


	//DEBUG

/*	temp = 0;

	fprintf(flog, "%d\t", CT->Total);

	for(i=0;i<QuantizedNChilds; i++)
	{
		fprintf(flog, "%d\t", CT->Pixels[i]);
		
		if(CT->Children[i])
			temp += GetCodeSizeSemi(CT->Children[i]);
	}

	fprintf(flog, "adapt = %0.5f, semi = %0.5f\n", result, temp);
*/

	if(CT!=NULL)
		DeallocateTree(CT);
	CT=NULL;

	free(Config);
	free(Pixels);
	free(ConfBitSize);
	return result;
}


void FindOptimalTemplatePosition(ULONG RealIndex, ULONG FactIndex, ULONG ***PixelsCounts, MULTITREE *CT, TEMPLATEFT *FT, int x_size, int y_size)
{
	ChildListItem *pCurChildItem=0;
	PixelsListItem *pCurPixelsItem=0;

	int TemplateSize;
	double MinValue;
	double CurrentValue;
	int MinPosition;
	int ShiftX, ShiftY;

	TEMPLATEFT *FFT;
	MULTITREE *CCT;

	ULONG sum;
	ULONG i,j,k;

	ULONG total;

	ULONG *Pixels;

	FFT = NULL;
	CCT = NULL;
	Pixels = NULL;

	TemplateSize = ( x_size*y_size + (x_size+1)*(y_size-1));

	MinValue = (double)MAXLLONG;


	for(i = 0;i<(unsigned)TemplateSize;i++)
	{

		sum = 0;
		for(j = 0; j<(unsigned)QuantizedNChilds; j++)
			for(k = j*NColors; k<(j+1)*NColors; k++)
				sum += PixelsCounts[i][FactIndex][k];

		if(sum)
			CurrentValue = GetNodeCodeLength(PixelsCounts[i][FactIndex]);
		else
			continue;
		
		if(CurrentValue < MinValue)
		{
			MinPosition = i;
			MinValue = CurrentValue;
		}
	}


	//DEBUG 
//	CloseLogFile();

	RestoreLeafByIndex(CT, FT, RealIndex, &CCT, &FFT);

	for(i = 0; i < (unsigned)QuantizedNChilds; i++)
	{

		sum = 0;
		for(j = i*NColors; j < (i+1)*NColors; j++)
			sum += PixelsCounts[MinPosition][FactIndex][j];
			

		if(sum)
		{
//			if(!CCT->Children)
//				CreateChildren(CCT);
			if(!FFT->Children)
				CreateChildrenFT(FFT);
			
			//CCT->Children[i] = NewTreeNode();
			//CCT->Children[i]->ChildPosition = CCT->ChildPosition + 1;
			pCurChildItem = AddChildListItem(CCT,i,CCT->ChildPosition + 1);
			FFT->Children[i] = NewFTNode();

			total = 0;

			for(j = i*NColors; j < (i+1)*NColors; j++)
			{
				//CCT->Children[i]->Pixels[j - i*NColors] = PixelsCounts[MinPosition][FactIndex][j];
				AddPixelsListItem(pCurChildItem->pNode,j - i*NColors,PixelsCounts[MinPosition][FactIndex][j]);
				total += PixelsCounts[MinPosition][FactIndex][j];
			}

			//CCT->Children[i]->Total = total;
			pCurChildItem->pNode->Total = total;

			Index2Shifts(MinPosition,x_size, y_size, &ShiftX, &ShiftY);

			FFT->ShiftX = ShiftX;
			FFT->ShiftY = ShiftY;
		}
	}

	if(CCT == CT)
	{
		Pixels = (ULONG*)malloc(sizeof(ULONG)*NColors);
		memset(Pixels,0, sizeof(ULONG)*NColors);
		total = 0;
		
		for(j = 0;j<(unsigned)NColors;j++)
		{
			//for(i = 0; i<(unsigned)QuantizedNChilds; i++)
			for(pCurChildItem=CT->pChildren;pCurChildItem;pCurChildItem=pCurChildItem->pNext)
			{
				//if(CCT->Children[i])
					//Pixels[j] += CCT->Children[i]->Pixels[j];
				pCurPixelsItem = FindPixelsItem(pCurChildItem->pNode,j);
				if(pCurPixelsItem)
					Pixels[j] += pCurPixelsItem->Counter;
			}
			total += Pixels[j];
		}

		DeallocatePixelsList(CCT);
		for(j = 0; j<(unsigned)NColors; j++)
			if(Pixels[j])
				AddPixelsListItem(CCT,j,Pixels[j]);//CCT->Pixels[j] = Pixels[j];

		CCT->Total = total;
		free(Pixels);
	}
}

void UpdatePixelsCountsFT(MULTITREE *CT, TEMPLATEFT *FT, IMAGE* GrayImage, int *lut, int x, int y, int StartDepth,
						  int EndDepth, int refPixelN, int TemplateSize, int x_size, int y_size,
						  ULONG range, int step, ULONG ***PixelsCounts)
{
	int       i;
	int       pixelN;
	TEMPLATEFT *FFT;
	MULTITREE *CCT;

	int refx, refy;
	int ShiftX;
	int ShiftY;

	ULONG ChildIndex;
	char *Template;
	ULONG index;

	ChildListItem *pCurItem=0;

	index = 0;
	ChildIndex = 0;

	Template = (char*)malloc(TemplateSize);
	memset(Template, 0, sizeof(char)*TemplateSize);

	refPixelN = lut[refPixelN];
	FFT = FT;
	CCT = CT;

	for(i=1;i<=StartDepth;i++)
	{

		ShiftX = FFT->ShiftX;
		ShiftY = FFT->ShiftY;
		
		refx = x + ShiftX;
		refy = y + ShiftY;
		
		if((refx>0 && (unsigned)refx <= GrayImage->ImageSizeX) && (refy>0 && (unsigned)refy<=GrayImage->ImageSizeY))
			pixelN = lut[GetGrayPixel(GrayImage,refx,refy)];
		else
			pixelN = 0;

		index = Shifts2Index(x_size, y_size, ShiftX, ShiftY);
		Template[index] = 1;

		if(QUANTIZE_CHILDS)
			pixelN = ChildsLut[pixelN];


		if(FFT->Children)
		{
			if(FFT->Children[pixelN])
				FFT = FFT->Children[pixelN];
		}

		if(CCT->pChildren)
		{
			pCurItem = FindChildItem(CCT,pixelN);
			if(pCurItem/*CCT->Children[pixelN]*/)
				CCT = pCurItem->pNode;//CCT->Children[pixelN];
		}
	}

	ChildIndex = CCT->Context;

	if(ChildIndex < (step+1)*range && ChildIndex >= step*range)
	{
		for(i=0;i<TemplateSize;i++)
		{
			if(Template[i])
				continue;

			Index2Shifts(i, x_size, y_size, &ShiftX, &ShiftY);

			refx = x + ShiftX;
			refy = y + ShiftY;
			
			if((refx>0 && (unsigned)refx <= GrayImage->ImageSizeX) && (refy>0 && (unsigned)refy<=GrayImage->ImageSizeY))
				pixelN = lut[GetGrayPixel(GrayImage,refx,refy)];
			else
				pixelN = 0;

			if(QUANTIZE_CHILDS)
				pixelN = ChildsLut[pixelN];

			index = pixelN * NColors + refPixelN;

			PixelsCounts[i][ChildIndex - step*range][index] += 1;
		}
	}

	free(Template);
}


int CheckChilds(TEMPLATEFT *FT)
{
	int i;
	
	if(FT->Children == NULL)
		return 1;
	else
	{

		//if all Childs are leaves then 0. othewise 1
		for(i = 0; i < QuantizedNChilds; i++)
		{
			if(FT->Children[i])
			{
				if(FT->Children[i]->Children)
					return 0;
			}
		}
		return 1;
	}
}

void WriteFT(TEMPLATEFT* FT, MULTITREE *CT, int BitLength, FLOW *f, int x_size, int y_size, int *counter)
{
	int i;
	int Index;

	ChildListItem *pCurItem=0;

	//DEBUG
//	if(CT->Context == 3046 || CT->Context == 3047 || CT->Context == 3048 || CT->Context == 3049)
//		i = i;

	if(CT->ChildPosition <= FreeTreeDepth)
	{
		if(CT->pChildren)
		{
			for(i = 0; i < QuantizedNChilds; i++)
			{
				if(FT->Children)
				{
					pCurItem = FindChildItem(CT,i);
					if(pCurItem /*CT->Children[i]*/ && FT->Children[i])
					{
						WriteFT(FT->Children[i], pCurItem->pNode/*CT->Children[i]*/, BitLength, f, x_size, y_size, counter);
					}
				}
			}
		}

		Index = Shifts2Index(x_size, y_size, FT->ShiftX, FT->ShiftY);
		f->OutputValue(f, Index, BitLength);

		//DEBUG
//		fprintf(flog, "%d %d\n", Index, CT->Context);
		
		(*counter) += BitLength;
	}
}


//the CT must be read first
void ReadFT(TEMPLATEFT *FT, MULTITREE *CT, FLOW *f, int BitLength, int x_size, int y_size)
{
//	int i;
	int Index;
	int ShiftX, ShiftY;
	int flag;

	ChildListItem *pCurItem=0;

	//DEBUG
//	if(CT->Context == 3046 || CT->Context == 3047 || CT->Context == 3048 || CT->Context == 3049)
//		i = i;

	if(CT->ChildPosition > FreeTreeDepth)
		return;

	if(CT->pChildren)
	{
		flag = 0;

		//for(i=0;i<QuantizedNChilds;i++)
		for(pCurItem=CT->pChildren;pCurItem;pCurItem=pCurItem->pNext)
		{
			//if(CT->Children[i])
			{
				if(pCurItem->pNode->pChildren/*CT->Children[i]->Children*/)
				{
					flag = 1;
					break;
				}
			}
		}
		
		if(flag)
		{
			//for(i=0;i<QuantizedNChilds;i++)
			for(pCurItem=CT->pChildren;pCurItem;pCurItem=pCurItem->pNext)
			{
				if(pCurItem->pNode->ChildPosition <= FreeTreeDepth/*CT->Children[i] && CT->Children[i]->ChildPosition <= FreeTreeDepth*/)
				{

					if(FT->Children == NULL)
						CreateChildrenFT(FT);

					FT->Children[pCurItem->PixelID/*i*/] = NewFTNode();
					ReadFT(FT->Children[pCurItem->PixelID/*i*/], pCurItem->pNode/*CT->Children[i]*/,f, BitLength, x_size, y_size);
				}
			}
		}

		Index = f->InputValue(f,BitLength);

		//DEBUG
//		fprintf(flog, "%d %d\n", Index, CT->Context);

		Index2Shifts(Index, x_size, y_size, &ShiftX, &ShiftY);

		FT->ShiftX = ShiftX;
		FT->ShiftY = ShiftY;
	}

}

void DeallocateChildrenFT(TEMPLATEFT *FT)
{
  int i;

  if (FT->Children!=NULL)
  {
    for(i=0;i<QuantizedNChilds;i++)
    {
      if (FT->Children[i]!=NULL)
      {
        DeallocateFT(FT->Children[i]);
      }
    }
	  deallocate(FT->Children);
	  FT->Children = NULL;
  }
}


int PruneFTAccordingCT(TEMPLATEFT *FT, MULTITREE *CT, int FreeTreeDepth)
{
	int i;
	int flag;

	ChildListItem *pCurItem=0;

	//DEBUG
//	if(CT->Context == 3046 || CT->Context == 3047 || CT->Context == 3048 || CT->Context == 3049)
//		i = i;


	if(CT->ChildPosition <= FreeTreeDepth)
	{
		if(CT->pChildren)
		{
			//check: are all children leaves or not
			flag = 1;
			//for(i=0;i<QuantizedNChilds;i++)
			for(pCurItem=CT->pChildren;pCurItem;pCurItem=pCurItem->pNext)
			{
//				if(CT->Children[i])
				{
					if(pCurItem->pNode->pChildren/*CT->Children[i]->Children*/)
					{
						flag = 0;
						break;
					}
				}
			}

			if(flag && FT->Children)
			{
				//for(i=0;i<QuantizedNChilds;i++)
				for(pCurItem=CT->pChildren;pCurItem;pCurItem=pCurItem->pNext)
					if(FT->Children[pCurItem->PixelID/*i*/])
					{
						DeallocateFT(FT->Children[pCurItem->PixelID/*i*/]);
						FT->Children[pCurItem->PixelID/*i*/] = NULL;
					}
			}

			
			//for(i=0;i<QuantizedNChilds;i++)
			for(pCurItem=CT->pChildren;pCurItem;pCurItem=pCurItem->pNext)
			{
				if(/*CT->Children[i] &&*/ FT->Children[pCurItem->PixelID/*i*/])
				{
					if(FT->Children[pCurItem->PixelID/*i*/]->ShiftX == 0 && FT->Children[pCurItem->PixelID/*i*/]->ShiftY == 0)
					{
						DeallocateFT(FT->Children[pCurItem->PixelID/*i*/]);
						FT->Children[pCurItem->PixelID/*i*/] = NULL;
					}
					else if(FT->Children[pCurItem->PixelID/*i*/])
					{
						flag = PruneFTAccordingCT(FT->Children[pCurItem->PixelID/*i*/],pCurItem->pNode/*CT->Children[i]*/, FreeTreeDepth);
						if(flag)
						{
							deallocate(FT->Children[pCurItem->PixelID/*i*/]);
							FT->Children[pCurItem->PixelID/*i*/] = NULL;
						}
					}
				}
				else
				{
					if(FT->Children[pCurItem->PixelID/*i*/])
					{
						DeallocateFT(FT->Children[pCurItem->PixelID/*i*/]);
						FT->Children[pCurItem->PixelID/*i*/] = NULL;
					}
				}
			}
			
			flag = 1;

			for(i=0;i<QuantizedNChilds;i++)
			{
				if(FT->Children[i])
				{
					flag = 0;
					break;
				}
			}

			if(flag)
			{
				deallocate(FT->Children);
				FT->Children = NULL;
			}
		}
		else
		{
			if(FT->Children)
			{
				DeallocateChildrenFT(FT);
				FT->Children = NULL;
				return 1;
			}
		}
	}

	return 0;
}


void GetClosestUnusedTemplatePosition(char *Template, int *ShiftX, int *ShiftY, 
									  int x_size, int y_size, int cur_depth, char* j_used)
{
#ifdef TEMPLATE_2D
    int  tJBIGX[31] = {30,-1,0,-1,+1,-2,0,-2,+2,-1,+1,-2,+2,-3,0,-3,+3,-1,+1,-3,+3,-2,+2,-4, 0,-4,+4,-1,+1,-3,+3 };
    int  tJBIGY[31] = {30,0,-1,-1,-1,0,-2,-1,-1,-2,-2,-2,-2,0,-3,-1,-1,-3,-3,-2,-2,-3,-3, 0,-4,-1,-1,-4,-4,-3,-3 };

//	int  tJBIGX[31] = {30,-1,  0, -2, +1, -1, 0, -3, +2, -2, +1, -1, 0, -4, +3, -3, +2, -2, +1, -1, 0, -5, +4, -4, +3, -3, +2, -2, +1, -1, 0,};
//    int  tJBIGY[31] = {30, 0, -1,  0, -1, -1, -2, 0, -1, -1, -2, -2, -3, 0, -1, -1, -2, -2, -3, -3, -4, 0, -1, -1, -2, -2, -3, -3, -4, -4, -5};

#else 
    int  tJBIGX[31] = {30,-1,-2,-3,-4,-5, -6,-7,-8,-9,-10,-11,-12,-13, -14,-15, -16,-17, -18,-19,-20,-21,-22, -23, -24,-25,-26,-27,-28,-29,-30};
    int  tJBIGY[31] = {30,0,0,0,0,0,00,00,00,00,00,00,00,0,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00 };
#endif
	
	int i, j;
	int TemplateSize;
	int SX;
	int SY;
	int flag;

	TemplateSize = (x_size*y_size)+(x_size + 1)*(y_size -1);

	for(i = 1; i < 31; i++)
	{
		if(j_used[i])
			continue;
		
		flag = 0;

		for(j=0;j<TemplateSize;j++)
		{
			if(Template[j])
			{
				Index2Shifts(j,x_size, y_size, &SX, &SY);
				if(SX == tJBIGX[i] && SY == tJBIGY[i])
				{
					flag = 1;
		//			j_used[i] = 1;
					break;
				}
			}
		}

		if(!flag && !j_used[i])
		{
			(*ShiftX) = tJBIGX[i];
			(*ShiftY) = tJBIGY[i];
			j_used[i] = 1;
			break;
		}
	}
}


void SetRootFT(TEMPLATEFT *FT)
{
	RootFT = FT;
}


MULTITREE *SetContextImageHybrid2(MULTITREE *CT, IMAGE* GrayImage, int *lut, int x, int y, int StartDepth, int EndDepth, int refPixelN, 
								 int x_size, int y_size)
{
  int       i;
  int       pixelN;
  MULTITREE *CCT;
  TEMPLATEFT *FFT;

  int refx, refy;
  

#ifdef TEMPLATE_2D
    int  tJBIGX[31] = {30,-1,0, -1,+1,-2, 0,-2,+2,-1,+1,-2,+2,-3, 0,-3,+3,-1,+1,-3,+3,-2,+2,-4, 0,-4,+4,-1,+1,-3,+3 };
    int  tJBIGY[31] = {30,0, -1,-1,-1, 0,-2,-1,-1,-2,-2,-2,-2, 0,-3,-1,-1,-3,-3,-2,-2,-3,-3, 0,-4,-1,-1,-4,-4,-3,-3 };

//	int  tJBIGX[31] = {30,-1,  0, -2, +1, -1, 0, -3, +2, -2, +1, -1, 0, -4, +3, -3, +2, -2, +1, -1, 0, -5, +4, -4, +3, -3, +2, -2, +1, -1,  0};
//    int  tJBIGY[31] = {30, 0, -1,  0, -1, -1, -2, 0, -1, -1, -2, -2, -3, 0, -1, -1, -2, -2, -3, -3, -4, 0, -1, -1, -2, -2, -3, -3, -4, -4, -5};

#else 
    int  tJBIGX[31] = {30,-1,-2,-3,-4,-5, -6,-7,-8,-9,-10,-11,-12,-13, -14,-15, -16,-17, -18,-19,-20,-21,-22, -23, -24,-25,-26,-27,-28,-29,-30};
    int  tJBIGY[31] = {30,0,0,0,0,0,00,00,00,00,00,00,00,0,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00,00 };
#endif


	char *Template;
	int TemplateSize;
	int ShiftX, ShiftY;
	int index;
	char j_used[31];

	ChildListItem *pCurItem=0;

	FFT = RootFT;
	CCT = Root;

	TemplateSize = (x_size*y_size) + (x_size+1)*(y_size-1);

	Template = malloc(TemplateSize);

	memset(Template,0,TemplateSize);
	memset(j_used, 0, 31);
	refPixelN = lut[refPixelN];

	for(i = 1; i <= StartDepth;i++)
	{
		if(CCT->ChildPosition <= FreeTreeDepth)
		{
			refx = x + FFT->ShiftX;
			refy = y + FFT->ShiftY;
			
			index = Shifts2Index(x_size, y_size, FFT->ShiftX, FFT->ShiftY);

			Template[index] = 1;
			
			if((refx>0 && (unsigned)refx <= GrayImage->ImageSizeX) && (refy>0 && (unsigned)refy<=GrayImage->ImageSizeY))
				pixelN = lut[GetGrayPixel(GrayImage,refx,refy)];
			else
				pixelN = 0;

			if(QUANTIZE_CHILDS)
				pixelN = ChildsLut[pixelN];

			if(CCT->pChildren)
			{
				pCurItem = FindChildItem(CCT,pixelN);
				if(pCurItem/*CCT->Children[pixelN]*/)
					CCT = pCurItem->pNode;//CCT = CCT->Children[pixelN];
				else
					return NULL;
			}
			else
				return NULL;
			
			if(i<FreeTreeDepth && FFT->Children)
				if(FFT->Children[pixelN])
					FFT = FFT->Children[pixelN];
		}
		else
		{
			GetClosestUnusedTemplatePosition(Template, &ShiftX, &ShiftY, x_size, y_size, i, j_used);

			refx = x + ShiftX;
			refy = y + ShiftY;
			
			if((refx>0 && (unsigned)refx <= GrayImage->ImageSizeX) && (refy>0 && (unsigned)refy<=GrayImage->ImageSizeY))
				pixelN = lut[GetGrayPixel(GrayImage, refx, refy)];
			else
				pixelN = 0;

			if(QUANTIZE_CHILDS)
				pixelN = ChildsLut[pixelN];

			if(CCT->pChildren)
			{
				pCurItem = FindChildItem(CCT,pixelN);
				if(pCurItem/*CCT->Children[pixelN]*/)
					CCT = pCurItem->pNode;//CCT = CCT->Children[pixelN];
				else
					return NULL;
			}
			else
				return NULL;
		}
		
		if(!QUANTIZE_CHILDS)
		{
			if(LocalPath[i-1] != pixelN)
			{
				free(Template);
				return NULL;
			}
		}
		else
		{
			if(LocalPath[i-1] != ChildsLut[pixelN])
			{
				free(Template);
				return NULL;
			}
		}
	}

	CCT = CT;

	for (i = StartDepth+1; i <= EndDepth; i++)
	{
		if(CCT->ChildPosition <= FreeTreeDepth)
		{
			refx = x + FFT->ShiftX;
			refy = y + FFT->ShiftY;
			
			index = Shifts2Index(x_size, y_size, FFT->ShiftX, FFT->ShiftY);

			Template[index] = 1;
			
			if((refx>0 && (unsigned)refx <= GrayImage->ImageSizeX) && (refy>0 && (unsigned)refy<=GrayImage->ImageSizeY))
				pixelN = lut[GetGrayPixel(GrayImage,refx,refy)];
			else
				pixelN = 0;

			if(QUANTIZE_CHILDS)
				pixelN = ChildsLut[pixelN];

			if(i<FreeTreeDepth && FFT->Children)
				if(FFT->Children[pixelN])
					FFT = FFT->Children[pixelN];
		}
		else
		{
			GetClosestUnusedTemplatePosition(Template, &ShiftX, &ShiftY, x_size, y_size, i, j_used);

			refx = x + ShiftX;
			refy = y + ShiftY;
			
			if((refx>0 && (unsigned)refx <= GrayImage->ImageSizeX) && (refy>0 && (unsigned)refy<=GrayImage->ImageSizeY))
				pixelN = lut[GetGrayPixel(GrayImage, refx, refy)];
			else
				pixelN = 0;

			if(QUANTIZE_CHILDS)
				pixelN = ChildsLut[pixelN];
		}

		if(i > FreeTreeDepth+1)
		{
			IncreasePixelsListItem(CT,refPixelN);//CCT->Pixels[refPixelN]++;
			CCT->Total += 1;
			CCT->ChildPosition = i;
		}

//		if (CCT->Children==NULL) CreateChildren(CCT);

		pCurItem = FindChildItem(CCT,pixelN);
 
		if (pCurItem/*CCT->Children[pixelN]*/==NULL)
		{
			AddChildListItem(CCT,pixelN,i+1);

		  //CCT->Children[pixelN] = NewTreeNode();
		  //CCT->Children[pixelN]->ChildPosition = i+1;

		  //assert(CCT->Children[pixelN]!=NULL);
		}
		//CCT=CCT->Children[pixelN];
		CCT = pCurItem->pNode;
	}
	
	if(i > FreeTreeDepth+1)
	{
		IncreasePixelsListItem(CCT,refPixelN);//CCT->Pixels[refPixelN]++;
		CCT->Total += 1;
		CCT->ChildPosition = i;
	}

	free(Template);

  return (CCT);
}



void ConstructTreeImageFTreeHybrid2(MULTITREE *CT, IMAGE *GrayImage, int *lut,int StartDepth, int EndDepth, int x_size, int y_size)
{
	MULTITREE *CCT;
	ULONG     x,y;
	int       pixelN;

	if(QuietLevel>2)
	{
		SetProgressVisibility(1);
		OpenProgressWindow();
	}

	ImageRewind(GrayImage);

	for(y=1;y<=GrayImage->ImageSizeY;y++)
	{

		for(x = 1; x <= GrayImage->ImageSizeX; x++)
		{
			pixelN = GetGrayPixel(GrayImage,x,y);
			CCT = SetContextImageHybrid2(CT, GrayImage,lut,x,y, StartDepth, EndDepth, pixelN, x_size, y_size);

			if(CCT)
				if(!CCT->Total)
					x = x;
		}
		
		if(QuietLevel>2) 
			UpdateProgressBar(y, GrayImage->ImageSizeY);
	}

	if(QuietLevel>2) CloseProgressWindow();
		ImageRewind(GrayImage);
}
