#if !defined(__CT_H)
#define __CT_H

#include "..\Modules\owntypes.h"
#include "..\Modules\image.h"
#include "..\Modules\genio.h"
#include "..\Modules\cb.h"
#include "..\Modules\mapseg.h"

struct idChildListItem
{
	struct idMULTITREE *pNode;
	int PixelID;
	//ULONG Counter;
	struct idChildListItem *pNext;
};

struct idPixelsListItem
{
	int PixelID;
	ULONG Counter;
	struct idPixelsListItem *pNext;
};

struct idMULTITREE
{
	struct idPixelsListItem *pMostProbablePixelsItem;
//  struct idMULTITREE  **Children;
//  ULONG               *Pixels;
  struct idChildListItem *pChildren;
  struct idPixelsListItem *pPixels;
  int                 ChildPosition;
  ULONG               Total;
  ULONG               Context;
};

struct idTemplateFT
{
	struct idTemplateFT  **Children;
	char ShiftX;
	char ShiftY;
};

struct idClusterProperties
{
	int SizeX;
	int SizeY;
};


typedef struct idMULTITREE MULTITREE;
typedef struct idChildListItem ChildListItem;
typedef struct idPixelsListItem PixelsListItem;
typedef struct idTemplateFT TEMPLATEFT;
typedef struct idClusterProperties CLPR;


#define NUMBER_M 0
#define CONST 0.00000
#define SEMI 0
#define FULL_SEARCH 0
#define TREE_COMPLETE 0

#define TEMPLATE_2D 1

#define QUANTIZE_CHILDS 0
#define FREE_TREE 0


#define LINE_SIZE_X 0
#define LINE_SIZE_Y 0

//========= Child List Routines ======================
ChildListItem *FindChildItem(MULTITREE *CT, int PixelID);
PixelsListItem *FindPixelsItem(MULTITREE *CT, int PixelID);
ChildListItem* AddChildListItem(MULTITREE *CT, int PixelID, int Position);
void RemoveChildItem(MULTITREE *CT, int PixelID);
PixelsListItem* AddPixelsListItem(MULTITREE *CT, int PixelID, int Counter);
void DeallocateChildList(MULTITREE *CT);
void DeallocatePixelsList(MULTITREE *CT);
//====================================================


MULTITREE *NewTreeNode();
MULTITREE *GetVarContext(MULTITREE *CT, BYTE *Data, int x);
void ConstructTree(MULTITREE *CT, BYTE* Data, int DataLength, int depth);
MULTITREE *SetContext(MULTITREE *CT, BYTE* Data, int x, int depth, int refPixelN);
void PrintVarContext(MULTITREE *CT, BYTE *Data, int x);
void ReadMTree(MULTITREE *MT, FLOW *inFlow, ULONG* NModels);
void WriteMTree(FLOW *outFlow, MULTITREE *CT);
//void PruneTree(MULTITREE *CT, int *StartDepth, double *Entropy);
void SetNumbers(MULTITREE *CT, int *ContextNumber);
void FillRoot(MULTITREE *CT);
void PrintTree(MULTITREE *CT);
void DeallocateTree(MULTITREE *CT);
double GetEntropySemi(ULONG *Pixels, double total);


void SetNColors(int N);
void SetDelta(double N);
void SetRoot(MULTITREE* r);
void SetQuantizedNChilds(int N, int *CL);
void SetFreeTreeParam(int FTD, int BL);
double GetEntropyAdapt(ULONG *Pixels, double total);

void WriteMT(MULTITREE *CT, FLOW *flow, int *counter);
void ReadMT(MULTITREE *CT, FLOW *flow, int *counter);
void SetCoeff(double N);

double PruneTreeOptimal(MULTITREE *CT);
void CheckChildren(MULTITREE *CT);
void CheckTree(MULTITREE *CT);
double GetCodeSizeAdapt(MULTITREE *CT);
double GetCodeSizeSemi(MULTITREE *CT);

void ConstructTreeImage(MULTITREE *CT, IMAGE *GrayImage, int *lut, int depth, int index);

MULTITREE *SetContextImage(MULTITREE *CT, IMAGE* GrayImage, int *lut, 
						   int x, int y, int depth, int refPixelN, int StartX, int StartY, int EndX, int EndY);


void OpenLogFile();
void CloseLogFile();

MULTITREE *SetContextImage2(MULTITREE *CT, IMAGE* GrayImage, int *lut, int x, int y, 
							int StartDepth, int EndDepth, int refPixelN, int StartX, int StartY, int EndX, int EndY);
void ConstructTreeImage2(MULTITREE *CT, IMAGE *GrayImage, int *lut, int StartDepth, int EndDepth, int index);



double PruneTreeOptimalWithConstructing(MULTITREE *CT, IMAGE *GrayImage, int *lut, int StartDepth, int EndDepth, int TemplateX, int TemplateY);

void ZeroLut(MULTITREE* CT);
double GetFromLut(BYTE *Config, MULTITREE* CT);
void Put2Lut(BYTE* Config, double Val, MULTITREE* CT);
int GetPaletteNumber(MULTITREE* CT);


void AddOnSubtree(MULTITREE* CT, ULONG* counter, int StartDepth);
void ReconstructTree(int StartDepth);
void StoreSubtree(MULTITREE *CT);
void InitTmpTrees(int StartDepth);

void SetAlreadyBuiltDepth(int num1, int num2);
void ConstructTreeImage3(MULTITREE *CT, IMAGE *GrayImage, int *lut, int StartDepth, int EndDepth, int index);


/*REDUCED TREE*/

int GetReducedNChilds(MULTITREE* CT);

/*
int GetReducedNChilds(MULTITREE_REDUCED* CT);
MULTITREE_REDUCED *SetContextReduced(MULTITREE_REDUCED *CT, BYTE* Data, int x, int depth, int refPixelN);
void CreateChildrenReduced(MULTITREE_REDUCED *CT, int NumberOfChildren);
MULTITREE_REDUCED *NewTreeNodeReduced(int LocalNColors, int LocalNChilds);
*/

/*FREE TREE*/
TEMPLATEFT *NewFTNode();
void DeallocateFT(TEMPLATEFT *FT);
void CreateChildrenFT(TEMPLATEFT *FT);
void DefineRange(ULONG ChildsNumber, int x_size, int y_size, int *range);
int Shifts2Index(int x_size, int y_size, int ShiftX, int ShiftY);
void Index2Shifts(int index, int x_size, int y_size, int *ShiftX, int *ShiftY);
void InitLeavesIndexes(MULTITREE *CT, ULONG* index);
void DefaultLeavesIndexes(MULTITREE *CT);
void RestoreLeafByIndex(MULTITREE *CT, TEMPLATEFT *FT, ULONG SearchIndex, MULTITREE **resCT, TEMPLATEFT **resFT);
void ConstructTreeImageFTree(MULTITREE *CT, TEMPLATEFT *FT, IMAGE *GrayImage, int *lut, int depth, int x_size, int y_size);
double GetNodeCodeLength(ULONG *PixelsCounts);
void FindOptimalTemplatePosition(ULONG RealIndex, ULONG FactIndex,  ULONG ***PixelsCounts, MULTITREE *CT, TEMPLATEFT *FT, int x_size, int y_size);
void UpdatePixelsCountsFT(MULTITREE *CT, TEMPLATEFT *FT, IMAGE* GrayImage, int *lut, int x, int y, int StartDepth,
						  int EndDepth, int refPixelN, int TemplateSize, int x_size, int y_size,
						  ULONG range, int step, ULONG ***PixelsCounts);


void WriteFT(TEMPLATEFT* FT, MULTITREE *CT, int BitLength, FLOW *f, int x_size, int y_size, int *counter);
void ReadFT(TEMPLATEFT *FT, MULTITREE *CT, FLOW *f, int BitLength, int x_size, int y_size);

void DeallocateChildrenFT(TEMPLATEFT *FT);
int PruneFTAccordingCT(TEMPLATEFT *FT, MULTITREE *CT, int FreeTreeDepth);

void GetClosestUnusedTemplatePosition(char *Template, int *ShiftX, int *ShiftY, 
									  int x_size, int y_size, int cur_depth, char* j_used);

MULTITREE *SetContextImageHybrid(MULTITREE *CT, IMAGE* GrayImage, int *lut, int x, int y, int EndDepth, int refPixelN, 
								 TEMPLATEFT *FT, int x_size, int y_size);
void ConstructTreeImageFTreeHybrid(MULTITREE *CT, TEMPLATEFT *FT, IMAGE *GrayImage, int *lut, int EndDepth, int x_size, int y_size);

void SetRootFT(TEMPLATEFT *FT);
MULTITREE *SetContextImageHybrid2(MULTITREE *CT, IMAGE* GrayImage, int *lut, int x, int y, int StartDepth, int EndDepth, int refPixelN, 
								 int x_size, int y_size);
void ConstructTreeImageFTreeHybrid2(MULTITREE *CT, IMAGE *GrayImage, int *lut,int StartDepth, int EndDepth, int x_size, int y_size);
void CheckTheTree(MULTITREE *CT);



//independent encoding
void GetClusterBounds(IMAGE *GrayImage, int index, int *StartX, int *StartY, int *EndX, int *EndY);
int GetNumberOfClustersMT(IMAGE *GrayImage);
void SetClusterSizes(int SizeX, int SizeY);
#endif /* __ACFNEW_H */
