/*-------------------------------------------------------------------*/
/* GENCODER.C      Pavel Kopyloff                                    */
/*                 Eugene Ageenko                                    */
/*                                                                   */
/*                                                                   */
/* Generic coder.                                                    */
/* Supports: QM, MQ, Range coders.                                  */
/*                                                                   */
/*-------------------------------------------------------------------*/

#define  ProgName       "GENCODER"
#define  VersionNumber  "Version L.02"
#define  LastUpdated    "21.09.2000"

#define IDString        "GENCODER02"

#include "gencoder.h"
#include "mqcoder.h"
#include "qmcoder.h"
#include "rngcoder.h"
#include "../Modules/memctrl.h"

typedef enum {MQ=0, QM=1, RANGE=2} CODERTYPE;

/*======================= Preparing part ============================*/

YESNO GENPrepareCoder(GENCODER *Coder, int EncoderType, int Action)
{
  Coder->Coder = NULL;

  switch (EncoderType)
  {
  case MQ:
#ifdef ENTROPY
    Coder->Encode      = (TEncode)MQEncodeStats;//(MQENC *, int , void*, double*);
#else 
    Coder->Encode      = (TEncode)MQEncode;//(MQENC *, int , void*);
#endif
    Coder->Decode      = (TDecode)MQDecode;//(MQDEC *, void*);

    switch (Action)
    {
    case ENCODING:
      Coder->InitCoder = (TInitCoder)MQInitEncoder;//(MQENC *, FILE *);
      Coder->Flush     = (TFlush)MQFlushEncoder;//(MQENC*);
      if((Coder->Coder = (MQENC*) allocate(sizeof(MQENC)))==NULL) return(NO);
      break;

    case DECODING:
      Coder->InitCoder = (TInitCoder)MQInitDecoder;//(MQDEC *, FILE *);
      Coder->Flush     = (TFlush)MQFlushDecoder;//(MQDEC*);
      if((Coder->Coder = (MQDEC*) allocate(sizeof(MQDEC)))==NULL) return(NO);
      break;
    }
    break;

  case QM:
    Coder->Encode      = (TEncode)QMEncode;//(QMENC *, int , void*);
    Coder->Decode      = (TDecode)QMDecode;//(QMDEC *, void*);
    Coder->Flush       = (TFlush)QMFlush;//(QMENC*);

    switch (Action)
    {
    case ENCODING:
      Coder->InitCoder = (TInitCoder)QMInitEncoder;//(QMENC *, FILE *);
      if((Coder->Coder = (QMENC*) allocate(sizeof(QMENC)))==NULL) return(NO);
      break;

    case DECODING:
      Coder->InitCoder = (TInitCoder)QMInitDecoder;//(QMDEC *, FILE *);
      if((Coder->Coder = (QMDEC*) allocate(sizeof(QMDEC)))==NULL) return(NO);
      break;
    }
    break;

  case RANGE:
    Coder->Encode      = (TEncode)RCEncode;//(QMENC *, int , void*);
    Coder->Decode      = (TDecode)RCDecode;//(QMDEC *, void*);

    switch (Action)
    {
    case ENCODING:
      Coder->InitCoder = (TInitCoder)RCInitEncoder;//(QMENC *, FILE *);
      Coder->Flush     = (TFlush)RCFlushEncoder;//(QMENC*);
      if((Coder->Coder = (RCENC*) allocate(sizeof(RCENC)))==NULL) return(NO);
      break;

    case DECODING:
      Coder->InitCoder = (TInitCoder)RCInitDecoder;//(QMDEC *, FILE *);
      Coder->Flush     = (TFlush)RCFlushDecoder;//(QMENC*);
      if((Coder->Coder = (RCENC*) allocate(sizeof(RCENC)))==NULL) return(NO);
      break;
    }
    break;

    default: return(NO);
  }
  return (YES);
}

/*======================= Deinitializing ============================*/

void GENDoneCoder(GENCODER *Coder)
{
  deallocate(Coder->Coder); 
  Coder->Coder = NULL;
}
