/*-------------------------------------------------------------------*/
/* GENMODEL.C      Pavel Kopylov   (justas@cs.joensuu.fi)            */
/*                                                                   */
/*                                                                   */
/* Generic modelling module                                          */
/*-------------------------------------------------------------------*/

#define  ProgName       "GENMODEL"
#define  VersionNumber  "Version L.02"
#define  LastUpdated    "21.09.2000"


#include <stdio.h>
#include "genmodel.h"
#include "memory.h"   // for memset
#include "qmodel.h"
#include "qsmodel.h"

typedef enum {MQ=0, QM=1, RANGE=2} CODERTYPE;

// ------------------- Forward-adaptive modelling ------------------

static double QMFastAttack[13] = {
  .30891, .14590, .06891, .03255, .01537, .00726, .00343, .00162, .00076,
  .00036, .00017, .00008, .00004};

static int QMFastAttackStateNumber[14] = {
  0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13};

//Linear Interpolation
static double FirstAttackStateBounds[13] = {
  .391863705, .207022136, .094141584, .043516659, .019361847, .008828931,
  .004361417, .002255332, .001189679, .000638284, .000314779, .000153245,
 5.12055E-05};


//Cubic spline method
static double MQFastAttack[13] = {
  .411570397, .210875253, .093169734, .043988716, .0192818,   .008717996,
  .004338346, .002254334, .00118039,  .00065058,  .000313119, .000155059,
  6.51268E-05};

static int MQFastAttackStateNumber[14] = {
  0, 1, 2, 3, 4, 5, 38, 39, 40, 41, 42, 43, 44, 45};

/*======================= Modelling part ============================*/
    
YESNO GENPrepareModel(GENMODEL *pGModel, int EncoderType)
{
  pGModel->Model       = NULL;
  switch (EncoderType)
  {
    case MQ:
      pGModel->InitModel   = (TInitModel)InitQModel;
      pGModel->ReInitModel = (TReInitModel)ReInitQModel;
      pGModel->DoneModel   = (TDoneModel)DoneQModel;
      pGModel->CreateInitialModel = (TCreateInitialModel)CreateInitialQModel;
      pGModel->StoreModel  = (TStoreModel) StoreInitialQModel;
      pGModel->LoadModel   = (TLoadModel) LoadInitialQModel;

      pGModel->FastAttack  = MQFastAttack;
      pGModel->FastAttackStateNumber  = MQFastAttackStateNumber;
      pGModel->Size        = sizeof(QMODEL);
      break;

    case QM:
      pGModel->InitModel   = (TInitModel)InitQModel;
      pGModel->ReInitModel = (TReInitModel)ReInitQModel;
      pGModel->DoneModel   = (TDoneModel)DoneQModel;
      pGModel->CreateInitialModel = (TCreateInitialModel)CreateInitialQModel;
      pGModel->StoreModel  = (TStoreModel) StoreInitialQModel;
      pGModel->LoadModel   = (TLoadModel) LoadInitialQModel;

      pGModel->FastAttack  = QMFastAttack;
      pGModel->FastAttackStateNumber  = QMFastAttackStateNumber;
      pGModel->Size        = sizeof(QMODEL);
      break;
    case RANGE:
      pGModel->InitModel   = (TInitModel)InitQSModel;
      pGModel->ReInitModel = (TReInitModel)ReInitQSModel;
      pGModel->DoneModel   = (TDoneModel)DoneQSModel;
      pGModel->CreateInitialModel = (TCreateInitialModel)CreateInitialQSModel;
      pGModel->StoreModel  = (TStoreModel) StoreInitialQSModel;
      pGModel->LoadModel   = (TLoadModel) LoadInitialQSModel;

      pGModel->FastAttack  = NULL;
      pGModel->Size        = sizeof(QSMODEL);
      break;

    default :
      return (NO);
  }
  return (YES);
}    
