/*-------------------------------------------------------------------*/
/* Project parser  Pavel Kopylov                                     */
/*       
/*Last update */                                                           
/*-------------------------------------------------------------------*/

#define ProgName        "Project"
#define VersionNumber   "Version 0.05"
#define LastUpdated     "09.10.2001"


#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "project.h"
#include "memctrl.h"
#include "file.h"
#include "interfc.h"
#include "color.h"

void trim(char*);
YESNO GetStringValue(char*, char*, char*);
YESNO GetIntValue(char*, int*, char*);
YESNO GetFloatValue(char*, double*, char*);


/*==== Project functions ============================================*/
void  CreateEmptyProject(PROJECTDATA *Project)
{
  Project->Layers=0;
  Project->Models =0;
  Project->Trees=0;
  Project->Colors=0;
  Project->ProjectParams=NULL;
  Project->Model=NULL;
  Project->Tree=NULL;
  Project->Color=NULL;
  Project->BackGround=0;
  Project->XSize=0;
  Project->YSize=0;
  Project->LeftBottomX=0;
  Project->LeftBottomY=0;
  Project->RightTopX=0;
  Project->RightTopY=0;
  Project->Scale=0;
  Project->Rotation=0;
  Project->InputName[0]=0;
  Project->OutputName[0]=0;
  Project->ProjectName[0]=0;
}


/*-------------------------------------------------------------------*/
YESNO ReadProject(PROJECTDATA *Project)
{
  FILE            *f;
  UINT32          k, i, j, m, n, o;
  char            line[127], parameters[127], *tmpStr;
  int             r,g,b;
  
  f = fopen(Project->ProjectName, "rt");
  if (f == NULL) return NO;


  Project->Layers = 0;  Project->Trees = 0;  Project->Models = 0; Project->Colors=0;
  while(!feof(f))
  {
    fscanf(f, " #%*[^\n]\n");                                     // comment
    k=fscanf(f, "%[^#\n]\n", line);
    if (k>0 && k!=EOF)
    {
      trim(line);
      tmpStr = strlwr(line);
      
      if (GetStringValue(tmpStr, Project->InputName, "input=") == YES);        // Input name
      else
      if (GetStringValue(tmpStr, Project->OutputName, "output=") == YES);      // Output name
      else
      if (GetIntValue(tmpStr, &(Project->BackGround), "background=") == YES);   // Background color
      else
      if (GetIntValue(tmpStr, &(Project->XSize), "xsize=") == YES);              // X - size
      else
      if (GetIntValue(tmpStr, &(Project->YSize), "ysize=") == YES);              // Y - size
      else
      if (GetFloatValue(tmpStr, &(Project->LeftBottomX), "leftbottomx=") == YES);// Map corners
      else
      if (GetFloatValue(tmpStr, &(Project->LeftBottomY), "leftbottomy=") == YES);
      else
      if (GetFloatValue(tmpStr, &(Project->RightTopX), "righttopx=") == YES);
      else
      if (GetFloatValue(tmpStr, &(Project->RightTopY), "righttopy=") == YES);
      else
      if (GetFloatValue(tmpStr, &(Project->Scale), "scale=") == YES);            // Meters per pixel
      else
      if (GetFloatValue(tmpStr, &(Project->Rotation), "rotation=") == YES);      // Rotation
      else
        if (strstr(tmpStr,"layer")!=NULL)
        {
          strcpy(tmpStr, (tmpStr+5));
          j = sscanf(tmpStr, "%d", &i);
          if (j>0) Project->Layers++;
        }
      else
        if (strstr(tmpStr,"tree")!=NULL)
        {
          strcpy(tmpStr, (tmpStr+4));
          j = sscanf(tmpStr, "%d", &i);
          if (j>0) Project->Trees++;
        }
      else
        if (strstr(tmpStr,"model")!=NULL)
        {
          strcpy(tmpStr, (tmpStr+5));
          j = sscanf(tmpStr, "%d", &i);
          if (j>0) Project->Models++;
        }
      else
        if (strstr(tmpStr,"color")!=NULL) /*old variant of the code*/
        {
          strcpy(tmpStr, (tmpStr+5));
          j = sscanf(tmpStr, "%d", &i);
          if (j>0) Project->Colors++;
        }
    } 
  }                    

  if (Project->Layers==0) return NO;

  //Allocate memory
  Project->ProjectParams = NULL;
  if (Project->Layers>0)
  {
    Project->ProjectParams = (PROJECTPARAMETERS *)allocate(sizeof(PROJECTPARAMETERS)*(Project->Layers));
    if (Project->ProjectParams == NULL) return NO;
  }

  Project->Tree = NULL;
  if (Project->Trees>0)
  {
    Project->Tree = (TREEPARAMETERS *)allocate(sizeof(TREEPARAMETERS)*(Project->Trees));
    if (Project->Tree == NULL) return NO;
  }

  Project->Model = NULL;
  if (Project->Models>0)
  {
    Project->Model = (MODELPARAMETERS *)allocate(sizeof(MODELPARAMETERS)*(Project->Models));
    if (Project->Model == NULL) return NO;
  }

  Project->Color = NULL;
  if (Project->Colors>0)
  {
    Project->Color = (COLORPARAMETERS *)allocate(sizeof(COLORPARAMETERS)*(Project->Colors));
    if (Project->Color == NULL) return NO;
  }

  //Parsing
  k=0; j=0; m=0; n=0; o=0;
  rewind(f);
  while(!feof(f))
  {
    fscanf(f, " #%*[^\n]\n"); // comment
    k=fscanf(f, "%[^#\n]\n", line);
    if (k>0 && k!=EOF)
    {
      tmpStr = strlwr(line);
      i=0;                                 
      if (Project->Layers>0)
      if (strstr(tmpStr,"layer")!=NULL)
      {
        strcpy(tmpStr, (tmpStr+5));
        k=sscanf(tmpStr, "%d", &i);
        if (k>0)
        {
          tmpStr = strchr(tmpStr, '=')+1;
          Project->ProjectParams[j].LayerIndex = i;

          k=sscanf(tmpStr, "%s %[^#\n]\n", Project->ProjectParams[j].LayerFileName, parameters);
          if (k<1) return(NO);           // no name ???
          trim(Project->ProjectParams[j].LayerFileName);

          sscanf(parameters, "%d , %d", &(Project->ProjectParams[j].LayerType), 
                                        &(Project->ProjectParams[j].LayerDependency));
          j++;
        }
      }
      if (Project->Trees>0)
      if (strstr(tmpStr,"tree")!=NULL)
      {
        strcpy(tmpStr, (tmpStr+4));
        k=sscanf(tmpStr, "%d", &i);
        if (k>0)
        {
          tmpStr = strchr(tmpStr, '=')+1;
          Project->Tree[m].LayerIndex = i;

          k=sscanf(tmpStr, "%s %[^#\n]\n", Project->Tree[m].LayerFileName, parameters);
          if (k<1) return(NO);           // no name ???
          m++;
        }
      }
      if (Project->Models>0)
      if (strstr(tmpStr,"model")!=NULL)
      {
        strcpy(tmpStr, (tmpStr+5));
        k=sscanf(tmpStr, "%d", &i);
        if (k>0)
        {
          tmpStr = strchr(tmpStr, '=')+1;
          Project->Model[n].LayerIndex = i;
          n++;
        }
      }
    if (Project->Colors>0)
      if (strstr(tmpStr,"color")!=NULL)
      {
        strcpy(tmpStr, (tmpStr+5));
        k=sscanf(tmpStr, "%d", &i);
        if (k>0)
        {
          tmpStr = strchr(tmpStr, '=')+1;
          Project->Color[o].LayerIndex = i;
          k=sscanf(tmpStr, "%d, %d, %d", &r, &g, &b);
          Project->Color[o].Color = RGB(r,g,b);
          if (k<1) return(NO);           // no name ???
          o++;
        }
      }
    }
  }

  fclose(f);
  return YES;
}


/*-------------------------------------------------------------------*/
YESNO SaveProject(PROJECTDATA *Project)
{
  FILE      *f;
  UINT       i;
  BYTE		r,g,b;


  if (Project->ProjectName[0]==0) PickFileName(Project->OutputName, Project->ProjectName);
  CheckFileName(Project->ProjectName, FormatNamePRJ);

  f = fopen(Project->ProjectName, "wt");
  if (f == NULL) return NO;

  fprintf(f, "# Map project\n\n");
  fprintf(f, "Input=%s\n", Project->InputName);
  fprintf(f, "Output=%s\n", Project->OutputName);
  fprintf(f, "Background=%d\n", Project->BackGround);
  fprintf(f, "XSize=%d\n", Project->XSize);
  fprintf(f, "YSize=%d\n", Project->YSize);
  fprintf(f, "LeftBottomX=%lf\n", Project->LeftBottomX);
  fprintf(f, "LeftBottomY=%lf\n", Project->LeftBottomY);
  fprintf(f, "RightTopX=%lf\n", Project->RightTopX);
  fprintf(f, "RightTopY=%lf\n", Project->RightTopY);
  fprintf(f, "Scale=%f\n", Project->Scale);
  fprintf(f, "Rotation=%lf\n\n", Project->Rotation);

  if (Project->ProjectParams != NULL)
  {
    fprintf(f, "# Layer section\n");
    fprintf(f, "# Layers:\n# LayerNumber, LayerType, DependentFrom, Number of colours per layert\n\n");
    fprintf(f, "# SubLayer, Colors\n");
    for(i=0; i<Project->Layers; i++)
    {
      if (Project->ProjectParams[i].LayerFileName[0]!=0)
      {
        fprintf(f, "Layer%d=%s, %d, %d\n", Project->ProjectParams[i].LayerIndex, 
                                           Project->ProjectParams[i].LayerFileName,
                                           Project->ProjectParams[i].LayerType,
                                            Project->ProjectParams[i].LayerDependency);
												 //Project->ColorInfo[i].NumberOfColors);
/*			  fprintf(f, "Layer%d, %d, %d, %d\n", Project->ProjectParams[i].LayerIndex,
							  Project->ProjectParams[i].LayerType,
                                                        Project->ProjectParams[i].LayerDependency);
//												  Project->ColorInfo[i].NumberOfColors);*/

/*			  for(j=0;j<Project->ColorInfo[i].NumberOfColors;j++)  
				  fprintf(f, "SubLayer_%d_%d, %d %d %d\n", Project->ProjectParams[i].LayerIndex,
														 j, 
														 (BYTE)MISSRED(Project->ColorInfo[i].Colors[j]),
														 (BYTE)MISSGREEN(Project->ColorInfo[i].Colors[j]),
														 (BYTE)MISSBLUE(Project->ColorInfo[i].Colors[j]));*/
		  }
	  }
  }

  if (Project->Tree != NULL)
  {
	  fprintf(f, "\n# Trees section\n");
	  fprintf(f, "# TreeNumber=TreeName\n");
	  for(i=0; i<Project->Trees; i++)
	  {
		fprintf(f, "Tree%d=%s\n", Project->Tree[i].LayerIndex, 
								  Project->Tree[i].LayerFileName);
	  }
  }

  if (Project->Model != NULL)
  {
	  fprintf(f, "\n# Models section\n");
	  fprintf(f, "# ModelNumber\n");
	  for(i=0; i<Project->Models; i++)
	  {
		fprintf(f, "Model%d\n", Project->Model[i].LayerIndex);
	  }
  }

  if (Project->Color != NULL)
  {
	  fprintf(f, "\n# Color section\n");
	  fprintf(f, "# LayerNumber=Color\n");
	  for(i=0; i<Project->Layers; i++)
	  {
        r = (BYTE)RED(Project->Color[i].Color);
        g = (BYTE)GREEN(Project->Color[i].Color);
        b = (BYTE)BLUE(Project->Color[i].Color);
		fprintf(f, "Color%d=%d,%d,%d\n", Project->Color[i].LayerIndex,r,g,b);
	  }
  }

  fprintf(f, "\n# End");

  fclose(f);
  return YES;
}


/*-------------------------------------------------------------------*/
YESNO DoneProject(PROJECTDATA *Project)
{
  if (Project->ProjectParams != NULL) deallocate(Project->ProjectParams);
  if (Project->Tree          != NULL) deallocate(Project->Tree);
  if (Project->Model         != NULL) deallocate(Project->Model);
  if (Project->Color         != NULL) deallocate(Project->Color);
  
  return YES;
}

/*==== List functions ===============================================*/
YESNO ReadList(char *InputName, int *Size, int **List)
{
  int             i, k;
  FILE            *f;
  char            line[1024], *tmpStr;
  int             *tmpList;

  f = fopen(InputName, "rt");
  if (f == NULL) return NO;

  //Get the SIZE
  while(!feof(f))
  {
    fscanf(f, " #%*[^\n]\n");                           // Comment
    k=fscanf(f, "%[^#\n]\n", line);
    if (k>0 && k!=EOF)
    {
      trim(line);
      tmpStr = strlwr(line);
      
      if (strstr(tmpStr, "size="))                      // Array Size
      {
        tmpStr  = strrchr(tmpStr, '=');  tmpStr++;
        if (tmpStr[0]!=0) sscanf(tmpStr, "%d", Size);
        else *Size = 0;
      }
    }
  }

  tmpList = (int*)allocate(sizeof(int)*((*Size)*((*Size)+1)));   // "Size" beginning values,
  if (tmpList==NULL) return NO;                                  // and Size*Size array values

  rewind(f);
  i=0;
  while(!feof(f))
  {
    fscanf(f, " #%*[^\n]\n");                                 // Comment
    fscanf(f, "%[^#\n]\n", line);
    if( sscanf(line, "%d", &k) > 0)
    {
      tmpList[i] = k;
      i++;
    }
  }
  fclose(f);


  (*List) = tmpList;
  return YES;
}


/*-------------------------------------------------------------------*/
YESNO DoneList(int *List)
{
  if (List != NULL) deallocate(List);

  return YES;
}


/*-------------------------------------------------------------------*/
void PrintProject(PROJECTDATA *Project, int QuietLevel)
{
  UINT  i;


  if (QuietLevel >= 1)
  {
    PrintMessage("Input file:       %s\n", Project->InputName);
    PrintMessage("Output file:      %s\n", Project->OutputName);
    PrintMessage("Background color: %d\n", Project->BackGround);

    PrintMessage("Layers:\n");
    for (i=0;i<Project->Layers; i++) 
      PrintMessage("  %d=%s, %d, %d\n", Project->ProjectParams[i].LayerIndex, 
                                        Project->ProjectParams[i].LayerFileName,
                                        Project->ProjectParams[i].LayerType,
                                        Project->ProjectParams[i].LayerDependency);
    PrintMessage("\nTrees:\n");
    for (i=0;i<Project->Trees; i++) 
      PrintMessage("  %d=%s\n", Project->Tree[i].LayerIndex, 
                                Project->Tree[i].LayerFileName);

    PrintMessage("\nModels:\n");
    for (i=0;i<Project->Models; i++) 
      PrintMessage("  %d=%s\n", Project->Model[i].LayerIndex);
  }
}



/*-------------------------------------------------------------------*/
YESNO DeleteItem(PROJECTDATA* Project, ULONG MaxNode)
{
  PROJECTPARAMETERS *TmpParams;
  TREEPARAMETERS    *TmpTrees;
  MODELPARAMETERS   *TmpModels;
  COLORPARAMETERS   *TmpColors;
  ULONG             Params, Trees, Models, Colors;
  ULONG             i,j;

  

  Params = Trees = Models = Colors = 0;
  for(i=0;i<Project->Layers;i++)
  {
    if(Project->ProjectParams[i].LayerIndex != MaxNode) Params++;
  }

  for(i=0;i<Project->Trees;i++)
  {
    if(Project->Tree[i].LayerIndex != MaxNode) Trees++;
  }

  for(i=0;i<Project->Models;i++)
  {
    if(Project->Model[i].LayerIndex != MaxNode) Models++;
  }

  for(i=0;i<Project->Colors;i++)
  {
    if(Project->Color[i].LayerIndex != MaxNode) Colors++;
  }

  //Allocate memory for the NEW project structure
  TmpParams = NULL;
  if (Params>0)
  TmpParams = (PROJECTPARAMETERS*)allocate(sizeof(PROJECTPARAMETERS)*Params);
  if (TmpParams == NULL) return NO;

  TmpTrees = NULL;
  if (Trees>0)
  TmpTrees = (TREEPARAMETERS*)allocate(sizeof(TREEPARAMETERS)*Trees);
  if (TmpTrees == NULL) return NO;

  TmpModels = NULL;
  if (Models>0)
  TmpModels = (MODELPARAMETERS*)allocate(sizeof(MODELPARAMETERS)*Models);
  if (TmpModels == NULL) return NO;

  TmpColors = NULL;
  if (Colors>0)
  TmpColors = (COLORPARAMETERS*)allocate(sizeof(COLORPARAMETERS)*Colors);
  if (TmpColors == NULL) return NO;


  //Copy
  j=0;
  for(i=0;i<Project->Layers;i++)
    if(Project->ProjectParams[i].LayerIndex != MaxNode)
    {
      strcpy(TmpParams[j].LayerFileName, Project->ProjectParams[i].LayerFileName);
      TmpParams[j].LayerIndex      = j;
      TmpParams[j].LayerType       = Project->ProjectParams[i].LayerType;
      TmpParams[j].LayerDependency = j;
      j++;
    }

  j=0;
  for(i=0;i<Project->Trees;i++)
    if(Project->Tree[i].LayerIndex != MaxNode)
    {
      strcpy(TmpTrees[j].LayerFileName, Project->Tree[i].LayerFileName);
      TmpTrees[j].LayerIndex = j;
      j++;
    }

  j=0;
  for(i=0;i<Project->Models;i++)
    if(Project->Model[i].LayerIndex != MaxNode)
    {
      TmpModels[j].LayerIndex = j;
      j++;
    }

  j=0;
  for(i=0;i<Project->Colors;i++)
    if(Project->Color[i].LayerIndex != MaxNode)
    {
      TmpColors[j].LayerIndex = j;
      TmpColors[j].Color      = Project->Color[i].Color;
      j++;
    }

  DoneProject(Project);
  Project->ProjectParams = TmpParams;
  Project->Tree          = TmpTrees;
  Project->Model         = TmpModels;
  Project->Color         = TmpColors;

  Project->Layers = Params;
  Project->Trees  = Trees;
  Project->Models = Models;
  Project->Colors = Colors;

  return YES;
}


/*==== Service functions ============================================*/
void trim(char *source)
{
  char *tmpStr;

  tmpStr = strpbrk(source, " \r\n\t,;");
  if (tmpStr != NULL) tmpStr[0]=0;
}


/*-------------------------------------------------------------------*/
YESNO GetStringValue(char *Source, char *Variable, char *section)
{
  if (strstr(Source, section))                      
  {
    Source = strrchr(Source, '=');  Source++;
    if (Source[0]!=0) strcpy(Variable, Source);
    else Variable[0] = 0;
    return YES;
  }
  else return NO;
}


/*-------------------------------------------------------------------*/
YESNO GetIntValue(char *Source, int *Variable, char *section)
{
  if (strstr(Source, section))                      
  {
    Source  = strrchr(Source, '=');  Source++;
    if (Source[0]!=0) sscanf(Source, "%d", Variable);
    else (*Variable) = 0;
    return YES;
  }
  else return NO;
}


/*-------------------------------------------------------------------*/
YESNO GetFloatValue(char *Source, double *Variable, char *section)
{
  if (strstr(Source, section))                      
  {
    Source  = strrchr(Source, '=');  Source++;
    if (Source[0]!=0) sscanf(Source, "%lf",Variable);
    else (*Variable) = 0;
    return YES;
  }
  else return NO;
}