#if ! defined(__COLOR_H)
#define __COLOR_H

#include "owntypes.h"

typedef enum {A=0, C=1, D50=2, D55=3, D65=4, E=5, D75=6, F2=7, F7=8, F10=9} WHITEPOINT;

#define DELAB 5.0 //2.9
#define DELUV 5.0 //2.3

typedef ULONG RGBPIXEL;

#ifndef RGB
#define RGB(r,g,b) ((RGBPIXEL)(((RGBPIXEL)((r)&0xFF))|(((RGBPIXEL)((g)&0xFF))<<8)|(((RGBPIXEL)((b)&0xFF))<<16))) 
#endif

#define RED(rgb)   ((rgb)&0xFF)
#define GREEN(rgb) (((rgb)>>8)&0xFF)
#define BLUE(rgb)  (((rgb)>>16)&0xFF)
#define RGBSIZE    sizeof(RGBPIXEL)

#define  BIN_BLACK   1
#define  BIN_WHITE   0
#define  GRAY_BLACK  0
#define  GRAY_WHITE  255
#define  RGB_BLACK   RGB(0,0,0)
#define  RGB_WHITE   RGB(0xFF,0xFF,0xFF)

void RGB_to_XYZ(BYTE r, BYTE g, BYTE b, double *xcap, double *ycap, double *zcap);
void XYZ_to_LAB(WHITEPOINT wp, double xcap, double ycap, double zcap, double *lstar, double *astar, double *bstar);
void XYZ_to_LUV(WHITEPOINT wp, double xcap, double ycap, double zcap, double *lstar, double *ustar, double *vstar);
void LAB_to_XYZ(WHITEPOINT wp, double lstar, double astar, double bstar, double *xcap, double *ycap, double *zcap);
void LUV_to_XYZ(WHITEPOINT wp, double lstar, double ustar, double vstar, double *xcap, double *ycap, double *zcap);
void XYZ_to_RGB(double xcap, double ycap, double zcap, BYTE *r, BYTE *g, BYTE *b);

void GetWhitepoint(WHITEPOINT wp, double *xcapn, double *ycapn, double *zcapn);

// High level
double DeltaE76(double l1, double a1, double b1, double l2, double a2, double b2);
double DeltaE94(double l1, double a1, double b1, double l2, double a2, double b2);
double ColorDistance76(WHITEPOINT wp, BYTE ar, BYTE ag, BYTE ab, BYTE br, BYTE bg, BYTE bb);
double ColorDistance94(WHITEPOINT wp, BYTE ar, BYTE ag, BYTE ab, BYTE br, BYTE bg, BYTE bb);

#endif /* __COLOR_H */
