/*-------------------------------------------------------------------*/
/* COLORFILT.C     Pavel Kopylov                                     */
/*                                                                   */
/*-------------------------------------------------------------------*/

#define ProgName        "COLORFILT"
#define VersionNumber   "0.01"
#define LastUpdated     "29.08.2004"


#include <math.h>
#include <stdlib.h>            
#include "colorfilt.h"
#include "interfc.h"
#include "shape.h"
#include "memctrl.h"

int       QuietLevel;

BYTE Gamma[256]=  // GI = I ^ (1/CRT Gamma);   CRT Gamma = 2.2
{0, 0x15, 0x1d, 0x22, 0x27, 0x2b, 0x2f, 0x32, 0x36, 0x39, 0x3b, 0x3e, 0x40, 0x43, 0x45, 0x47,
 0x49, 0x4b, 0x4d, 0x4f, 0x51, 0x53, 0x55, 0x56, 0x58, 0x5a, 0x5b, 0x5d, 0x5e, 0x60, 0x61, 0x63,
 0x64, 0x65, 0x67, 0x68, 0x69, 0x6b, 0x6c, 0x6d, 0x6f, 0x70, 0x71, 0x72, 0x73, 0x75, 0x76, 0x77,
 0x78, 0x79, 0x7a, 0x7b, 0x7c, 0x7e, 0x7f, 0x80, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88,
 0x89, 0x8a, 0x8b, 0x8b, 0x8c, 0x8d, 0x8e, 0x8f, 0x90, 0x91, 0x92, 0x93, 0x94, 0x95, 0x95, 0x96,
 0x97, 0x98, 0x99, 0x9a, 0x9a, 0x9b, 0x9c, 0x9d, 0x9e, 0x9f, 0x9f, 0x0a0, 0x0a1, 0x0a2, 0x0a2, 0x0a3,
 0x0a4, 0x0a5, 0x0a6, 0x0a6, 0x0a7, 0x0a8, 0x0a9, 0x0a9, 0x0aa, 0x0ab, 0x0ab, 0x0ac, 0x0ad, 0x0ae, 0x0ae, 0x0af,
 0x0b0, 0x0b0, 0x0b1, 0x0b2, 0x0b3, 0x0b3, 0x0b4, 0x0b5, 0x0b5, 0x0b6, 0x0b7, 0x0b7, 0x0b8, 0x0b9, 0x0b9, 0x0ba,
 0x0bb, 0x0bb, 0x0bc, 0x0bd, 0x0bd, 0x0be, 0x0bf, 0x0bf, 0x0c0, 0x0c0, 0x0c1, 0x0c2, 0x0c2, 0x0c3, 0x0c4, 0x0c4,
 0x0c5, 0x0c5, 0x0c6, 0x0c7, 0x0c7, 0x0c8, 0x0c8, 0x0c9, 0x0ca, 0x0ca, 0x0cb, 0x0cb, 0x0cc, 0x0cd, 0x0cd, 0x0ce,
 0x0ce, 0x0cf, 0x0d0, 0x0d0, 0x0d1, 0x0d1, 0x0d2, 0x0d2, 0x0d3, 0x0d4, 0x0d4, 0x0d5, 0x0d5, 0x0d6, 0x0d6, 0x0d7,
 0x0d7, 0x0d8, 0x0d9, 0x0d9, 0x0da, 0x0da, 0x0db, 0x0db, 0x0dc, 0x0dc, 0x0dd, 0x0dd, 0x0de, 0x0de, 0x0df, 0x0e0,
 0x0e0, 0x0e1, 0x0e1, 0x0e2, 0x0e2, 0x0e3, 0x0e3, 0x0e4, 0x0e4, 0x0e5, 0x0e5, 0x0e6, 0x0e6, 0x0e7, 0x0e7, 0x0e8,
 0x0e8, 0x0e9, 0x0e9, 0x0ea, 0x0ea, 0x0eb, 0x0eb, 0x0ec, 0x0ec, 0x0ed, 0x0ed, 0x0ee, 0x0ee, 0x0ef, 0x0ef, 0x0f0,
 0x0f0, 0x0f1, 0x0f1, 0x0f2, 0x0f2, 0x0f3, 0x0f3, 0x0f3, 0x0f4, 0x0f4, 0x0f5, 0x0f5, 0x0f6, 0x0f6, 0x0f7, 0x0f7,
 0x0f8, 0x0f8, 0x0f9, 0x0f9, 0x0fa, 0x0fa, 0x0fa, 0x0fb, 0x0fb, 0x0fc, 0x0fc, 0x0fd, 0x0fd, 0x0fe, 0x0fe, 0x0ff};

struct idSORTELEM
{
  double  val;
  int     pos;
};
typedef struct idSORTELEM SORTELEM;

int compare(const void *elem1, const void *elem2 );
int round(double x);

/* ------------------------------------------------------------------*/
void Convolution(IMAGE *Source, IMAGE *Destination, int ContextSize, int QL)
{
  ULONG     y, x;
  SLONG     w;
  RGBPIXEL  pixel, NewPixel;
  double    L;
  int       i, GrayPixel, CS;
  IMAGE     tmpImage;
  YESNO     ret;


  QuietLevel = QL;
  if (QuietLevel>=2) PrintMessage("Convolution\n");

  SetImageSize(&tmpImage, Source->ImageSizeX, Source->ImageSizeY);
  ret=ImageInit("", &tmpImage, OUTPUT, YES, AUTOMATIC, GRAYMODE, MEMORY, 0);
  if (ret==NO) return;

  SetProgressVisibility(1);
  OpenProgressWindow();

  for (y=1;y<=Source->ImageSizeY;y++)
  {
    for (x=1;x<=Source->ImageSizeX;x++)
    {
      pixel=GetRGBPixelEx(Source, x, y);
      L=(float)(RED(pixel))*0.3 + (float)(GREEN(pixel))*0.586 + (float)(BLUE(pixel))*0.115;
      GrayPixel=(BYTE)L;
      
      PutGrayPixel(&tmpImage, x, y, GrayPixel);
    }
    if (QuietLevel>=2) UpdateProgressBar(y, (Source->ImageSizeY)*2);
  }

  TemplateX = TemplateArray[0][0];
  TemplateY = TemplateArray[0][1];
  TemplateM = TemplateArray[0][2];
  CS = min(ContextSize, TemplateArray[0][0][0]);

  
  ImageRewind(&tmpImage);
  for (y=1;y<=tmpImage.ImageSizeY;y++)
  {
    for (x=1;x<=tmpImage.ImageSizeX;x++)
    {
      GrayPixel=GetGrayPixel(&tmpImage, x+TemplateX[37], y+TemplateY[37])*TemplateM[37];
      w = GrayPixel;

      for(i=1;i<CS;i++)
      {
        GrayPixel=GetGrayPixel(&tmpImage, x+TemplateX[i], y+TemplateY[i])*TemplateM[i];
        w += GrayPixel;
      }
      GrayPixel=127+w/4;
      if (GrayPixel<0)GrayPixel=0;
      if (GrayPixel>255)GrayPixel=255;

      NewPixel=RGB((BYTE)GrayPixel,(BYTE)GrayPixel,(BYTE)GrayPixel);
      PutRGBPixelEx(Destination, x, y, NewPixel);
    }
    if (QuietLevel>=3) UpdateProgressBar(y+Source->ImageSizeY, (Source->ImageSizeY)*2);
  }
  CloseProgressWindow();

  ImageDone(&tmpImage);
}

/* ------------------------------------------------------------------*/
void Border(IMAGE *Source, IMAGE *Destination, int ContextSize, int QL)
{
  ULONG     y, x;
  RGBPIXEL  pixel;
  double    L, LA, LC;
  int       i, GrayPixel, CS;


  QuietLevel = QL;
  if (QuietLevel>=2) PrintMessage("Lightness activity\n");

  SetProgressVisibility(1);
  OpenProgressWindow();

  TemplateX = TemplateArray[0][0];
  TemplateY = TemplateArray[0][1];
  TemplateM = TemplateArray[0][2];
  CS = min(ContextSize, TemplateArray[0][0][0]);

  for (y=1;y<=Source->ImageSizeY;y++)
  {
    for (x=1;x<=Source->ImageSizeX;x++)
    {
      pixel=GetRGBPixelEx(Source, x+TemplateX[37], y+TemplateY[37]);
      LA=(float)(RED(pixel))*0.3 + (float)(GREEN(pixel))*0.586 + (float)(BLUE(pixel))*0.115;
      LC=LA;

      for(i=1;i<CS;i++)
      {
        pixel=GetRGBPixelEx(Source, x+TemplateX[i], y+TemplateY[i]);
        L=(float)(RED(pixel))*0.3 + (float)(GREEN(pixel))*0.586 + (float)(BLUE(pixel))*0.115;
        LA += L;
      }
      LA /= CS;
      L = LC-LA;
      if (L<0) L=-L;

      GrayPixel=(BYTE)L;
      
      PutGrayPixel(Destination, x, y, GrayPixel);
    }
    if (QuietLevel>=2) UpdateProgressBar(y, Source->ImageSizeY);
  }
}

/* ----------------------------------------------------------------- */
void ColorActivity(IMAGE *Source, IMAGE *Destination, int ContextSize, int QL)
{
  ULONG     y, x;
  RGBPIXEL  pixel, NewPixel;
  int       i, GrayPixel, CS;
  double    accL, accA, accB;
  double    Avg, AL, AA, AB, mL, mA, mB;
  BYTE      r,g,b;
  double    X,Y,Z,L,A,B;


  QuietLevel  = QL;
  if (QuietLevel>=2) PrintMessage("Color activity\n");

  TemplateX = TemplateArray[0][0];
  TemplateY = TemplateArray[0][1];
  TemplateM = TemplateArray[0][2];
  CS = min(ContextSize, TemplateArray[0][0][0]);

  SetProgressVisibility(1);
  OpenProgressWindow();
  for (y=1;y<=Source->ImageSizeY;y++)
  {
    for (x=1;x<=Source->ImageSizeX;x++)
    {
      pixel=GetRGBPixelEx(Source, x+TemplateX[37], y+TemplateY[37]);
      r = (BYTE)RED(pixel);
      g = (BYTE)GREEN(pixel);
      b = (BYTE)BLUE(pixel);
      if (r==0 && g==0 && b==0) { accL=0; accA=0; accB=0; }
      else
      {
        RGB_to_XYZ(r, g, b, &X, &Y, &Z);
        XYZ_to_LAB(D65, X, Y, Z, &accL, &accA, &accB);
      }
      AL=accL; AA=accA; AB=accB;

      for(i=1;i<CS;i++)
      {
        pixel=GetRGBPixelEx(Source, x+TemplateX[i], y+TemplateY[i]);
        r = (BYTE)RED(pixel);
        g = (BYTE)GREEN(pixel);
        b = (BYTE)BLUE(pixel);
        if (r==0 && g==0 && b==0) { L=0; A=0; B=0; }
        else
        {
          RGB_to_XYZ(r, g, b, &X, &Y, &Z);
          XYZ_to_LAB(D65, X, Y, Z, &L, &A, &B);
        }

        accL += L;
        accA += A;
        accB += B;
      }
      mL = accL / CS;
      mA = accA / CS;
      mB = accB / CS;

      Avg = DeltaE76(mL, mA, mB, AL, AA, AB);

      GrayPixel = round(Avg);
      if (GrayPixel<0)GrayPixel=0;
      if (GrayPixel>255)GrayPixel=255;

      NewPixel=RGB(GrayPixel,GrayPixel,GrayPixel);
      PutRGBPixelEx(Destination, x, y, NewPixel);
    }
    if (QuietLevel>=2) UpdateProgressBar(y, Source->ImageSizeY);
  }
  CloseProgressWindow();
}


/* ----------------------------------------------------------------- */
void GammaCorrection(IMAGE *Source, IMAGE *Destination, int QL)
{
  ULONG     y, x;
  RGBPIXEL  pixel, NewPixel;
  int       r,g,b;


  QuietLevel  = QL;
  if (QuietLevel>=2) PrintMessage("Gamma correction\n");

  SetProgressVisibility(1);
  OpenProgressWindow();

  for (y=1;y<=Source->ImageSizeY;y++)
  {
    for (x=1;x<=Source->ImageSizeX;x++)
    {
      pixel=GetRGBPixelEx(Source, x, y);
      r = RED(pixel);
      g = GREEN(pixel);
      b = BLUE(pixel);
      
      NewPixel=RGB(Gamma[r],Gamma[g],Gamma[b]);
      PutRGBPixelEx(Destination, x, y, NewPixel);
    }
    if (QuietLevel>=3) UpdateProgressBar(y, Source->ImageSizeY);
  }
  CloseProgressWindow();
}


/* ----------------------------------------------------------------- */
double ImageERR(IMAGE *Source, IMAGE *Destination, int QL)
{
  ULONG     y, x, acc=0;
  RGBPIXEL  pixel, pixel1;

  QuietLevel  = QL;
  if (QuietLevel>=2) PrintMessage("ERR\n");

  SetProgressVisibility(1);
  OpenProgressWindow();

  for (y=1;y<=Source->ImageSizeY;y++)
  {
    for (x=1;x<=Source->ImageSizeX;x++)
    {
      pixel=GetRGBPixelEx(Source, x, y);
      pixel1=GetRGBPixelEx(Destination, x, y);
      if(pixel!=pixel1)
      {
        acc ++;
      }
    }
    if (QuietLevel>=3) UpdateProgressBar(y, Source->ImageSizeY);
  }
  CloseProgressWindow();
  if (QuietLevel>=3) PrintMessage("Changed: %u pixels\n", acc);

  return ((double)acc/(double)(Source->ImageSizeY)/(double)(Source->ImageSizeX));
}


/* ----------------------------------------------------------------- */
double ImageDE(IMAGE *Source, IMAGE *Destination, int QL)
{
  ULONG     y, x, acc1=0;
  RGBPIXEL  pixel, pixel1;
  double    acc=0, val;

  QuietLevel  = QL;
  if (QuietLevel>=2) PrintMessage("DE\n");

  SetProgressVisibility(1);
  OpenProgressWindow();

  for (y=1;y<=Source->ImageSizeY;y++)
  {
    for (x=1;x<=Source->ImageSizeX;x++)
    {
      pixel=GetRGBPixelEx(Source, x, y);
      pixel1=GetRGBPixelEx(Destination, x, y);
      if(pixel!=pixel1)
      {
        val = ColorDistance76(D65, (BYTE)RED(pixel), (BYTE)GREEN(pixel), (BYTE)BLUE(pixel), 
                                   (BYTE)RED(pixel1), (BYTE)GREEN(pixel1), (BYTE)BLUE(pixel1));
        acc += val;
        acc1++;
      }
    }
    if (QuietLevel>=3) UpdateProgressBar(y, Source->ImageSizeY);
  }
  CloseProgressWindow();
  if (QuietLevel>=3) PrintMessage("Changed: %u pixels\n", acc1);

  return (acc/Source->ImageSizeY/Source->ImageSizeX);
}


/* ----------------------------------------------------------------- */
void CromaFilter(IMAGE *Source, IMAGE *Destination, int Template, int ContextSize, int value, int QL)
{
  ULONG     y, x;
  RGBPIXEL  pixel, NewPixel;
  BYTE      r,g,b;
  double    X,Y,Z,L,A,B;
  IMAGE     LImage, AImage, BImage, AAImage, BBImage;
  YESNO     ret;
  RGBPIXEL  LL, AA, BB;


  QuietLevel = QL;
  if (QuietLevel>=2) PrintMessage("Chromacity filter\n");

  SetImageSize(&LImage, Source->ImageSizeX, Source->ImageSizeY);
  ret=ImageInit("L.pgm", &LImage, OUTPUT, YES, AUTOMATIC, RGBMODE, DISK, 10);
  if (ret==NO) return;
  SetImageSize(&AImage, Source->ImageSizeX, Source->ImageSizeY);
  ret=ImageInit("A.pgm", &AImage, OUTPUT, YES, AUTOMATIC, RGBMODE, DISK, 10);
  if (ret==NO) return;
  SetImageSize(&BImage, Source->ImageSizeX, Source->ImageSizeY);
  ret=ImageInit("B.pgm", &BImage, OUTPUT, YES, AUTOMATIC, RGBMODE, DISK, 10);
  if (ret==NO) return;

  // Reconstruct into L*, a*, b* planes
  SetProgressVisibility(1);
  OpenProgressWindow();

  for (y=1;y<=Source->ImageSizeY;y++)
  {
    for (x=1;x<=Source->ImageSizeX;x++)
    {
      pixel=GetRGBPixelEx(Source, x, y);
      r = (BYTE)RED(pixel);
      g = (BYTE)GREEN(pixel);
      b = (BYTE)BLUE(pixel);

      RGB_to_XYZ(r, g, b, &X, &Y, &Z);
      XYZ_to_LAB(D65, X, Y, Z, &L, &A, &B);

      LL = (RGBPIXEL)roundint(L*10000);
      AA = (RGBPIXEL)roundint((A+127)*10000);
      BB = (RGBPIXEL)roundint((B+127)*10000);
      
      PutRGBPixel(&LImage, x, y, LL);
      PutRGBPixel(&AImage, x, y, AA);
      PutRGBPixel(&BImage, x, y, BB);
    }
    if (QuietLevel>=3) UpdateProgressBar(y, (Source->ImageSizeY)*2);
  }
  ImageDone(&LImage);
  ImageDone(&AImage);
  ImageDone(&BImage);


  // Filter chromacity planes a* and b*
  ret=ImageInit("A.pgm", &AImage, INPUT, YES, AUTOMATIC, RGBMODE, DISK, 10);
  if (ret==NO) return;
  ret=ImageInit("B.pgm", &BImage, INPUT, YES, AUTOMATIC, RGBMODE, DISK, 10);
  if (ret==NO) return;
  SetImageSize(&AAImage, Source->ImageSizeX, Source->ImageSizeY);
  ret=ImageInit("AA.pgm", &AAImage, OUTPUT, YES, AUTOMATIC, RGBMODE, DISK, 10);
  if (ret==NO) return;
  SetImageSize(&BBImage, Source->ImageSizeX, Source->ImageSizeY);
  ret=ImageInit("BB.pgm", &BBImage, OUTPUT, YES, AUTOMATIC, RGBMODE, DISK, 10);
  if (ret==NO) return;

  VectorMedian(&AImage, &AAImage, Template, ContextSize, QL,0);
  VectorMedian(&BImage, &BBImage, Template, ContextSize, QL,0);

  ImageDone(&AImage);
  ImageDone(&BImage);
  ImageDone(&AAImage);
  ImageDone(&BBImage);


  // Reconstruct back to RGB
  ret=ImageInit("L.pgm", &LImage, INPUT, YES, AUTOMATIC, RGBMODE, DISK, 10);
  if (ret==NO) return;
  ret=ImageInit("AA.pgm", &AImage, INPUT, YES, AUTOMATIC, RGBMODE, DISK, 10);
  if (ret==NO) return;
  ret=ImageInit("BB.pgm", &BImage, INPUT, YES, AUTOMATIC, RGBMODE, DISK, 10);
  if (ret==NO) return;
  
  for (y=1;y<=Source->ImageSizeY;y++)
  {
    for (x=1;x<=Source->ImageSizeX;x++)
    {
      LL = GetRGBPixel(&LImage, x, y);
      AA = GetRGBPixel(&AImage, x, y);
      BB = GetRGBPixel(&BImage, x, y);

      L = ((double)LL/10000.0);
      A = ((double)AA/10000.0-127);
      B = ((double)BB/10000.0-127);

      LAB_to_XYZ(D65, L, A, B, &X, &Y, &Z);
      XYZ_to_RGB(X, Y, Z, &r, &g, &b);

      NewPixel=RGB(r,g,b);
      PutRGBPixelEx(Destination, x, y, NewPixel);
    }
    if (QuietLevel>=3) UpdateProgressBar(y+Source->ImageSizeY, (Source->ImageSizeY)*2);
  }
  CloseProgressWindow();

  ImageDone(&LImage);
  ImageDone(&AImage);
  ImageDone(&BImage);

  remove("L.pgm");
  remove("A.pgm");
  remove("B.pgm");
  remove("AA.pgm");
  remove("BB.pgm");
}


/* ----------------------------------------------------------------- */
void VectorMedian(IMAGE *Source, IMAGE *Destination, int Template, int ContextSize, int QL, int Adaptive)
{
  ULONG     y, x;
  RGBPIXEL  pixel;


  QuietLevel  = QL;
  if (QuietLevel>=2)
  {
    if (Adaptive==0) PrintMessage("Vector Median Filter\n");
    else if (Adaptive==1) PrintMessage("Center Weighted Vector Median Filter\n");
      else if (Adaptive==2) PrintMessage("Rank-controlled Vector Median Filter\n");
  }
  SetProgressVisibility(1);
  OpenProgressWindow();

  for (y=1;y<=Source->ImageSizeY;y++)
  {
    for (x=1;x<=Source->ImageSizeX;x++)
    {
      if (Adaptive==0) pixel = LocalMedian(Source, x, y, Template, ContextSize);
      else if (Adaptive==1) pixel = CWLocalMedian(Source, x, y, Template, ContextSize);
        else if (Adaptive==2) pixel = RCLocalMedian(Source, x, y, Template, ContextSize);
      PutRGBPixelEx(Destination, x, y, pixel);
    }
    if (QuietLevel>=3) UpdateProgressBar(y, Source->ImageSizeY);
  }
  CloseProgressWindow();
}


/* ----------------------------------------------------------------- */
RGBPIXEL LocalMedian(IMAGE *Source, ULONG x, ULONG y, int Template, int ContextSize)
{
  SORTELEM  *lookup;
  RGBPIXEL  pixel;
  BYTE      r,g,b;
  int       i, CS, j;

  
  TemplateX = TemplateArray[Template][0];
  TemplateY = TemplateArray[Template][1];
  CS        = min(ContextSize, TemplateArray[Template][0][0]);
  lookup    = allocate(sizeof(SORTELEM)*(CS+1));

  // Calculate dE
  pixel=GetRGBPixelEx(Source, x, y);
  r = (BYTE)RED(pixel);
  g = (BYTE)GREEN(pixel);
  b = (BYTE)BLUE(pixel);
  
  lookup[0].val = sqrt((double)(r*r)+(double)(g*g)+(double)(b*b));
  lookup[0].pos = 0;
  for(i=1; i<=CS; i++)
  {
    pixel=GetRGBPixelEx(Source, x+TemplateX[i], y+TemplateY[i]);
    r = (BYTE)RED(pixel);
    g = (BYTE)GREEN(pixel);
    b = (BYTE)BLUE(pixel);
    lookup[i].val = sqrt((double)(r*r)+(double)(g*g)+(double)(b*b));
    lookup[i].pos = i;
  }

  // Sort 
  qsort(lookup, CS+1, sizeof(SORTELEM), compare);

  j=lookup[(CS+1)/2].pos;
  if (j==0)
  {
    pixel=GetRGBPixelEx(Source, x, y);
  }
  else pixel=GetRGBPixelEx(Source, x+TemplateX[j], y+TemplateY[j]);

  deallocate(lookup);

  return (pixel);
}


/* ----------------------------------------------------------------- */
RGBPIXEL CWLocalMedian(IMAGE *Source, ULONG x, ULONG y, int Template, int ContextSize)
{
  double    *D, *W, n1, n, d1, dn, w, ra, ga, ba, t;
  RGBPIXEL  pixel, *X;
  BYTE      r,g,b,r1,g1,b1;
  int       i, CS, j, Debug=0;


  //if (x==3 && y==3) Debug=1;
  //if (x==33 && y==10) Debug=1;
  //if (x==48 && y==42) Debug=1;
  
  TemplateX = TemplateArray[Template][0];
  TemplateY = TemplateArray[Template][1];
  CS        = 8;//min(ContextSize, TemplateArray[Template][0][0]);
  D         = allocate(sizeof(double)*(CS+1));
  W         = allocate(sizeof(double)*(CS+1));
  X         = allocate(sizeof(RGBPIXEL)*(CS+1));
  
  
  // Calculate Xi
  X[0] =GetRGBPixelEx(Source, x, y);
  for(i=1; i<=CS; i++) X[i] = GetRGBPixelEx(Source, x+TemplateX[i], y+TemplateY[i]);

  if (Debug)
  {
    printf("\n(%d,%d,%d)\t(%d,%d,%d)\t(%d,%d,%d)\n",(BYTE)RED(X[7]),(BYTE)GREEN(X[7]),(BYTE)BLUE(X[7]), 
                                                  (BYTE)RED(X[4]),(BYTE)GREEN(X[4]),(BYTE)BLUE(X[4]), 
                                                  (BYTE)RED(X[8]),(BYTE)GREEN(X[8]),(BYTE)BLUE(X[8]));

    printf("(%d,%d,%d)\t(%d,%d,%d)\t(%d,%d,%d)\n",(BYTE)RED(X[3]),(BYTE)GREEN(X[3]),(BYTE)BLUE(X[3]), 
                                                  (BYTE)RED(X[0]),(BYTE)GREEN(X[0]),(BYTE)BLUE(X[0]), 
                                                  (BYTE)RED(X[1]),(BYTE)GREEN(X[1]),(BYTE)BLUE(X[1]));

    printf("(%d,%d,%d)\t(%d,%d,%d)\t(%d,%d,%d)\n\n",(BYTE)RED(X[6]),(BYTE)GREEN(X[6]),(BYTE)BLUE(X[6]), 
                                                  (BYTE)RED(X[2]),(BYTE)GREEN(X[2]),(BYTE)BLUE(X[2]), 
                                                  (BYTE)RED(X[5]),(BYTE)GREEN(X[5]),(BYTE)BLUE(X[5]));
  }
  // Calculate Di
  for (i=0; i<=CS; i++)
  {
    r = (BYTE)RED(X[i]);
    g = (BYTE)GREEN(X[i]);
    b = (BYTE)BLUE(X[i]);
    n = sqrt(r*r+g*g+b*b);
    D[i]=0;

    for (j=0; j<=CS; j++)
    {
      if (Debug && i==5 && j==5)
      {
        i=i; j=j;
      }

      r1 = (BYTE)RED(X[j]);
      g1 = (BYTE)GREEN(X[j]);
      b1 = (BYTE)BLUE(X[j]);
      n1 = sqrt(r1*r1+g1*g1+b1*b1);

      if (n!=0 && n1!=0) t=(r*r1+g*g1+b*b1)/(n*n1);
      else t=0;
      
      if (t>1) t=1;
      D[i] = D[i] + acos(t);
    }
  }
  if (Debug)
  {
    printf("\n%f\t%f\t%f\n",D[7], D[4], D[8]);
    printf("\n%f\t%f\t%f\n",D[3], D[0], D[1]);
    printf("\n%f\t%f\t%f\n\n",D[6], D[2], D[5]);
  }

  // Calculate Dn & D1
  d1=D[0];
  dn=D[0];
  for (i=1; i<=CS; i++)
  {
    if (D[i]<d1) d1=D[i];
    if (D[i]>dn) dn=D[i];
  }

  // Calculate Wi
  if (d1!=dn)
  {
    w=0;
    for (i=0; i<=CS; i++)
    {
      W[i] = (dn-D[i])/(dn-d1);
      w+=W[i];
    }
    if (Debug)
    {
      printf("\n%f\t%f\t%f\n",W[7], W[4], W[8]);
      printf("\n%f\t%f\t%f\n",W[3], W[0], W[1]);
      printf("\n%f\t%f\t%f\n\n",W[6], W[2], W[5]);
    }


    // Calculate Y^
    ra=0;
    ga=0;
    ba=0;
    for (i=0; i<=CS; i++)
    {
      r = (BYTE)RED(X[i]);
      g = (BYTE)GREEN(X[i]);
      b = (BYTE)BLUE(X[i]);

      ra += r*W[i];
      ga += g*W[i];
      ba += b*W[i];
    }
    ra/=w;
    ga/=w;
    ba/=w;

    r=(BYTE)ra;
    g=(BYTE)ga;
    b=(BYTE)ba;
    pixel=RGB(r,g,b);
  }
  else
  {
    pixel=X[0];
  }

  deallocate(X);
  deallocate(W);
  deallocate(D);

  return (pixel);
}


/* ----------------------------------------------------------------- */
RGBPIXEL RCLocalMedian(IMAGE *Source, ULONG x, ULONG y, int Template, int ContextSize)
{
  SORTELEM  *lookup;
  RGBPIXEL  pixel, pixelc;
  BYTE      r,g,b, rc,gc,bc;
  int       i, CS, j, R, Debug=0;
  double    Tol, Val, ra, ga, ba;

  
  //if (x==8 && y==9) Debug=1;

  R=5; Tol=60; Val=0;// Seee Lukac

  TemplateX = TemplateArray[Template][0];
  TemplateY = TemplateArray[Template][1];
  CS        = 8;   //min(ContextSize, TemplateArray[Template][0][0]);
  lookup    = allocate(sizeof(SORTELEM)*(CS));

  // Calculate dE
  if (Debug==1)
  {
    x=x;
  }
  pixelc=GetRGBPixelEx(Source, x, y);
  rc = (BYTE)RED(pixelc);
  gc = (BYTE)GREEN(pixelc);
  bc = (BYTE)BLUE(pixelc);
  
  for(i=1; i<=CS; i++)
  {
    pixel=GetRGBPixelEx(Source, x+TemplateX[i], y+TemplateY[i]);
    r = (BYTE)RED(pixel);
    g = (BYTE)GREEN(pixel);
    b = (BYTE)BLUE(pixel);
    lookup[i-1].val = sqrt((double)((rc-r)*(rc-r))+(double)((gc-g)*(gc-g))+(double)((bc-b)*(bc-b)));  // Try CIE DL
    lookup[i-1].pos = i;
  }

  // Sort
  qsort(lookup, CS, sizeof(SORTELEM), compare);
  if (Debug==1)
  {
    for(i=0; i<CS; i++)
    {
      printf("(%f,%d)\t",lookup[i].val,lookup[i].pos);
    }
    printf("\n");
  }

  // Calculate Val
  ra=0;
  ga=0;
  ba=0;
  for(i=1; i<=R; i++)
  {
    j = lookup[i].pos;
    pixel=GetRGBPixelEx(Source, x+TemplateX[j], y+TemplateY[j]);
    r = (BYTE)RED(pixel);
    g = (BYTE)GREEN(pixel);
    b = (BYTE)BLUE(pixel);

    ra += r;
    ga += g;
    ba += b;
  }
  ra/=(double)R;
  ga/=(double)R;
  ba/=(double)R;

  Val = sqrt((((double)rc-ra)*((double)rc-ra))+
             (((double)gc-ga)*((double)gc-ga))+
             (((double)bc-ba)*((double)bc-ba)));  // Try CIE DL

  if (Val>=Tol)
  {
    j=lookup[0].pos;
    if (j>0) 
    {
      pixel=GetRGBPixelEx(Source, x+TemplateX[j], y+TemplateY[j]);
    }
    else pixel = pixelc;
  }
  else pixel=pixelc;

  deallocate(lookup);

  return (pixel);
}

/* ----------------------------------------------------------------- */
int compare(const void *elem1, const void *elem2)
{
  SORTELEM  *e1, *e2;

  e1 = (SORTELEM*)elem1;
  e2 = (SORTELEM*)elem2;

  if (e1->val<e2->val) return -1;
  else
    if (e1->val==e2->val) return 0;
    else
      if (e1->val>e2->val) return 1;

  return 0;
}

/* ------------------------------------------------------------------*/
void FSErrorDiffusion(IMAGE *Source, IMAGE *Destination, int QL)
{
  ULONG     y, x;
  RGBPIXEL  pixel;
  int       GrayPixel, error, BinPixel, i;
  IMAGE     tmpImage;
  YESNO     ret;
  double    L;


  TemplateX = TemplateArray[3][0];
  TemplateY = TemplateArray[3][1];
  TemplateM = TemplateArray[3][2];

  QuietLevel = QL;
  if (QuietLevel>=2) PrintMessage("Floyd-Stteinberg Error Diffusion\n");

  SetImageSize(&tmpImage, Source->ImageSizeX, Source->ImageSizeY);
  ret=ImageInit("", &tmpImage, OUTPUT, YES, AUTOMATIC, GRAYMODE, MEMORY, 0);
  if (ret==NO) return;

  SetProgressVisibility(1);
  OpenProgressWindow();

  for (y=1;y<=Source->ImageSizeY;y++)
  {
    for (x=1;x<=Source->ImageSizeX;x++)
    {
      pixel=GetRGBPixelEx(Source, x, y);
      L=(float)(RED(pixel))*0.3 + (float)(GREEN(pixel))*0.586 + (float)(BLUE(pixel))*0.115;
      GrayPixel=(BYTE)L;
      
      PutGrayPixel(&tmpImage, x, y, GrayPixel);
    }
    if (QuietLevel>=2) UpdateProgressBar(y, (Source->ImageSizeY)*2);
  }
  ImageRewind(&tmpImage);


  for (y=2;y<tmpImage.ImageSizeY;y++)
  {
    for (x=2;x<tmpImage.ImageSizeX;x++)
    {
      GrayPixel=GetGrayPixel(&tmpImage, x, y);
      if (GrayPixel<128)
      {
        error=GrayPixel;
        BinPixel=1;
      }
      else
      {
        error=GrayPixel-255;
        BinPixel=0;
      }
      PutBitPixel(Destination, x, y, BinPixel);

      for (i=1;i<=4;i++)
      {
        GrayPixel=GetGrayPixel(&tmpImage, x+TemplateX[i], y+TemplateY[i]);
        GrayPixel+=error*TemplateM[i]/16;
        if (GrayPixel>255) GrayPixel=255;
        if (GrayPixel<0) GrayPixel=0;
        PutGrayPixel(&tmpImage, x+TemplateX[i], y+TemplateY[i], GrayPixel);
      }
    }
    if (QuietLevel>=3) UpdateProgressBar(y+Source->ImageSizeY, (Source->ImageSizeY)*2);
  }
  CloseProgressWindow();

  ImageDone(&tmpImage);
}


int round(double x)
{
  int      tmp;
  double   buff;


  tmp = (int)x;
  buff= x-(double)tmp;

  if(buff>=0.5) tmp++;

  return (tmp);
}