/*-------------------------------------------------------------------*/
/* GENIO.C         Pavel Kopylov                                     */
/*                 Eugene Ageenko                                    */
/*                                                                   */
/* General Input/Output routines                                     */
/* Modefied by Viktor Veis 12.11.2001                                */
/* Add TellFlow, SeekFlow, buffer and string routines.               */
/*-------------------------------------------------------------------*/
#define  ModuleName     "GENIO"
#define  VersionNumber  "Version L.03"
#define  LastUpdated    "26.10.2001"

#include <stdio.h>
#include <assert.h>
#include <string.h>
#include "genio.h"
#include "interfc.h"

int  InputBit(FLOW*);
void OutputBit(FLOW*, int bit);
int  InputValue(FLOW*, int bits);
void OutputValue(FLOW*, int x, int bits);
void FlushInput(FLOW*);
void FlushOutput(FLOW*);
int InputFlowBit(FLOW*);
void OutputFlowBit(FLOW*, int);
void FlushFlowInput(FLOW*);
void FlushFlowOutput(FLOW*);
int InputFlowValue(FLOW* f, int);
void OutputFlowValue(FLOW*, int, int);

#define DEBUG


// ---- File IO

// in case of failure, function lowers Status flag to 0
// if Status is set, function does nothing
void OutputByteToFile(FLOW* f, int x)
{
  if(f->Flow==NULL) return;

  if (f->Status)
  {
    if (x != (putc(x, (FILE*)(f->Flow)))) f->Status = 0;

#ifdef DEBUG  
  else f->CurrCnt++;
#endif

  }
}

// in case of failure, function lowers Status flag to 0
// if Status is set, function does nothing
int InputByteFromFile(FLOW* f)
{
  int x;

  if(f->Flow==NULL) return 0;

  if (f->Status) 
  {
    if ((x=getc((FILE*)(f->Flow))) != EOF) 
    {

#ifdef DEBUG  
      f->CurrCnt++;
#endif

      return x;
    }
    else 
      {f->Status = 0; return 0;}
  }
  else return 0;
}

// ---- Memory
void OutputByteToMem(FLOW* f, int x)
{
  if(f->Flow==NULL) return;
  
  if (f->Status) {
    if (f->CurrCnt < f->MaxCnt)
    {
      ((BYTE*)(f->Flow))[f->CurrCnt] = (BYTE)x;
      f->CurrCnt++;
    }
    else f->Status = 0;
  }
}

int InputByteFromMem(FLOW* f)
{
  int x;

  if(f->Flow==NULL) return 0;
  
  if (f->Status) {
    if (f->CurrCnt < f->MaxCnt)
    {
      x = (int)((BYTE*)(f->Flow))[f->CurrCnt];
      f->CurrCnt++;
      return x;
    }
    else {f->Status=0; return 0;}
  }  
  else return 0;
}

// returns 1 if Ok, 0 if not
int SeekFile(FLOW *f, long offset, int origin)
{
  if(f->Flow==NULL) return 0;
  if(!f->Status) return 0;
  return ((fseek(f->Flow,offset,origin)) == 0 ? 1 : 0);
}


// returns -1L in case of error
long TellFile(FLOW* f)
{
  if(f->Flow==NULL) return -1L;
  if(!f->Status) return -1L;
  return (ftell(f->Flow));
}

int SeekMem(FLOW* f, long offset, int origin)
{
  if(f->Flow==NULL) return 0;
  if(!f->Status) return 0;
  switch (origin)
  {
	case SEEK_SET:
		f->CurrCnt = offset;
		break;
	case SEEK_CUR:
		f->CurrCnt+=offset;
		break;
	case SEEK_END:
		f->CurrCnt=f->MaxCnt+offset;
		break;
	default: return 0;
  }
  return 1;
}

long TellMem(FLOW* f)
{
  if(f->Flow==NULL) return -1L;
  if(!f->Status) return -1L;
  return (long)f->CurrCnt;
}

int  SeekFlow(FLOW *f, long offset, int origin)
{
  return(f->SeekFlow(f, offset, origin));
}

long TellFlow(FLOW *f)
{
  return(f->TellFlow(f));
}



// INITIALIZATION 
void InitFlow(FLOW* f, FLOWTYPE type, void* ptr, ULONG size)
{
  f->Status      = 1; // operational
  f->FlowType    = type;
  f->Flow        = ptr;
  f->CurrCnt     = 0;
  f->MaxCnt      = size;
  f->BitsInQueue = 0;
  f->BitQueue    = 0;
  switch (type)
  {
  case F_FILE:
      f->OutputByte = (TOutputByte) OutputByteToFile;
      f->InputByte  = (TInputByte) InputByteFromFile;
      f->SeekFlow   = (TSeekFlow) SeekFile;
      f->TellFlow   = (TTellFlow) TellFile;
      break;
  case F_MEMORY:
      f->OutputByte = (TOutputByte) OutputByteToMem;
      f->InputByte  = (TInputByte) InputByteFromMem;
      f->SeekFlow   = (TSeekFlow) SeekMem;
      f->TellFlow   = (TTellFlow) TellMem;
      break;
  }
  // BitStream
  f->InputBit   = (TInputBit)   InputFlowBit;
  f->OutputBit  = (TOutputBit)  OutputFlowBit;
  f->InputValue = (TInputValue) InputFlowValue;
  f->OutputValue= (TOutputValue)OutputFlowValue;
  f->FlushInput = (TFlushInput) FlushFlowInput;
  f->FlushOutput= (TFlushOutput)FlushFlowOutput;
}




/*-------------------------------------------------------------------*/
int ReadULongFromFlow(FLOW* f, ULONG* value, int bytecount)
{
  int    i;
  int    c;
  ULONG  res = 0;

  for(i=0; i<bytecount; i++)
  {
    c = f->InputByte(f);
    if (!f->Status) break;
    c <<= (8*i);
    res |= c;
  }

  *value = res;
  return i;
}


/*-------------------------------------------------------------------*/
int ReadSLongFromFlow(FLOW* f, long* value, int bytecount)
{
  int    i;
  long   c=0, res=0, resc;

  for(i=0; i<bytecount; i++)
  {
    c = f->InputByte(f);
    if (!f->Status) return i;
    resc = c << (8*i);
    res |= resc;
  }

  if (c & 0x80)
  {
    for(i; i<sizeof(long); i++)
    {
      c = 0xFF<<(8*i);
      res |= c;
    }
  }

  *value = res;
  return bytecount;
}


/*-------------------------------------------------------------------*/
int WriteULongToFlow(FLOW* f, ULONG value, int bytecount)
{                       
  int    i;
  ULONG  c,b;

  b = value;
  for(i=0; i<bytecount; i++)
  {
    c = b & 0xFF;
    f->OutputByte(f,c);
    if (!f->Status) break;
    b >>=8;
  }
  return (i);
}


/*-------------------------------------------------------------------*/
int WriteSLongToFlow(FLOW* f, long value, int bytecount)
{                       
  int    i;
  long   c,b;

  b = value;
  for(i=0; i<bytecount; i++)
  {
    c = b & 0xFF;
    f->OutputByte(f,c);
    if (!f->Status) break;
    b >>=8;
  }
  return (i);
}


/*-------------------------------------------------------------------*/
int InputFlowBit(FLOW* f)
{
  int bit;

  if( f->BitsInQueue == 0 )
  {
    f->BitQueue = f->InputByte(f);
    if(f->BitQueue == EOF)
    {
      ErrorMessage("Unexpected EOF in input file.");
      return( EOF );
    }
    f->BitsInQueue = 8;
  }
  bit = (f->BitQueue & (0x80)) >> 7;
  f->BitQueue <<= 1;
  (f->BitsInQueue)--;

  return( bit );
}


/*-------------------------------------------------------------------*/
void OutputFlowBit(FLOW* f, int bit)
{
  assert( bit == 0 || bit == 1 );

  f->BitQueue <<= 1;
  f->BitQueue |= bit;
  if( ++(f->BitsInQueue) == 8 )
  {
    f->OutputByte(f,f->BitQueue);
    f->BitsInQueue = 0;
    f->BitQueue = 0;
  }
}


/*-------------------------------------------------------------------*/
void FlushFlowInput(FLOW* f)
{
  f->BitsInQueue = 0;
}


/*-------------------------------------------------------------------*/
void FlushFlowOutput(FLOW* f)
{
  if(f->BitsInQueue==0)
  {
    return;
  }
  while((f->BitsInQueue)++ < 8)
  {
    f->BitQueue <<= 1;
  }
  f->OutputByte(f,f->BitQueue);
  f->BitsInQueue = 0;
  f->BitQueue = 0;
}


/*-------------------------------------------------------------------*/
int InputFlowValue(FLOW* f, int bits)
{
  int i;
  int x    = 0;
  int mask = 0x01;

  for(i=1; i<=bits; i++)
  {
    if(InputFlowBit(f))  x |= mask;
    mask <<= 1;
  }
  return( x );
}


/*-------------------------------------------------------------------*/
void OutputFlowValue(FLOW* f, int x, int bits)
{
  int i;

  for(i=1; i<=bits; i++)
  {
    OutputFlowBit(f, x & 0x01);
    x >>= 1;
  }
}

/*-------------------------------------------------------------------*/
BYTE *GetCurrentFlowPointer(FLOW *f)
{
	if (f->FlowType!=F_MEMORY) return 0;
	return ((BYTE *)f->Flow)+f->CurrCnt;
}

/*----------------------- Buffer and string I/O -------------------- */

/*All functions here return the number of read (written) bytes*/

int ReadBufferFromFlow(void *buffer, int size, FLOW *f)
{
	int i;
	BYTE *buf=(BYTE*)buffer;

	for(i=0; i<size; i++)
	{
		buf[i] = (BYTE)f->InputByte(f);
		if (!f->Status) break;
	}
	return i;
}

/*-------------------------------------------------------------------*/

int WriteBufferToFlow(void *buffer, int size, FLOW *f)
{
	int i;
	BYTE *buf=(BYTE*)buffer;

	for(i=0; i<size; i++)
	{
		f->OutputByte(f,buf[i]);
		if (!f->Status) break;
	}
	return i;
}

/*-------------------------------------------------------------------*/

int ReadStringFromFlow(char *st, int max_size, FLOW *f)
{
	int i;

	for(i=0; i<max_size-1; i++)
	{
		st[i] = (char)f->InputByte(f);
		if (st[i]==0) return i+1;
		if (!f->Status) break;
	}
	st[i]=0;
	return i+1;
}

/*-------------------------------------------------------------------*/

int WriteStringToFlow(char *st, FLOW *f)
{
	return WriteBufferToFlow(st,strlen(st)+1,f);
}
