#ifndef GENIO_H
#define GENIO_H

#include "owntypes.h"

typedef enum {F_FILE=0, F_MEMORY=1} FLOWTYPE;


struct idFLOW
{
  FLOWTYPE FlowType;
  void *Flow;
  int Status;

  ULONG MaxCnt;
  ULONG CurrCnt;
  ULONG BitsInQueue;
  ULONG BitQueue;

  void (*OutputByte) (struct idFLOW*, int);
  int  (*InputByte) (struct idFLOW*);
  int  (*SeekFlow) (struct idFLOW*, ULONG offset, int origin);
  long (*TellFlow) (struct idFLOW*);

  // BITSTREAM
  int  (*InputBit)(struct idFLOW*);
  void (*OutputBit)(struct idFLOW*, int);
  int  (*InputValue)(struct idFLOW*, int);
  void (*OutputValue)(struct idFLOW*, int, int);
  void (*FlushInput)(struct idFLOW*);
  void (*FlushOutput)(struct idFLOW*);
};

typedef struct idFLOW FLOW;

typedef void  (*TOutputByte) (FLOW*, int);
typedef int   (*TInputByte) (FLOW*);
typedef int   (*TSeekFlow) (FLOW*, ULONG, int);
typedef long  (*TTellFlow) (FLOW*);


/*------------------------------------------------------------------ */
void InitFlow(FLOW* f, FLOWTYPE type, void* ptr, ULONG size);
int  ReadULongFromFlow(FLOW* f, ULONG* value, int bytecount);
int  ReadSLongFromFlow(FLOW* f, long* value, int bytecount);
int  WriteULongToFlow(FLOW* f, ULONG value, int bytecount);
int  WriteSLongToFlow(FLOW* f, long value, int bytecount);

BYTE *GetCurrentFlowPointer(FLOW *f);

#define WRITEUINT(f, value, bytes) {if(WriteULongToFlow((f),(ULONG)(value),(bytes))!=(bytes)) return(NO); else;}
#define WRITESINT(f, value, bytes) {if(WriteSLongToFlow((f),(long)(value),(bytes))!=(bytes)) return(NO); else;}
// this macro requires one temporary variable ULONG tmp_ulong
#define READUINT(f, value, bytes) {if(ReadULongFromFlow((f),&tmp_ulong,(bytes))!=(bytes)) return(NO); else (value)=tmp_ulong;}
// this macro requires one temporary variable long tmp_long
#define READSINT(f, value, bytes) {if(ReadSLongFromFlow((f),&tmp_long,(bytes))!=(bytes)) return(NO); else (value)=tmp_long;}

int  SeekFlow(FLOW *f, long offset, int origin);
long TellFlow(FLOW *f);

/*----------------------- Buffer and string I/O -------------------- */
/*All functions here return the number of read (written) bytes*/
int ReadBufferFromFlow(void *buffer, int size, FLOW *f);
int WriteBufferToFlow(void *buffer, int size, FLOW *f);
int ReadStringFromFlow(char *st, int max_size, FLOW *f);
int WriteStringToFlow(char *st, FLOW *f);

/* ----------------------- Bit stream I/O -------------------------- */
typedef int  (*TInputBit)(FLOW*);
typedef void (*TOutputBit)(FLOW*, int);
typedef int  (*TInputValue)(FLOW*, int);
typedef void (*TOutputValue)(FLOW*, int, int);
typedef void (*TFlushInput)(FLOW*);
typedef void (*TFlushOutput)(FLOW*);

void OutputByteToFile(FLOW*, int);
int  InputByteFromFile(FLOW*);
void OutputByteToMem(FLOW*, int);
int  InputByteFromMem(FLOW*);
int  SeekFile(FLOW*, long, int);
long TellFile(FLOW*);
int  SeekMem(FLOW*, long, int);
long TellMem(FLOW*);


#endif
