#if ! defined(__IMAGE_H)
#define __IMAGE_H

#include "file.h"
#include "pgm.h"
#include "owntypes.h"
#include "color.h"

/*---------------------  Image data structure  -----------------------*/

#define  LINESMAX         20000
#define  LINESAUTOMAX     128
#define  LINESMIN         10
#define  NAMEMAXLENGTH    127
#define  SAFEZONE         5


typedef  enum  { MEMORY=0, DISK=1, TEMPORARY=2}               TARGETTYPE;
typedef  enum  { WHOLEIMAGE=0, BUFACCESS=1, AUTOMATIC=2 }     BUFFERING;
typedef  enum  { BINMODE=0, GRAYMODE=1, RGBMODE=2}            MODETYPE;

typedef  struct {
  ULONG      Left;
  ULONG      Right;
  ULONG      Upper;
  ULONG      Lower;
} BOUNDS;

typedef struct
{
  FILE*          FilePointer;
  TARGETTYPE     TargetType;
  ULONG          ImageSizeX;
  ULONG          ImageSizeY;
  ULONG          ImageBytesX;
  BOUNDS         Bound;
  FILETYPE       FileType;
  FILEMODE       FileMode;
  BUFFERING      Buffering;
  MODETYPE       ModeType;
  ULONG          DefaultColor;
  int            Max;
  ULONG          LinesInBuffer;
  ULONG          ScrollSize;
  ULONG          FirstBufLine;
  ULONG          LastBufLine;
  ULONG          BufCount;
  ULONG          LinesReadTotal;
  BYTE           **data;
}IMAGE;


/*-----------------------  General routines  -------------------------*/
// Initializes image, buffer size (in lines) is passed to function
YESNO ImageInit(char InputName[], IMAGE* image, FILEMODE mode, int AllowOverWrite, 
                BUFFERING Buffering, MODETYPE ModType, TARGETTYPE TargetType,
                ULONG BufferSize);

YESNO ImageDone(IMAGE* image);
void FlushImage(IMAGE* image);
YESNO ImageRewind(IMAGE* image);

YESNO SetImageSize(IMAGE* image, ULONG x, ULONG y);
YESNO SetDefaultColor(IMAGE* image, ULONG color);
YESNO SetImageBounds(IMAGE* image, ULONG bx, ULONG by, ULONG cx, ULONG cy);
YESNO EmptyWholeImage(IMAGE* image);

YESNO ExportImage(IMAGE* image, BYTE* target, ULONG Color);
YESNO ExportImageEx(IMAGE* image, BYTE* target, ULONG Color, ULONG BackColor);
YESNO ExportImageBin(IMAGE* image, BYTE* target, int Plane);
YESNO CopyImage(IMAGE* destination, IMAGE* Source, char* Name, TARGETTYPE TargetType);
YESNO DumpImage(IMAGE* image, char* Name);


/*------------------  Pixel level  -------------------------*/

int       GetBitPixel(IMAGE* image, ULONG x, ULONG y);
void      PutBitPixel(IMAGE* image, ULONG x, ULONG y, int pixel);
int       GetGrayPixel(IMAGE* image, ULONG x, ULONG y);
void      PutGrayPixel(IMAGE* image, ULONG x, ULONG y, int pixel);
RGBPIXEL  GetRGBPixel(IMAGE* image, ULONG x, ULONG y);
void      PutRGBPixel(IMAGE* image, ULONG x, ULONG y, RGBPIXEL pixel);

/*------------------  Universal routines -------------------*/
int       GetPixel(IMAGE* image, ULONG x, ULONG y);
void      PutPixel(IMAGE* image, ULONG x, ULONG y, int pixel);
RGBPIXEL  GetRGBPixelEx(IMAGE* image, ULONG x, ULONG y);
void      PutRGBPixelEx(IMAGE* image, ULONG x, ULONG y, RGBPIXEL pixel);


#endif /* __IMAGE_H */
