/*-------------------------------------------------------------------*/
/* MEMCTRL.C       Timo Kaukoranta                                   */
/*                 Eugene Ageenko (corrections)                      */
/*                                                                   */
/* Memory control; allocation and deallocation                       */
/*                                                                   */
/*-------------------------------------------------------------------*/

#define ProgName        "MEMCTRL"
#define VersionNumber   "Version L.03"
#define LastUpdated     "04.04.00"

/*
#define VersionNumber   "Version 0.04"
#define LastUpdated     "27.8.96"
*/

/* ----------------------------------------------------------------- */

#include "memctrl.h"
#include "interfc.h"


/*========================= Memory handling =========================*/


#if ! defined(NO_MEMORYCONTROLLER)

#include <stdio.h>
#include <stdlib.h>

#if defined(AMOUNTCONTROL)
unsigned long _allocated    = 0L; /* # of bytes that are currently allocated. */
unsigned long _maxallocated = 0L; /* Max. # of bytes that has been allocated */
#endif
unsigned long _nallocated   = 0L; /* # of allocation calls */
unsigned long _ndeallocated = 0L; /* # of deallocation calls */


/* ----------------------------------------------------------------- */


void memoryreport(void)
{
#if defined(AMOUNTCONTROL)
  PrintErrorMessage("Currently allocated:%lu",_allocated);
  PrintErrorMessage("Max. allocated:     %lu",_maxallocated);
#endif
  PrintErrorMessage("Allocation calls:   %lu",_nallocated);
  PrintErrorMessage("Deallocation calls: %lu",_ndeallocated);
  PrintErrorMessage("Difference in calls:%lu",_nallocated - _ndeallocated);
}


/* ----------------------------------------------------------------- */


unsigned long usedmemory(void)
{
#if defined(AMOUNTCONTROL)
  return(_maxallocated);
#else
  return(0);
#endif
}


/* ----------------------------------------------------------------- */


int checkmemory(void)
{
  if( _nallocated != _ndeallocated
#if defined(AMOUNTCONTROL)
      || _allocated != 0L
#endif
     )
  {
    PrintErrorMessage("Memory handling stinks:");
    memoryreport();
    return(-1);
  }
  else
  {
    return(0);
  }
}


/*-------------------------------------------------------------------*/


void* allocatememory(size_t nbytes, char* file, int line)
{
#if defined(AMOUNTCONTROL)
  /* sizeof(size_t) is multiplied by 2 because otherwise the word */
  /* alignment do not work always. Or something like that...      */
  size_t* temp;
  temp  = (size_t*)malloc(nbytes + 2*sizeof(size_t));

  if( temp == NULL )
  {
    ErrorMessage("allocate: Out of memory in file `%s' in line %i. "
                    "Requiring %li", file, line, nbytes);
    checkmemory();

#if defined(EXIT_WHEN_MEMORYCONTROLLER_FAILS)
    exit( 1 );
#else
    return( NULL );
#endif

  }
  *temp = nbytes;
  _allocated += (unsigned long)nbytes;
  if( _allocated > _maxallocated )
  {
    _maxallocated = _allocated;
  }
  _nallocated++;
  return( (void*)(temp + 2) );

#else /* ! defined(AMOUNTCONTROL) */

  size_t* temp = (size_t*)malloc(nbytes);

  if( temp == NULL )
  {
    ErrorMessage("allocate: Out of memory in file `%s' in line %i. "
                    "Requiring %i", file, line, nbytes);
    checkmemory();

#if defined(EXIT_WHEN_MEMORYCONTROLLER_FAILS)
    exit( 1 );
#else
    return( NULL );
#endif

  }
  _nallocated++;
  return( (void*)(temp) );
#endif
}


/*-------------------------------------------------------------------*/


void deallocate(void* block)
{
#if defined(AMOUNTCONTROL)
  size_t* temp;
#endif

  if( block != NULL )
  {
#if defined(AMOUNTCONTROL)
    temp = (size_t*)(block) - 2;
    _allocated -= (unsigned long) *(temp);
    free(temp);
#else
    free(block);
#endif
    _ndeallocated++;
  }
}

#endif   /* ! defined(NO_MEMORYCONTROLLER) */

