/***********************************************
**
** Filename: HistoData.cpp
**   
**  HistoData Class 
**
**
** Date  
**    27.02.2003   
**    01.03.2004   Polar Quantization -> See CHistoDataPolar
**    22.04.2004   Product Scala Quantization -> See CHistoDataXY
**
** Written by: 
**    Alexander Kolesnikov 
**
** Email:
**   koles@cs.joensuu.fi
**
**  Copyright (c) 2003, 2004  Alexander Kolesnikov 

**
**********************************************/
#include "stdafx.h"

#include "HistoData.h"


/************************************************************
* 
************************************************************/

void heapSort(double *numbers, ULONG array_size)
{
  long i;
  double temp;

  for (i = (array_size / 2)/*-1*/; i >= 0; i--)
  {
	  siftDown(numbers, i, array_size-1);
  }

  for (i = array_size-1; i >= 1; i--)
  {
    temp = numbers[0];
    numbers[0] = numbers[i];
    numbers[i] = temp;
    siftDown(numbers, 0, i-1);
  }
}


void siftDown(double *numbers, ULONG root, ULONG bottom)
{
  ULONG done, maxChild;
  double temp;

  done = 0;
  while ((root*2 <= bottom) && (!done))
  {
    if (root*2 == bottom)
      maxChild = root * 2;
    else if (numbers[root * 2] > numbers[root * 2 + 1])
      maxChild = root * 2;
    else
      maxChild = root * 2 + 1;

	if (numbers[root] < numbers[maxChild])
    {
      temp = numbers[root];
      numbers[root] = numbers[maxChild];
      numbers[maxChild] = temp;
      root = maxChild;
    }
    else
      done = 1;
  }
}


CHistoData::CHistoData() 
{
   int terr = InitMemory(0);
}


CHistoData::CHistoData(long int size) 
{
    int      terr = InitMemory(size);
	long int n;

    for(n = 0; n < m_nSize; n++)
	{
        m_pX[n]     = (double)n*(double)n;
	    m_pHisto[n] = 1/m_nSize; // (long int)(rand());
	    m_pE2[n]    = 0; 
	}

    // Make One Big Cluster for the data;
	InitClusters();

	//
	CalcMoments();
}

CHistoData::CHistoData(DArray *pX) 
{
	int       terr=0;
	long int  size_in, count=0, i, j, jj, n;
	int       run, len;
    NArray    *pHisto_tmp = NULL;
//	double    tmp;
	DArray    *pBuff_tmp=NULL;

	//
	size_in = pX->size;

	terr = InitMemory(size_in);
	if(terr != 0) 
	{
		return;
	}

	///////////////////////////////
	pBuff_tmp = CreateDArray(pX->size);
	if(pBuff_tmp == NULL)
	{
        FreeMemory();

		return;
	}

    for(i = 0; i < size_in; i++)
	{
       pBuff_tmp->value[i] = pX->value[i];
	}

	heapSort(pBuff_tmp->value,size_in);

	// How many unique values?
    count = 0;
	for(n = 1; n <= size_in-1; n++) 
	{
	   if(pBuff_tmp->value[n] != pBuff_tmp->value[n-1]) 
	   {
		   count += 1;
	   }
	}
    
	m_nSize = count + 1;

	////////////////////////////////////////
    // Copy data
	for(n = 0; n <= m_nSize-1; n++) 
	{
	   m_pHisto[n] = 1;
	}

	m_pX[0] = pBuff_tmp->value[0];
    jj  = 1;
    run = 0;
    for(n = 1; n <= size_in-1; n++) 
	{
	   if(pBuff_tmp->value[n] != pBuff_tmp->value[n-1]) 
	   {
		  m_pX[jj] = pBuff_tmp->value[n];
          
		  if(run == 1) // Stop the Run
		  {
		     m_pHisto[jj-1] = len;
			 run            = 0;
             len            = 0;
		  }
		  jj += 1;
	   }
	   else 
	   {
		   if(run == 0)
		   {
			   run = 1;
			   len = 2;
		   }
		   else // if(run == 1)
		   {
               len += 1;
		   }

	   }
	}//for(n

    // Make One Big Cluster for the data;
	InitClusters();

	//
	CalcMoments();

	// Free memory
    DestroyDArray(&pBuff_tmp);
}


CHistoData::CHistoData(DArray *pData, NArray *pHisto) 
{
    int      terr = 0;
	long int n, size = pData->size;

	//m_nSize = (size >= 2) ? size : 2;
    terr = InitMemory(size);
	if(terr != 0) return;

    for(n = 0; n < m_nSize; n++)
	{
        m_pX[n]     = pData->value[n];
	    m_pHisto[n] = (unsigned long int)pHisto->value[n];
	    m_pE2[n]    = 0; 
	}

    // Make One Big Cluster for the data;
	InitClusters();

	//
	CalcMoments();
}

/*
CHistoData::CHistoData(ULONG *pHisto, long int size) 
{
	int      terr = 0;
    long int n, max_n;
    
	if((pHisto == NULL) && (size < 0)) return;
   
    // Max:
	max_n = size-1;
	for(n = size-1; n >= 0; n--)
	{
       if(pHisto[n] != 0) 
	   {
		   max_n = n;

		   break;
	   }
	}

    ////
    m_nSize = (max_n+1 >= 1) ? max_n+1 : 1;
	
	terr = InitMemory(m_nSize);
	if(terr != 0) return;

    // Make One Big Cluster for the data;
	InitClusters();
}
*/

CHistoData::CHistoData(double *pX, long int size) 
{
	int      terr=0, run, len;
	long int size_in, count=0, i, j, jj, n;

    NArray   *pHisto_tmp = NULL;
	DArray   *pBuff_tmp=NULL;

	size_in = size;

	///////////////////////////////
	pBuff_tmp = CreateDArray(size_in);
	if(pBuff_tmp == NULL)
	{
        FreeMemory();
		return;
	}

	// Copy Data
    for(i = 0; i < size_in; i++)
	{
       pBuff_tmp->value[i] = pX[i];
	}


	heapSort(pBuff_tmp->value,pBuff_tmp->size);
	count = 0;
	for(n = 1; n <= size_in-1; n++) 
	{
	   if(pBuff_tmp->value[n] != pBuff_tmp->value[n-1]) 
	   {
		   count += 1;
	   }
	}
    
	m_nSize = count + 1;

	terr = InitMemory(m_nSize);
	if(terr != 0) 
	{
		return;
	}

	////////////////////////////////////////
    // Copy data
	for(n = 0; n < m_nSize; n++) 
	{
	   m_pHisto[n] = 1;
	}

	m_pX[0] = pBuff_tmp->value[0];
    jj  = 1;
    run = 0;
    for(n = 1; n <= size_in-1; n++) 
	{
	   if(pBuff_tmp->value[n] != pBuff_tmp->value[n-1]) 
	   {
		  m_pX[jj] = pBuff_tmp->value[n];

          
		  if(run == 1) // Stop the Run
		  {
		     m_pHisto[jj-1] = len;
			 run            = 0;
             len            = 0;
		  }
		  jj += 1;
	   }
	   else 
	   {
		   if(run == 0)
		   {
			   run = 1;
			   len = 2;
		   }
		   else // if(run == 1)
		   {
               len += 1;
		   }

	   }
	}//for(n

    // Make One Big Cluster for the data;
	InitClusters();

	//
	CalcMoments();

	// Free memory
    DestroyDArray(&pBuff_tmp);
	IsErrorEst = 0;
}


CHistoData::CHistoData(double *pX, unsigned long int *pHisto, 
					   long int size) 
{
	int       terr=0;
    long int  n, i, j, jj, SS=0, len, run, count, size_full;
//	double    tmp;
	DArray    *pBuff_tmp = NULL;

	if((pX == NULL) || (pHisto == NULL) || (size < 0))
	{
		return;
	}

	////////////////////////////////////////////////
    for(n = 0, SS = 0; n < size; n++)
	{
       SS += pHisto[n];
	 //  printf(">><< n=%d) H=%ld, SS=%ld\n", n, pHisto[n], SS);
	}
    size_full = SS;
		
	///////////////////////////////
	pBuff_tmp = CreateDArray(size_full);
	if(pBuff_tmp == NULL)
	{
		return;
	}

    for(n = 0, jj=0; n < size; n++)
	{
	   for(i = 0; i < (long int)pHisto[n]; i++)
	   {
           pBuff_tmp->value[jj] = pX[n];
		   jj += 1;
	   }//for(i
	}//for(n

	// Sort input data
/*	for(i = 0; i <= size_full-2; i++)
	{
        for(j = i+1; j <= size_full-1; j++)
		{
           if(pBuff_tmp->value[i] > pBuff_tmp->value[j])
		   {
              tmp = pBuff_tmp->value[j];
			  pBuff_tmp->value[j] = pBuff_tmp->value[i];
			  pBuff_tmp->value[i] = tmp;
		   }
		}//for(j
	}//for(i*/

	/*for(i = 0; i < size_full; i++)
	{
       printf("<S> i=%ld) X=%.12f \n", 
		           i, pBuff_tmp->value[i]); 
	}
    */
	heapSort(pBuff_tmp->value,pBuff_tmp->size);
	// How many unique values?
    count = 0;
	
	count = size_full;

	for(n = 1; n < size_full; n++) 
	{
	 /*  if(pBuff_tmp->value[n] != pBuff_tmp->value[n-1]) 
	   {
		   count += (long int)1;
	   }
	   */
	   if(pBuff_tmp->value[n] == pBuff_tmp->value[n-1]) 
	   {
		   count -= 1;
	//	   printf("===> n=%ld) %.12f %.12f  \n", 
	//		   n, pBuff_tmp->value[n], pBuff_tmp->value[n-1]);
	   }
	}
    
	m_nSize = count; // + 1;

	//printf("m_nSize=%ld\n", m_nSize) ;

	terr = InitMemory(m_nSize);
	if(terr != 0) 
	{
		return;
	}

	////////////////////////////////////////
    m_pX[0] = pBuff_tmp->value[0];
    jj  = 1;
	//run = 1;
	len = 1;

    for(n = 1; n <= size_full-1; n++) 
	{
	   if(pBuff_tmp->value[n-1] != pBuff_tmp->value[n]) 
	   {
		  m_pX[jj] = pBuff_tmp->value[n];
          m_pHisto[jj-1] = len;

		 /* if(run == 1) // Stop the Run
		  {
			 run = 0;
		  }
		  */

		  jj  += 1;
		  len  = 1;
	   }
	   else   // IF(pBuff_tmp->value[n] == pBuff_tmp->value[n-1]))
	   {
		   len += 1;

	       /*printf("RL n=%ld) jj=%ld, len=%d, %.12f %.12f \n", 
				   n, jj, 
				   len, pBuff_tmp->value[n-1], pBuff_tmp->value[n]);
		   */
	   }

	   if(n == size_full-1) 
	   {
		 m_pHisto[jj-1] = len;
	   }
	}//for(n

/*	for(i = 0; i < m_nSize; i++)
	{
		printf("*** i=%d) X=%.6f, H=%ld\n",  i, m_pX[i], m_pHisto[i]);
	}
*/	
    // Make One Big Cluster for the data;
	InitClusters();

	//
	CalcMoments();

    // Free memory
    DestroyDArray(&pBuff_tmp);
}

/**********************************************************
*   Destroy HistoData:
*     Free memory and free struct pHisto;
*
*    Use after CreateHistoData() method
*  
**********************************************************/
CHistoData::~CHistoData()
{
    FreeMemory();
}

/**********************************************
*  Reallocate memory for arrays x[], h[], and ind[] of the HistoData;
*
*   Use this functon after CreateHistoData() only!
*
*   Parameters:
*    (HistoData*) pHisto    Histogram 
*    (int)        size      size of HistoData

*   Return code of error 
*      0: OK
*     -1: ERROR
*************************************************/
int CHistoData::ReallocateMemory(long int size) 
{  
   int      terr = 0;
   long int len=0;

   if(size < 0)
   {
	   return -1;
   }
   
   FreeMemory();          

   // Allocate memory 
   m_nSize = size;

   terr = InitMemory(m_nSize);
   if(terr != 0)
   {
	   FreeMemory();          

	   return -1;
   }

   // Make One Big Cluster for the data;
   InitClusters();
   
   return terr;  
}


/*******************************************************************
*  PRIVATE method
*  Allocate memory for arrays x[], h[], and ind[] of the HistoData;
*
*   Use this functon for object HistoData;
*   Don't use after using CreateHistoData(), because
*   the method CreateHistoData() envokes InitMemoryOfHistoData 
*
*   Parameters:
*    (HistoData*) pHisto    pointer to HistoData 
*    (int)        size     size of HistoData

*   Return code of error 
*      0: OK
*     -1: ERROR
*************************************************/
int CHistoData::InitMemory(long int size) 
{  
   int terr=0;

   if(size < 0) return -1;

   m_nSize = size; 
   
   
   // Allocate memory 
   m_pX     = new double[m_nSize]; 
 
   // Statistics
   m_pHisto = new unsigned long int[m_nSize]; 
   m_pProby = new double[m_nSize]; 
   m_pE2    = new double[m_nSize]; 

   //
   m_pMomentZero = new double[m_nSize]; 
   m_pMomentOne  = new double[m_nSize]; 
   m_pMomentTwo  = new double[m_nSize]; 

   // Cluster data
   m_pClusterStart     = new long int[m_nSize];  
   m_pClusterIndex     = new ULONG[m_nSize];  
   m_pClusterNumber    = new ULONG[m_nSize];  
   m_pClusterMeans     = new double[m_nSize];  
   m_pClusterVariances = new double[m_nSize];  
   m_pClusterThrs      = new double[m_nSize];  

   
   
   // A.A.
   m_pClustersList = NULL;
   m_pClustersThr  = NULL;
   IsErrorEst = 0;

   /* Check the memory allocation */
   if((m_pHisto == NULL)         || (m_pProby == NULL) || 
      (m_pClusterStart == NULL)  || (m_pClusterIndex == NULL) || 
	  (m_pClusterNumber == NULL) || (m_pClusterMeans == NULL) || 
	  (m_pMomentTwo == NULL)     || (m_pX == NULL) || 
	  (m_pMomentZero == NULL)    || (m_pMomentOne == NULL) ||
	  (m_pE2 == NULL)            || (m_pClusterThrs == NULL) ||
	  (m_pClusterVariances == NULL) )
   {
	   FreeMemory(); 

	   terr = -1;
   }


   return terr;
}

/***********************************************
*  PRIVATE method
**
*  Free memory of the HistoData 
*    Free memory for x, h, and  ind arrays;
*    Set size to -1;
* 
*
*  Parameter:
*    (HistoData*)  pHisto   
*
*  Return:
*      none
***********************************************/
void CHistoData::FreeMemory() 
{
      if(m_pX != NULL)
	  {
	     delete [] m_pX;
	     m_pX = NULL;
	  }

      if(m_pE2 != NULL)
	  {
	     delete [] m_pE2;
	     m_pE2 = NULL;
	  }

      if(m_pHisto != NULL)
	  {
	     delete [] m_pHisto;
	     m_pHisto = NULL;
	  }

	  if(m_pProby != NULL)
	  {
	     delete [] m_pProby;
	     m_pProby = NULL;
	  }

	  // Cluster data
      if(m_pClusterStart != NULL)
	  {
	     delete [] m_pClusterStart;
	     m_pClusterStart = NULL;
	  }

	  // Cluster data
      if(m_pClusterIndex != NULL)
	  {
	     delete [] m_pClusterIndex;
	     m_pClusterIndex = NULL;
	  }

      if(m_pClusterNumber != NULL)
	  {
	     delete [] m_pClusterNumber;
	     m_pClusterNumber = NULL;
	  }

      if(m_pClusterMeans != NULL)
	  {
	     delete [] m_pClusterMeans;
	     m_pClusterMeans = NULL;
	  }

      if(m_pClusterVariances != NULL)
	  {
	     delete [] m_pClusterVariances;
	     m_pClusterVariances = NULL;
	  }

      if(m_pClusterThrs != NULL)
	  {
	     delete [] m_pClusterThrs;
	     m_pClusterThrs = NULL;
	  }

	  // Cumulative moments
	  if(m_pMomentZero != NULL)
	  {
	     delete [] m_pMomentZero;
	     m_pMomentZero = NULL;
	  }

	  if(m_pMomentOne != NULL)
	  {
	     delete [] m_pMomentOne;
	     m_pMomentOne = NULL;
	  }

	  if(m_pMomentTwo != NULL)
	  {
	     delete [] m_pMomentTwo;
	     m_pMomentTwo = NULL;
	  }

	  
	  // A.A.
	  if(m_pClustersList != NULL)
	  {
		delete [] m_pClustersList;
		m_pClustersList = NULL;
	  }

	  if(m_pClustersThr != NULL)
	  {
		delete [] m_pClustersThr;
		m_pClustersThr = NULL;
	  }
}

int CHistoData::InitClusters()
{
   int       terr=0;
   long int  n;

   if((m_nClustersNumber < 0) || (m_pX == NULL) || 
	  (m_pProby == NULL)  )
   {
       return -1;
   }

   // One Cluster;
   m_nClustersNumber  = 1;
   m_pClusterStart[0] = 0; 

   for(n = 0; n < m_nSize; n++)
   {
	   m_pClusterIndex[n]     = 0;
	   m_pE2[n]               = 0;
	   m_pClusterThrs[n]      = 0;
   	   m_pClusterMeans[n]     = 0;
   	   m_pClusterVariances[n] = 0;
   }//for(n
   m_pClusterIndex[0] = 1;

   return terr;
}


/*********************************************************
*   Make copy 
*  
*
*   Return code of error  
*      0: OK
*     -1: ERROR
*
**********************************************************/
/*int CHistoData::Copy(CHistoData* pHisto_dst)
{
	/ *
	int           err, k, size_src, size_dst;
    ULONG     *ph_s, *ph_d; 
    ULONG     *pn_s, *pn_d;

    if((pHisto_src == NULL) || (pHisto_dst == NULL))   
	{
       return -1;
	}
     
	size_src = pHisto_src->size;
    if(size_src  < 0)
	{
       return -1;
	}

	size_dst = pHisto_dst->size;
	if(size_src != size_dst)
	{
	   err = ReallocateMemoryOfHistoData(pHisto_dst, size_src);
	   if(err != 0)
	   {
		  return -1;
	   }
	}

	ph_s = (ULONG*)pHisto_src->h; 
	pn_s = (ULONG*)pHisto_src->ind; 
	ph_d = (ULONG*)pHisto_dst->h; 
	pn_d = (ULONG*)pHisto_dst->ind; 

	for(k = 0; k < size_src; k++) 
	{
	   *ph_d++ = *ph_s++; 
	   *pn_d++ = *pn_s++; 
	}

	pHisto_dst->clusters_number = pHisto_src->clusters_number;  
*
	return 0; //err;
}
*/

long int CHistoData::GetSize()
{
	return m_nSize;
}

/******************************************************************
*
** 
******************************************************************/
double*  CHistoData::GetThresholds(/*long int Ms*/)
{

    int terr = CalcClustersList();
	if(terr != 0) return NULL;
	
	terr = CalcThresholds();
	if(terr != 0) return NULL;

    return m_pClusterThrs;
} 

/******************************************************************
*
** 
******************************************************************/
double*  CHistoData::GetMeans(long int Ms)
{
//    int terr = CalcClustersList();
//	if(terr != 0) return NULL;

	long int i;

	long int current_index = 0;


	for(i=2;i<Ms;i++)
		current_index+=i;

	for(i=current_index;i<current_index+Ms;i++)
		m_pClusterMeans[i-current_index] = m_pClustersList[i];

    return m_pClusterMeans;
} 

/******************************************************************
*
** 
******************************************************************/
double* CHistoData::GetVariances()
{
    int terr = CalcClustersList();
	if(terr != 0) return NULL;
	
	terr = CalcVariances();
	if(terr != 0) return NULL;

    return  m_pClusterVariances;
} 




/************************************************************************
**  Type Dump of the HistoData 
**
**
** Return code of error 
**    0: OK
**   -1: ERROR
*************************************************************************/
int CHistoData::TypeHistoData()
{  
    long int k, h;
    int      ind;	  
	double   C0, C1, C2, p, x;

    for(k = 0; k < m_nSize; k++)
	{
		h   =  m_pHisto[k];
		p   =  m_pProby[k];
	    x   =  m_pX[k];
	    ind =  m_pClusterIndex[k];
		C0  =  m_pMomentZero[k];
		C1  =  m_pMomentOne[k];
		C2  =  m_pMomentTwo[k];

		//if(h != 0)
		{
            printf("k=%ld) x=%f C0=%f, C1=%f, C2=%f\n", 
				    k, x, C0, C1, C2);
		}
	}
	
	double  d_myu, d_omega, Q;

    for(long int j = 0; (j < 100) && (j < m_nSize); j++)
 	{
          for(long int i = 0; (i < 100) && (i < m_nSize); i++)
		  {
			  if(i <= j)
			  {
  
			    if(i==0)
				{
			       d_myu   = m_pMomentOne[j];
                   d_omega = m_pMomentZero[j];
				}
				else
				{
			       d_myu   = m_pMomentOne[j]  -  m_pMomentOne[i-1];
                   d_omega = m_pMomentZero[j] -  m_pMomentZero[i-1];
				}
 
			     Q = d_myu*d_myu/d_omega;
			  }
			  else
			  {
				  Q = 0;
			  }

	          //printf("%.4f ", Q);
		  }
		  //printf("\n");

	}

	return 0;
}

/*void CHistoData::Type()
{  
    long int n, j;

	CalcClustersList();

    for(n = 0; n < m_nSize; n++)
	{
		j = m_pClusterNumber[n]; 

        printf("n=%ld) x=%.4f (%d); #Cj=%d at %ld: <Myu_j>=%.12f \n", 
				    n, m_pX[n], m_pClusterIndex[n], 
					j, 
					m_pClusterStart[j], 
					m_pClusterMeans[j]);
	}//for(n

	printf("m_nClustersNumber: %d\n", m_nClustersNumber);

}*/

/******************************************************************
**  Calc list of segments 
**  Call the function after for any changes of approximating
** 
** Return code of error 
**    0: OK
**   -1: ERROR
******************************************************************/
int CHistoData::CalcClustersList()
{  
   long int  n, jj, n1, n2;
   int     k;
   double  d_omega, d_myu;

   for(n = 0; n < m_nSize; n++)
   {
      m_pClusterStart[n] = 0;
   }//for(n   

   m_nClustersNumber = 0;

   jj = 0;
   for(n = 0; n < m_nSize; n++)
   {
       if((m_pClusterIndex[n] == 1))
	   {
           m_pClusterStart[jj] = n;
		   //printf("j=%ld: n=%ld\n", jj, n);
		   jj++;
	   }//if(

   }//for(n   

   m_nClustersNumber = jj;

   if(m_nClustersNumber >= 1)
   {
      // Partition & Means
      for(k = 0; k < m_nClustersNumber; k++)
	  {
      	 n1 = m_pClusterStart[k];  

    	 if(k <= m_nClustersNumber-2)
		 {
	         n2 = m_pClusterStart[k+1]-1;   
		 }
	     else
		 {
            n2 = m_nSize-1;    
		 }

		 /////////////////////////////////////////////
		 if(n1 == 0) 
		 {
		    d_myu   = m_pMomentOne[n2];
		    d_omega =  m_pMomentZero[n2];
		 }
		 else
		 {
		    d_myu   = m_pMomentOne[n2] - m_pMomentOne[n1-1];
		    d_omega =  m_pMomentZero[n2] - m_pMomentZero[n1-1];
		 }

		 m_pClusterMeans[k] = d_myu/d_omega;

		 for(n = n1; n <= n2; n++)
		 {
             m_pClusterNumber[n] = k;
    	 }//for(n
	  }//for(k   

//////////////////////////////////////////////
      CalcThresholds();
      CalcVariances();
	  CalcMeans();

      double  x, p, myu_x, E2=0;
	  int     j_x;

      double S_p = 0;

  	for(long int i = 0; i < m_nSize; i++)
	{
        x     = m_pX[i];
	    j_x   = m_pClusterNumber[i];
	    myu_x = m_pClusterMeans[j_x];
		p     = m_pProby[i];
        E2    += p*(x - myu_x)*(x - myu_x);

        //printf("%ld) in %ld: X=%.8f->%.8f; (%.16f) E2=%.12f\n", 
		//	      i, j_x,    x, myu_x,  p, E2);

		S_p += p;

	}//for(i

	//printf("S_p=%.f\n", S_p);
  }
   else
   {
      for(n = 0; n < m_nSize; n++)
	  {
          m_pClusterNumber[n] = 0;
	  }//for(n   

   }

  return 0;
}


/******************************************************************
**  Type Dump of Clusters
**
** 
** Return code of error 
**    0: OK
**   -1: ERROR
******************************************************************/
int CHistoData::TypeClustersList()
{
	 int       k;
	 long int  n1, n2;
	 double    S_sigma=0, S_myu=0; //, myu, sigma, ;

	 printf("Size: %2d; Clusters: %d; \n", m_nSize, m_nClustersNumber);

     for(k = 0; k < m_nClustersNumber; k++)
	 {
	     n1 = m_pClusterStart[k];

		 if(k <= m_nClustersNumber-2)
		 {
	        n2 = m_pClusterStart[k+1]-1; 
		 }
		 else
		 {
            n2 = m_nSize-1;
		 }

		 printf("j=%3d) [%5ld, %5ld]->%3ld, Thr=%12.4f: Myu=%10.4f; Sgm=%16.12f\n", 
			      k, n1, n2, 
				  n2-n1+1, 
				  m_pClusterThrs[k], 
				  m_pClusterMeans[k],
				  m_pClusterVariances[k]);

		 S_sigma += m_pClusterVariances[k];
	 }
     printf("Sigma=%.12f\n", S_sigma);
	 printf("---------------------------\n");

	 return 0;
}

/******************************************************************
**  Type Dump of Clusters
**
** 
** Return code of error 
**    0: OK
**   -1: ERROR
******************************************************************/
void CHistoData::TypeThresholds()
{
	 long int  k;

	 //printf("Histogram size: %2d; thresholds number: %d: \n", m_nSize, m_nClustersNumber-1);

     for(k = 0; k < m_nSize; k++)
	 {
		 if(m_pClusterIndex[k] == 1)
		 {
		    printf("%.12f ", m_pX[k]);
		 }
		 //printf("k=%d) Ind=%d \n", k, m_pClusterIndex[k]);
	 }

	 printf("\n");
}
 

int CHistoData::CalcThresholds()
{
     int   terr =0;
	 long int  jj, k;

	 if((m_pX == NULL) || (m_nSize <= 0) || (m_nClustersNumber <= 0) || 
        (m_pClusterIndex == NULL) )
	 {
		 return -1;
	 }  

	 ///////////////////////////////////////////
	 jj = 0;
     for(k = 0; k < m_nSize; k++)
	 {
		 if(m_pClusterIndex[k] == 1)
		 {
		    m_pClusterThrs[jj] = m_pX[k];
			jj +=1;
		 }
	 }

	 return terr;
}
 

int CHistoData::CalcMeans()
{
     int      terr =0;
	 long int jj, k;
	 double   val;

	 if((m_pX == NULL) || (m_nSize <= 0) || (m_nClustersNumber <= 0))
	 {
		 return -1;
	 }  

	 ///////////////////////////////////////////
	 jj = 0;
     for(k = 0; k < m_nClustersNumber; k++)
	 {
		 terr = CalcMeansForCluster(k, &val);   
		 m_pClusterMeans[k] = val;
		 //printf("k=%d: Means = %.8f \n", k, val);
	 }

	 return terr;
}
 
int CHistoData::CalcVariances()
{
     int      terr =0;
	 long int jj, k;
	 double    val;

	 if((m_pX == NULL) || (m_nSize <= 0) || (m_nClustersNumber <= 0))
	 {
		 return -1;
	 }  

	 ///////////////////////////////////////////
	 jj = 0;
     for(k = 0; k < m_nClustersNumber; k++)
	 {
		 terr = CalcVarianceForCluster(k, &val);   
		 m_pClusterVariances[k] = val;
	 }

	 return terr;
}


/***************************************************************
*  Load Histo Data data from ASCII file
*
*
**  File format:
**  DARA    - label
**  x1
**  x2
**   ....
**
**  Parameters:
**    char*       filename  -- input filename
**    HistoData*  pHisto
**
** Return code of error 
**    0: OK
**   -1: ERROR
***************************************************************/
int CHistoData::Load(char* filename)
{   int       err=0;
    long int  k, size;
    FILE*     pFile = NULL; 
	float     b;

    if(filename == NULL) 
	{
	   return -1;
	}

    pFile = OpenForReadingAndCheckLabel(filename, "DATA");
	if(pFile == NULL) 
    {
	   return -1;
    }

    // Load number of points; 
	fscanf(pFile, "%d", &size);
    if(size < 0) 
    {
  	    return -1;
    } 

	err = ReallocateMemory(size);
	if(err != 0)
    {
  	    return -1;
    } 

    for(k = 0; k < size; k++) 
    {
		fscanf(pFile, "%f", &b);
		m_pX[k] = (double)b;
	}

    //
    for(k = 0; k < size; k++) 
    {
		m_pHisto[k] = 1;
	}

/*
	fscanf(pFile, "%f", &b);
	m_pX[0] = (double)b/1000;
	m_pClusterIndex[0] = 1;
	fscanf(pFile, "%d", &c1);

    for(k = 1; k < size; k++) 
    {
		fscanf(pFile, "%f", &b);
		m_pX[k] = (double)b/1000;
		//m_pX[k] = (double)k;
		fscanf(pFile, "%d", &c2);

		if(c2 == c1)
		{
			m_pClusterIndex[k] = 1;
		}
		else
		{
            m_pClusterIndex[k] = 0;
		}
		c1 = c2;
	}//for(k	
*/
    m_nSize = size;

	//m_pClusterIndex[m_nSize-1] = 0;

    CalcClustersList();
	CalcMoments();

    return err;
}


/***************************************************************
*  Load Histo Data data from ASCII file
*
*
**  File format:
**  DARA    - label
**  x1
**  x2
**   ....
**
**  Parameters:
**    char*       filename  -- input filename
**    HistoData*  pHisto
**
** Return code of error 
**    0: OK
**   -1: ERROR
***************************************************************/
int CHistoData::LoadPartition(char* filename)
{   int      err=0, c1, c2;
    long int k, size;
    FILE*    pFile = NULL; 
	float    b;
	double   dE;

    if(filename == NULL) 
	{
	   return -1;
	}

    pFile = OpenForReadingAndCheckLabel(filename, "PARTITION");
	if(pFile == NULL) 
    {
	   return -1;
    }

    // Load number of points; 
	fscanf(pFile, "%d", &size);
    if(size < 0) 
    {
  	    return -1;
    } 

	err = ReallocateMemory(size);
	if(err != 0)
    {
  	    return -1;
    } 

	fscanf(pFile, "%f", &b);
	m_pX[0] = (double)b;
	m_pClusterIndex[0] = 1;
	fscanf(pFile, "%d", &c1);

    for(k = 1; k < size; k++) 
    {
		fscanf(pFile, "%f", &b);
		m_pX[k] = (double)b;
		//m_pX[k] = (double)k;
		fscanf(pFile, "%d", &c2);

		if(c2 != c1)
		{
			m_pClusterIndex[k] = 1;
		}
		else
		{
            m_pClusterIndex[k] = 0;
		}
		c1 = c2;
	}//for(k	

    m_nSize = k;

    m_pClusterIndex[m_nSize-1] = 0;

	CalcMoments();

/*    for(k = 0; k < size; k++) 
    {
       printf("k=%ld) X=%.1f ind=%ld \n", k, m_pX[k], m_pClusterIndex[k]); 
	}
*/
	///
	CalcClustersList();
	CalcVariance(&dE);
	printf("Loaded Partition: MSE = %.12f \n\n", dE);

    return err;
}


/**************************************************************************
*  Open file 'filename' for reading and check label
*
*  Return error code:
*     0: OK
**************************************************************************/
FILE* CHistoData::OpenForReadingAndCheckLabel(char* filename, char *label)
{
   FILE*  pFile = NULL;
   char   szLabel[32];

   if(filename == NULL)
   {
	   return NULL;
   }

   // Open file 
   pFile = fopen(filename, "rt");
   if(pFile == NULL) 
   {
	   return NULL;
   }

	// Load & Check the Label 
   	fscanf(pFile, "%s", szLabel);
	if(strcmp(label, szLabel) != 0) 
    {
	   return NULL;
    }

	return pFile;
}


/***********************************************************************************
* 1) Calculate Variance for the HistoData
*
*
*  Result:
*   1. (double *)var_val  -- Variance value
*
*  Return error code: 
*    0  -- OK
*    -1 -- ERROR
*************************************************************************************/
int CHistoData::CalcVariance(double *var_val)
{
    int       err=0;
    long int  k, M;
	double    value;

	M = m_nClustersNumber;

    *var_val = 0;

	if(m_nClustersNumber < 0)
    {
	    *var_val = -1;

		return -1;
	}

	for(k = 0; k < M; k++)
	{
        err = CalcVarianceForCluster(k, &value);
		//err = CalcVarianceForCluster_B(k, &value);

	    if(err == 0) 
		{
		 	  *var_val += value;
		}
		else 
		{
            err      = -1;
		    *var_val = -1;

              break;
		}//
        //printf("cluster #%2ld) n_start: %ld: %.12f -> %.12f \n", 
		  //     k, this->m_pClusterStart[k], value, *var_val);		  		  

	}//for(k

    return err;
}

/*****************************************************************
*  PRIVATE methods
*****************************************************************/ 

/*************************************************************************
*  Calc variance for Cluster
*
*  Parameters:
*    1. int k     -- Cluster number
*
*  Result:
*    2. (double*) var_val      -- variance value
*
*  Return code of error:
*    0: OK
*   -1: ERROR
*******************************************************************************/
int CHistoData::CalcVarianceForCluster(int k, double *var_val)
{   
	int       err=0;
	long int  n1, n2;
	double    d_omega, d_myu, d_lambda;

	*var_val = 1.0e10;

	if(m_pClusterStart == NULL)
	{
		return -1;
	}

	n1 = m_pClusterStart[k];  

	if(k <= m_nClustersNumber-2)
	{
	   n2 = m_pClusterStart[k+1]-1;   
	}
	else
	{
       n2 = m_nSize-1;    
	}

	if(n1 == n2)
	{
   	   *var_val = 0;  

	 //  printf("A: k=%ld) [%ld, %ld]->%.12f \n", k, n1, n2, *var_val);

	   return 0;
	}

    d_omega  = m_pMomentZero[n2]; 
	d_myu    = m_pMomentOne[n2]; 
	d_lambda = m_pMomentTwo[n2]; 

	if(n1 > 0)
	{
       d_omega  -= m_pMomentZero[n1-1]; 
	   d_myu    -= m_pMomentOne[n1-1]; 
	   d_lambda -= m_pMomentTwo[n1-1]; 
	}
	
    if(d_omega != 0)
	{
	   *var_val = d_lambda - (d_myu*d_myu)/d_omega;
	}
	else
	{
       *var_val = -1;
        err = -3;
	}

    return err;
}

/*************************************************************************
*  Calc variance for Cluster
*
*  Parameters:
*    1. int k         -- Cluster number
*
*  Result:
*    2. (double*) var_val      -- Means value
*
*  Return code of error:
*    0: OK
*   -1: ERROR
*******************************************************************************/
int CHistoData::CalcMeansForCluster(int k, double *var_val)
{   
	int       err=0;
	long int  n1, n2;
	double    d_omega, d_myu;

	*var_val = 1.0e10;

	n1 = m_pClusterStart[k];  

	if(k <= m_nClustersNumber-2)
	{
	   n2 = m_pClusterStart[k+1]-1;   
	}
	else
	{
       n2 = m_nSize-1;    
	}

    d_omega  = m_pMomentZero[n2]; 
	d_myu    = m_pMomentOne[n2]; 

	if(n1 > 0)
	{
       d_omega  -= m_pMomentZero[n1-1]; 
	   d_myu    -= m_pMomentOne[n1-1]; 
	}
	
    *var_val = d_myu/d_omega;

    return err;
}

int CHistoData::CalcVarianceForCluster_B(int k, double *var_val)
{
	int       err=0;
	long int  n1, n2, N1, n;
	double    omega, myu;

	*var_val = 1.0e10;

	n1 = m_pClusterStart[k];  

	if(k <= m_nClustersNumber-2)
	{
	   n2 = m_pClusterStart[k+1]-1;   
	}
	else
	{
       n2 = m_nSize-1;    
	}

	if(n1 == n2)
	{
   	   *var_val = 0;  
	    myu = m_pX[n1];
        printf("B: k=%ld)  [%ld, %ld]->%.4f (myu=%.3f)\n", k, n1, n2, *var_val, myu);	   
		
		return 0;
	}

	N1 = 0;
	myu = 0;
	for(n = n1; n <= n2; n++)
	{
       myu += m_pX[n]*m_pHisto[n];
	   N1  += 1;
	}
//	myu = myu/(double)N1;

	omega = 0;
	for(n = n1; n <= n2; n++)
	{
       omega += (m_pX[n]-myu)*(m_pX[n]-myu)*m_pHisto[n];
	}

	*var_val = omega; ///m_nSize;

    printf("B: k=%ld) [%ld, %ld]->%.4f (myu=%.3f) \n", k, n1, n2, *var_val, myu);

    return err;
}


int CHistoData::CalcVarianceFromPToP(double x1, double x2, double *var_val)
{   
	long int n1=-1, n2=-1;
	double   omega_1, myu_1, lambda_1, omega_2, myu_2, lambda_2, 
		     d_omega, d_myu, d_lambda;

	*var_val = -1;

	n1 = FindIndex(x1);
	n2 = FindIndex(x2);

	if((n1 == -1) || (n2 == -1))
	{
		return -1;
	}

	if(n1 == n2)
	{
   	   *var_val = 0;  

	   return 0;
	}

    // CalcVarianceFromPToP(i+1, n, &E2);
	omega_2  = m_pMomentZero[n2+1]; 
	myu_2    = m_pMomentOne[n2+1]; 
	lambda_2 = m_pMomentTwo[n2+1]; 

	if(n1 >= 0)
	{
        omega_1  = m_pMomentZero[n1]; 
		myu_1    = m_pMomentOne[n1]; 
	    lambda_1 = m_pMomentTwo[n1]; 
    }
	else
	{
        omega_1  = 0; 
		myu_1    = 0; 
	    lambda_1 = 0; 
	}

    d_omega  = omega_2  - omega_1; 
	d_myu    = myu_2    - myu_1;  
	d_lambda = lambda_2 - lambda_1; 


	if(d_omega != 0)
	{
	   *var_val = d_lambda - (d_myu*d_myu)/d_omega; // - (d_myu*d_myu)/d_omega; 
	}
	else
	{
       *var_val = -1;
	}

	if(*var_val != -1)
	{
	  // printf("---- n=[%ld,%ld]: Om:(%.6f,%.6f) Myu:(%.6f,%.6f) Lambda:(%.6f,%.6f) => Sigma=%.6f\n", 
		//         n1, n2, omega_1, omega_2,  myu_1, myu_2, lambda_1, lambda_2, *var_val);
	}

    return 0;
}



/******************************************************************
* Save Thresholds file 'filename'
*
* Return code of error 
*    0: OK
*   -1: ERROR
*******************************************************************/
int CHistoData::SaveThresholds(char* filename)
{ 
	/*
	  int      err=0;
	double   S=0; 
    FILE*    pFile = NULL;
    char     THRESHOLDS_LABEL[] = "THRESHOLDS";

    if((pHisto == NULL) || (filename == NULL))
	{
		 return -1;
	}

    if(pHisto->size < 0) 
	{
		 return -1;
	}

	// Open file; 
	pFile = fopen(filename, "w");
    if(pFile == NULL)
	{
		 return -1;
	}

	err = CalcClustersList(pHisto);
	if(err != 0)
	{
		 return -1;
	}

	// Save Label 
    fprintf(pFile, "%s\n", THRESHOLDS_LABEL);

    // Save number of Thresholds; 
	fprintf(pFile, "%d\n", pHisto->clusters_number-1);

    // Save values 
    for(int k = 0; k < pHisto->clusters_number-1; k++) 
	{
        fprintf(pFile, "%ld\n", pHisto->cluster_end[k]);
	}
*/
    return 0; // err;
}


/////////////////////////////////////////////////////////
int CHistoData::CalcMoments()
{
	int     err=0;
	long int  n, jj=0;
	double  p, S0=0.0, x; 
    DArray  *pBuff_X=NULL, *pBuff_H=NULL;

	//SortData();
 
    // MomentZero for the whole Histo;
	for(n = 0, S0=0.0; n < m_nSize; n++)
	{
	    S0 += (double)m_pHisto[n];
	}

	// Probability
	for(n = 0; n < m_nSize; n++)
	{
       m_pProby[n] = (double)m_pHisto[n] / S0;
	}

	// Moments
    p = m_pProby[0];
    x = m_pX[0];

	m_pMomentZero[0] = p;
	m_pMomentOne[0]  = p*x;
	m_pMomentTwo[0]  = p*x*x;

	n = 0;

/*    printf("n=%d) %.8f: (%.8f %.8f %.8f)\n", 
		       n, x , m_pMomentZero[n], m_pMomentOne[n], m_pMomentTwo[n]);
*/

	for(n = 1; n <= m_nSize-1; n++)
	{
	   p = m_pProby[n];
	   x = m_pX[n];

       m_pMomentZero[n] = p         + m_pMomentZero[n-1];
       m_pMomentOne[n]  = p * x     + m_pMomentOne[n-1] ;
       m_pMomentTwo[n]  = p * x * x + m_pMomentTwo[n-1] ;

	   m_pClusterNumber[n] = 0;

/*	   printf("n=%ld) %.8f: (%.8f %.8f %.8f)\n", 
		       n, x , m_pMomentZero[n], m_pMomentOne[n], m_pMomentTwo[n]);
*/	 
	}//for(n

	return 0;
}

double CHistoData::CalcMoment(int num)
{
	int     err=0;
	double  a;
   	
    if(num == 1) 
	{
        a = m_pMomentOne[m_nSize-1];
	}
	else
	{
		a = m_pMomentTwo[m_nSize-1];
	}

	return a;
}


int CHistoData::SortData()
{  
	long int          i, j;
	unsigned long int tmp2;
	double            tmp;

	for(i = 0; i <= m_nSize-2; i++)
	{
        for(j = i+1; j <= m_nSize-1; j++)
		{
           if(m_pX[i] > m_pX[j])
		   {
              tmp = m_pX[j];
			  m_pX[j] = m_pX[i];
			  m_pX[i] = tmp;
		   
              tmp = m_pProby[j];
			  m_pProby[j] = m_pProby[i];
			  m_pProby[i] = tmp;

              tmp2 = m_pHisto[j];
			  m_pHisto[j] = m_pHisto[i];
			  m_pHisto[i] = tmp2;
		  }

		}//for(j
	}//for(i

	return 0;
}

/*
int CHistoData::SortData(double *pX, long int size)
{  
	long int i, j;
	double   tmp;

	for(i = 0; i <= size-2; i++)
	{
        for(j = i+1; j <= size-1; j++)
		{
           if(pX[i] > pX[j])
		   {
              tmp   = pX[j];
			  pX[j] = pX[i];
			  pX[i] = tmp;
		   }
		}//for(j
	}//for(i

	return 0;
}

*/	
void CHistoData::MongeTest(int *res)
{
	long int   i, j;
	int     count=0;
    double  a1, a2, a3, a4, b1, b2;


	for(i = 0; i < m_nSize-2; i++) 
	{
		for(j = i+1; j < m_nSize-1; j++) 
		{
			CalcVarianceFromPToP(i,   j+1, &a1);
			CalcVarianceFromPToP(i+1, j, &a2);
			CalcVarianceFromPToP(i, j, &a3);
			CalcVarianceFromPToP(i+1, j+1, &a4);
            b1 = a1 + a2;
			b2 = a3 + a4;


			if((a1 != -1) && (a2 != -1) && (a3 != -1) && (a4 != -1))
			{
  			 //  printf("i=%d, j=%d) [%.6f, %.6f][%.6f, %.6f]  %.6f and %.6f ", 
			//	       i, j, a1, a2, a3, a4, b1, b2);

			   if(b1 >= b2)
			   {
               //   printf(" ==> OK!\n\n");
				  
			   }
			   else
			   {

                if((b1 != 0.0) || (b2 != 0.0))
				{
  			      printf("i=%d, j=%d)  %.6f and %.6f from [%.6f, %.6f][%.6f, %.6f]   \n", 
			   	       i,j, b1,b2,  a1,a2,a3,a4);

				   count +=1;
				}
			   }

			}

			
		}//for(j
	}//for(i


	*res = (count == 0) ? 1 : 0;

	printf("Count=%d\n", count);
}


long int CHistoData::GetClustersNumber()
{
	return m_nClustersNumber;
}

void CHistoData::CalcMatrix()
{
	long int  i, j;
	double  a;

	for(i = 0; i < m_nSize-2; i++) 
	{

        for(j = 0; j <= i; j++) 
		{
  			printf(" 0 "); 
		}//for(j

		for(j = i+1; j < m_nSize-1; j++) 
		{
			CalcVarianceFromPToP(i,   j+1, &a);
  			printf("%.6f ", a);
		}//for(j

		printf("\n"); 

	}//for(i

}


long int CHistoData::FindIndex(double x)
{
	long int j, n=-1; 
	double   d;

	for(j = 0; j <= m_nSize-2; j++)
	{
	   d = (m_pX[j] + m_pX[j+1])/2;

	   if((m_pX[j] <= x) && (x < d))
	   {
		   n = j;

		   break;
	   }

	   if((d <= x) && (x <= m_pX[j+1]))
	   {
		   n = j+1;

		   break;
	   }

	}//for(j

	return n;
}



int CHistoData::CalcUniformClustersList()
{  
   long int  n, jj, n1, n2;
   int     k;
//   double  d_omega, d_myu;

   for(n = 0; n < m_nSize; n++)
   {
      m_pClusterStart[n] = 0;
   }//for(n   

   m_nClustersNumber = 0;

   jj = 0;

   for(n = 0; n < m_nSize; n++)
   {
       if((m_pClusterIndex[n] == 1))
	   {
           m_pClusterStart[jj] = n;
		   jj++;
	   }//if(

   }//for(n   

   m_nClustersNumber = jj;

   if(m_nClustersNumber >= 1)
   {
      // Partition & Means
      for(k = 0; k < m_nClustersNumber; k++)
	  {
      	 n1 = m_pClusterStart[k];  

    	 if(k <= m_nClustersNumber-2)
		 {
	         //n2 = m_pClusterStart[k+1]-1;   
			 n2 = m_pClusterStart[k+1]-1;   
		 }
	     else
		 {
            n2 = m_nSize-1;    
		 }


//		 m_pClusterMeans[k] = d_myu/d_omega;
		 double step = (m_pX[m_nSize-1] - m_pX[0])/(double)m_nClustersNumber;
		 m_pClusterMeans[k] = ((m_pX[0] + k*step) + (m_pX[0] + (k+1)*step))/(double)2;

		 for(n = n1; n < n2; n++)
		 {
             m_pClusterNumber[n] = k;
    	 }//for(n
	  }//for(k   

	  m_pClusterNumber[m_nSize-1] = m_nClustersNumber-1;

//////////////////////////////////////////////
	  CalcThresholds();
//    CalcUniformVariances();
//	  CalcMeans();

/*      double  x, p, myu_x, E2=0;
	  int     j_x;

      double S_p = 0;

  	for(long int i = 0; i < m_nSize; i++)
	{
        x     = m_pX[i];
	    j_x   = m_pClusterNumber[i];
	    myu_x = m_pClusterMeans[j_x];
		p     = m_pProby[i];
        E2    += p*(x - myu_x)*(x - myu_x);

        //printf("%ld) in %ld: X=%.8f->%.8f; (%.16f) E2=%.12f\n", 
		//	      i, j_x,    x, myu_x,  p, E2);

		S_p += p;

	}//for(i

	//printf("S_p=%.f\n", S_p);*/
  }
   else
   {
      for(n = 0; n < m_nSize; n++)
	  {
          m_pClusterNumber[n] = 0;
	  }//for(n   

   }

  return 0;
}


int CHistoData::CalcUniformVarianceForCluster(int k, double *var_val)
{   
	int       err=0;
	long int  n1, n2;
	double    d_omega, d_myu, d_lambda;

	*var_val = 1.0e10;

	if(m_pClusterStart == NULL)
	{
		return -1;
	}

	n1 = m_pClusterStart[k];  

	if(k <= m_nClustersNumber-2)
	{
	   n2 = m_pClusterStart[k+1]-1;   
	}
	else
	{
       n2 = m_nSize-1;    
	}

	if(n1 == n2)
	{
   	   *var_val = 0;  

	 //  printf("A: k=%ld) [%ld, %ld]->%.12f \n", k, n1, n2, *var_val);

	   return 0;
	}

	double step = (m_pX[m_nSize-1] - m_pX[0])/(double)m_nClustersNumber;
	double Centr = ((m_pX[0] + k*step) + (m_pX[0] + (k+1)*step))/(double)2;

	if(n2<m_nSize-1)
		n2 = n2-1;

    d_omega  = m_pMomentZero[n2]; 
	d_myu    = m_pMomentOne[n2]; 
	d_lambda = m_pMomentTwo[n2]; 

	if(n1 > 0)
	{
       d_omega  -= m_pMomentZero[n1-1]; 
	   d_myu    -= m_pMomentOne[n1-1]; 
	   d_lambda -= m_pMomentTwo[n1-1]; 
	}
	
    if(d_omega != 0)
	{
	   *var_val = d_lambda - 2*d_myu*Centr + (Centr*Centr)*d_omega;
	}
	else
	{
       *var_val = -1;
        err = -3;
	}

    return err;
}


int CHistoData::CalcUniformVariance(double *var_val)
{
    int       err=0;
    long int  k, M;
	double    value;

	M = m_nClustersNumber;

    *var_val = 0;

	if(m_nClustersNumber < 0)
    {
	    *var_val = -1;

		return -1;
	}

	for(k = 0; k < M; k++)
	{
        err = CalcUniformVarianceForCluster(k, &value);
		//err = CalcVarianceForCluster_B(k, &value);

	    if(err == 0) 
		{
		 	  *var_val += value;
		}
		else 
		{
            err      = -1;
		    *var_val = -1;

              break;
		}//
        //printf("cluster #%2ld) n_start: %ld: %.12f -> %.12f \n", 
		  //     k, this->m_pClusterStart[k], value, *var_val);		  		  

	}//for(k

    return err;
}

/******************************************************************

                            E N D
******************************************************************/