/***********************************************
**
** Filename: HistoData.h
**   
** Basic functions for HistoData object
**
**
** Date  
**    
**	  27.02.2003   
**    01.03.2004  Polar Quantization   
**    17.09.2004  General Codebook
**
** Written by: 
**    Alexander Kolesnikov 

**
**
** Email:
**   koles@cs.joensuu.fi
**
**
**  Copyright (c) 2003,2004 Alexander Kolesnikov 
**
**********************************************/


#ifndef HISTODATA_H
#define HISTODATA_H

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>

typedef unsigned long int ULONG;

#include "VectSTD.h"
#include "SearchBasic.h"

class CHistoData 
{
public:

  /**********************************************/
  /* Constructor, Destructor, Copy method       */
  /**********************************************/
	   
	CHistoData();
    CHistoData(long int size);
    CHistoData(unsigned long int *pHisto, long int size); 

	CHistoData(DArray *pX); 
//	CHistoData(DArray *pPhi, DArray *pR); 
	CHistoData(DArray *pX, NArray *pHisto); 

	//
    CHistoData(double *pX, unsigned long int *pHisto, long int size); 
	CHistoData(double *pX, long int size); 
	
	CHistoData(CHistoData& HistoData);

	virtual ~CHistoData();

	int    Copy(CHistoData* pHisto_dst);

    int    InitMemory(long int size);
    void   FreeMemory(); 
    int    ReallocateMemory(long int size); 

    ///////////////////////
	// L O A D
    ////////////////////
    int    Load(char* filename);
	int    LoadPartition(char* filename);
    int    SaveThresholds(char* filename);
    FILE*  OpenForReadingAndCheckLabel(char* filename, char *label);

    ///////////////////////
	// G E T
	////////////////////////
    long int   GetSize();
	long int   GetClustersNumber();

    double*    GetThresholds();
	double*    GetMeans(long int Ms);
	double*    GetVariances();
	long int*  GetClusterNumbers();
	
    ////////////////////////////////////////
	//   C A L C
    ////////////////////////////////////////
    int        CalcMoments();
	int        CalcMeans();
	int        CalcVariances();
	int        CalcThresholds();
    int        InitClusters();
	int        CalcClustersList();

	double     CalcMoment(int num);
 	int        CalcVariance(double *dErr);
    int        CalcVarianceForCluster(int num, double *value);	
	int        CalcVarianceForCluster_B(int num, double *value);	
	int        CalcVarianceFromPToP(double x1, double x2, double *value);
    int        CalcMeansForCluster(int num, double *value);	

	////////////////////////////////////////
	int       SortData();
	int       SortData(double *pX, long int size);
	long int  FindIndex(double x);

	////////////////////////////////////////
	// S E T
    ////////////////////////////////////////
    int    SetClusterStart(double x);

    int    Uniform(ULONG M, double *value);
	int    Uniform_old(ULONG M, double *value);
    int    MaxLloyd(ULONG M, double *value);

    void   MongeTest(int *res);
    int    MakeTest(long int size); 
    void   CalcMatrix();  // !!!

	////////////////////////////////////////
	// T Y P E
    ////////////////////////////////////////
//	void   Type();
    void   TypeThresholds();
    int    TypeHistoData();
    int    TypeClustersList();
    int    TypeBounds(Bounds *pB);

	////////////////////////////////////////
	//  Dyanmic Programming
    ////////////////////////////////////////
    int FullSearchOne2One(int Ms, int W, double *val);
    int FullSearchOne2Many(int Ms, int W, double *val);

    //int BoundedSearchOne2Many(int Ms, int W, double *val);
    int CalcBounds(Bounds *pB);
    int CalcBoundsOne2One(Bounds *pB);
    int CalcBoundsOne2Many(Bounds *pB);

    //////////////////////////////////////////////////////////
    // Dynamic Programming for MSE Min
	//////////////////////////////////////////////////////////
	int DP_MSE(ULONG Ms, int version, int type, double *val);
 
    int DP_MSE_min(Bounds *pBnd, DTable  *pD2, NTable *pNodes);
    int DP_MSE_min_row(Bounds *pBnd, DTable  *pD2, NTable *pNodes);
    int DP_MSE_min_row_right(Bounds *pBnd, DTable  *pD2, NTable *pNodes);

	/*implemented by Akimov Alexander*/
	
	
	
	// Weak time-complexity reduciing: for m*n*log(n) 
	int DP_MSE_min_row_reduced(Bounds *pBnd, DTable  *pD2, NTable *pNodes);

	
	
	
	//SMAWK algorithm!!! 
	//all this procedures are used in SMAWK-variant
	int DP_MSE_min_row_SMAWK(Bounds *pBnd, DTable  *pD2, NTable *pNodes);
	double GetMatrixValue(Bounds *pBnd, DTable  *pD2, long int row_index, long int col_index);

	int SMAWK_Reduce(Bounds *pBnd, DTable *pD2, long int Nr, long int Nc,
							 long int Level, long int **pointers, long int *summator);

	long int GetRealColumnIndex(Bounds *pBnd, long int **pointers, long int Level, long int index);
	long int GetRealColumnIndex(long *real_indexes, long int index);
	long int GetRealRowIndex(Bounds *pBnd, long int Level, long int index);
	int SMAWK_MinCompute(Bounds *pBnd, DTable *pD2, NTable *pNodes);
	int CalcUniformClustersList();
	int CalcUniformVarianceForCluster(int k, double *var_val);
	int CalcUniformVariance(double *var_val);
	long int m_nCurClusterInd;
// private:
 
	// Data
    long int  m_nSize;           // Number of points     
	double    *m_pX;             // Points

     // Statistic data
    ULONG     *m_pHisto;         // Histogram data 
	double    *m_pProby;          // Probability    

	// Moments
	double   *m_pMomentZero;     // Cumulative Moment Zero    
	double   *m_pMomentOne;      // Cumulative Moment One    
    double   *m_pMomentTwo;      // Cumulative Moment Two    
	
	double  *m_pE2;               // MSE  

    ////////////////////////////
	// Cluster data
    long int  m_nClustersNumber; // Number of Clusters 
	
	long int     *m_pClusterStart;      // Start point of the Cluster (Threshold) 
	ULONG      *m_pClusterIndex;      // Index data: 1=ClusterStart;      
	ULONG      *m_pClusterNumber;     // Cluster Nodes;       ???
	double   *m_pClusterMeans;      // Cluster Means;      
	double   *m_pClusterVariances;  // Cluster Variances;      
    double   *m_pClusterThrs;       // Cluster Thresholds;      


	//added by AA
	double *m_pClustersList;
	double *m_pClustersThr;
	int IsErrorEst;

};


void heapSort(double *numbers, ULONG array_size);
void siftDown(double *numbers, ULONG root, ULONG bottom);

//#define CB_NUM 10


#endif /* HISTODATA_H */
