/***********************************************
**
** Filename: HistoDataPolar.cpp
**   
**  HistoData Class 
**
**
** Date  
**    27.02.2003   
**    01.03.2004   Polar Quantization
**    20.04.2004   new Class HistoDataPolar
**    11.05.2004   Continue
**    29.08.2004   Accelerating of quantizaing
**
** Written by: 
**    Alexander Kolesnikov 
**
** Email:
**   koles@cs.joensuu.fi
**
**  Copyright (c) 2003, 2004  Alexander Kolesnikov 
**
**
**********************************************/
#include "stdafx.h"
#include "../Modules/owntypes.h"
#include "../Modules/param.h"



extern "C"
{
	#include "../Modules/cb.h"
}

#include "HistoDataPolar.h"

//#define  PI 3.1415926535897931

#define  MIN_VAL     1.0e35
#define  UNDEF_VAL   -99

/************************************************************
* 
************************************************************/
/* UNDER CONSTRUCTON */ 
CHistoDataPolar::CHistoDataPolar() 
{
   m_nSize      = 0;
   
   m_pR    = NULL; 
   m_pPhi  = NULL; 

   m_pHData_R   = NULL;
   m_pHData_Phi = NULL;

   m_pRadiusMomentZero = NULL; 
   m_pRadiusMomentOne  = NULL; 
   m_pRadiusMomentTwo  = NULL; 

   m_pPhaseMomentZero = NULL; 
   m_pPhaseMomentOne  = NULL; 
   m_pPhaseMomentTwo  = NULL; 

   m_pPolarMomentZero = NULL; 
   m_pPolarMomentOne  = NULL; 
   m_pPolarMomentTwo  = NULL; 

   m_pE2 = NULL;	     
}  


CHistoDataPolar::CHistoDataPolar(long int size) 
{
	InitMemory(m_nSize);
}


/*****************************************************
**
**
**
**
**
*****************************************************/
/* UNDER CONSTRUCTON */ 
CHistoDataPolar::CHistoDataPolar(DArray *pR, DArray *pPhi) 
{
	ULONG     size_in, count=0, i, j, n;
    NArray    *pHisto_tmp = NULL;
	double    tmp;

	//
	size_in = pPhi->size;
    m_pPhi  = NULL; 
   
	// SortByPhase();
	for(i = 0; i <= size_in-2; i++)
	{
        for(j = i+1; j <= size_in-1; j++)
		{
           if(pPhi->value[i] > pPhi->value[j])
		   {
              tmp = pPhi->value[j];
			  pPhi->value[j] = pPhi->value[i];
			  pPhi->value[i] = tmp;

              tmp = pR->value[j];
			  pR->value[j] = pR->value[i];
			  pR->value[i] = tmp;
		   }

		}//for(j
	}//for(i

	// How many unique values?
    count = 0;
	for(n = 1; n <= size_in-1; n++) 
	{
	   if(pPhi->value[n] != pPhi->value[n-1]) 
	   {
		   count += 1;
	   }
	}
    
	m_nSize = count + 1;

    ////////////////////////////////////////
	// Create the object
	InitMemory(m_nSize);
    m_pHData_R->CalcMoments();
	m_pHData_Phi->CalcMoments();
}

/*****************************************************
**
**
**
**
**
*****************************************************/
/* UNDER CONSTRUCTON */ 
CHistoDataPolar::CHistoDataPolar(DPolyline *pPoly)
{
	int       err = 0;
	long int  n; 
	double    dx, dy, r, phi;

	m_nSize = pPoly->size - 1;
	
	if((pPoly == NULL) || (pPoly->size < 0))
	{
		return;
	}

	// ???
    InitMemory(m_nSize);	

	if(err != 0)
	{
		return;
	}

    for(n = 0; n <= m_nSize-1; n++)
	{
       dx = pPoly->x[n+1] - pPoly->x[n];
	   dy = pPoly->y[n+1] - pPoly->y[n];

	   r  = sqrt(dx*dx + dy*dy);

	   // 1. Radius 
	   m_pR[n] = r;

	//   printf("%8f ", r);

	   if((dy >= 0) && (dx >= 0))
	   {
	       phi = asin(dy/r);
	   }

	   if((dy < 0) && (dx >= 0))
	   {
		   phi = 2*PI - asin((-dy)/r);
	   }

	   if((dy >= 0) && (dx < 0))
	   {
	       phi = PI - asin(dy/r);
	   }
	   if((dy < 0) && (dx < 0))
	   {
		   phi = PI + asin((-dy)/r);
	   }

	   // 2. Phase
	   m_pPhi[n] = phi;
	}

	// Hist is tmp
	// Make analysis of Histo  XXX
	for(n = 0; n < m_nSize; n++)
	{
		m_pHisto[n] = 1;
	}
  
	// Use Probability, not Histo!
	for(n = 0; n < m_nSize; n++)
	{
		m_pProby[n] = 1.0/m_nSize;
	}

	// 
	m_pHData_R   = new CHistoData(m_pR, m_nSize);
    m_pHData_Phi = new CHistoData(m_pPhi, m_nSize);

	// 
    SortDataByPhase();

   	CalcMoments(); 

}//CHistoDataPolar(DPolyline *pPoly)

CHistoDataPolar::CHistoDataPolar(double *pR, double *pPhi, long int size)
{
	int       err = 0;
	long int  n; 

	m_nSize = size;
	
    InitMemory(m_nSize);	

	if(err != 0)
	{
		return;
	}

    for(n = 0; n <= m_nSize-1; n++)
	{
	   // 1. Radius 
	   m_pR[n] = pR[n];

	   // 2. Phase
	   m_pPhi[n] = pPhi[n];
	}

	// Hist is tmp
	// Make analysis of Histo  XXX
	for(n = 0; n < m_nSize; n++)
	{
		m_pHisto[n] = 1;
	}
  
	// Use Probability, not Histo!
	for(n = 0; n < m_nSize; n++)
	{
		m_pProby[n] = 1.0/m_nSize;
	}

	// 
	m_pHData_R   = new CHistoData(m_pR, m_nSize);
    m_pHData_Phi = new CHistoData(m_pPhi, m_nSize);

	// 
    SortDataByPhase();

    CalcMoments(); 
}


CHistoDataPolar::CHistoDataPolar(double *pR, double *pPhi, 
								 unsigned long int *pHisto, 
								 long int size)
{
	int               err = 0;
	long int          n; 
	unsigned long int SS = 0;

	m_nSize = size;

    InitMemory(m_nSize);	
	if(err != 0)
	{
		return;
	}

    for(n = 0, SS = 0; n < m_nSize; n++)
	{
		SS += pHisto[n];
	}

	printf("SS=%ld", SS);
    
	for(n = 0; n < m_nSize; n++)
	{
	   // 1. Radius 
	   m_pR[n] = pR[n];

	   // 2. Phase
	   m_pPhi[n] = pPhi[n];

	   // 3. Histogram
	   	m_pHisto[n] = pHisto[n];

		// 4. Probability
		m_pProby[n] = (double)m_pHisto[n]/(double)SS;
	}

	// 
	m_pHData_R   = new CHistoData(m_pR,   m_pHisto, m_nSize);
    m_pHData_Phi = new CHistoData(m_pPhi, m_pHisto, m_nSize);

	// 
    SortDataByPhase();

    CalcMoments(); 
}

/**********************************************************
*   Destroy HistoData:
*     Free memory and free struct pHisto;
*
*    Use after CreateHistoData() method
*  
**********************************************************/
/* UNDER CONSTRUCTON */ 
CHistoDataPolar::~CHistoDataPolar()
{
    FreeMemory();
}

/**********************************************
*  Reallocate memory for arrays x[], h[], and ind[] of the HistoData;
*
*   Use this functon after CreateHistoData() only!
*
*   Parameters:
*    (HistoData*) pHisto    Histogram 
*    (int)        size      size of HistoData

*   Return code of error 
*      0: OK
*     -1: ERROR
*************************************************/
/* UNDER CONSTRUCTON */ 
int CHistoDataPolar::ReallocateMemory(ULONG size) 
{    ULONG  len=0;

   m_nSize   = (size > 0) ? size : 1;
   
   FreeMemory();          

   // Allocate memory 
   
   InitMemory(m_nSize);

/*     // One Cluster;
      m_nClustersNumber_R  = 1;
      m_pClusterStart_R[0] = 0; 
 	   

      for(ULONG n = 0; n < m_nSize; n++)
	  {
		   m_pE2[n]           = 0;
	  }
      m_pClusterIndex_R[0] = 1;
  */ 
      return 0;  
}


/*******************************************************************
*  PRIVATE method
*  Allocate memory for arrays x[], h[], and ind[] of the HistoData;
*
*   Use this functon for object HistoData;
*   Don't use after using CreateHistoData(), because
*   the method CreateHistoData() envokes InitMemoryOfHistoData 
*
*   Parameters:
*    (HistoData*) pHisto    pointer to HistoData 
*    (int)        size     size of HistoData

*   Return code of error 
*      0: OK
*     -1: ERROR
*************************************************/
/* UNDER CONSTRUCTON */ 
int CHistoDataPolar::InitMemory(ULONG size) 
{  int    terr=0;

   m_nSize = size;
   if(size < 0) return -1;

   // Allocate memory 
   m_pR   = new double[m_nSize]; 
   m_pPhi = new double[m_nSize]; 
 
   // Statistics
   m_pHisto = new ULONG[m_nSize]; 
   m_pProby = new double[m_nSize]; 
   m_pE2    = new double[m_nSize]; 

   //
   m_pPhaseMomentZero = new double[m_nSize]; 
   m_pPhaseMomentOne  = new double[m_nSize]; 
   m_pPhaseMomentTwo  = new double[m_nSize]; 

   m_pPolarMomentZero = new double[m_nSize]; 
   m_pPolarMomentOne  = new double[m_nSize]; 
   m_pPolarMomentTwo  = new double[m_nSize]; 

   /* Check the memory allocation */
   if((m_pHisto == NULL)           || (m_pProby == NULL) || 
	  (m_pR == NULL)               || (m_pPhi == NULL) || 
	  (m_pPhaseMomentZero == NULL) || (m_pPhaseMomentOne == NULL) ||
	  (m_pPhaseMomentTwo == NULL)  ||
	  (m_pPolarMomentZero == NULL) || (m_pPolarMomentOne == NULL) ||
	  (m_pPolarMomentTwo == NULL)  ||
	  (m_pE2 == NULL)   )  
   {
	   FreeMemory(); 

	   terr = -1;
   }

    return 0;
}


/***********************************************
*  PRIVATE method
**
*  Free memory of the HistoData 
*    Free memory for x, h, and  ind arrays;
*    Set size to -1;
* 
*
*  Parameter:
*    (HistoData*)  pHisto   
*
*  Return:
*      none
***********************************************/
/* UNDER CONSTRUCTON */ 
void CHistoDataPolar::FreeMemory() 
{
      if(m_pHData_R != NULL)
	  {
	     m_pHData_R->FreeMemory();
		 delete m_pHData_R;
	     m_pHData_R = NULL;
	  }

      if(m_pHData_Phi != NULL)
	  {
	     m_pHData_Phi->FreeMemory();
		 delete m_pHData_Phi;
	     m_pHData_Phi = NULL;
	  }

	  /////////////////////////////
      if(m_pR != NULL)
	  {
	     delete [] m_pR;
	     m_pR = NULL;
	  }

      if(m_pPhi != NULL)
	  {
	     delete [] m_pPhi;
	     m_pPhi = NULL;
	  }

	  /////////////////////////////
      if(m_pE2 != NULL)
	  {
	     delete [] m_pE2;
	     m_pE2 = NULL;
	  }

	  // Phase
      if(m_pPhaseMomentZero != NULL)
	  {
	     delete [] m_pPhaseMomentZero;
	     m_pPhaseMomentZero = NULL;
	  }

      if(m_pPhaseMomentOne != NULL)
	  {
	     delete [] m_pPhaseMomentOne;
	     m_pPhaseMomentOne = NULL;
	  }

      if(m_pPhaseMomentTwo != NULL)
	  {
	     delete [] m_pPhaseMomentTwo;
	     m_pPhaseMomentTwo = NULL;
	  }

	  // Polar
      if(m_pPolarMomentZero != NULL)
	  {
	     delete [] m_pPolarMomentZero;
	     m_pPolarMomentZero = NULL;
	  }

      if(m_pPolarMomentOne != NULL)
	  {
	     delete [] m_pPolarMomentOne;
	     m_pPolarMomentOne = NULL;
	  }

      if(m_pPolarMomentTwo != NULL)
	  {
	     delete [] m_pPolarMomentTwo;
	     m_pPolarMomentTwo = NULL;
	  }

}

/*********************************************************
*   Make copy 
*  
*
*   Return code of error  
*      0: OK
*     -1: ERROR
*
**********************************************************/
int CHistoDataPolar::Copy(CHistoDataPolar* pHisto_dst)
{
	/*
	int           err, k, size_src, size_dst;
    ULONG     *ph_s, *ph_d; 
    ULONG     *pn_s, *pn_d;

    if((pHisto_src == NULL) || (pHisto_dst == NULL))   
	{
       return -1;
	}
     
	size_src = pHisto_src->size;
    if(size_src  < 0)
	{
       return -1;
	}

	size_dst = pHisto_dst->size;
	if(size_src != size_dst)
	{
	   err = ReallocateMemoryOfHistoData(pHisto_dst, size_src);
	   if(err != 0)
	   {
		  return -1;
	   }
	}

	ph_s = (ULONG*)pHisto_src->h; 
	pn_s = (ULONG*)pHisto_src->ind; 
	ph_d = (ULONG*)pHisto_dst->h; 
	pn_d = (ULONG*)pHisto_dst->ind; 

	for(k = 0; k < size_src; k++) 
	{
	   *ph_d++ = *ph_s++; 
	   *pn_d++ = *pn_s++; 
	}

	pHisto_dst->clusters_number = pHisto_src->clusters_number;  
*/
	return 0; //err;
}


/************************************************************************
**  Type Dump of the HistoData 
**
**
** Return code of error 
**    0: OK
**   -1: ERROR
*************************************************************************/
/*int CHistoDataPolar::TypeHistoData()
{  
	m_pHData_R->TypeHistoData();
    m_pHData_Phi->TypeHistoData();
 
	return 0;
}
*/

void CHistoDataPolar::Type()
{  
/*    ULONG   n, j;

	CalcClustersList();

    for(n = 0; n < m_nSize; n++)
	{
		j = m_pClusterNumber[n]; 

        printf("n=%ld) r=%.4f (%d); #Cj=%d at %ld: <Myu_j>=%.12f \n", 
				    n, m_pR[n], m_pClusterIndex[n], 
					j, 
					m_pClusterStart[j], 
					m_pClusterMeans[n]);
	}//for(n

	printf("m_nClustersNumber: %d\n", m_nClustersNumber);
	*/

//	m_pHData_R->Type();
//	m_pHData_Phi->Type();
}
/*
int CHistoDataPolar::SetClusterStart(double x)
{
   int err=0;

   //m_pClusterIndex[n1] = 1;

   return err;
}
*/

/********************************************************
**  Make Bounding Corridor: the main function 
**
**    1) (DPolyline*)  pPoly -- polyline
**
**  Return error code: 
**    0:  OK
**   -1:  ERROR
********************************************************/
/*
int CHistoDataPolar::CalcBounds(Bounds *pB)
{   int   err = 0;


    if(pB == NULL)
	{
		return -1;
	}

    if(pB->type == 1)
	{
	   err = CalcBoundsOne2One(pB);
	}

    if(pB->type == 2)
	{
	   err = CalcBoundsOne2Many(pB);
	}

    if(pB->type == 3)
	{
	   //err = CalcBoundsOne2Many_RS(pB, W);
	}


    return err;
}
*/

/******************************************************************
**  Calc list of segments 
**  Call the function after for any changes of approximating
** 
** Return code of error 
**    0: OK
**   -1: ERROR
******************************************************************/
// UNDER CONSTRUCTION 
/*
int CHistoDataPolar::CalcClustersList()
{  
   ULONG   n, jj, n1, n2;
   int     k;
   double  d_omega, d_myu;

   for(n = 0; n < m_nSize; n++)
   {
      m_pClusterStart[n] = 0;
   }//for(n   

   m_nClustersNumber = 0;

   jj = 0;
   for(n = 0; n < m_nSize; n++)
   {
       if((m_pClusterIndex[n] == 1))
	   {
           m_pClusterStart[jj] = n;
		   //printf("j=%ld: n=%ld\n", jj, n);
		   jj++;
	   }//if(

   }//for(n   

   m_nClustersNumber = jj;

   if(m_nClustersNumber >= 1)
   {
      // Partition & Means
      for(k = 0; k < m_nClustersNumber; k++)
	  {
      	 n1 = m_pClusterStart[k];  

    	 if(k <= m_nClustersNumber-2)
		 {
	         n2 = m_pClusterStart[k+1]-1;   
		 }
	     else
		 {
            n2 = m_nSize-1;    
		 }

		 /////////////////////////////////////////////
		 if(n1 == 0) 
		 {
		    d_myu   = m_pMomentOne_R[n2];
		    d_omega =  m_pMomentZero_R[n2];
		 }
		 else
		 {
		    d_myu   = m_pMomentOne_R[n2] - m_pMomentOne_R[n1-1];
		    d_omega =  m_pMomentZero_R[n2] - m_pMomentZero_R[n1-1];
		 }

         m_pClusterMeans[k]  = d_myu/d_omega;

		 for(n = n1; n <= n2; n++)
		 {
             m_pClusterNumber[n] = k;

		 }//for(n
		 
	  }//for(k   
   }


   return 0;
}

*/
/******************************************************************
**  Type Dump of Clusters
**
** 
** Return code of error 
**    0: OK
**   -1: ERROR
******************************************************************/
int CHistoDataPolar::TypeClustersList()
{
//	 ULONG   n1;
	 double  S_sigma=0, S_myu=0; // myu, 

/*	 printf("Size: %2d; Clusters: %d; \n", m_nSize, m_nClustersNumber);

     for(k = 0; k < m_nClustersNumber; k++)
	 {
		  printf("k=%3d) Start = %3d \n", k, m_pClusterStart[k]);
	 }
     printf("********************\n");


     for(k = 0; k < m_nClustersNumber; k++)
	 {
	     n1 = m_pClusterStart[k];

		 if(k <= m_nClustersNumber-2)
		 {
	        n2 = m_pClusterStart[k+1]-1; 
		 }
		 else
		 {
            n2 = m_nSize-1;
		 }

         CalcVarianceForCluster(k, &sigma);
		 CalcMeansForCluster(k, &myu);
         
		 printf("j=%3d) [%5ld, %5ld] -> %3ld: Sigma = %16.12f;  Myu = %16.12f\n", 
			      k, n1, n2, n2-n1+1, sigma, myu);

		 S_sigma += sigma;
	 }
     printf("Sigma=%.12f\n", S_sigma);
	 printf("---------------------------\n");
*/

	 printf("===============================================\n");
     printf("--> Radius: \n");
	 m_pHData_R->CalcClustersList();
     m_pHData_R->TypeClustersList();

	 printf("--> Phase:\n");
	 m_pHData_Phi->CalcClustersList();
	 m_pHData_Phi->TypeClustersList();
     printf("===============================================\n");

	 return 0;
}

/******************************************************************
**  Type Dump of Clusters
**
** 
** Return code of error 
**    0: OK
**   -1: ERROR
******************************************************************/
/*
void CHistoDataPolar::TypeThresholds()
{
	 ULONG   k;

	 //printf("Histogram size: %2d; thresholds number: %d: \n", m_nSize, m_nClustersNumber-1);
/ *
     for(k = 0; k < m_nSize; k++)
	 {
		 if(m_pClusterIndex_R[k] == 1)
		 {
		    printf("%.12f ", m_pR[k]);
		 }
		 //printf("k=%d) Ind=%d \n", k, m_pClusterIndex[k]);
	 }

	 printf("\n");
* /

}
 */


/***************************************************************
*  Load Histo Data data from ASCII file
*
*
**  File format:
**  DARA    - label
**  x1
**  x2
**   ....
**
**  Parameters:
**    char*       filename  -- input filename
**    HistoData*  pHisto
**
** Return code of error 
**    0: OK
**   -1: ERROR
***************************************************************/
/*
int CHistoDataPolar::Load(char* filename)
{   int       err=0, fields_number;
    long int  k, size, c;
    FILE*     pFile = NULL; 
	float     b;

    if(filename == NULL) 
	{
	   return -1;
	}

    pFile = OpenForReadingAndCheckLabel(filename, "POLAR_DATA");
	if(pFile == NULL) 
    {
	   return -1;
    }

    // Load number of points; 
	fscanf(pFile, "%d", &size);
    if(size < 0) 
    {
  	    return -1;
    } 

	fscanf(pFile, "%d", &fields_number);
    if(fields_number < 0) 
    {
  	    return -1;
    } 

	// Reallocate memory
	err = ReallocateMemory(size);
	if(err != 0)
    {
  	    return -1;
    } 

    for(k = 0; k < size; k++) 
    {
		fscanf(pFile, "%f", &b);
		m_pR[k] = (double)b;

		fscanf(pFile, "%f", &b);
		m_pPhi[k] = (double)b;

		if(fields_number == 3)
		{
   	    	fscanf(pFile, "%d", &c);
			m_pHisto[k] = (unsigned long int)c;
		}
	}

    m_nSize = k;

	// 
	m_pHData_R   = new CHistoData(m_pR,   m_pHisto, m_nSize);
    m_pHData_Phi = new CHistoData(m_pPhi, m_pHisto, m_nSize);

	// 
    SortDataByPhase();

    CalcMoments(); 

    return err;
}
*/

/**************************************************************************
*  Open file 'filename' for reading and check label
*
*  Return error code:
*     0: OK
**************************************************************************/
/*
FILE* CHistoDataPolar::OpenForReadingAndCheckLabel(char* filename, char *label)
{
   FILE*  pFile = NULL;
   char   szLabel[32];

   if(filename == NULL)
   {
	   return NULL;
   }

   // Open file 
   pFile = fopen(filename, "rt");
   if(pFile == NULL) 
   {
	   return NULL;
   }

	// Load & Check the Label 
   	fscanf(pFile, "%s", szLabel);
	if(strcmp(label, szLabel) != 0) 
    {
	   return NULL;
    }

	return pFile;
}
*/
/***********************************************************************************
* 1) Calculate Variance for the HistoData
*
*
*  Result:
*   1. (double *)var_val  -- Variance value
*
*  Return error code: 
*    0  -- OK
*    -1 -- ERROR
*************************************************************************************/
int CHistoDataPolar::CalcVariancePhase(double *var_val)
{
    int     err=0;
//    ULONG   k, M;
//	double  value;
/*
	M = m_nClustersNumber;

    *var_val = 0;

	if(m_nClustersNumber < 0)
    {
	    *var_val = -1;

		return -1;
	}

	for(k = 0; k < M; k++)
	{
        err = CalcVarianceForClusterPhase(k, &value);

	    if(err == 0) 
		{
		 	  *var_val += value;
		}
		else 
		{
            err      = -1;
		    *var_val = -1;

              break;
		}//
       // printf("cluster #%2ld) n_start: %ld: %.12f -> %.12f \n", 
		 //      k, this->m_pClusterStart[k], value, *var_val);		  		  
	}//for(k
*/
    return err;
}




/*****************************************************************
*  PRIVATE methods
*****************************************************************/ 

/*************************************************************************
*  Calc variance for Cluster
*
*  Parameters:
*    1. int j                -- number of the end   point of the Cluster
*
*  Result:
*    2. (double*) var_val      -- variance value
*
*  Return code of error:
*    0: OK
*   -1: ERROR
*******************************************************************************/
/*
int CHistoDataPolar::CalcVarianceForCluster(int k, double *var_val)
{   
	int       err=0;
	ULONG     n1, n2;
	double    d_omega, d_myu, d_lambda;

	*var_val = 1.0e10;

	n1 = m_pClusterStart[k];  

	if(k <= m_nClustersNumber-2)
	{
	   n2 = m_pClusterStart[k+1]-1;   
	}
	else
	{
       n2 = m_nSize-1;    
	}

	if(n1 == n2)
	{
   	   *var_val = 0;  

	 //  printf("A: k=%ld) [%ld, %ld]->%.12f \n", k, n1, n2, *var_val);

	   return 0;
	}

    d_omega  = m_pMomentZero[n2]; 
	d_myu    = m_pMomentOne[n2]; 
	d_lambda = m_pMomentTwo[n2]; 

	if(n1 > 0)
	{
       d_omega  -= m_pMomentZero[n1-1]; 
	   d_myu    -= m_pMomentOne[n1-1]; 
	   d_lambda -= m_pMomentTwo[n1-1]; 
	}
	
    if(d_omega != 0)
	{
	   *var_val = d_lambda - (d_myu*d_myu)/d_omega;
	}
	else
	{
       *var_val = -1;
        err = -3;
	}

   // printf("A: k=%ld) [%ld, %ld]->%.12f \n", k, n1, n2, *var_val);

    return err;
}
*/
/*
int CHistoDataPolar::CalcVarianceForCluster_B(int k, double *var_val)
{
	int       err=0;
	ULONG     n1, n2, N1, n;
	double    omega, myu;

	*var_val = 1.0e10;

	n1 = m_pClusterStart[k];  

	if(k <= m_nClustersNumber-2)
	{
	   n2 = m_pClusterStart[k+1]-1;   
	}
	else
	{
       n2 = m_nSize-1;    
	}

	if(n1 == n2)
	{
   	   *var_val = 0;  
	    myu = m_pR[n1];
        printf("B: k=%ld)  [%ld, %ld]->%.4f (myu=%.3f)\n", k, n1, n2, *var_val, myu);	   
		
		return 0;
	}

	N1 = 0;
	myu = 0;
	for(n = n1; n <= n2; n++)
	{
       myu += m_pR[n]*m_pHisto[n];
	   N1  += 1;
	}
//	myu = myu/(double)N1;

	omega = 0;
	for(n = n1; n <= n2; n++)
	{
       omega += (m_pR[n]-myu)*(m_pR[n]-myu)*m_pHisto[n];
	}

	*var_val = omega; ///m_nSize;

    printf("B: k=%ld) [%ld, %ld]->%.4f (myu=%.3f) \n", k, n1, n2, *var_val, myu);

    return err;
}
*/
/* ********************
int CHistoDataPolar::CalcVarianceForClusterPhase(int k, double *var_val)
{
	int       err=0;
	ULONG     n1, n2, N1, n;
	double    omega, myu;

	*var_val = 1.0e10;

	n1 = m_pClusterStart[k];  

	if(k <= m_nClustersNumber-2)
	{
	   n2 = m_pClusterStart[k+1]-1;   
	}
	else
	{
       n2 = m_nSize-1;    
	}

	if(n1 == n2)
	{
   	   *var_val = 0;  
	    myu = m_pPhi[n1];
        printf("B: k=%ld)  [%ld, %ld]->%.4f (myu=%.3f)\n", k, n1, n2, *var_val, myu);	   
		
		return 0;
	}

	N1 = 0;
	myu = 0;
	for(n = n1; n <= n2; n++)
	{
       myu += m_pPhi[n];
	   N1  += 1;
	}
	myu = myu/(double)N1;

	omega = 0;
	for(n = n1; n <= n2; n++)
	{
       omega += (1 - cos(m_pPhi[n]-myu));
	}

	*var_val = omega;

    printf("B: k=%ld) [%ld, %ld]->%.4f (myu=%.3f) \n", k, n1, n2, *var_val, myu);

    return err;
}

*****************/



/******************************************************************
* Save Thresholds file 'filename'
*
* Return code of error 
*    0: OK
*   -1: ERROR
*******************************************************************/
/*
int CHistoDataPolar::SaveThresholds(char* filename)
{ 
    int      err=0;
	double   S=0; 
    FILE*    pFile = NULL;
    char     THRESHOLDS_LABEL[] = "THRESHOLDS";

    if((pHisto == NULL) || (filename == NULL))
	{
		 return -1;
	}

    if(pHisto->size < 0) 
	{
		 return -1;
	}

	// Open file; 
	pFile = fopen(filename, "w");
    if(pFile == NULL)
	{
		 return -1;
	}

//	err = CalcClustersList(pHisto);
	if(err != 0)
	{
		 return -1;
	}

	// Save Label 
    fprintf(pFile, "%s\n", THRESHOLDS_LABEL);

/ *    // Save number of Thresholds; 
	fprintf(pFile, "%d\n", pHisto->clusters_number-1);

    // Save values 
    for(int k = 0; k < pHisto->clusters_number-1; k++) 
	{
        fprintf(pFile, "%ld\n", pHisto->cluster_end[k]);
	}
* /
    return 0; // err;
}

*/

/////////////////////////////////////////////////////////
int CHistoDataPolar::CalcMoments()
{   int     er= 0;
    long int  n;
    double  r, r2, phi, p, S0;

    m_pHData_Phi->CalcMoments();
	m_pHData_R->CalcMoments();

	////////////////////////////////////////
/*	for(n = 0, S0=0.0; n < m_nSize; n++)
	{
	    S0 += (double)m_pHisto[n];
	}

	// Probability
	for(n = 0; n < m_nSize; n++)
	{
       m_pProby[n] = (double)m_pHisto[n] / S0;
	}
*/
	//////////////////////////////////////////
	// Cumulants
//	SortDataByPhase(); // XXX ???

    p   = m_pProby[0];
	phi = m_pPhi[0];
	r   = m_pR[0];

    r2  = r*r;

	m_pPhaseMomentZero[0]  = p;             
	m_pPhaseMomentOne[0]   = p * phi;    
	m_pPhaseMomentTwo[0]   = p * phi * phi; 

    m_pPolarMomentZero[0]  = r2 * p;             
	m_pPolarMomentOne[0]   = r2 * p * phi;    
	m_pPolarMomentTwo[0]   = r2 * p * phi * phi; 

	n = 0;
	/*
	printf("PHI: n=%d) %.7f  (%.7f %.7f %.7f - %.7f %.7f)\n", 
		     n, m_pX[n], m_pMomentZero[n], m_pMomentOne[n], m_pMomentTwo[n],
			 m_pMomentThree[n], m_pMomentFour[n]);
    */
	for(n = 1; n <= m_nSize-1; n++)
	{
       phi = m_pPhi[n];
	   r   = m_pR[n];
	   r2  = r*r;
	   p   = m_pProby[n];	   

   	   m_pPhaseMomentZero[n] = p             + m_pPhaseMomentZero[n-1];
       m_pPhaseMomentOne[n]  = p * phi       + m_pPhaseMomentOne[n-1] ;
       m_pPhaseMomentTwo[n]  = p * phi * phi + m_pPhaseMomentTwo[n-1] ;

	   m_pPolarMomentZero[n] = r2 * p             + m_pPolarMomentZero[n-1];
       m_pPolarMomentOne[n]  = r2 * p * phi       + m_pPolarMomentOne[n-1] ;
       m_pPolarMomentTwo[n]  = r2 * p * phi * phi + m_pPolarMomentTwo[n-1] ;
	   
	//   m_pClusterNumber[n] = 0;

	   /*printf("PHI: n=%d) %.7f   [%.7f %.7f %.7f - %.7f %.7f]\n", 
		     n, m_pX[n], m_pMomentZero[n], m_pMomentOne[n], m_pMomentTwo[n], 
			 m_pMomentThree[n], m_pMomentFour[n]);
        */
	}//for(n

	return 0;
}

/*
int CHistoDataPolar::SortData()
{  
	ULONG i, j;
	double tmp;

	for(i = 0; i <= m_nSize-2; i++)
	{
        for(j = i+1; j <= m_nSize-1; j++)
		{
           if(m_pR[i] > m_pR[j])
		   {
              tmp = m_pR[j];
			  m_pPhi[j] = m_pR[i];
			  m_pR[i] = tmp;
		   }

		}//for(j
	}//for(i


/*	for(i = 0; i <= m_nSize-2; i++)
	{
    //    if(m_pX[i] == m_pX[i+1])
            printf("i=%ld)  %.6f \n", i, m_pX[i]);
	}
	printf("\n");
* /	
	return 0;
}
*/
/*
int CHistoDataPolar::SortPointsByPhase()
{  
	long int i, j;

	for(i = 0; i <= m_nSize-2; i++)
	{
        for(j = i+1; j <= m_nSize-1; j++)
		{
           if(m_pPoints[i].phase > m_pPoints[j].phase)
		   {
              SwapPolarPoints(&m_pPoints[i], &m_pPoints[j]);
		   }
		}//for(j
	}//for(i

	/*
	for(i = 0; i <= m_nSize-1; i++)
	{
        if(m_pX[i] == m_pX[i+1])
            printf("i=%ld)  %10.6f  %12.6f \n", i, pPhi->value[i], pR->value[i]);
	}
	printf("\n");
	*/
	
//	return 0;
//}
	

// Swap Polar Points
/*void CHistoDataPolar::SwapPolarPoints(DPolarPoint *p1, DPolarPoint *p2)
{
    DPolarPoint *p;

	p  = p1;
	p1 = p2;
	p2 = p;
}
*/

void CHistoDataPolar::SortDataByPhase()
{  
/*	long int          i, j;
	unsigned long int tmp2;
	double            tmp;

	for(i = 0; i <= m_nSize-2; i++)
	{
        for(j = i+1; j <= m_nSize-1; j++)
		{
           if(m_pPhi[i] > m_pPhi[j])
		   {
              tmp       = m_pPhi[j];
			  m_pPhi[j] = m_pPhi[i];
			  m_pPhi[i] = tmp;

              tmp     = m_pR[j];
			  m_pR[j] = m_pR[i];
			  m_pR[i] = tmp;

              tmp         = m_pProby[j];
	          m_pProby[j] = m_pProby[i];
	          m_pProby[i] = tmp;

              tmp2         = m_pHisto[j];
	          m_pHisto[j] = m_pHisto[i];
	          m_pHisto[i] = tmp2;
		   }

		}//for(j
	}//for(i
*/
	heapSortPhase();
}

void CHistoDataPolar::SortDataByRadius()
{  
/*	long int i, j;
	double tmp;

	for(i = 0; i <= m_nSize-2; i++)
	{
        for(j = i+1; j <= m_nSize-1; j++)
		{
           if(m_pR[i] > m_pR[j])
		   {
              tmp      = m_pR[j];
			  m_pR[j]  = m_pR[i];
			  m_pR[i]  = tmp;

		      tmp        = m_pPhi[j];
			  m_pPhi[j]  = m_pPhi[i];
			  m_pPhi[i]  = tmp;

              tmp         = m_pProby[j];
	          m_pProby[j] = m_pProby[i];
	          m_pProby[i] = tmp;
		   }

		}//for(j
	}//for(i
*/
	heapSortRadius();
}


/*****************************************************************************
**
**  Cost Function for R:  E_r = <(r_i-R_j)^2>
**  Parameters:
**
**  Return error code: 
**     0: OK
**    -1: ERROR
**********************************************************************************/
int CHistoDataPolar::DP_MSE_PolarQ_1(ULONG M, int nType, double *error_value)
{ 
	int         err=0, nVers=2;
	ULONG Mr, Mphi, m_r, m_phi, m_r_min, m_phi_min;
    double      R1, R2, E_1, E_2, E_min, E_curr, a1, a2;
    CHistoData  *pHData_R=NULL, *pHData_Phi=NULL;

	//////////////////////////////////////
	R1 = pHData_R->CalcMoment(1);
    printf("R1 = %.12f \n", R1);
	R2 = pHData_R->CalcMoment(2);
	printf("R2= %.12f (%.12f) \n", R2, sqrt(R2));

	////////////////////
    Mr = (int)(3*sqrt((double)M));
	pHData_R->DP_MSE(Mr, nVers, nType, &E_1);

	////////////////////
	Mphi = (int)(3*sqrt((double)M));
	pHData_Phi->DP_MSE(Mphi, nVers, nType, &E_2);

	E_min = 1.0e35;

	for(M = 300; M <= 10000;  M += 100)
	{
        for(m_r = 30; m_r <= Mr; m_r++)
		{
		    m_phi = (int)((double)M/(double)m_r);

   		    if(m_phi <= Mphi)
			{
			   E_1 = pHData_R->m_pE2[m_r];  

			   //pHData_Phi->MakeUniformClusters(m_phi);
	           //pHData_Phi->CalcVariance(&E_2);

	           E_2 = R2 * pHData_Phi->m_pE2[m_phi];  

		       E_curr = E_1 +  R2 * E_2; 
			
			   if(E_curr < E_min) 
			   {
                   E_min     = E_curr;
			       m_r_min   = m_r;
			       m_phi_min = m_phi;
			       a1 = E_1;
  			       a2 =  R2 * E_2;
			   }
	
		       //printf("m_r=%d, m_phi=%d) %.12f + %.12f = %.12f \n", 
			     //   m_r, m_phi, E_1, E_2, E_curr);
			}
		}//for/(m_x

         printf("M=%d) E2: %.10f for M_r=%d and M_phi=%d; R2=%.3f\n",
		     M, E_min, m_r_min, m_phi_min, R2);

	}//fopr(M

  	// Free memory

    return 0;
}

/*****************************************************************************
**  Quantization on Phase space taking into account Radius-coordinate of the point
**
**  Parameters:
**   nType:       1=Single-state; 2=Multi-state
**  Return error code: 
**     0: OK
**    -1: ERROR
**********************************************************************************/
int CHistoDataPolar::DP_Phase_MSE_min(ULONG Mphi, int nType, double *error_value)
{ 
	int         err=0, flag_pnode=1, verbose=1, i1=0, nVers=2;
	long int    n, N, ni=0, nn, undef_val = -99999, ind, ind1;
	ULONG m;
    DTable      *pD2=NULL; 
	double      E2; //, E_1, E_2;
    NTable      *pNodes=NULL;
    Bounds      *pBnd=NULL;

	*error_value = 1.0e20;

    //////////////////////////
	// Use m_pHData_Phi to keep data for Phase!!!

	/////////////////////////////////////////////////////
    N = m_pHData_Phi->m_nSize;

     // Allocate memory 
	pNodes = CreateNTable(N+1, Mphi+1);
	if(pNodes == NULL)
	{   
		 return -1;
	}

    SetNTableToValue(pNodes, undef_val);

    // Allocate memory for D
	pD2 = CreateDTable(N+1, Mphi+1);

	if(pD2 == NULL)
	{
		 DestroyNTable(&pNodes); 

		 return -1;
	}

    // Allocate memory for the Bounding Corridor
	pBnd = CreateBounds(nType, N, Mphi, Mphi);
	if(pBnd == NULL)
	{
		 DestroyNTable(&pNodes); 
         DestroyDTable(&pD2); 

		 return -1;
	}
	
	// Calculate the Bounding Corridor 
	err = m_pHData_Phi->CalcBounds(pBnd);
	if(err != 0)
	{
		 DestroyNTable(&pNodes); 
         DestroyDTable(&pD2); 

		 return -1;
	}

	////////////////////////////////////////////
    // DP search in state space for polar coordinates (r,phi)
	DP_Phase_search(pBnd, pD2, pNodes);
    ////////////////////////////////////////////

	// Init Cluster Index
	
	long int upper_bound;
	
	

	if(m_pHData_Phi->m_pClustersList != NULL)
		delete [] m_pHData_Phi->m_pClustersList;

	m_pHData_Phi->m_pClustersList = new double[(UINT)((Mphi)*(Mphi+1)/2)];

	long int cur_index = 0;

	for(upper_bound = 2; upper_bound <= Mphi;upper_bound++)
	{
	
		m_pHData_Phi->m_nClustersNumber = upper_bound;
		
		for(n = 0; n < N; n++)
		{
		    m_pHData_Phi->m_pClusterIndex[n] = 0;
		}
		m_pHData_Phi->m_pClusterIndex[0] = 1;

		// Make output solution
		ind = N - 1;  

		for(m = upper_bound; m >= 1; m--)
		{
		   nn   = ind * pNodes->size_x + (m-pBnd->bottom[ind]);
		   ind1 = pNodes->data[nn];
		   //printf("m=%ld) %ld  -> %ld \n", m, ind, ind1);

		   if(ind1 == undef_val)  {
			   fprintf(stderr, "ERROR with pNode!!! m=%d;  ind: %d,   ind1: %d, \n", 
					   m, ind, ind1);
			   err = -1;

			   break;
		   }
			
			
		   if(ind+1<N)
			m_pHData_Phi->m_pClusterIndex[ind+1] = 1;

		   ind = ind1;
		}//for(m

		m_pHData_Phi->m_pClusterIndex[0] = 1;
		m_pHData_Phi->CalcClustersList();
		
		
		for(long int i=0;i<upper_bound;i++)
			m_pHData_Phi->m_pClustersList[cur_index++] = m_pHData_Phi->m_pClusterMeans[i];
	}
	//////////////
	nn = (N-1) * pD2->size_x + (Mphi - pBnd->bottom[N-1]);
    double E2_a  = pD2->data[nn]; 
	
	double E2_b; 
	m_pHData_Phi->CalcVariance(&E2_b);

	if((ULONG)m_pHData_Phi->m_nClustersNumber != Mphi)
	{
		printf("WARNNING! MSE: %d -> %d \n", Mphi, m_pHData_Phi->m_nClustersNumber);
	}
	
	*error_value = E2_a; 

    //////////////////////////////////////////
	
	if(pBnd->type == 2)
	{ 
		printf("Bound Type: %d\n", pBnd->type);
		for(m = 1; m <= Mphi; m++)
		{
	         nn = (N-1) * pD2->size_x + (m - pBnd->bottom[N-1]);
             E2 = pD2->data[nn]; 
             m_pE2[m] = E2;
			 //printf("m=%d) E2 = %.12f\n", m, E2);
		}//for(m
	}

	// Free memory  
 	DestroyBounds(&pBnd);
	DestroyDTable(&pD2);
	DestroyNTable(&pNodes);

    return 0;
}


/**/
//The given function  creates a new order of rows processing
//Time complexity of the function is about O(N), where the N is the
//number of something
/**/


void SetAddArraysPolar(long N, long *new_order, long *new_left_array, long *new_right_array)
{
	int level_index = 1;

	new_order[0] = (N-1)/2;
	new_left_array[0] = -1;
	new_right_array[0] = -1;

	new_order[1] = 0;
	new_left_array[1] = -1;
	new_right_array[1] = new_order[0];


	new_order[2] = N-1;
	new_left_array[2] = new_order[0];
	new_right_array[2] = -1;

	long current_index = 3;
	long divider = 2;
	long i;
	
	long left_index;
	long right_index;

	long *used;
	__int64 new_index;

	used = new long[N];

	memset(used,0,N*sizeof(long));
	
	for(i=0;i<3;i++)
		used[new_order[i]] = 1;

	while(current_index<N)
	{
		divider = divider*2;

		for(i=1;i<divider;i+=2)
		{
			new_index = ((__int64)(N-1) * i)/divider;
			
			if(used[new_index])
				continue;
			else used[new_index] = 1;
			
			new_order[current_index] = new_index;
			
			left_index = ((__int64)(N-1) * (i-1))/divider;
			right_index = ((__int64)(N-1) * (i+1))/divider;


			new_left_array[current_index] = left_index;
			new_right_array[current_index] = right_index;

			current_index++;

			if(current_index>=N)
				break;
		}
	}

	delete used;
}


////////////////////////////////////////////////////////////
//  2)
//  PolarQ: 
//  Quant. Method:  DP_MSE_min for Phase  : 
//  Loops  Order :  m-n! (good)
//  Search Width : 1 pass from new_left till End
//  
////////////////////////////////////////////////////////////
int CHistoDataPolar::DP_Phase_search(Bounds *pBnd, DTable  *pD2, NTable *pNodes)
{   int        err=0; 
    long int  m, nn, nn1,  N, Ms, undef_val = -99999, i_argmin, new_left = -1,
	          new_right = -1, n, n1, n2, i, I1, I2;		
    double    val_min, val_curr, E1, E2, d_myu, d_omega, d_lambda, d_nyu, d_pi,
		      omega, myu, lambda, nyu, pi, means;  

	if((pBnd == NULL) || (pD2 == NULL) || (pNodes == NULL))
  	{
		return -1;
	}

	// No points in the HistoData 
	if(m_nSize < 0)
  	{
		return -1;
	}

	// Number of the points 
	N = m_nSize;   
	
	// Desired number of Clusters 
	Ms = pBnd->size_m;  //m_nClustersNumber;
	Ms = (Ms < N) ? Ms : N-1;

	/////////////////////////
    // m=1;
   	// nn = 0 * pD2->size_x + 0 for (0,0);; 
	pD2->data[0] = 0;

	m  = 1;
    n1 = pBnd->left[m];
	n2 = pBnd->right[m];
	//printf("\n*****  m=%d) %d -- %d ****\n", m, n1, n2);

	for(n = n1; n <= n2; n++) 
	{
	   //CalcVarianceFromPToP(0, n, &E2);
 	   omega  = m_pPolarMomentZero[n]; 
  	   myu    = m_pPolarMomentOne[n]; 
	   lambda = m_pPolarMomentTwo[n]; 

	   pi      = m_pPhaseMomentZero[n]; 
	   nyu     = m_pPhaseMomentOne[n]; 
	   
	   //E2 = lambda - (myu*myu)/omega;
	   means = nyu/pi;
	   E2 = lambda - 2*myu*means + omega*means*means;
 
 	   nn = n * pNodes->size_x + (m-pBnd->bottom[n]);
	   pNodes->data[nn] = 0;

       nn = n * pD2->size_x + (m - pBnd->bottom[n]);
       pD2->data[nn] = E2;

	   //printf("E[0,%d] = %.12f\n", n, E2);
	}


	long *new_order;
	long *new_left_array;
	long *new_right_array;
	long row_index;


	new_order = new long[N];
	new_left_array = new long[N];
	new_right_array = new long[N];


	///////////////////////////////////////
	// Other m's.
	for(m = 2; m <= Ms; m++) 
	{   
  	    n1 = pBnd->left[m];
		n2 = pBnd->right[m];

		//printf("\n*****  m=%d) n = %d -- %d ****\n", m, n1, n2);
		new_left  = -1;

		memset(new_order,0,sizeof(long)*N);
		memset(new_right_array,0,sizeof(long)*N);
		memset(new_left_array,0,sizeof(long)*N);

		SetAddArraysPolar(n2-n1+1,new_order,new_left_array,new_right_array);

		for(n = n1; n <= n2; n++) 
		{
			row_index = n1 + new_order[n-n1];

			i_argmin = -99;
		    val_min  =  1.0e35;

		    I1 = pBnd->left[m-1];
		    I2 = row_index-1;

//			if(new_left  != -1) I1 = new_left;

			new_left = new_left_array[n-n1];
			new_right = new_right_array[n-n1];
			
			if(new_left != -1)
			{
				new_left += n1;
				nn = new_left * pNodes->size_x + (m-pBnd->bottom[new_left]);
				I1 = max(pNodes->data[nn],I1);
			}

			if(new_right != -1)
			{
				new_right += n1;
				nn = new_right * pNodes->size_x + (m-pBnd->bottom[new_right]);
				I2 = min(pNodes->data[nn],I2);
			}

			omega  = m_pPolarMomentZero[row_index]; 
            myu    = m_pPolarMomentOne[row_index]; 
			lambda = m_pPolarMomentTwo[row_index]; 

	        pi     = m_pPhaseMomentZero[row_index]; 
		    nyu    = m_pPhaseMomentOne[row_index]; 

            ////////////////////////////////////////
  		    for(i = I1; i <= I2; i++) 
			{ 
			    // Min{ D[i][m-1] + d(i,n); } 
                // (m-1) Clusters for [I1,i] points;
			    nn1 = i * pD2->size_x + (m-1 - pBnd->bottom[i]);  
                E1  = pD2->data[nn1];

			    // One Cluster for [i+1,n] points; 
			    //CalcVarianceFromPToP(i+1, n, &E2);

  	  		    if(row_index == i+1)
				{
	               E2 = 0.0;
				}
				else //  i = n-1;
				{ 
                   d_omega  = omega  - m_pPolarMomentZero[i]; 
  	               d_myu    = myu    - m_pPolarMomentOne[i];  
				   d_lambda = lambda - m_pPolarMomentTwo[i];  

	               d_pi     = pi     - m_pPhaseMomentZero[i]; 
				   d_nyu    = nyu    - m_pPhaseMomentOne[i]; 

 			 	   //E2 = d_lambda - (d_myu*d_myu)/d_omega;
  			       means = d_nyu/d_pi;
				   E2 = d_lambda - 2*d_myu * means + d_omega *means*means;
				}
		
	            val_curr = E1 + E2;
                if(val_curr < val_min) 
				{
			            val_min  = val_curr;
                        i_argmin = i;
			   	        new_left = i;
				}//if

				//if(n > 2800)
				//{
 			    // 	printf("n=%ld; i=%ld)  D[%d][%ld] + E[%ld][%ld] = D[%ld][%ld]: %.6f+%.6f = %.6f  (>=%.6f)\n", 
				//	    n, i,            m-1, i,       i+1, n,         m,n,  
                  //              E1,      E2, val_curr, val_min) ;
				//}

			}//for(i

		    /*	if(n > 2800) 
			{
  	           printf("m=%ld, n=%ld: E_min = %.12f for I=%ld \n", 
 		       m, n, val_min, i_argmin);			
			   printf("\n");
			}
             */
		     nn = row_index * pNodes->size_x + (m-pBnd->bottom[row_index]);
	         pNodes->data[nn] = i_argmin;

		     nn = row_index * pD2->size_x + (m - pBnd->bottom[row_index]);
			 pD2->data[nn] = val_min;
		}//for(n

	 }//for(m


	 delete[] new_order;
	 delete[] new_left_array;
	 delete[] new_right_array;

     return 0;
}

/*****************************************************************************
**   2nd approximation
**
**  Parameters:
**    1) (int )        Ms           -- 
**    2) (double *)    error_value --  
**
**  Return error code: 
**     0: OK
**    -1: ERROR
*********************************************************************************/
int CHistoDataPolar::SPQ(ULONG Mr, ULONG Mphi, double *error_value)
{
	int      err=0;
	int      nType=1, nVers=2;
    double   E2, E2_R, R2=1, E2_Phi;

	///////////////////
	// 1. Radius
	//////////////////////////////////
	m_pHData_R->DP_MSE(Mr, nVers, nType, &E2_R);
	m_pHData_R->CalcClustersList();

    printf("\n Radius: Mr=%d;  mse = %16.12f \n", 
	         m_pHData_R->GetClustersNumber(), E2_R);
	
	//////////////////////////////////
	// 2. Phase
	//////////////////////////////////

	err = DP_Phase_MSE_min(Mphi, nType, &E2_Phi);
    m_pHData_Phi->CalcClustersList();

    printf(" Phase: Mhi=%d;  mse = %16.12f \n", 
	         m_pHData_R->GetClustersNumber(), E2_Phi);
	
	/////////////////
    E2 = E2_R + E2_Phi;
	printf("PolarQ2: E_r=%f + E_phi=%f  ->  E2=%.12f \n", 
		    E2_R, E2_Phi, E2);

    *error_value = E2_R; // E2; 

	return 0;
}

/*****************************************************************************
**   2nd approximation
**
**  Parameters:
**    1) (int )        Ms           -- 
**    2) (double *)    error_value --  
**
**  Return error code: 
**     0: OK
**    -1: ERROR
*********************************************************************************/
int SavePolarData(int iteration, long r_num, double *r_centroids, 
								 long phi_num, double *phi_centroids)
{
	char FileName[128] = { '\0' };
	char str[2] = {'\0'};
	char temp_str[128] = {'\0'};
	double mask;
	double temp_num;

	FILE *f;

	strcpy(FileName,"zzzTemp1\0");
	itoa(iteration,temp_str,10);
	strcat(FileName,temp_str);
	strcat(FileName,".cb\0");

	f = fopen(FileName,"w");
	fprintf(f,"%d\n",r_num);
	
	for(int i = 0;i<r_num;i++)
	{
		temp_num = (double)r_centroids[i];
		fprintf(f,"%.0f\n",temp_num);
	}
	
	fclose(f);
	
	strcpy(FileName,"zzzTemp2\0");
	itoa(iteration,temp_str,10);
	strcat(FileName,temp_str);
	strcat(FileName,".cb\0");

	f = fopen(FileName,"w");
	fprintf(f,"%d\n",phi_num);
	mask = (double)pow(10,PHI_SCALE);
	
	for(i = 0;i<phi_num;i++)
	{
		temp_num = (double)(phi_centroids[i]*mask);
		fprintf(f,"%.0f\n",temp_num);
	}
	
	fclose(f);
	return 1;
}




int CHistoDataPolar::SPQ(ULONG M, ULONG *Mr, ULONG *Mphi, double *error_value)
{
	int      err=0;
	ULONG      nType=2, nVers=2, Mphi_max, Mr_max, m_r, m_phi, m_r_min, m_phi_min;
    double   E2_R, R2=1, E2_Phi, E_min, E_1, E_2, E_curr;

	Mr_max   = (ULONG)(sqrt(3*(double)M));
	Mphi_max = (ULONG)(sqrt(3*(double)M));

	m_pHData_R->DP_MSE(Mr_max, nVers, nType, &E2_R);

    printf("\nRadus: Mr=%d;  mse = %16.12f \n", 
	         m_pHData_R->GetClustersNumber(), E2_R);

//	m_pHData_R->CalcClustersList();

	//////////////////////////////////
    err = DP_Phase_MSE_min(Mphi_max, nType, &E2_Phi);
	    
	int i;
	for(i=0;i<CB_NUMBER;i++)
	{
		long M_cur;

		M_cur = ((i+1)*M)/CB_NUMBER;

		Mr_max   = (ULONG)(sqrt(3*(double)M_cur));
		Mphi_max = (ULONG)(sqrt(3*(double)M_cur));
		E_min = 1.0e35;

		for(m_r = 2; m_r <= Mr_max; m_r++)
		{
			m_phi = (ULONG)((double)M_cur/(double)m_r);

   			if(m_phi <= Mphi_max)
			{
			   E_1 = m_pHData_R->m_pE2[m_r];  
			   E_2 = m_pE2[m_phi];  

			   E_curr = E_1 +  E_2; 
				
			   if(E_curr < E_min) 
			   {
				  E_min     = E_curr;
				  m_r_min   = m_r;
				  m_phi_min = m_phi;
			   }
		
				   //printf("m_r=%d, m_phi=%d) %.12f + %.12f = %.12f \n", 
					 //   m_r, m_phi, E_1, E_2, E_curr);
				}
		}//for/(m_x

		printf("M=%d) E2: %.10f for M_r=%d and M_phi=%d;\n",
				 M_cur, E_min, m_r_min, m_phi_min);

		double *p_temp;

		double *r_centroids;
		double *phi_centroids;

		r_centroids = new double[m_r_min];
		phi_centroids = new double[m_phi_min];

		p_temp = GetMeansR(m_r_min);
		memcpy((r_centroids),p_temp,(m_r_min)*sizeof(double));
		p_temp = GetMeansPhi(m_phi_min);
		memcpy((phi_centroids),p_temp,(m_phi_min)*sizeof(double));

		SavePolarData(i,m_r_min,r_centroids,m_phi_min,phi_centroids);
		
		delete [] r_centroids;
		delete [] phi_centroids;
	}



	if((m_r_min != -1) && (m_phi_min != -1))
	{
//        err = SPQ(m_r_min, m_phi_min, error_value);

		*error_value = E_min;
	    *Mr    = m_r_min;
	    *Mphi  = m_phi_min;
	}
	else
	{
        err = -1;
	}

	return err;
}


/*
int CHistoDataPolar::DP_MSE_PolarQ_2(int Mphi, int nType, double *error_value)
{ 
	int         err=0, m, flag_pnode=1, verbose=1, i1=0, nVers=2;
	long int    n, N, ni=0, nn, undef_val = -99999, ind, ind1;
    DTable      *pD2=NULL; 
	double      E2; //, E_1, E_2;
    NTable      *pNodes=NULL;
    Bounds      *pBnd=NULL;

    CHistoData  *pHData_Phi=NULL;

	*error_value = 1.0e20;

    /////////////////////////////////////////////////////
	pHData_Phi = new CHistoData(); // m_pPhi, m_pR
	if(err != 0)
	{
		printf("Sorry, can't create pPhi\n");
	    return -1;
	}

	N = pHData_Phi->m_nSize;   

     // Allocate memory 
	pNodes = CreateNTable(N+1, Mphi+1);
	if(pNodes == NULL)
	{   
		 return -1;
	}

    SetNTableToValue(pNodes, undef_val);

    // Allocate memory for D
	pD2 = CreateDTable(N+1, Mphi+1);

	if(pD2 == NULL)
	{
		 DestroyNTable(&pNodes); 

		 return -1;
	}

    // Allocate memory for the Bounding Corridor
	pBnd = CreateBounds(nType, N, Mphi, Mphi);
	if(pBnd == NULL)
	{
		 DestroyNTable(&pNodes); 
         DestroyDTable(&pD2); 

		 return -1;
	}
	
	// Calculate the Bounding Corridor 
	err = pHData_Phi->CalcBounds(pBnd);
	if(err != 0)
	{
		 DestroyNTable(&pNodes); 
         DestroyDTable(&pD2); 

		 return -1;
	}

	////////////////////////////////////////////

	pHData_Phi->DP_MSE_search_Phi(pBnd, pD2, pNodes);

	// Init Cluster Index
	for(n = 0; n < N; n++)
	{
        pHData_Phi->m_pClusterIndex[n] = 0;
	}
	pHData_Phi->m_pClusterIndex[0] = 1;

    // Make output solution
    ind = N - 1;  

    for(m = Mphi; m >= 1; m--)
	{
       nn   = ind * pNodes->size_x + (m-pBnd->bottom[ind]);
       ind1 = pNodes->data[nn];
	   //printf("m=%ld) %ld  -> %ld \n", m, ind, ind1);

	   if(ind1 == undef_val)  {
           fprintf(stderr, "ERROR with pNode!!! m=%d;  ind: %d,   ind1: %d, \n", 
			       m, ind, ind1);
		   err = -1;

		   break;
	   }

	   pHData_Phi->m_pClusterIndex[ind+1] = 1;

	   ind = ind1;
	}//for(m

	pHData_Phi->m_pClusterIndex[0] = 1;

    pHData_Phi->CalcClustersList();
    pHData_Phi->TypeClustersList();

	//////////////
	nn = (N-1) * pD2->size_x + (Mphi - pBnd->bottom[N-1]);
    double E2_a  = pD2->data[nn]; 
	
	double E2_b; 
	pHData_Phi->CalcVariance(&E2_b);

	if(pHData_Phi->m_nClustersNumber != Mphi)
	{
		printf("WARNNING! MSE: %d -> %d \n", Mphi, pHData_Phi->m_nClustersNumber);
	}
	
	*error_value = E2_a; 

    //////////////////////////////////////////
	
	if(pBnd->type == 2)
	{ 
		//printf("Bound Type: %d\n", pBnd->type);
		for(m = 1; m <= Mphi; m++)
		{
	         nn = (N-1) * pD2->size_x + (m - pBnd->bottom[N-1]);
             E2 = pD2->data[nn]; 
             pHData_Phi->m_pE2[m] = E2;
			 //printf("m=%d) E2 = %.12f\n", m, E2);
		}//for(m
	}

	// Free memory  
 	DestroyBounds(&pBnd);
	DestroyDTable(&pD2);
	DestroyNTable(&pNodes);

    return 0;
}

*/

//////////////////////////////////////////////////////////////////
//
//
//
//////////////////////////////////////////////////////////////////
/*
int CHistoDataPolar::CalcFunctional_2(double *value)
{
	int       i, j;
	long int  j_r, j_phi, N;
    double    E2_b, E2_c, E2, E_r, E_phi, E_phi_b, E_phi_c, r, phi, myu_r, myu_phi, 
		      b, R2;

	R2 = pHData_R->CalcMoment(2);

	E_r   = 0; 
	E_phi = 0; 
	E_phi_b = 0;
	E_phi_c = 0;
	b = 0;

	N     = (int)pHData_R->m_nSize;

	for(i = 0; i < N; i++)
	{
        r     = pHData_R->m_pX[i];
	    j_r   = pHData_R->m_pClusterNumber[i];
	    myu_r = pHData_R->m_pClusterMeans[j_r];

        // Find index of the R in unsorted input data 
        for(j = 0; j < pR->size; j++) 
		{
           if(r == pR->value[j])  break;
		}

		// Get the correspondent Phi from unsorted input data 
        phi = pPhi->value[j];

        // Find index in sorted data 
        for(j = 0; j < pR->size; j++) 
		{
           if(phi == pHData_Phi->m_pX[j])  break;
		}

		// Get Cluster Number & Means
	    j_phi   = pHData_Phi->m_pClusterNumber[j];
	    myu_phi = pHData_Phi->m_pClusterMeans[j];

		E_r   += (r - myu_r)*(r - myu_r);
		E_phi += r*myu_r*(phi - myu_phi)*(phi - myu_phi);

		E_phi_b += r * r *(phi - myu_phi)*(phi - myu_phi);
		E_phi_c += R2 *(phi - myu_phi)*(phi - myu_phi);
		
		//a = sin((phi - myu_phi)/2);
		//E_phi += r * myu_r * 4 *a * a;
		b  +=  (r-myu_r) * r *(phi - myu_phi)*(phi - myu_phi); 
	}//for(n

	E_r     /= (double)N;
	E_phi   /= (double)N;
	E_phi_b /= (double)N;
	E_phi_c /= (double)N;
	b        /= (double)N;
	
	E2   = E_r + E_phi;
	E2_b = E_r + E_phi_b;
	E2_c = E_r + E_phi_c;

    printf("Functional: %.12f + %.12f =  %.12f \n", E_r, E_phi, E2);
	printf("          : %.12f + %.12f =  %.12f (%.8f)\n", E_r, E_phi_b, E2_b, b);
	printf("          : %.12f + %.12f =  %.12f \n", E_r, E_phi_c, E2_c);

	*value = E2;

	return 0;
}
*/

int CHistoDataPolar::CalcFunctional(double *value)
{
	//int      i, j;
	long int  j_r, j_phi, i, j;
    double    E2_b, E2_c, E2_d, E2, E_r, E_phi, E_phi_b, E_phi_c,E_phi_d,
		      r, phi, myu_r, myu_phi, b, R2, R1, p;
	
	R1 = m_pHData_R->CalcMoment(1);
	R2 = m_pHData_R->CalcMoment(2);
	
	E_r     = 0; 
	E_phi   = 0; 
	E_phi_b = 0;
	E_phi_c = 0;
	E_phi_d = 0;
	b = 0;

	m_pHData_R->CalcClustersList();
	m_pHData_Phi->CalcClustersList();

	/*for(i = 0; i < N; i++)
	{
         printf("n=%ld) X=%.8f; j=%ld; Myu=%.6f; \n", 
			      i, 
				  m_pHData_R->m_pX[i], 
				  m_pHData_R->m_pClusterNumber[i], 
				  m_pHData_R->m_pClusterMeans[i]);
	}
*/
	for(i = 0; i < m_pHData_R->m_nSize; i++)
	{
        r     = m_pHData_R->m_pX[i];
	    j_r   = m_pHData_R->m_pClusterNumber[i];
	    myu_r = m_pHData_R->m_pClusterMeans[j_r];
		p     = m_pHData_R->m_pProby[i];
        E_r   += p*(r - myu_r)*(r - myu_r);

/*        printf("%ld) %ld: r=%.8f, myu_r=%.8f: E_r=%.6f\n", 
			      i, 
				  m_pHData_R->m_pClusterNumber[i], 
				  r,  
                  m_pHData_R->m_pClusterMeans[j_r],
				  E_r
			  );
*/
		
	}

	printf("\n**************\nm_pHData_R->TypeClustersList(); \n");

	
    printf("ISE for Radius: %.8f\n", E_r);

	//////////////////////////////////////////////////////
/*	for(i = 0; i < m_pHData_Phi->m_nSize; i++)
	{
        j_phi   = m_pHData_Phi->m_pClusterNumber[i];

	    printf("%ld) Phi=%.8f; Ri=%.6f) j=%ld; Myu=%.6f; \n", 
			      i, 
				  m_pHData_Phi->m_pX[i], 
				  m_pR[i],
				  m_pHData_Phi->m_pClusterNumber[i], 
				  m_pHData_Phi->m_pClusterMeans[j_phi]
				  );
	}
*/

	//for(i = 0; i < m_pHData_Phi->m_nSize; i++)
	for(i = 0; i < m_nSize; i++)
	{
        phi     = m_pPhi[i];
		p       = m_pHData_Phi->m_pProby[i];
        r       = m_pR[i];

        for(j = 0; j < m_pHData_R->m_nSize; j++)
		{
            if(r == m_pHData_R->m_pX[j])
			{
				j_r   = m_pHData_R->m_pClusterNumber[j];
	            myu_r = m_pHData_R->m_pClusterMeans[j_r];
		
				break;
			}
		}

        for(j = 0; j < m_pHData_Phi->m_nSize; j++)
		{
            if(phi == m_pHData_Phi->m_pX[j])
			{
				j_phi   = m_pHData_Phi->m_pClusterNumber[j];
	            myu_phi = m_pHData_Phi->m_pClusterMeans[j_phi];
		
				break;
			}
		}

		E_phi   +=  p * r*myu_r*(phi - myu_phi)*(phi - myu_phi);
		E_phi_b +=  p * r * r *(phi - myu_phi)*(phi - myu_phi);
		E_phi_c +=  p * R2 *(phi - myu_phi)*(phi - myu_phi);
		E_phi_d +=  p * R1* R1 *(phi - myu_phi)*(phi - myu_phi);
		
		b  +=  p * (r-myu_r) * r *(phi - myu_phi)*(phi - myu_phi); 

/*	    printf("%ld) P=(%.8f, %.6f) C=(%ld,%ld) Myu=(%.6f, %.6f)\n", 
			      i, 
				  m_pPhi[i], m_pR[i], 
				  j_phi, j_r, 
				  myu_phi, myu_r 
				  );

*/
	}//for(n

	/*E_phi   /= (double)pHData_Phi->m_nSize;
	E_phi_b /= (double)pHData_Phi->m_nSize;
	E_phi_c /= (double)pHData_Phi->m_nSize;
	b       /= (double)pHData_Phi->m_nSize;
	*/
	E2   = E_r + E_phi;
	E2_b = E_r + E_phi_b;
	E2_c = E_r + E_phi_c;  // s R2
	E2_d = E_r + E_phi_d;  // s R2

    printf("E_r+<r*myu_r*E_phi>: %.12f + %.12f =  %.12f \n", E_r, E_phi, E2);
	printf("E_r+<r*r*E_phi>    : %.12f + %.12f =  %.12f (%.8f)\n", E_r, E_phi_b, E2_b, b);
	printf("E_r+<R2>*E_phi     : %.12f + %.12f =  %.12f \n", E_r, E_phi_c, E2_c);
	printf("E_r+<R>^2*E_phi    : %.12f + %.12f =  %.12f \n", E_r, E_phi_d, E2_d);

	*value = E2;

	return 0;
}


int CHistoDataPolar::DP_Product_One(ULONG Mr, ULONG Mphi, double *dE2)
{   int err = 0;
    int          nVers=2, nType=1;
	double       R1, R2, E2_R, E2_Phi, E2; 

	R1 = m_pHData_R->CalcMoment(1);
    printf("R1 = %.12f \n", R1);

	R2 = m_pHData_R->CalcMoment(2);
	printf("R2= %.12f (%.12f) \n", R2, sqrt(R2));

	////////////////////
	m_pHData_R->DP_MSE(Mr,     nVers, nType, &E2_R);
    m_pHData_R->CalcClustersList();

	m_pHData_Phi->DP_MSE(Mphi, nVers, nType, &E2_Phi);
	m_pHData_Phi->CalcClustersList();

    E2 =   E2_R + R2 * E2_Phi; 

	printf("E2_r=%f + E2_phi=%f (phi=%f) = (E2=%.12f) \n\n", 
		      E2_R, R2*E2_Phi, E2_Phi, E2);

	*dE2 = E2;

	return err;
}

int CHistoDataPolar::DP_Product_Many(ULONG M, ULONG *Mr, ULONG *Mphi, double *E2)
{   int err = 0;
    int          nVers=2, nType=2, m_r, m_phi, m_r_min, m_phi_min, Mr_max, Mphi_max;
	double       R1, R2, E2_R, E2_Phi, E_min, E_1, E_2, E_curr, a1, a2; 
    CHistoData   *pHData_R=NULL, *pHData_Phi=NULL;
    clock_t     t1, t2;

    pHData_R = new CHistoData(m_pR, m_nSize);
	if(pHData_R == NULL)
	{
		return -1;
	}

	pHData_Phi = new CHistoData(m_pPhi, m_nSize);
	if(pHData_Phi == NULL)
	{
		return -1;
	}

	R1 = pHData_R->CalcMoment(1);
    printf("R1 = %.12f \n", R1);

	R2 = pHData_R->CalcMoment(2);
	printf("R2= %.12f (%.12f) \n", R2, sqrt(R2));

	////////////////////
	t1 = clock();

    Mr_max = (int)(2*sqrt((double)M));
	pHData_R->DP_MSE(Mr_max, nVers, nType, &E2_R);

	Mphi_max = (int)(2*sqrt((double)M));
	pHData_Phi->DP_MSE(Mphi_max, nVers, nType, &E2_Phi);

	E_min = 1.0e35;

    for(m_r = 10; m_r <= Mr_max; m_r++)
	{
		m_phi = (int)((double)M/(double)m_r);

		if(m_phi <= Mphi_max)
		{
			E_1 = pHData_R->m_pE2[m_r];  
	        E_2 = R2 * pHData_Phi->m_pE2[m_phi];  

		    E_curr = E_1 + E_2; 
			
			if(E_curr < E_min) 
			{
               E_min     = E_curr;
			   m_r_min   = m_r;
			   m_phi_min = m_phi;
			   a1 = E_1;
			   a2 = E_2;
			}
	
		    /*printf("m_r=%d, m_phi=%d) %.12f + %.12f = %.12f \n", 
			        m_r, m_phi, E_1, E_2, E_curr);
					*/
		}
	}//for/(m_x

	t2 = clock();
    double dT = ((double) (t2 - t1)) / (double) CLOCKS_PER_SEC;

    printf("E2: %.10f for M_r=%d and M_phi=%d; R2=%.3f (T=%.3f) \n",
	E_min, m_r_min, m_phi_min, R2, dT);

	nType = 1;
	m_pHData_R->DP_MSE(m_r_min,     nVers, nType, &E2_R);
    m_pHData_R->CalcClustersList();

	m_pHData_Phi->DP_MSE(m_phi_min, nVers, nType, &E2_Phi);
	m_pHData_Phi->CalcClustersList();

	*Mr   = m_r_min;
	*Mphi = m_phi_min;
	*E2   = E_min; 

	return err;
}



int CHistoDataPolar::MaxLloyd(ULONG M, ULONG *Mr, ULONG *Mphi, double *E2)
{
    ULONG     terr = 0, version=2, type=1, m_r, m_phi, Mr_max, Mphi_max, m_r_min, m_phi_min;
	double  E_1, E_2, E_curr, E_min, R2;

    if((m_pHData_R == NULL) || (m_pHData_Phi == NULL))
	{
		return -1;
	}

    R2 = m_pHData_R->CalcMoment(2);

	Mr_max = (ULONG)(2*sqrt((double)M));
	Mphi_max = (ULONG)(2*sqrt((double)M));

	E_min = 1.0e35;


    for(m_r = (ULONG)(sqrt((double)M)/(double)4); m_r <= Mr_max; m_r++)
	{
		m_phi = (ULONG)((double)M/(double)m_r);

		if(m_phi <= Mphi_max)
		{
			terr = m_pHData_R->MaxLloyd(m_r, &E_1);
			terr = m_pHData_Phi->MaxLloyd(m_phi, &E_2);

			E_curr = E_1 + E_2*R2; 
			
			if(E_curr < E_min) 
			{
               E_min     = E_curr;
			   m_r_min   = m_r;
			   m_phi_min = m_phi;
			}
	
	//	    printf("m_r=%d, m_phi=%d) %.12f + %.12f = %.12f \n", 
	//		        m_r, m_phi, E_1, E_2*R2, E_curr);
					
		}
	}//for/(m_x

    terr = m_pHData_R->MaxLloyd(m_r_min, &E_1);
	terr = m_pHData_Phi->MaxLloyd(m_phi_min, &E_2);

	m_pHData_R->CalcClustersList();  
	m_pHData_Phi->CalcClustersList();  

	*E2   = E_min;
    *Mr   = m_r_min;
	*Mphi = m_phi_min;

	return terr;
}

int CHistoDataPolar::MaxLloyd(ULONG Mr, ULONG Mphi, double *E2)
{
    ULONG     terr = 0, version=2, type=1;
	double  E2_1, E2_2, R2;

    if((m_pHData_R == NULL) || (m_pHData_Phi == NULL))
	{
		return -1;
	}

	R2 = m_pHData_R->CalcMoment(2);

	terr = m_pHData_R->MaxLloyd(Mr, &E2_1);
	
    terr = m_pHData_Phi->MaxLloyd(Mphi, &E2_2);

	m_pHData_R->CalcClustersList();  
	m_pHData_Phi->CalcClustersList();  

	*E2 = E2_1 + E2_2*R2;

	return terr;
}

/************************************************
** Unconstrained Polar Quantization
**
**  Given number of clusters Mr and total number of clusters Mphi.
**  Find optimal allocation of clusters number among rings
**  Uniform quantization by Phase!
**  Find MSE min UPQ 
**
**  Data should be sorted by Radius!
************************************************/
int CHistoDataPolar::UPQ(ULONG M, ULONG Mr, double *E2, ULONG *phi_c, int mode)
{
    ULONG         err=0, nVers=2, nType=2, Mx=100, j, m1, m2, m0, k, cell_k, m, n;
	long int    i; 
	double      E2_R, phi, delta, c, b, a, S, dE2, myu_r;
		        //S_r1, S_r2;
	char        *fname=NULL;
	CHistoData  *pHData=NULL;
	DArray      **ppPhi=NULL, **ppE2_phi=NULL; //, *pRc=NULL;  
	NArray      *pM_min=NULL, *pM_max=NULL, *pM_num=NULL; 

	pM_num = CreateNArray(Mr); 
	pM_min = CreateNArray(Mr); 
	pM_max = CreateNArray(Mr); 
	
	m_pHData_R->DP_MSE(Mr, nVers, nType, &E2_R);
//	m_pHData_R->CalcClustersList();
    //m_pHData_R->TypeClustersList();

	///////////////////////////////
	// Allocate memory for Rings
	ppPhi    = (DArray**)malloc(Mr*sizeof(DArray*)); 	
    ppE2_phi = (DArray**)malloc(Mr*sizeof(DArray*)); 	
 
	SortDataByRadius();

	// Construct Rings
	err = ConstructRings(Mr, ppPhi);
 
	long Max = 0;;

    ////
	m0 = (ULONG)(M/Mr);
	m1 = (ULONG)(1);
    m2 = (ULONG)(M-(Mr-1)*m1);
//	m2 = (ULONG)(m_nSize-1);
    for(k = 0; k < Mr; k++)
	{
		n = ppPhi[k]->size;

		pM_num->value[k] = m0;
		pM_min->value[k] = m1;
		pM_max->value[k] = n-1;
//		pM_max->value[k] = m2;
				// znum
		if(pM_num->value[k] >= n-1)
		{
		   pM_num->value[k] = n-1;
		}

		// m
		if(pM_max->value[k] > n-1)
		{
			pM_max->value[k] = n-1;
		}

	    if(pM_min->value[k] > pM_max->value[k])
		{
			pM_min->value[k] = pM_max->value[k];
		}

		if(pM_min->value[k] > pM_num->value[k])
		{
			pM_min->value[k] = pM_num->value[k];
		}
/*xxxxxxxx*/
		if(pM_max->value[k] > m2)
		{
			pM_max->value[k] = m2;
		}
	}//for(k
   
    for(k = 0; k < Mr; k++)
	{
	   ppE2_phi[k] = CreateDArray(m2-m1+1);

    //   printf("k=%3d) m=(%2d, %2d); Cnt=%2d]  ", 
	  //   k, m1, m2, ppPhi[j]->size);

       myu_r = m_pHData_R->m_pClusterMeans[k];

       for(m = m1; m <= m2; m++)
	   {   
		   if(mode)
			delta = PI/(2*m);
		   else
			delta = 2*PI/m; 
           S = 0;
           for(i = 0; i < ppPhi[k]->size; i++)
           {
				 phi    = ppPhi[k]->value[i]; 
                 cell_k = (int)((phi + delta/2)/delta);
			     c      = cell_k * delta;

				 a = 2*sin((phi-c)/2)* myu_r;
				 b = a*a;
				 //a = (phi-c) * myu_r;
				 S += b;
		   }//for(i
    	   ppE2_phi[k]->value[m-m1] = S / m_nSize; 

       //    printf("[%d]=%.4f ", m-m1, ppE2_phi[k]->value[m-m1]);

	   }//for(m
      // printf("\n");

	}//for(k

	// Allocation of resources !!!
	double E2_phi_1 = 0, E2_phi_2 = 0;
    err = PostOptimization(ppE2_phi, pM_min, pM_num, pM_max, Mr, &E2_phi_1);

	int S_m = 0;
    /*for(k = 0; k < Mr; k++)
	{
		S_m += pM_num->value[k];
	}
	printf("S_m=%d \n", S_m);
	*/

	S_m=0;
	E2_phi_2 = 0;
	int counter = 0;
    for(k = 0; k < Mr; k++)
	{
        m =  pM_num->value[k];
		m1 = pM_min->value[k];
		
		phi_c[k] = m;
		counter+=m;
		S = ppE2_phi[k]->value[m-m1];
		E2_phi_2 += S;
        S_m += m;

//       printf("k=%3d;  E2=%12.8f (%12.8f)  \n", 
//		      k,  S, E2_phi_2);

	}//for(j

	dE2 = E2_R + E2_phi_2; 
	*E2 = dE2;

//	printf("Mr=%3d; M=%3d:  E2=%12.4f (=%12.4f+%16.4f) \n", 
//		Mr, S_m, dE2, E2_R, E2_phi_1);

  	// Free memory
	DestroyNArray(&pM_num);
	DestroyNArray(&pM_min);
	DestroyNArray(&pM_max);

	for(j = 0; j < Mr; j++)
	{
	    DestroyDArray(&ppPhi[j]);
	    DestroyDArray(&ppE2_phi[j]);
	}//for(j

	free(ppPhi);    ppPhi = NULL; 
	free(ppE2_phi); ppE2_phi=NULL;

	return 0;
}

/************************************************
**  Unconstrained Polar Quantization
**
**  Given number of clustewrs Mr and total number of clusters Mphi.
**  Find optimal allocation of clusters number among rings
**  Optimal quantization by Phase!
**  Find MSE min UPQ 
**
**  Data should be sorted by Radius!
************************************************/
int CHistoDataPolar::UPQ_Full(ULONG M, ULONG Mr, double *E2, CODEBOOK *CB)
{
    ULONG         err=0, nVers=2, nType=2, Mx=100, i, j, m1, m2, m0, k, S_m=0;
	long int    m, n; 
	double      E2_R, S, dE2, myu_r;
	char        *fname=NULL;
	CHistoData  *pHData=NULL, **ppHData_Ring=NULL;
	DArray      **ppPhi=NULL, **ppE2_phi=NULL;   
	NArray      *pM_min=NULL, *pM_max=NULL, *pM_num=NULL; 

	pM_num = CreateNArray(Mr); 
	pM_min = CreateNArray(Mr); 
	pM_max = CreateNArray(Mr); 
	
	//m_pHData_R->DP_MSE(Mr, nVers, nType, &E2_R);
	
	///////////////////////////////
	// Allocate memory for Rings
	ppHData_Ring = (CHistoData**)malloc(Mr * sizeof(CHistoData*)); 	 
	ppPhi        = (DArray**)malloc(Mr * sizeof(DArray*)); 	
    ppE2_phi     = (DArray**)malloc(Mr * sizeof(DArray*)); 	
 
	SortDataByRadius();

	// Construct Rings
	err = ConstructRings(Mr, ppPhi);

	m1 = (int)(  1);

//	m0 = (int)(M/Mr);
//  m2 = (int)((6.0*M)/Mr);

	m0 = (int)(M - Mr + 1);
	m2 = (int)(M - Mr + 1);


    for(k = 0; k < Mr; k++)
	{
	    n = ppPhi[k]->size;

		pM_num->value[k] = m0;
		pM_min->value[k] = m1;
		pM_max->value[k] = n-1;

		// num
		if(pM_num->value[k] >= n-1)
		{
		   pM_num->value[k] = max(n-1,1);
		}

		// m
		if(pM_max->value[k] >= n-1)
		{
			pM_max->value[k] = max(n-1,1);
		}

	    if(pM_min->value[k] > pM_max->value[k])
		{
			pM_min->value[k] = max(pM_max->value[k],1);
		}

		if(pM_max->value[k] > pM_num->value[k])
		{
			pM_max->value[k] = pM_num->value[k];
		}

//   	printf("k=%d)  %d < %d < %d (%d)\n", 
//			k, pM_min->value[k], pM_num->value[k], pM_max->value[k], n);

	}

	// Optimal Quantization of Rings

	for(k = 0; k < Mr; k++)
	{
		int a = ppPhi[k]->size;

	   ppHData_Ring[k] = new CHistoData(ppPhi[k]);
       ppHData_Ring[k]->DP_MSE(pM_max->value[k], nVers, nType, E2);

	   if(pM_max->value[k] >= ppHData_Ring[k]->m_nSize-1)
	   {
			pM_max->value[k] =  max(ppHData_Ring[k]->m_nSize-1,1);
	   }
	}//for(k

	// Get Quant erorrs
    for(k = 0; k < Mr; k++)
	{
	   m1 = pM_min->value[k];
       m2 = pM_max->value[k];
       ppE2_phi[k] = CreateDArray(m2 - m1 + 1);

	   printf("k=%3d) m=(%2d, %2d): ", k , m1, m2);

	   //for(i = 0; i < mppHData_Ring[k]->m_nSize; i++)
	   for(i = 0; i < m2-m1+1; i++)
	   {
		  myu_r = m_pHData_R->m_pClustersList[(Mr-1)*Mr/2 + k];
		  
          ppE2_phi[k]->value[i] = myu_r * myu_r * ppHData_Ring[k]->m_pE2[i + m1];

	//	  printf("[%d] %.12f ", i, ppE2_phi[k]->value[i]);
	   }//for(i

	   printf("\n");

	}//for(k

	// Allocation of resources !!!
	double E2_phi_1 = 0, E2_phi_2 = 0;

	if(Mr == 1)
	{
		pM_num->value[0] =	M;
	}
	else
	    err = PostOptimization(ppE2_phi, pM_min, pM_num, pM_max, Mr, &E2_phi_1);
	   //MakeConfigurationDP(M,ppE2_phi,pM_num,Mr);

	//if(err == 0)
	//{

	  S_m=0;
	  E2_phi_2 = 0;
      for(k = 0; k < Mr; k++)
	  {
          m =  pM_num->value[k];
	      m1 = pM_min->value[k];

		  if(m < ppE2_phi[k]->size)
			S = ppE2_phi[k]->value[m-1];
		  else
			S = 0;

		  E2_phi_2 += S;
          S_m += m;

		  printf("k=%3d) m=%3d; Cnt=%4d, E2 = %12.4f\n", 
 			     k,     m,     ppPhi[k]->size, S);
	
	  }//for(j

	   
	  E2_R = this->m_pHData_R->m_pE2[Mr];
//	  E2_R = 0;
	  dE2 = E2_R + E2_phi_2; 

	   printf("Mr=%3d; M=%3d:  E2=%12.4f (=%12.4f+%16.4f) \n", 
		Mr, S_m, dE2, E2_R, E2_phi_2);
	//}

	   (*E2)= dE2;


	//Keep the data in teh CB structue
	CreateNewTrainingSet(CB,M,2,1,sizeof(VECTORELEMENT),0,100,0,"UDAFF.COM");
	int cur_index = 0; 
	int shift_1;
	int shift_2;

	shift_1 = (Mr-1)*Mr/2;

	
	for(i=0;i<Mr;i++)
	{
		m =  pM_num->value[i];
		shift_2 = (m-1)*m/2;

		for(j=0;j<m;j++)
		{
			CB->Book[cur_index].vector[0] = (VECTORELEMENT)(m_pHData_R->m_pClustersList[shift_1+i]);
			CB->Book[cur_index].vector[1] = (VECTORELEMENT)(ppHData_Ring[i]->m_pClustersList[shift_2 + j]*pow(10,PHI_SCALE));
			cur_index++;
		}
	}



  	// Free memory
	DestroyNArray(&pM_num);
	DestroyNArray(&pM_min);
	DestroyNArray(&pM_max);

	for(j = 0; j < Mr; j++)
	{
	    DestroyDArray(&ppPhi[j]);
	    DestroyDArray(&ppE2_phi[j]);
		delete ppHData_Ring[j];
	
	}//for(j

	 
	
	delete [] ppHData_Ring;

	free(ppPhi);    ppPhi = NULL; 
	free(ppE2_phi); ppE2_phi=NULL;

	return 0;
}



/********************************************************
**
**  PostOptimization
**
**  Parameters:
**
**   Return Error code
**    0:  OK
**   -1:  ERROR
**
********************************************************/
int CHistoDataPolar::PostOptimization(DArray **ppE2, NArray *pM_min, NArray *pM_num, 
									  NArray *pM_max, long int rings_num, double *dE2)
{ 



    long int        err=0, M1, M2, ind, W_curr, W_prev, ksi, x, k, x_min, ksi0, S_W_prev=0, 
               S_W_curr=0, k0, H = 2, ksi_1, ksi_2, verbose=0, x1, new_left_x=-1; 
	long int   K, M, nn; 
    double     d_min, d, F, D;
	DTable     *pD2=NULL; 
	DArray     *pE2_curr=NULL;  // *pE2_prev=NULL, 
    NTable     *pNodes=NULL;
	
	if((ppE2 == NULL) || (rings_num <= 0))
  	{
		return -1;
	}

	K  = rings_num;   // Number of Rings
    M  = 0;           // Total number of Clusters;  
	M1 = 0; 
	M2 = 0;

	for(k = 0; k < rings_num; k++)
	{
		M  += pM_num->value[k];
		M1 += pM_min->value[k];
        M2 += pM_max->value[k];

	//	printf("k=%d)  %d < %d < %d\n", 
	//		k, pM_min->value[k], pM_num->value[k], pM_max->value[k]);
	}//for(k

    /* Allocate memory */
	pNodes = CreateNTable(K, M2);
	if(pNodes == NULL)
	{   
		 return -1;
	}
	//SetNTableToValue(pNodes, UNDEF_VAL);

    /* Allocate memory for D */
	H = 2;
	pD2 = CreateDTable(H, M2+1);
	if(pD2 == NULL)
	{
		 DestroyNTable(&pNodes); 

		 return -1;
	}
	SetDTableToValue(pD2, MIN_VAL);

    /**************/

  	/**************/
	pD2->data[0] = 0;
    S_W_curr     = 0;

	/*  k = 0; */
    //pPoly_curr = ppPoly_src[0];
	pE2_curr = (DArray*)ppE2[0];

    //W_curr = pPoly_curr->segms_num_max - pPoly_curr->segms_num_min + 1; 
	W_curr = pM_max->value[0] - pM_min->value[0] + 1; 
    S_W_curr += W_curr; 

    if(verbose==1)
	{
	   printf("\n************** k = 0: W = %d  ****************\n", W_curr);
	}

    for(ksi = 0; ksi < S_W_curr; ksi++)
	{
		nn = 0 * pD2->size_x + ksi;
		//pD2->data[nn] = pPoly_curr->E2[ksi];
		pD2->data[nn] = pE2_curr->value[ksi];
 
        nn = 0 * pNodes->size_x + ksi;
        pNodes->data[nn] = ksi;

		//printf(" D[0][%2d] =%10.12f  \n", ksi, pD2->data[nn]);

	}//for(ksi
    //printf("\n");

	S_W_prev = S_W_curr;
	ksi0  = M - M1;
	ksi_1 = 0;
    ksi_2 = ksi0;
	ksi_2 = (ksi_2 >= ksi0+1) ? ksi0+1 : ksi_2;

	/**  k = 1 .. K-2; **/
//    for(k = 1; k <= K-2; k++)
	for(k = 1; k <= K-2; k++)
	{
        //pPoly_curr = ppPoly_src[k];
		//W_curr = pPoly_curr->segms_num_max - pPoly_curr->segms_num_min + 1; 
		pE2_curr = (DArray*)ppE2[k];
        W_curr = pM_max->value[k] - pM_min->value[k] + 1; 
   
        //pPoly_prev = ppPoly_src[k-1];
        //W_prev = pPoly_prev->segms_num_max - pPoly_prev->segms_num_min + 1; 

		//pE2_prev = (DArray*)ppE2[k-1];
		W_prev= pM_max->value[k-1] - pM_min->value[k-1] + 1; 

	    if(verbose == 1) 
		{
	       printf("\n ******* k = %d: Range for ksi: [%d, %d] **********\n", 
		                 k, ksi_1, ksi_2);
		}

		new_left_x = -1;
	    for(ksi = ksi_1; ksi <= ksi_2; ksi++)
		{
	  	   // Find min
           d_min = MIN_VAL; 
           x_min = UNDEF_VAL;

           x1 = 0;
           
	 	   for(x = x1; x < W_curr; x++) 
		   {
		 	  if( (0 <= ksi-x ) && (ksi-x < S_W_prev)) 
			  {
			     //F = pPoly_curr->E2[x];
                 F = pE2_curr->value[x];
		         nn = ((k-1)%H)*pD2->size_x + (ksi-x);
		         D  = pD2->data[nn];

			     if(D != MIN_VAL)
				 {
  			       d = D + F;
			       if(d < d_min)
				   {
                      d_min = d;
			          x_min = x;
					  
					  new_left_x = x;

				   }

				   if(verbose==1)
				   {
				      printf("d[%d] = (F[%d][%d]=%.2f) + (D[%d][%d]=%.2f) = %.2f  >= %.6f (%d)\n", 
		                        ksi,    k,  x,    F,     k-1, ksi-x,    D,    d, d_min, x_min);
				   }

				}//if(D
			  }//if(ksi

			}//for(x

           nn = k * pNodes->size_x + ksi;
           pNodes->data[nn] = x_min;

           nn = ((k)%H) * pD2->size_x + ksi;
           pD2->data[nn] = d_min;

		   if(d_min != MIN_VAL)
		   {
				if(verbose==1)
				{
		            printf("*** D[%d][%d] =%.2f;  x_min = %2d\n", 
		                    k, ksi, d_min, x_min);

                    F = pE2_curr->value[x_min];
		            nn = ((k-1)%H)*pD2->size_x + (ksi-x_min);
		            D  = pD2->data[nn];

  			        printf(" -> D[%d][%d] = (F[%d][%d]=%.2f) + (D[%d][%d]=%.2f) = %.2f \n\n", 
                                k, ksi,      k,  x_min,  F,    k-1, ksi-x_min, D, d_min);


				}
		   }
		}//for(ksi

	    S_W_prev  = S_W_curr;
  	    S_W_curr += W_curr;

		// Clear row of pD2
		for(x = 0; x < pD2->size_x; x++)
		{
           nn = ((k-1)%H) * pD2->size_x + x;
           pD2->data[nn] = MIN_VAL;
		}//for(x
	}//for(k
	
	/** k  = K-1; */ 
    //pPoly_curr = ppPoly_src[K-1];
    //W_curr = pPoly_curr->segms_num_max - pPoly_curr->segms_num_min + 1; 
    pE2_curr = ppE2[K-1];
    W_curr = pM_max->value[K-1] - pM_min->value[K-1] + 1; 

//    printf("\n**** K-1 = %d: W = %d; (%d, %d) \n", 
//		     K-1, W_curr, pM_min[K-1], pM_max[K-1]);

    //pPoly_prev = ppPoly_src[K-2];
    //W_prev = pPoly_prev->segms_num_max - pPoly_prev->segms_num_min + 1; 
    //pE2_prev = ppE2[K-2];
    W_prev = pM_max->value[K-2] - pM_min->value[K-2] + 1; 

    d_min = MIN_VAL;
    x_min = UNDEF_VAL;
	
	ksi0  = M - M1;
    k0    = K - 1;

	if(verbose == 1)
	{
	   printf("k0 = %d; Ksi0 = %d;  M=%d, M1=%d; S_W_prev=%d \n",  
	           k0, ksi0, M, M1, S_W_prev);
	}

	for(x = 0; x < W_curr; x++)
	{
	  if(verbose == 1)
	  {
         printf("x = %d) W_curr = %d; ksi0-x=%d;   S_W_prev=%d; \n", 
			     x, W_curr, ksi0-x, S_W_prev);
	  }
	  
	  if((0 <= ksi0-x) && (ksi0-x < S_W_prev))
	  {
		   //F = pPoly_curr->E2[x];
		   F = pE2_curr->value[x];
	       nn = ((k0-1)%H) *pD2->size_x + (ksi0-x);
	       D  = pD2->data[nn];
		   if(D != MIN_VAL)
		   {
  		      d = D + F;

		      if(d < d_min) 
			  {
                 d_min = d;
		         x_min = x;
			  }

			  if(verbose==1)
			  {
				  printf("d[%d] = (F[%d][%d]=%.2f) + (D[%d][%d]=%.2f) = %.2f  >= %.6f (%d)\n", 
		          ksi0,     k0,   x,    F,     k0-1, ksi0-x, D,    d, d_min, x_min);
			  }
		   }//if(D
	  }
	}//for(x

    nn = k0 * pNodes->size_x + ksi0;
    pNodes->data[nn] = x_min;

    nn = ((k0)%H) * pD2->size_x + ksi0;
    pD2->data[nn] = d_min;

   	if(d_min != MIN_VAL)
	{
     	if(verbose == 1)
		{
           printf("*** D[%d][%d] =%.2f;  x_min = %2d\n", k0, ksi0, d_min, x_min);

                    F = pE2_curr->value[x_min];
		            nn = ((k-1)%H)*pD2->size_x + (ksi0-x_min);
		            D  = pD2->data[nn];

  			        printf(" -> D[%d][%d] = (F[%d][%d]=%.2f) + (D[%d][%d]=%.2f) = %.6f \n\n", 
                                k, ksi0,      k,  x_min,  F,    k-1, ksi0-x_min, D, d_min);


		}
	}

    if(x_min == UNDEF_VAL)
    {
            printf("ERROR before BackTrack x_min = %d \n", x_min);
    }
    else
    {      
       // GetPolylineByPathNumber(ppPoly_src[K-1], x_min);
       // CalcNodesList(ppPoly_src[K-1]);

	   /* Backtrack; */
       ind = ksi0; 
	   ind = ind - x_min;

	   pM_num->value[K-1] = x_min + pM_min->value[K-1];

	   int  S_x = 0;

//	   printf("k=%2d) %2d\n", K-1, pM_num->value[K-1]);
       
	   S_x += pM_num->value[K-1];

       for(k = K-2; k >= 0; k--)
	   {
           nn    = k * pNodes->size_x + ind;
           x_min = pNodes->data[nn]; 

           if(x_min == UNDEF_VAL)  
		   {
		      printf("ERROR in BackTrack!!!  k=%d) x_min = %d \n", k, x_min);
              err = -1;

		        break;
		   }

		   pM_num->value[k] = x_min + pM_min->value[k];
		   S_x += pM_num->value[k];
//		   printf("k=%2d) %2d   \n", k, pM_num->value[k]);
           ind = ind - x_min;
	   }//for(k

	   printf("\n");
	   printf("S_x= %d \n", S_x);

       // Final result 
       nn  = ((K-1)%H) * pD2->size_x + ksi0;
       *dE2 = pD2->data[nn];
	   
	 //  printf("dE2=%.4f for (%d,%d)\n", pD2->data[nn], K-1, ksi0);

	}
	
	// Free memory  
    DestroyDTable(&pD2);
	DestroyNTable(&pNodes);

    return err;
}


int CHistoDataPolar::ConstructRings(ULONG Mr, DArray **ppPhi)
{
	ULONG       terr=0,  k, cnt; 
	long int  n1, n2, i, ii, j;
	double    Thr_1, Thr_2, r, tmp;


	int shift= (Mr-1)*Mr;

	// Construct Rings
	int cur_index = 0;
	for(k = 0; k < Mr; k++)
	{
/*		n1 = m_pHData_R->m_pClusterStart[k];

		if(Mr == 1)
		{
			n2 = m_pHData_R->m_nSize-1;
		}
		else if(k <= m_pHData_R->m_nClustersNumber-2)
		{
	       n2 = m_pHData_R->m_pClusterStart[k+1]-1;   
		}
	    else
		{  
           n2 = m_pHData_R->m_nSize-1;    
		}

		if(n2 < 0)
		{
			break;
		}*/

//		Thr_1 = m_pHData_R->m_pX[n1];
//		Thr_2 = m_pHData_R->m_pX[n2];

		Thr_1 = m_pHData_R->m_pClustersThr[shift + (cur_index++)];
		Thr_2 = m_pHData_R->m_pClustersThr[shift + (cur_index++)];

	    cnt = 0;
	    for(i = 0; i < m_nSize; i++)
		{
           //r = m_pHData_R->m_pX[i];
		   r = m_pR[i];
           if((Thr_1 <= r) && (r <= Thr_2))  
		   {
			   cnt++;
		   }
		}//for(i

		ppPhi[k] = CreateDArray(cnt);

		// Get data for Rings
	    for(i = 0, ii=0; i < m_nSize; i++)
		{
			r = m_pR[i];
            if((Thr_1 <= r) && (r <= Thr_2))
			{
			   ppPhi[k]->value[ii] = m_pPhi[i];
			   ii += 1;
			}
		}//for(i
	}//for(k
	  
	// Sort Phases in Rings
    for(k = 0; k < Mr; k++)
	{

	//bubble sorting		
/*		for(i = 0; i <= ppPhi[k]->size-1; i++)
	   {
          for(j = i+1; j <= ppPhi[k]->size-1; j++)
		  {
		     if(ppPhi[k]->value[i] > ppPhi[k]->value[j])
			 {
                tmp                = ppPhi[k]->value[j];
			    ppPhi[k]->value[j] = ppPhi[k]->value[i];
			    ppPhi[k]->value[i] = tmp;
			 }
		  }//for(j
	   }//for(i*/

		heapSort(ppPhi[k]->value, ppPhi[k]->size);


	}//for(k


	return terr;
}

///////////////////////////////////////////////////////////////
// G E T
///////////////////////////////////////////////////////////////

long int CHistoDataPolar::GetClustersNumberR()
{
	return m_pHData_R->GetClustersNumber();
}

long int CHistoDataPolar::GetClustersNumberPhi()
{
	return m_pHData_Phi->GetClustersNumber();
}
///////////////////////////////////////////////////////////////

double* CHistoDataPolar::GetThresholdsR()
{
	return m_pHData_R->GetThresholds(); 
}

double* CHistoDataPolar::GetThresholdsPhi()
{
	return m_pHData_Phi->GetThresholds(); 
}

///////////////////////////////////////////////////////////////
double* CHistoDataPolar::GetMeansR(unsigned long Mr)
{
	return m_pHData_R->GetMeans(Mr);
}

double* CHistoDataPolar::GetMeansPhi(unsigned long Mphi)
{
	return  m_pHData_Phi->GetMeans(Mphi);
}

///////////////////////////////////////////////////////////////
double* CHistoDataPolar::GetVariancesR()
{
   return m_pHData_R->GetVariances();
}

double*  CHistoDataPolar::GetVariancesPhi()
{
   return m_pHData_Phi->GetVariances();
}

void CHistoDataPolar::heapSortRadius()
{
	
   ULONG array_size;

   long i;

   double temp;
   ULONG tmp1;

	array_size = m_nSize;

  for (i = (array_size / 2)/*-1*/; i >= 0; i--)
    siftDownRadius(i, array_size-1);

  for (i = array_size-1; i >= 1; i--)
  {
    temp = m_pR[0];
    m_pR[0] = m_pR[i];
    m_pR[i] = temp;

    temp = m_pPhi[0];
    m_pPhi[0] = m_pPhi[i];
    m_pPhi[i] = temp;

    temp = m_pProby[0];
    m_pProby[0] = m_pProby[i];
    m_pProby[i] = temp;

    tmp1 = m_pHisto[0];
    m_pHisto[0] = m_pHisto[i];
    m_pHisto[i] = tmp1;

    siftDownRadius(0, i-1);
  }
}

void CHistoDataPolar::siftDownRadius(ULONG root, ULONG bottom)
{
  ULONG done, maxChild, tmp1;
  double temp;

  done = 0;
  while ((root*2 <= bottom) && (!done))
  {
    if (root*2 == bottom)
      maxChild = root * 2;
    else if (m_pR[root * 2] > m_pR[root * 2 + 1])
      maxChild = root * 2;
    else
      maxChild = root * 2 + 1;

    if (m_pR[root] < m_pR[maxChild])
    {
		temp = m_pR[root];
		m_pR[root] = m_pR[maxChild];
		m_pR[maxChild] = temp;

		temp = m_pPhi[root];
		m_pPhi[root] = m_pPhi[maxChild];
		m_pPhi[maxChild] = temp;

		temp = m_pProby[root];
		m_pProby[root] = m_pProby[maxChild];
		m_pProby[maxChild] = temp;

		tmp1 = m_pHisto[root];
		m_pHisto[root] = m_pHisto[maxChild];
		m_pHisto[maxChild] = tmp1;

		root = maxChild;
    }
    else
      done = 1;
  }
}

void CHistoDataPolar::heapSortPhase()
{
	
   ULONG array_size;

   long i;

   double temp;
   ULONG tmp1;

	array_size = m_nSize;

  for (i = (array_size / 2)/*-1*/; i >= 0; i--)
    siftDownPhase(i, array_size-1);

  for (i = array_size-1; i >= 1; i--)
  {
    temp = m_pR[0];
    m_pR[0] = m_pR[i];
    m_pR[i] = temp;

    temp = m_pPhi[0];
    m_pPhi[0] = m_pPhi[i];
    m_pPhi[i] = temp;

    temp = m_pProby[0];
    m_pProby[0] = m_pProby[i];
    m_pProby[i] = temp;

    tmp1 = m_pHisto[0];
    m_pHisto[0] = m_pHisto[i];
    m_pHisto[i] = tmp1;

    siftDownPhase(0, i-1);
  }
}

void CHistoDataPolar::siftDownPhase(ULONG root, ULONG bottom)
{
  ULONG done, maxChild, tmp1;
  double temp;

  done = 0;
  while ((root*2 <= bottom) && (!done))
  {
    if (root*2 == bottom)
      maxChild = root * 2;
    else if (m_pPhi[root * 2] > m_pPhi[root * 2 + 1])
      maxChild = root * 2;
    else
      maxChild = root * 2 + 1;

    if (m_pPhi[root] < m_pPhi[maxChild])
    {
		temp = m_pR[root];
		m_pR[root] = m_pR[maxChild];
		m_pR[maxChild] = temp;

		temp = m_pPhi[root];
		m_pPhi[root] = m_pPhi[maxChild];
		m_pPhi[maxChild] = temp;

		temp = m_pProby[root];
		m_pProby[root] = m_pProby[maxChild];
		m_pProby[maxChild] = temp;

		tmp1 = m_pHisto[root];
		m_pHisto[root] = m_pHisto[maxChild];
		m_pHisto[maxChild] = tmp1;

		root = maxChild;
    }
    else
      done = 1;
  }
}


/********************************************************
**
**  PostOptimization(implemented by akimov)
**
**  Parameters:
**
**	Return E2
********************************************************/



double CHistoDataPolar::MakeConfigurationDP(int CurrentMaxNumber, DArray **Errors, NArray *Config, int NumberOfRings)
{
	int i,m,n,k;

	//Number of centroids
	//needed to add to all 
	// rings
	int RealMaxNumber = CurrentMaxNumber - NumberOfRings;

	//By default we add one centroid to each 
	//ring
	for(i=0;i<NumberOfRings;i++)
		Config->value[i] = 1;

	
	//if there is no centreo
	if(RealMaxNumber == 0)
	{
		double Sum = 0;

		for(i=0;i<NumberOfRings;i++)
		{
			m = Config->value[i]-1;
			Sum += Errors[i]->value[m];
		}

		return Sum;
	}

	if(NumberOfRings == 1)
	{
		Config->value[0] = CurrentMaxNumber;
		return Errors[NumberOfRings]->value[CurrentMaxNumber-1];
	}


	double **MinVal;
	int **MinArg;

	MinVal = NULL;
	MinArg = NULL;


	MinVal = (double**)malloc(sizeof(double*)*(NumberOfRings-1));
	MinArg = (int**)malloc(sizeof(int*)*(NumberOfRings-1));

	for(i=0;i<(NumberOfRings-1);i++)
	{
		MinVal[i] = (double*)malloc(sizeof(double)*(RealMaxNumber + 1));
		MinArg[i] = (int*)malloc(sizeof(int)*(RealMaxNumber + 1));
	}



	//DP resource distribution problem
//	printf("\nDP configuration for %d clusters", NumberOfRings);
	
	for(n = 1; n < NumberOfRings;n++)
	{
		int Start;

		Start = (n==NumberOfRings-1?RealMaxNumber:0);
		
		for(m = Start; m <= RealMaxNumber;m++)
		{
			double min_val = (double)MAXLLONG;
			int arg_min = -1;

			for(k=0; k<=m; k++)
			{
				double temp_val = 0;
				
				if(n==1)
				{
					if(m-k>=Errors[0]->size)
						temp_val += 0;
					else
						temp_val += Errors[0]->value[m-k];
				}
				else
					temp_val += MinVal[n - 2][m-k];
					
				if(k>=Errors[n]->size)
					temp_val += 0;
				else
					temp_val += Errors[n]->value[k];
				
				if(temp_val < min_val)
				{
					min_val = temp_val;
					arg_min = k;
				}
			}

			MinVal[n-1][m] = min_val;
			MinArg[n-1][m] = arg_min;

//			if(n-1>=(NumberOfClusters-1) || m>RealMaxNumber)
//				m = m;

//			printf("\nfor %d clusters optimal distribution is: %d  vs. %d",n+1, m-arg_min+1, arg_min+1);
		}
	}


	


	n = NumberOfRings - 1;
	k = RealMaxNumber;
	int sum = 0;


	while(n>0)
	{
		Config->value[n] = MinArg[n-1][k] + 1;
		sum += Config->value[n];

		k = k - MinArg[n-1][k];
		n = n-1;
	}

	Config->value[0] = CurrentMaxNumber - sum;


	for(i=0;i<NumberOfRings-1;i++)
	{
		delete [] MinVal[i];
		delete [] MinArg[i];
	}

	free(MinVal);
	free(MinArg);
	

	double Sum = 0;

	for(i=0;i<NumberOfRings;i++)
		Sum += Errors[i]->value[Config->value[i]-1];

	return Sum;
}
/******************************************************************


                            E N D
******************************************************************/