/*******************************************************************************
**  Filename: SearchBasic.cpp
**  
** 
**  Dynamical Programmming  for Quantization
**  
**  Author:
**     Alexander Kolesnikov
**
**  26.02.2003   New file for ReducedSearchBasic.cpp
**  01.03.2004   Polar Quantization start
**  16.03.2004   Polar Quantization: 2nd works
**  22.04.2004   New Classes: Polar and XY
**  17.09.2004   DP_MSE is changed to keep info about all
**               intermediate codebooks
**
**	
**
**  Email:
**   koles@cs.joensuu.fi
**   kolesnikov@iae.nsk.su
**   akimov@cs.joensuu.fi
**
**  Copyright (c) 2003,2004  Alexander Kolesnikov 
**
**
********************************************************************************/
  
/*********************************************************************************
** List of methods;
**  
**
*********************************************************************************/

#include "stdafx.h"

#include "HistoData.h"
#include "HistoDataPolar.h"
#include "SearchBasic.h"

#include "../Modules/owntypes.h"


long int counter = 0;
//FILE *reduce_file;
//FILE *back_dir_file;


/*****************************************************************************
**
**
**  Parameters:
**    1) (int )        Ms           -- 
**    2) (double *)    error_value --  
**
**  Return error code: 
**     0: OK
**    -1: ERROR
**********************************************************************************/
int CHistoData::DP_MSE(ULONG Ms, int nVers, int nType, double *error_value)
{ 
	int       err=0, m, flag_pnode=1, verbose=1, i1=0;
	long int  N, ni=0, nn, undef_val = -99999, ind, ind1;
    DTable    *pD2=NULL; 
	double    E2;
    NTable    *pNodes=NULL;
    Bounds    *pBnd=NULL;

	*error_value = 1.0e20;

   // Number of the vertices 
	N = m_nSize;   

	// Number of Clusters
	m_nClustersNumber = Ms;

	Ms = (Ms <= N) ? Ms : N-1;



     // Allocate memory 
	pNodes = CreateNTable(N+1, Ms+1);
	if(pNodes == NULL)
	{   
		 return -1;
	}

    SetNTableToValue(pNodes, undef_val);

    // Allocate memory for D
	pD2 = CreateDTable(N+1, Ms+1);

	if(pD2 == NULL)
	{
		 DestroyNTable(&pNodes); 

		 return -1;
	}

    // Allocate memory for the Bounding Corridor
	pBnd = CreateBounds(nType, N, Ms, Ms);
	if(pBnd == NULL)
	{
		 DestroyNTable(&pNodes); 
         DestroyDTable(&pD2); 

		 return -1;
	}
	
	// Calculate the Bounding Corridor 
	err = CalcBounds(pBnd);
	if(err != 0)
	{
		 DestroyNTable(&pNodes); 
         DestroyDTable(&pD2); 

		 return -1;
	}

    //this->TypeBounds(pBnd);
	
    ////////////////////////////////////////////////

	// 1) Full search with n-m order
	if(nVers == 1) DP_MSE_min(pBnd, pD2, pNodes);

	// 2) Monge 1-pass search with Left-End width

	//here we check the corectness of the following algorithms.
//	if(nVers == 2) DP_MSE_min(pBnd, pD2, pNodes);

//	if(nVers == 2) DP_MSE_min_row(pBnd, pD2, pNodes);

	if(nVers == 2) DP_MSE_min_row_reduced(pBnd, pD2, pNodes);

//	if(nVers == 2) DP_MSE_min_row_SMAWK(pBnd, pD2, pNodes);

	// 3) Monge 2-pass search with Left-Right width
	if(nVers == 3) DP_MSE_min_row_right(pBnd, pD2, pNodes);

	long int n;
	
	// Init Cluster Index
	
	long int upper_bound;
	long int cur_index_means = 0;
	long int cur_index_thr = 0;

	if(m_pClustersList != NULL)
		delete [] m_pClustersList;


	if(m_pClustersThr != NULL)
		delete [] m_pClustersThr;
	
	
	if(!IsErrorEst)
	{
		m_pClustersList = new double[(Ms*(Ms+1)/2)];
		m_pClustersThr  = new double[(Ms*(Ms+1))];
	}

	for(upper_bound = 1; upper_bound <= Ms; upper_bound++)
	{
		for(n = 0; n < m_nSize; n++)
		{
			m_pClusterIndex[n] = 0;
		}

		m_pClusterIndex[0] = 1;
//		m_pClustersThr[cur_index_thr++] = 0;

		// Make output solution
		ind = N - 1;  

		for(m = upper_bound; m >= 1; m--)
		{
		   nn   = ind * pNodes->size_x + (m-pBnd->bottom[ind]);
		   ind1 = pNodes->data[nn];

		   	if(ind1 == undef_val)
			{
			   fprintf(stderr, "ERROR with pNode!!! m=%d;  ind: %d,   ind1: %d, \n", 
					   m, ind, ind1);
				err = -1;
				break;
			}

		   m_nClustersNumber = upper_bound;
		   if(ind+1<m_nSize)
		   {
			 m_pClusterIndex[ind+1] = 1;
		   }
		
		   ind = ind1;
		}//for(m

		m_pClusterIndex[0] = 1;
		
		if(m_pClustersThr!=NULL)
			m_pClustersThr[cur_index_thr++] = m_pX[0];

		err = CalcClustersList();

		long int i;
		
		if(m_pClustersList!=NULL)
		{
			for(i=0;i<upper_bound;i++)
				m_pClustersList[cur_index_means++] = m_pClusterMeans[i];

			for(i=1;i<m_nSize;i++)
			{
				if(m_pClusterIndex[i])
				{
					m_pClustersThr[cur_index_thr++] = m_pX[i-1];
					m_pClustersThr[cur_index_thr++] = m_pX[i];
				}
			}
		}
		
		
		if(m_pClustersThr!=NULL)
			m_pClustersThr[cur_index_thr++] = m_pX[m_nSize-1];
		
		//////////////
		nn = (N-1) * pD2->size_x + (Ms - pBnd->bottom[N-1]);
		double E2_a  = pD2->data[nn]; 
		
//		if((ULONG)m_nClustersNumber != Ms)
//		{
//			printf("WARNNING! MSE: %d -> %d \n", Ms, m_nClustersNumber);
//		}
		
		*error_value = E2_a; 
		//////////////////////////////////////////
	}
		
	if(pBnd->type == 2)
	{ 
		//printf("Bound Type: %d\n", pBnd->type);
		m_pE2[0] = 1.0e35;
		for(m = 1; m <= Ms; m++)
		{
			 nn = (N-1) * pD2->size_x + (m - pBnd->bottom[N-1]);
			 E2 = pD2->data[nn]; 
			 m_pE2[m] = E2;
				 //printf("m=%d) E2 = %.12f\n", m, E2);
		}//for(m
	}
	// Free memory  
	DestroyBounds(&pBnd);
	DestroyDTable(&pD2);
	DestroyNTable(&pNodes);

    return 0;
}
////////////////////////////////////////////////////////////
//   1
//
//   Quant. Metod: DP_MSE_min
//   Loops  Order: m-n  (good)
//   Search Width: Full One pass
//
/////////////////////////////////////////////////////////////////
int CHistoData::DP_MSE_min(Bounds *pBnd, DTable  *pD2, NTable *pNodes)
{   int        err=0; 
    long int m, nn, nn1,  N, Ms, undef_val = -99999, i_argmin, 
		     n, n1, n2, i, I1, I2;		
    double   val_min, val_curr, E1, E2, d_myu, d_omega, d_lambda, 
		     omega_2, myu_2, lambda_2;  
	
	
	if((pBnd == NULL) || (pD2 == NULL) || (pNodes == NULL))
  	{
		return -1;
	}

	// No points in the HistoData 
	if(m_nSize < 0)
  	{
		return -1;
	}
	
	// Number of the points 
	N    = m_nSize;   

	// Desired number of Clusters 
	Ms = pBnd->size_m;  //m_nClustersNumber;
	Ms = (Ms < N) ? Ms : N-1;

	/////////////////////////
    // m=1;
	///////////////////////
   	// nn = 0 * pD2->size_x + 0 for (0,0);; 
	pD2->data[0] = 0;

	m  = 1;
    n1 = pBnd->left[m];
	n2 = pBnd->right[m];

	//printf("\n*****  m=%d) %d -- %d ****\n", m, n1, n2);

	for(n = n1; n <= n2; n++) 
	{
	   //CalcVarianceFromPToP(0, n, &E2);
 	   omega_2  = m_pMomentZero[n]; 
	   myu_2    = m_pMomentOne[n];   
	   lambda_2 = m_pMomentTwo[n]; 
	   
 	   if(omega_2 != 0)
	   {
	      E2 = lambda_2 - (myu_2 * myu_2)/ omega_2;
	   }
	   else
	   {
		  printf("WARNNING!!!\n");

          E2 = -1;
	   }

 	   nn = n * pNodes->size_x + (m-pBnd->bottom[n]);
	   pNodes->data[nn] = 0;

       nn = n * pD2->size_x + (m - pBnd->bottom[n]);
       pD2->data[nn] = E2;

	   //printf("n=%ld) D[1][%ld] = %.12f\n",
		 //      n, n, E2);
	}

	///////////////////////////////////////
	// Other m's.
	for(m = 2; m <= Ms; m++) 
	{   
  	    n1 = pBnd->left[m];
		n2 = pBnd->right[m];

		//printf("***** m=%ld: %ld -- %ld ***** \n", m , n1, n2);

		for(n = n1; n <= n2; n++) 
		{
            
			i_argmin = -99;
		    val_min  =  1.0e35;

		    I1 = pBnd->left[m-1];
 		    I2 = pBnd->right[m-1];
            I2 = (I2 <= n-1) ? I2 : n-1;

			omega_2  = m_pMomentZero[n]; 
            myu_2    = m_pMomentOne[n]; 
			lambda_2 = m_pMomentTwo[n]; 

            ////////////////////////////////////////
			//printf(" ------- I: %ld -- %ld -------\n", I1, I2);
  		    
			for(i = I1; i <= I2; i++) 
			{ 
				/////////////////////////////////////
			    // Min{ D[i][m-1] + d(i,n); } 
                // (m-1) Clusters for [I1,i] points;
			    nn1 = i * pD2->size_x + (m-1 - pBnd->bottom[i]);  
                E1  = pD2->data[nn1];

				////////////////////////////////////////
			    // One Cluster for [i+1,n] points; 
			    // CalcVarianceFromPToP(i+1, n, &E2);

                d_omega  = omega_2  - m_pMomentZero[i]; 
  	            d_myu    = myu_2    - m_pMomentOne[i];    
				d_lambda = lambda_2 - m_pMomentTwo[i];  

 			    E2 = d_lambda - d_myu*d_myu/d_omega;

                val_curr = E1 + E2;
                if(val_curr < val_min) 
				{
			       val_min   = val_curr;
                   i_argmin  = i;
				}//if(

				//printf("n=%ld; i=%ld)  D[%d][%ld] + E[%ld][%ld] = D[%ld][%ld]: %.6f+%.6f = %.6f  (>=%.6f)\n", 
				//	    n, i,            m-1, i,       i+1, n,         m,n,  
                  //                E1,      E2, val_curr, val_min) ;
			 }//for(i

		     nn = n * pNodes->size_x + (m-pBnd->bottom[n]);
	         pNodes->data[nn] = i_argmin;

		     nn = n * pD2->size_x + (m - pBnd->bottom[n]);
	         pD2->data[nn] = val_min;
		}//for(n
	 }//for(m

 	 n = N-1;
	 m = Ms;
	 nn = n * pD2->size_x + (m - pBnd->bottom[n]);

     return 0;
}

////////////////////////////////////////////////////////////
//  2)
//
//  Quant. Method:  DP_MSE_min
//  Loops  Order :  m-n! (good)
//  Search Width : 1 pass from new_left till End
//  
////////////////////////////////////////////////////////////
int CHistoData::DP_MSE_min_row(Bounds *pBnd, DTable  *pD2, NTable *pNodes)
{   int        err=0; 
    long int  m, nn, nn1,  N, Ms, undef_val = -99999, i_argmin, new_left = -1,
	          n, n1, n2, i, I1, I2;		
    double    val_min, val_curr, E1, E2, d_myu, d_omega, d_lambda, 
		      omega_2, myu_2, lambda_2;  
	
	

	
	if((pBnd == NULL) || (pD2 == NULL) || (pNodes == NULL))
  	{
		return -1;
	}

	// No points in the HistoData 
	if(m_nSize < 0)
  	{
		return -1;
	}
	
	// Number of the points 
	N = m_nSize;   
	
	// Desired number of Clusters 
	Ms = pBnd->size_m;  //m_nClustersNumber;
	Ms = (Ms < N) ? Ms : N-1;

	/////////////////////////
    // m=1;
   	// nn = 0 * pD2->size_x + 0 for (0,0);; 
	pD2->data[0] = 0;

	m  = 1;
    n1 = pBnd->left[m];
	n2 = pBnd->right[m];
	//printf("\n*****  m=%d) %d -- %d ****\n", m, n1, n2);

	for(n = n1; n <= n2; n++) 
	{
	   //CalcVarianceFromPToP(0, n, &E2);
 	   omega_2  = m_pMomentZero[n]; 
	   
 	   if(omega_2 != 0)
	   {
		  myu_2    = m_pMomentOne[n]; 
		  lambda_2 = m_pMomentTwo[n]; 
	      E2       = lambda_2 - (myu_2*myu_2)/omega_2;
	   }
	   else
	   {
          E2 = -1;
	   }

 	   nn = n * pNodes->size_x + (m-pBnd->bottom[n]);
	   pNodes->data[nn] = 0;

       nn = n * pD2->size_x + (m - pBnd->bottom[n]);
       pD2->data[nn] = E2;

	   //printf("E[1,%d] = %.4f\n", n, E2);
	}

	///////////////////////////////////////
	// Other m's.

	long count = 0;
	for(m = 2; m <= Ms; m++) 
	{   
  	    n1 = pBnd->left[m];
		n2 = pBnd->right[m];

		//printf("\n*****  m=%d) n = %d -- %d ****\n", m, n1, n2);
		new_left  = -1;
/*-------------------------------------------*/
/*Solving the problem for m clusters*/
		for(n = n1; n <= n2; n++) 
		{
            i_argmin = -99;
		    val_min  =  1.0e35;

		    I1 = pBnd->left[m-1];
		    I2 = n-1; 

			if(new_left  != -1) I1 = new_left;
		
			omega_2  = m_pMomentZero[n]; 
            myu_2    = m_pMomentOne[n]; 
			lambda_2 = m_pMomentTwo[n]; 

            ////////////////////////////////////////
  		    for(i = I1; i <= I2; i++) 
			{ 
			   count++;
			   
			    // Min{ D[i][m-1] + d(i,n); } 
                // (m-1) Clusters for [I1,i] points;
			    nn1 = i * pD2->size_x + (m-1 - pBnd->bottom[i]);  
                E1  = pD2->data[nn1];

			    // One Cluster for [i+1,n] points; 
			    //CalcVarianceFromPToP(i+1, n, &E2);
                d_omega  = omega_2  - m_pMomentZero[i]; 
	            d_myu    = myu_2    - m_pMomentOne[i];  
				d_lambda = lambda_2 - m_pMomentTwo[i];  

			    if(n == i+1)
				{
	                   E2 = 0.0;
				}
				else //  i = n-1;
				{ 
				 	E2 = d_lambda - (d_myu*d_myu)/d_omega;
				}

                val_curr = E1 + E2;
                if(val_curr < val_min) 
				{
			            val_min  = val_curr;
                        i_argmin = i;
			   	        new_left = i;
				}//if


			 }//for(i

		     nn = n * pNodes->size_x + (m-pBnd->bottom[n]);
	         pNodes->data[nn] = i_argmin;

		     nn = n * pD2->size_x + (m - pBnd->bottom[n]);
	         pD2->data[nn] = val_min;


		}//for(n

	 }//for(m

/*  	 n = N-1;
	 m = Ms;
	 nn = n * pD2->size_x + (m - pBnd->bottom[n]);
*/	

     return 0;
}

////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////

void SetAddArrays(long N, long *new_order, long *new_left_array, long *new_right_array)
{
	int level_index = 1;

	new_order[0] = (N-1)/2;
	new_left_array[0] = -1;
	new_right_array[0] = -1;

	new_order[1] = 0;
	new_left_array[1] = -1;
	new_right_array[1] = new_order[0];


	new_order[2] = N-1;
	new_left_array[2] = new_order[0];
	new_right_array[2] = -1;

	long current_index = 3;
	long divider = 2;
	long i;
	
	long left_index;
	long right_index;

	long *used;
	__int64 new_index;

	used = new long[N];

	memset(used,0,N*sizeof(long));
	
	for(i=0;i<3;i++)
		used[new_order[i]] = 1;

	while(current_index<N)
	{
		divider = divider*2;

		for(i=1;i<divider;i+=2)
		{
			new_index = ((__int64)(N-1) * i)/divider;
			
			if(used[new_index])
				continue;
			else used[new_index] = 1;
			
			new_order[current_index] = new_index;
			
			left_index = ((__int64)(N-1) * (i-1))/divider;
			right_index = ((__int64)(N-1) * (i+1))/divider;


			new_left_array[current_index] = left_index;
			new_right_array[current_index] = right_index;

			current_index++;

			if(current_index>=N)
				break;
		}
	}

	delete used;
}

////////////////////////////////////////////////////////////
//  21)
//
//  Quant. Method:  DP_MSE_min
//  Loops  Order :  m-n! (good)
//  time complexity : O(m*n*log(n))
//  
////////////////////////////////////////////////////////////
int CHistoData::DP_MSE_min_row_reduced(Bounds *pBnd, DTable  *pD2, NTable *pNodes)
{   int        err=0; 
    long int  m, nn, nn1,  N, Ms, undef_val = -99999, i_argmin, new_left = -1,
	          n, n1, n2, i, I1, I2;		
    double    val_min, val_curr, E1, E2, d_myu, d_omega, d_lambda, 
		      omega_2, myu_2, lambda_2;  
	


	
	if((pBnd == NULL) || (pD2 == NULL) || (pNodes == NULL))
  	{
		return -1;
	}

	// No points in the HistoData 
	if(m_nSize < 0)
  	{
		return -1;
	}
	
	// Number of the points 
	N = m_nSize;   
	
	// Desired number of Clusters 
	Ms = pBnd->size_m;  //m_nClustersNumber;
	Ms = (Ms < N) ? Ms : N;

	/////////////////////////
    // m=1;
   	// nn = 0 * pD2->size_x + 0 for (0,0);; 
	pD2->data[0] = 0;

	m  = 1;
    n1 = pBnd->left[m];
	n2 = pBnd->right[m];
	//printf("\n*****  m=%d) %d -- %d ****\n", m, n1, n2);

	for(n = n1; n <= n2; n++) 
	{
	   //CalcVarianceFromPToP(0, n, &E2);
 	   omega_2  = m_pMomentZero[n]; 
	   
 	   if(omega_2 != 0)
	   {
		  myu_2    = m_pMomentOne[n]; 
		  lambda_2 = m_pMomentTwo[n]; 
	      E2       = lambda_2 - (myu_2*myu_2)/omega_2;
	   }
	   else
	   {
          E2 = -1;
	   }

 	   nn = n * pNodes->size_x + (m-pBnd->bottom[n]);
	   pNodes->data[nn] = 0;

       nn = n * pD2->size_x + (m - pBnd->bottom[n]);
       pD2->data[nn] = E2;

	   //printf("E[1,%d] = %.4f\n", n, E2);
	}

	///////////////////////////////////////
	// Other m's.


	long *new_left_array;
	long *new_right_array;
	long *new_order;


	new_left_array = NULL;
	new_right_array = NULL;
	
	new_order = new long[N];
	new_right_array = new long[N];
	new_left_array = new long[N];



	long row_index;

	long new_right;

	int count = 0;
	counter = 0;
	for(m = 2; m <= Ms; m++) 
	{   
  	    n1 = pBnd->left[m];
		n2 = pBnd->right[m];
		

		/*fill in additional arrays*/
		/*such as new_order and so on*/
		memset(new_order,0,sizeof(long)*N);
		memset(new_right_array,0,sizeof(long)*N);
		memset(new_left_array,0,sizeof(long)*N);

		SetAddArrays(n2-n1+1,new_order,new_left_array,new_right_array);

		//printf("\n*****  m=%d) n = %d -- %d ****\n", m, n1, n2);
		new_left  = -1;
/*-------------------------------------------*/
/*Solving the problem for m clusters*/
		
		for(n = n1; n <= n2; n++) 
		{
			
			row_index = n1 + new_order[n-n1];

			i_argmin = -99;
		    val_min  =  1.0e35;

		    I1 = pBnd->left[m-1];
		    I2 = row_index-1;

//			if(new_left  != -1) I1 = new_left;

			new_left = new_left_array[n-n1];
			new_right = new_right_array[n-n1];
			
			if(new_left != -1)
			{
				new_left += n1;
				nn = new_left * pNodes->size_x + (m-pBnd->bottom[new_left]);
				I1 = max(pNodes->data[nn],I1);
			}

			if(new_right != -1)
			{
				new_right += n1;
				nn = new_right * pNodes->size_x + (m-pBnd->bottom[new_right]);
				I2 = min(pNodes->data[nn],I2);
			}

			omega_2  = m_pMomentZero[row_index]; 
            myu_2    = m_pMomentOne[row_index]; 
			lambda_2 = m_pMomentTwo[row_index]; 
			
			
//			if(this->IsErrorEst)
//				printf("\n%d\t%d\t%.4f\t%.4f",m,n,((double)n/(double)n2)*100,((double)m/(double)Ms)*100);

            ////////////////////////////////////////
  		    for(i = I1; i <= I2; i++) 
			{ 
			   counter++;
			    // Min{ D[i][m-1] + d(i,n); } 
                // (m-1) Clusters for [I1,i] points;
			    nn1 = i * pD2->size_x + (m-1 - pBnd->bottom[i]);  
                E1  = pD2->data[nn1];

			    // One Cluster for [i+1,n] points; 
			    //CalcVarianceFromPToP(i+1, n, &E2);
                d_omega  = omega_2  - m_pMomentZero[i]; 
	            d_myu    = myu_2    - m_pMomentOne[i];  
				d_lambda = lambda_2 - m_pMomentTwo[i];  

			    if(row_index == i+1)
				{
	                   E2 = 0.0;
				}
				else //  i = n-1;
				{ 
				 	E2 = d_lambda - (d_myu*d_myu)/d_omega;
				}

                val_curr = E1 + E2;
                if(val_curr < val_min) 
				{
			            val_min  = val_curr;
                        i_argmin = i;
			   	        new_left = i;
				}//if

				// if(n > 2800)
 			   	//printf("n=%ld; i=%ld)  D[%d][%ld] + E[%ld][%ld] = D[%ld][%ld]: %.6f+%.6f = %.6f  (>=%.6f)\n", 
				//	    n, i,            m-1, i,       i+1, n,         m,n,  
                  //                E1,      E2, val_curr, val_min) ;

			 }//for(i

		     nn = row_index * pNodes->size_x + (m-pBnd->bottom[row_index]);
	         pNodes->data[nn] = i_argmin;

		     nn = row_index * pD2->size_x + (m - pBnd->bottom[row_index]);
	         pD2->data[nn] = val_min;
		}//for(row_index


	 }//for(m

/*  	 n = N-1;
	 m = Ms;
	 nn = n * pD2->size_x + (m - pBnd->bottom[n]);
*/	

	/*for test usage only*/
//	FILE *f;

//	f = fopen("z_reduced_data.txt","w");
	
/*	for(m = 2; m <= Ms; m++)
	{
		n1 = pBnd->left[m];
		n2 = pBnd->right[m];
		for(n=n1;n<=n2;n++)
		{
			nn = n * pNodes->size_x + (m-pBnd->bottom[n]);
			i_argmin = pNodes->data[nn];
			nn = n * pD2->size_x + (m - pBnd->bottom[n]);
			val_min = pD2->data[nn];
//			fprintf(f,"m = %d, n = %d, i_argmin = %d, val_min = %f\n",m,n,i_argmin,val_min);
		}
	}

//	fclose(f);
	/*end of test usage*/

	printf("\nMs = %d\tcounter = %d\n",Ms,counter);

	delete[] new_order;
	delete[] new_left_array;
	delete[] new_right_array;
    return 0;
}






////////////////////////////////////////////////////////////
//  22)
//
//  Quant. Method:  DP_MSE_min
//  Loops  Order :  m-n! (good)
//  time complexity : O(m*n)
//  
////////////////////////////////////////////////////////////
//Some routines, which are necessary in SMAWK algorithm


double CHistoData::GetMatrixValue(Bounds *pBnd, DTable  *pD2, long int row_index, long int col_index)
{
	long nn1;
	double d_omega, d_myu, d_lambda;
	double omega_2, myu_2, lambda_2;
	long int m = m_nCurClusterInd;

	double E1, E2;

	long int I1, I2;

	long int n = row_index;
	long int i = col_index;

	if(col_index>row_index-1)
		return 10e35;

   // Min{ D[i][m-1] + d(i,n); } 
   // (m-1) Clusters for [0,i] points;


		omega_2  = m_pMomentZero[n]; 
        myu_2    = m_pMomentOne[n]; 
		lambda_2 = m_pMomentTwo[n]; 

	
		nn1 = i * pD2->size_x + (m-1 - pBnd->bottom[i]);  
		E1  = pD2->data[nn1];

		// One Cluster for [i+1,n] points; 
		//CalcVarianceFromPToP(i+1, n, &E2);
        d_omega  = omega_2  - m_pMomentZero[i]; 
	    d_myu    = myu_2    - m_pMomentOne[i];  
		d_lambda = lambda_2 - m_pMomentTwo[i];  



	if(col_index == row_index+1)
	{
		E2 = 0.0;
	}
	else //  i = n-1;
	{ 
	 	E2 = d_lambda - (d_myu*d_myu)/d_omega;

	}
	return E1 + E2;
}


long int CHistoData::GetRealColumnIndex(long *real_indexes, long int index)
{
	return real_indexes[index];
}

long int CHistoData::GetRealColumnIndex(Bounds *pBnd, long int **pointers, long int Level, long int index)
{
	long int cur_level = Level;
	long int cur_index = index;

	while(cur_level>0)
		cur_index = pointers[--cur_level][cur_index];

	return pBnd->left[m_nCurClusterInd-1] + cur_index;
}

long int CHistoData::GetRealRowIndex(Bounds *pBnd, long int Level, long int index)
{
	long int result;

	long int n1 = pBnd->left[m_nCurClusterInd];
	long int n2 = pBnd->right[m_nCurClusterInd];

	if(Level==0)
		result = n1+index;
	else
	{
		result = n1 + pow(2,Level)*(index+1) - 1;
	}

	return result;
}

void PointersFastInit(long int *summator, long int length, long int *pointers)
{
	memcpy(pointers,summator,sizeof(long int)*length);
}

void PointersFastUpdate(long int *summator, long int start, long int end, long int shift, long int *pointers)
{
	memcpy(pointers+start,summator+start+shift, sizeof(long int)*(end-start));
}

int CHistoData::SMAWK_Reduce(Bounds *pBnd, DTable *pD2, long int Nr, long int Nc,
							 long int Level, long int **pointers, long int *summator)
{
	long int nr, nc;
	
	nc = Nc; // local variables, describing number of rows
	nr = Nr; // and number of columns

	long int n1, n2;

	if(nc == nr)
		return 0;

	n1 = pBnd->left[m_nCurClusterInd];
	n2 = pBnd->right[m_nCurClusterInd];

	long int cur_index = 0;
	long int i,j;

	double A_cur, A_next;
	int was_deleted;

	long int row_index;
	long int column_index;
	long int column_index_next;
	long int out_next = nr;
	int two_back_steps = 0;

//	int *IsDeleted;
	int DelIndex;

//	IsDeleted = new int[Nc];

//	memset(IsDeleted,0,sizeof(int)*Nc);

	//		for test usage only
//	fprintf(reduce_file, "Level %d\n",Level);


	if(Level == 3)
		Level = Level;
	

	double time_RI = 0;
	double time_MV = 0;
	double time_X  = 0;
	double time_temp;
	double gen_time = 0;


//	double time_proc_start = clock();
	
	
   /*new variables*/	
	double prev_A_cur, prev_A_next;
	long int prev_row_index = 0;
	long int prev_column_index;
	long int shift = 0;


	was_deleted = 0;


	int step_back, step_forward, step_sameplace, start;


	start = 1;
	step_back = 0;
	step_forward = 0;
	step_sameplace = 0;

	gen_time = clock();


	while(nc > nr)
	{
		was_deleted = 0;
		


		if(!step_back)
		{
			if(prev_row_index != -1)
				prev_row_index = row_index;

			prev_column_index = column_index;
			prev_A_cur = A_cur;
			prev_A_next = A_next;
		}

		if(start)
		{
			row_index = GetRealRowIndex(pBnd,Level,cur_index);
			//real index of the row

			/*for test usage*/
//			time_temp = clock();
			column_index = GetRealColumnIndex(pBnd,pointers,Level,cur_index);
			if(cur_index < nr-1)
				column_index_next = GetRealColumnIndex(pBnd,pointers,Level,cur_index+1);
			else
				column_index_next = GetRealColumnIndex(pBnd,pointers,Level-1,out_next);
//			time_RI += (clock() - time_temp);


			/*for test usage*/
//			time_temp = clock();
			A_cur = GetMatrixValue(pBnd, pD2, row_index, column_index);
			A_next = GetMatrixValue(pBnd, pD2, row_index, column_index_next);
//			time_MV += (clock() - time_temp); 

			start = 0;
		}


		if(step_forward)
		{
			column_index = column_index_next;
			row_index = GetRealRowIndex(pBnd,Level,cur_index);
			
			/*for test usage*/
//			time_temp = clock();
			if(cur_index < nr-1)
				column_index_next = GetRealColumnIndex(pBnd,pointers,Level,cur_index+1);
			else
				column_index_next = GetRealColumnIndex(pBnd,pointers,Level-1,out_next);
//			time_RI += (clock() -time_temp);

			
			/*for test usage*/
//			time_temp = clock();
			A_cur = GetMatrixValue(pBnd, pD2, row_index, column_index);
			A_next = GetMatrixValue(pBnd, pD2, row_index, column_index_next);
//			time_MV += (clock() -time_temp);
			step_forward = 0;
		}

		if(step_back)
		{
			if(prev_row_index == -1)
			{
				start = 1;
				step_back = 0;
				continue;
			}
			
			row_index = prev_row_index;
			column_index = prev_column_index;
			
			A_cur = prev_A_cur;

			/*for test usage*/
//			time_temp = clock();
			A_next = GetMatrixValue(pBnd, pD2, row_index, column_index_next);
//			time_MV += (clock() -time_temp);

			step_back = 0;
			prev_row_index = -1;
		}


		if(step_sameplace)
		{

			
			/*for test usage*/
//			time_temp = clock();
			column_index_next = GetRealColumnIndex(pBnd,pointers,Level-1,out_next);
//			time_RI += (clock()-time_temp);
			
			/*for test usage*/
//			time_temp = clock();
			A_next = GetMatrixValue(pBnd, pD2, row_index, column_index_next);
//			time_MV += (clock()-time_temp);
			step_sameplace = 0;
		}

		counter++;
//		for test usage only
//		fprintf(reduce_file,"%d\t%d\t%d\t%.3f\t%.3f",row_index,column_index,
//												column_index_next,A_cur,A_next);



		if((A_cur <= A_next) && (cur_index == nr-1))
		{
//			IsDeleted[cur_index+1] = 1;
			DelIndex = cur_index+1;
			was_deleted = 1;

			step_sameplace = 1;
//			//		for test usage only
//			fprintf(reduce_file,"\tA_next deleted\n");
		}

		if((A_cur <= A_next) && (cur_index < nr-1))
		{
			cur_index += 1;
			step_forward = 1;
			//		for test usage only
//			fprintf(reduce_file,"\tPassed\n");
		}

		if(A_cur > A_next)
		{
//			IsDeleted[cur_index] = 1;
			DelIndex = cur_index;
			was_deleted = 1;
			if(cur_index>0)
				cur_index -= 1;
			step_back = 1;

			//		for test usage only
//			fprintf(reduce_file,"\tA_cur deleted\n");
		}



//		time_temp = clock();
		if(was_deleted)
		{
//			for(i = DelIndex;i<min(DelIndex+1,nc);i++)
//				if(IsDeleted[i])
//				{
					i = DelIndex;
					if(i < nr)
					{
						if(i < nr-1)
						{
							pointers[Level-1][i] = pointers[Level-1][i+1];

							shift++;

//							time_temp = clock();
							PointersFastUpdate(summator,i+1,nr-1,shift,pointers[Level-1]);
//							time_X += (clock() - time_temp);
						}
//						for(j = i+1;j<nr-1;j++)
//							pointers[Level-1][j]++;
						
						pointers[Level-1][nr-1] = out_next;
					}
//					else
//						pointers[Level-1][nr-1] = out_next;

					out_next++;
//					break;
//				}

			nc--;
		//	memset(IsDeleted,0,sizeof(int)*nc);
		}
//		time_X += (clock() - time_temp);
		
	}

//	delete [] IsDeleted;


//	gen_time = clock() - gen_time;
//	printf("Reduce : gen time=%3.6f, time RI=%3.6f, time MV=%3.6f, time X=%3.6f \n", gen_time, time_RI, time_MV, time_X);

	return 0;
}



int CHistoData::SMAWK_MinCompute(Bounds *pBnd, DTable *pD2, NTable *pNodes)
{

	long int m = m_nCurClusterInd;
	long int n1 = pBnd->left[m];
	long int n2 = pBnd->right[m];

	long int Level = 0;//starting level of calculations
	long int **pointers;

	long int MaxLevel = (long int)(log((double)(n2-n1+1))/LOG2);
	long int i,j;

	pointers = (long int**)malloc(sizeof(long int*)*MaxLevel);

	int ForwardDirection = 1;
	long int LevelNr = n2-n1+1;
	long int LevelNc = n2-n1+1;

	long int nn;

	double t1 = clock();
	double t2;

	long int *summator;

	summator = new long int[LevelNc];

	for(i = 0;i<LevelNr;i++)
		summator[i] = i;


	while(1)
	{
		if(ForwardDirection)
		{
			if(LevelNr == 1)
			{
			
				/*for test usage*/
				t2 = clock();
				printf("\nSMAWK - forward direction %3.4f\n",t2-t1);




				double val_min = 10e35;
				double val_cur;
				long int arg_min;

				long int column_index;
				long int row_index;

				row_index = GetRealRowIndex(pBnd,Level,0);

				long int left, right;

				if(m<pBnd->size_m)
				{
					left = 0;
					right = LevelNc;
				}
				else
				{
					left = m-1;
					right = n2;
				}

				for(i=left;i<right;i++)
				{
					if(m<pBnd->size_m)
						column_index = GetRealColumnIndex(pBnd,pointers,Level,i);
					else
						column_index = i;
					val_cur = GetMatrixValue(pBnd,pD2,row_index,column_index);

					if(val_cur < val_min)
					{
						val_min = val_cur;
						arg_min = i;
					}
				}

				nn = row_index * pNodes->size_x + (m_nCurClusterInd - pBnd->bottom[row_index]);
				pNodes->data[nn] = arg_min;

				nn = row_index * pD2->size_x + (m_nCurClusterInd - pBnd->bottom[row_index]);
				pD2->data[nn] = val_min;

				ForwardDirection = 0;
				counter +=1;


				/*for test usage*/
				t1 = clock();

			}
			else
			{
				Level++;
				LevelNr = (n2-n1+1)/pow(2,Level);
				
				pointers[Level-1] = new long int[LevelNr];
				//for(i=0;i<LevelNr;i++)
				//	pointers[Level-1][i] = i;
				PointersFastInit(summator,LevelNr,pointers[Level-1]);

				SMAWK_Reduce(pBnd,pD2,LevelNr, LevelNc, Level, pointers, summator);

//				if(LevelNr>1)
					LevelNc = LevelNr;
			}
		}
		else //  BackwardDirection
		{
			if(m == pBnd->size_m)
				break;
			
			Level--;

			long int LevelNrPrev;

			LevelNr = (n2-n1+1)/pow(2,Level);
			LevelNrPrev = (n2-n1+1)/pow(2,Level+1);

			long int new_left;
			long int new_right;
			double cur_val;
			double min_val = 10e35;
			long int arg_min;

			long int row_index;
			long int column_index;

		

			for(i=0;i<LevelNrPrev;i++)
			{
				row_index = GetRealRowIndex(pBnd,Level+1,i);
				nn = row_index * pNodes->size_x + (m_nCurClusterInd - pBnd->bottom[row_index]);
				pNodes->data[nn] = pointers[Level][pNodes->data[nn]];
			}


			// for test usage
//			fprintf(back_dir_file,"Level = %d\tLevelNr = %d\n",Level,LevelNr);

			for(i=0;i<LevelNr;i+=2)
			{
				
				long int left_index = -1;
				long int right_index = -1;

				if(i==0)
					new_left = 0;
				else
				{
					row_index = GetRealRowIndex(pBnd,Level,i-1);
					left_index = row_index;
					nn = row_index * pNodes->size_x + (m_nCurClusterInd - pBnd->bottom[row_index]);
					new_left = pNodes->data[nn];
				}

				if(i==LevelNr-1)
					new_right = LevelNr-1;
				else
				{
					row_index = GetRealRowIndex(pBnd,Level,i+1);
					right_index = row_index;
					nn = row_index * pNodes->size_x + (m_nCurClusterInd - pBnd->bottom[row_index]);
					new_right = pNodes->data[nn];
				}

				row_index = GetRealRowIndex(pBnd,Level,i);

				// for test usage
//				fprintf(back_dir_file,"new left = %d, new_right = %d, row_index = %d\t",left_index,right_index,row_index);

				
				for(j=new_left;j<=new_right;j++)
				{
					column_index = GetRealColumnIndex(pBnd,pointers,Level,j);
					//for test usage
//					fprintf(back_dir_file,"column_index = %d\t",column_index);
//					int col_ind = GetRealColumnIndex(pBnd,pointers,Level-1,j);
//					col_ind = GetRealColumnIndex(pBnd,pointers,Level+1,j);
					cur_val = GetMatrixValue(pBnd,pD2,row_index,column_index);

					if(cur_val<min_val)
					{
						min_val = cur_val;
						arg_min = j;
					}

					counter++;
				}

//				fprintf(back_dir_file,"\n");
			
				nn = row_index * pNodes->size_x + (m_nCurClusterInd - pBnd->bottom[row_index]);
				pNodes->data[nn] = arg_min;

				nn = row_index * pD2->size_x + (m_nCurClusterInd - pBnd->bottom[row_index]);
				pD2->data[nn] = min_val;

				min_val = 10e35;
			}


			if(Level == 0)
			{
				for(i=0;i<LevelNr;i++)
				{
					row_index = GetRealRowIndex(pBnd,Level,i);
					nn = row_index * pNodes->size_x + (m_nCurClusterInd - pBnd->bottom[row_index]);
					pNodes->data[nn] = GetRealColumnIndex(pBnd,pointers,Level,pNodes->data[nn]);
				}
			}


			delete [] pointers[Level];

			if(Level == 0)
				break;
		}

	}
	t2 = clock();
	printf("\nSMAWK - backward direction %3.4f\n",t2-t1);

	free(pointers);
	delete [] summator;
	
	return 0;
}


int CHistoData::DP_MSE_min_row_SMAWK(Bounds *pBnd, DTable  *pD2, NTable *pNodes)
{   int        err=0; 
    long int  m, nn, nn1,  N, Ms, undef_val = -99999, i_argmin, new_left = -1,
	          n, n1, n2, i, I1, I2;		
    double    val_min, val_curr, E1, E2, d_myu, d_omega, d_lambda, 
		      omega_2, myu_2, lambda_2;  
	
	if((pBnd == NULL) || (pD2 == NULL) || (pNodes == NULL))
  	{
		return -1;
	}
	
	counter = 0;

	// No points in the HistoData 
	if(m_nSize < 0)
  	{
		return -1;
	}
	
	// Number of the points 
	N = m_nSize;   
	
	// Desired number of Clusters 
	Ms = pBnd->size_m;  //m_nClustersNumber;
	Ms = (Ms < N) ? Ms : N-1;

	/////////////////////////
    // m=1;
   	// nn = 0 * pD2->size_x + 0 for (0,0);; 
	pD2->data[0] = 0;

	m  = 1;
    n1 = pBnd->left[m];
	n2 = pBnd->right[m];
	//printf("\n*****  m=%d) %d -- %d ****\n", m, n1, n2);

	
	
	for(n = n1; n <= n2; n++) 
	{
	   //CalcVarianceFromPToP(0, n, &E2);
 	   omega_2  = m_pMomentZero[n]; 
	   
 	   if(omega_2 != 0)
	   {
		  myu_2    = m_pMomentOne[n]; 
		  lambda_2 = m_pMomentTwo[n]; 
	      E2       = lambda_2 - (myu_2*myu_2)/omega_2;
	   }
	   else
	   {
          E2 = -1;
	   }

 	   nn = n * pNodes->size_x + (m-pBnd->bottom[n]);
	   pNodes->data[nn] = 0;

       nn = n * pD2->size_x + (m - pBnd->bottom[n]);
       pD2->data[nn] = E2;

	   //printf("E[1,%d] = %.4f\n", n, E2);
	}

	///////////////////////////////////////
	// Other m's.


/*	long int *columns;
	long int *rows;
	long int *min_indexes;

	columns = NULL;
	rows = NULL;
	min_indexes = NULL;
*/
	long int Nr, Nc;

//	reduce_file = fopen("zz_reduce_SMAWK.txt","w");
//	back_dir_file = fopen("zz_backdir_SMAWK.txt","w");


	for(m = 2; m <= Ms; m++) 
	{   
  	    n1 = pBnd->left[m];
		n2 = pBnd->right[m];

		//printf("\n*****  m=%d) n = %d -- %d ****\n", m, n1, n2);
		//new_left  = -1;
/*-------------------------------------------*/
/*Solving the problem for m clusters*/

	
		/*end of implememntation*/
		m_nCurClusterInd = m;
/*		if(columns != NULL)
			delete [] columns;
		if(rows != NULL)
			delete [] rows;
		if(min_indexes != NULL)
			delete [] min_indexes;

		columns = new long int [n2-n1+1];
		rows = new long int [n2-n1+1];
		min_indexes = new long int [n2-n1+1];

		for(n=n1;n<=n2;n++)
		{
			columns[n-n1] = n;
			rows[n-n1] = n;
		}
*/
		Nc = n2-n1+1;
		Nr = n2-n1+1;

		// for test usage
//		fprintf(reduce_file,"\nQuantization level %d\n",m);
//		fprintf(back_dir_file,"\nQuantization level %d\n",m);


		SMAWK_MinCompute(pBnd, pD2, pNodes);

/*		for(n = n1; n <= n2; n++) 
		{

			
/*            i_argmin = -99;
		    val_min  =  1.0e35;

		    I1 = pBnd->left[m-1];
		    I2 = n-1; 

			if(new_left  != -1) I1 = new_left;
		
			omega_2  = m_pMomentZero[n]; 
            myu_2    = m_pMomentOne[n]; 
			lambda_2 = m_pMomentTwo[n]; 

            ////////////////////////////////////////
  		    for(i = I1; i <= I2; i++) 
			{ 
			   
			    // Min{ D[i][m-1] + d(i,n); } 
                // (m-1) Clusters for [I1,i] points;
			    nn1 = i * pD2->size_x + (m-1 - pBnd->bottom[i]);  
                E1  = pD2->data[nn1];

			    // One Cluster for [i+1,n] points; 
			    //CalcVarianceFromPToP(i+1, n, &E2);
                d_omega  = omega_2  - m_pMomentZero[i]; 
	            d_myu    = myu_2    - m_pMomentOne[i];  
				d_lambda = lambda_2 - m_pMomentTwo[i];  

			    if(n == i+1)
				{
	                   E2 = 0.0;
				}
				else //  i = n-1;
				{ 
				 	E2 = d_lambda - (d_myu*d_myu)/d_omega;
				}

                val_curr = E1 + E2;
                if(val_curr < val_min) 
				{
			            val_min  = val_curr;
                        i_argmin = i;
			   	        new_left = i;
				}//if

				// if(n > 2800)
 			   	//printf("n=%ld; i=%ld)  D[%d][%ld] + E[%ld][%ld] = D[%ld][%ld]: %.6f+%.6f = %.6f  (>=%.6f)\n", 
				//	    n, i,            m-1, i,       i+1, n,         m,n,  
                  //                E1,      E2, val_curr, val_min) ;

			 }//for(i

		     nn = n * pNodes->size_x + (m-pBnd->bottom[n]);
	         pNodes->data[nn] = i_argmin;

		     nn = n * pD2->size_x + (m - pBnd->bottom[n]);
	         pD2->data[nn] = val_min;
//		}//for(n*/

	 }//for(m

/*  	 n = N-1;
	 m = Ms;
	 nn = n * pD2->size_x + (m - pBnd->bottom[n]);
*/	


	/*for test usage only*/
//	FILE *f;


//	fclose(reduce_file);
//	fclose(back_dir_file);

//	f = fopen("z_smawk_data.txt","w");
	
/*	for(m = 2; m <= Ms; m++)
	{
		n1 = pBnd->left[m];
		n2 = pBnd->right[m];


		for(n=n1;n<=n2;n++)
		{

			nn = n * pNodes->size_x + (m-pBnd->bottom[n]);
			i_argmin = pNodes->data[nn];
			nn = n * pD2->size_x + (m - pBnd->bottom[n]);
			val_min = pD2->data[nn];
//			fprintf(f,"m = %d, n = %d, i_argmin = %d, val_min = %f\n",m,n,i_argmin,val_min);
		}
	}
*/

	printf("\nMs = %d\tcounter = %d\n",Ms,counter);


//	fclose(f);




	/*end of test usage*/

     return 0;
}

///////////////////////////////////////////////////////////////////////////////////
//  3
//  
//  Quant. Method: DP_MSE_min
//  Loops   Order: m-n (good)
//
//  Two passes: Forward & Backward
//  Search  Width: 1) from new_left  till 1st max
//                 2) from new_right till new_left
//  
// 
// 
///////////////////////////////////////////////////////////////////////////////////
int CHistoData::DP_MSE_min_row_right(Bounds *pBnd, DTable  *pD2, NTable *pNodes)
{   int        err=0, flag_stop=0; 
    long int  m, nn, nn1,  N, Ms, undef_val = -99999, i_argmin, new_left = -1,
	          n, n1, n2, i, I1, I2, new_right=-1;		
    double    val_min, val_curr, E1, E2, d_myu, d_omega, d_lambda, 
		      omega_2, myu_2, lambda_2;  
	NArray    *pLefts=NULL, *pRights=NULL; 

	printf("DEBUG 1\n");

	if((pBnd == NULL) || (pD2 == NULL) || (pNodes == NULL))
  	{
		return -1;
	}

	// No points in the HistoData 
	if(m_nSize < 0)
  	{
		return -1;
	}
	
	// Number of the points 
	N = m_nSize;   
	
	// Desired number of Clusters 
	Ms = pBnd->size_m;  //m_nClustersNumber;
	Ms = (Ms < N) ? Ms : N-1;

	/////////////////////////////////////
	pLefts = CreateNArray(N+1);
	if(pLefts == NULL)
	{
       return -1;
	}

    pRights = CreateNArray(N+1);
	if(pRights == NULL)
	{
	   DestroyNArray(&pLefts);
	   
       return -1;
	}


	/////////////////////////
    // m=1;
   	// nn = 0 * pD2->size_x + 0 for (0,0);; 
	pD2->data[0] = 0;

	m  = 1;
    n1 = pBnd->left[m];
	n2 = pBnd->right[m];
	//printf("\n*****  m=%d) %d -- %d ****\n", m, n1, n2);

	for(n = n1; n <= n2; n++) 
	{
	   //CalcVarianceFromPToP(0, n, &E2);
 	   omega_2  = m_pMomentZero[n]; 
	   
 	   if(omega_2 != 0)
	   {
		  myu_2    = m_pMomentOne[n]; 
		  lambda_2 = m_pMomentTwo[n]; 
	      E2       = lambda_2 - (myu_2*myu_2)/omega_2;
	   }
	   else
	   {
          E2 = -1;
	   }

 	   nn = n * pNodes->size_x + (m-pBnd->bottom[n]);
	   pNodes->data[nn] = 0;

       nn = n * pD2->size_x + (m - pBnd->bottom[n]);
       pD2->data[nn] = E2;

	   //printf("E[1,%d] = %.4f\n", n, E2);
	}

	///////////////////////////////////////
	// Other m's.
	for(m = 2; m <= Ms; m++) 
	{   
  	    n1 = pBnd->left[m];
		n2 = pBnd->right[m];

		///////////////////////////////////////
		// 1st run
		//////////////////////////////////////
		//printf("\n*****  m=%d) n = %d -- %d ****\n", m, n1, n2);

		//new_left  = -1; // XXXXX
		for(n= n1; n <= n2; n++) 
		{
           pLefts->value[n] = pBnd->left[m-1];
		}

		for(n = n1; n <= n2; n++) 
		{
            i_argmin = -99;
		    val_min  =  1.0e35;
            flag_stop = 0;

		    I1 = pBnd->left[m-1];
		    I2 = n-1; 

			if(pLefts->value[n-1] > I1)
			{
                I1 = pLefts->value[n-1];
			}
			//if(new_left  != -1) I1 = new_left;  // XXXXX
		
			omega_2  = m_pMomentZero[n]; 
            myu_2    = m_pMomentOne[n]; 
			lambda_2 = m_pMomentTwo[n]; 

            ////////////////////////////////////////
  		    for(i = I1; i <= I2; i++) 
			{ 
			    // Min{ D[i][m-1] + d(i,n); } 
                // (m-1) Clusters for [I1,i] points;
			    nn1 = i * pD2->size_x + (m-1 - pBnd->bottom[i]);  
                E1  = pD2->data[nn1];

			    // One Cluster for [i+1,n] points; 
			    //CalcVarianceFromPToP(i+1, n, &E2);
                d_omega  = omega_2  - m_pMomentZero[i]; 
	            d_myu    = myu_2    - m_pMomentOne[i];  
				d_lambda = lambda_2 - m_pMomentTwo[i];  

  	            if(d_omega != 0)
				{
				    if(n == i+1)
					{
	                   E2 = 0.0;
					}
					else //  i = n-1;
					{ 
					 	E2 = d_lambda - (d_myu*d_myu)/d_omega;
					}
				 }
	             else
				 {
                       E2 = -1;
					   printf("**** E2==-1 !\n");
				 }

                 val_curr = E1 + E2;
                 if(val_curr < val_min) 
				 {
			            val_min          = val_curr;
                        i_argmin         = i;
			   	        //new_left = i;
						pLefts->value[n] = i;
						flag_stop        = 1;
				 }//if

				 if((val_curr > val_min) && (flag_stop==1))
				 {
					 flag_stop=0;
					 break;
				 }

				// if(n > 2800)
 			   	//printf("n=%ld; i=%ld)  D[%d][%ld] + E[%ld][%ld] = D[%ld][%ld]: %.6f+%.6f = %.6f  (>=%.6f)\n", 
				//	    n, i,            m-1, i,       i+1, n,         m,n,  
                  //                E1,      E2, val_curr, val_min) ;

			 }//for(i

			//if(n > 2800)
	        // printf("m=%ld, n=%ld: E_min = %.12f for I=%ld \n", 
 		    // m, n, val_min, i_argmin);			
			 //printf("\n");

		     nn = n * pNodes->size_x + (m-pBnd->bottom[n]);
	         pNodes->data[nn] = i_argmin;

		     nn = n * pD2->size_x + (m - pBnd->bottom[n]);
	         pD2->data[nn] = val_min;
		}//for(n

		/*printf("LEFTS\n");
		TypeNArray(pLefts);
		*/
		////////////////////////////////////////////////////
		// 2nd run
		////////////////////////////////////////////////////
		new_right = -1; // XXXXX
		for(n = n1; n <= n2; n++) 
		{
			pRights->value[n] = n-1;
		}
				
		//printf("m=%d) n = [%ld-%ld]\n", m, n1, n2);
		
		for(n = n2; n >= n1; n--) 
		{
            i_argmin  = -99;
		    val_min   =  0;

		    I1 = pBnd->left[m-1];
		    I2 = n-1; 

		    if(pLefts->value[n] > I1) 
			{
				I1 = pLefts->value[n];
			}

			if((I2 > new_right) && (new_right != -1)) 
			{
				I2 = new_right; // XXXXX
			}

			if((n < N-1) && (pRights->value[n+1] < n-1))
			{
			   //	I2 = pRights->value[n+1];
			}

            omega_2  = m_pMomentZero[n]; 
            myu_2    = m_pMomentOne[n]; 
			lambda_2 = m_pMomentTwo[n]; 

			if(n == n2)
			{
			   printf("m=%d, n=%ld) i = [%d - %d] \n", m, n, I1, I2);
			}

  		    for(i = I2; i >= I1; i--) 
			{ 
 			    // Min{ D[i][m-1] + d(i,n); } 
                // (m-1) Clusters for [I1,i] points;
			    nn1 = i * pD2->size_x + (m-1 - pBnd->bottom[i]);  
                E1  = pD2->data[nn1];

			    // One Cluster for [i+1,n] points; 
			    // CalcVarianceFromPToP(i+1, n, &E2);
	            d_omega  = omega_2  - m_pMomentZero[i]; 
	            d_myu    = myu_2    - m_pMomentOne[i];  
				d_lambda = lambda_2 - m_pMomentTwo[i];  

  	            if(d_omega != 0)
				{
	                  E2  = d_lambda - (d_myu*d_myu)/d_omega;
				}
	            else
				{
                      E2 = -1;
				}

                val_curr = E1 + E2;
                if(val_curr < val_min) 
				{
			        val_min           = val_curr;
                    i_argmin          = i;
					pRights->value[n] = i;
					new_right         = i;

				}//if(

			 }//for(i

		     nn = n * pNodes->size_x + m-pBnd->bottom[n];
	         pNodes->data[nn]  = i_argmin;

		     nn = n * pD2->size_x + m-pBnd->bottom[n];
	         pD2->data[nn] = val_min;

		     if(n == n2)
			 {
			    printf("  R_new=%ld \n", new_right);
			 }


		}//for(n


		///////////////////////////

	 }//for(m

	 printf("DEBUG 2\n");

	 // Free memory
	 DestroyNArray(&pLefts);
	 DestroyNArray(&pRights);

     return 0;
}



////////////////////////////////////////////////////////////////////////////////
/********************************************************
**  Make Bounding Corridor: the main function 
**
**    1) (DPolyline*)  pPoly -- polyline
**
**  Return error code: 
**    0:  OK
**   -1:  ERROR
********************************************************/
int CHistoData::CalcBounds(Bounds *pB)
{   int   err = 0;


    if(pB == NULL)
	{
		return -1;
	}

    if(pB->type == 1)
	{
	   err = CalcBoundsOne2One(pB);
	}

    if(pB->type == 2)
	{
	   err = CalcBoundsOne2Many(pB);
	}

    if(pB->type == 3)
	{
	   //err = CalcBoundsOne2Many_RS(pB, W);
	}


    return err;
}


/********************************************************
**  Quantization
**
**  Make Bounding Corridor for One to One 
**
**
**  Return error code: 
**    0:  OK
**   -1:  ERROR
********************************************************/
int CHistoData::CalcBoundsOne2One(Bounds *pB)
{ 
	long int   m, n, M, W, N;

    if(pB == NULL)
	{
		return -1;
	}

    N = m_nSize;
	if(N < 0)
	{
		return -1;
	}
   
    M = pB->size_m;
    W = pB->W;
	
	///******** LEFT & RIGHT *********
	pB->left[0]  = 0;  
	pB->right[0] = 0;  
    
 	for(m = 1; m <= M-1; m++)
	{
	   pB->left[m]  = m - 1;
	   pB->right[m] = m + (N - M-1); 
    }//for(m

    pB->left[M]  = N - 1;
    pB->right[M] = N - 1; 

	//********* TOP & BOTTOM *********

	// 1
    pB->top[0]    = 1;
  	pB->bottom[0] = 0;  // ???

	// 2:
 	for(n = 1; n <= M-2; n++)
	{
        pB->top[n]    = n+1;
    	pB->bottom[n] = 1; 
    }//for(n

	// 3:
 	for(n = M-1; n <= N-M; n++)
	{
        pB->top[n]    = M-1;
    	pB->bottom[n] = 1; 
    }//for(n

	// 4:
 	for(n = N-M+1; n <= N-2; n++)
	{
        pB->top[n]    = M-1;
    	pB->bottom[n] = n - (N - 1 - M); 
    }//for(n

	// 5
	pB->top[N-1]    = M;
    pB->bottom[N-1] = M; 

    return 0;
}

/********************************************************
**  Quantization
**
**  Make Bounding Corridor for One to Many 
**
*
**  Return error code: 
**    0:  OK
**   -1:  ERROR
********************************************************/
int CHistoData::CalcBoundsOne2Many(Bounds *pB)
{ 
	long int   m, n, M, W, N;

    if(pB == NULL)
	{
		return -1;
	}

    N = m_nSize;
	if(N < 0)
	{
		return -1;
	}
   
    M = pB->size_m;
    W = pB->W;
	
	///******** LEFT & RIGHT *********
	pB->left[0]  = 0;  
	pB->right[0] = 0;  
    
 	for(m = 1; m <= M-1; m++)
	{
	   pB->left[m]  = m - 1;
	   pB->right[m] = N - 1; 
    }//for(m

    pB->left[M]  = N - 1;
    pB->right[M] = N - 1; 

	//********* TOP & BOTTOM *********

	// 1
    pB->top[0]    = 1;
  	pB->bottom[0] = 0; 

	// 2:
 	for(n = 1; n <= M-2; n++)
	{
        pB->top[n]    = n+1;
    	pB->bottom[n] = 1; 
    }//for(n

	// 3:
 	for(n = M-1; n <= N-2; n++)
	{
        pB->top[n]    = M-1;
    	pB->bottom[n] = 1; 
    }//for(n

	// 4
	pB->top[N-1]    = M;
    pB->bottom[N-1] = 1; 

    return 0;
}



int CHistoData::Uniform_old(ULONG M, double *value)
{
    long int n1, n;
	double   dn;
	int      k, kk;

    if(M < 0)
	{
		return -1;
	}

    m_nClustersNumber = M; 
	for(n = 0; n < m_nSize; n++)
    {
       m_pClusterIndex[n]  = 0;
	   m_pClusterNumber[n] = 0;
	   m_pClusterMeans[n]  = 0;
	}

	dn = (double)m_nSize/(double)M;
    kk = 0;

    for(k = 0; k < M; k++)
	{
		n1 = (long int)((double)k*dn);
		m_pClusterIndex[n1] = 1;
        m_pClusterStart[kk] = n1;
		kk += 1;
	}

	CalcClustersList();

/*	CalcVariances();
	CalcMeans();
	CalcThresholds();
*/
	CalcVariance(value);

    return 0;
}


int CHistoData::Uniform(ULONG M, double *value)
{
    long n1, n;
	double   dn;
	int      k, kk;
	double step;

    if(M < 0)
	{
		return -1;
	}

    m_nClustersNumber = M; 
	for(n = 0; n < m_nSize; n++)
    {
       m_pClusterIndex[n]  = 0;
	   m_pClusterNumber[n] = 0;
	   m_pClusterMeans[n]  = 0;
	}


	dn = (double)(m_pX[m_nSize-1] - m_pX[0])/(double)M;
    kk = 0;

	int ci = 0;
	n1 = 0;

    for(k = 0; k < M; k++)
	{
		step = m_pX[0] + ((double)k*dn);
		
		while(m_pX[ci]<step && ci<m_nSize)
		{
			n1++;
			ci++;
		}

		m_pClusterIndex[n1] = 1;
        m_pClusterStart[kk] = n1;
		kk += 1;
	}



//	CalcClustersList();

	CalcUniformClustersList();



/*	CalcVariances();
	CalcMeans();
	CalcThresholds();
*/
	CalcUniformVariance(value);
    return 0;
}

int CHistoData::MakeTest(long int N)
{
    long int n;

    if(N < 0)
	{
		return -1;
	}

    m_nSize = N; 
	for(n = 0; n < m_nSize; n++)
    {
       m_pClusterIndex[n] = 0;
	}

    for(n = 0; n < m_nSize; n++)
	{
		m_pX[n] = (double)n*(double)n;
	}

	CalcMoments();

    return 0;
}

int CHistoData::MaxLloyd(ULONG M, double *value)
{
	int      k, cnt=0; 
	long int i, n_old, n_new;
	double   Th, R1, R2, val, val_old=-2;

    if(M < 0)
	{
		return -1;
	}

	Uniform_old(M, value);

	

	// Iterations 
    for(i = 0; i < 200;  i++)
	{ 
		cnt = 0;

		// Update all Clusters
		for(k = 0; k <= m_nClustersNumber-2; k++)
		{
			n_old = m_pClusterStart[k+1]; //[k+1]

		    CalcMeansForCluster(k, &R1); 
			CalcMeansForCluster(k+1, &R2); 

  
			
			Th = (R1 + R2)/2;  
			n_new = FindIndex(Th);

		
//			printf(" %ld -> %ld\n", n_old, n_new);
			m_pClusterIndex[n_old] = 0;
			m_pClusterIndex[n_new] = 1;

            if(n_new != n_old)
			{
				 cnt += 1;
			}
		}//for(k

	


		CalcClustersList();
		CalcVariance(&val);



	    //printf("i=%d) mse: %.12f \n", i, val);

		if((cnt == 0) || (val == val_old))
		{
			break;
		}
		val_old = val;
	}//for(i

   /* CalcThresholds();
    CalcVariances();
    CalcMeans();
*/
	*value = val;
//	delete [] this->m_pClusterIndex;

    return 0;
}

int CHistoData::TypeBounds(Bounds *pB)
{
	long int  n;
	int       m;

	if(pB == NULL)
	{
		return -1;
	}
	printf("**** Bounds: N: %ld; Ms: %ld; W = %d; Type: %ld\n", 
		    pB->size_n, pB->size_m, pB->W, pB->type);

	/* One2One Corridor Full Search */
	if(pB->type == 1) 
	{
	   printf("Left and Right \n");

  	   for(m = 0; m <= pB->size_m; m++)
	   {
		   if((pB->left[m] >=0 ) && (pB->right[m] >= 0))
		   {
               printf("m=%4ld) %4ld -- %4ld\n", 
			   m, pB->left[m], pB->right[m]);
		   }
	   }

	   printf("\nBottom and Top \n");

  	   for(n = 0; n < pB->size_n; n++)
	   {
		   if((pB->bottom[m] >=0 ) && (pB->top[m] >= 0))
		   {
                printf("n=%4ld) %4ld -- %4ld\n", 
			      n, pB->bottom[n], pB->top[n]);
		   }
	   }
	}

	/* Full Search in One2Many Corridor */
	if(pB->type == 2)
	{
	   printf("\nLeft  & Right \n",  pB->size_m);
  	   for(m = 0; m <= pB->size_m; m++)
	   {
          printf("%4ld) %4ld -- %4ld\n", m, pB->left[m], pB->right[m]);
	   }

	   printf("\nBottom  & Top \n",  pB->size_n);
  	   for(n = 0; n < pB->size_n; n++)
	   {
  //         printf("n=%4d) %4d -- %4d\n", n, pB->bottom[n], pB->top[n]);
	   }

	}

	return 0;
}






////////////////////////////////////////////////////////////
//  2)
//  PolarQ: 
//  Quant. Method:  DP_MSE_min for Phi : 
//  Loops  Order :  m-n! (good)
//  Search Width : 1 pass from new_left till End
//  
////////////////////////////////////////////////////////////
/*
int CHistoData::DP_MSE_search_Phi(Bounds *pBnd, DTable  *pD2, NTable *pNodes)
{   int        err=0; 
    long int  m, nn, nn1,  N, Ms, undef_val = -99999, i_argmin, new_left = -1,
	          n, n1, n2, i, I1, I2;		
    double    val_min, val_curr, E1, E2, d_myu, d_omega, d_lambda, d_nyu, d_pi,
		      omega, myu, lambda, nyu, pi, means;  

	//DArray      *pR, *pPhi;

	if((pBnd == NULL) || (pD2 == NULL) || (pNodes == NULL))
  	{
		return -1;
	}

	// No points in the HistoData 
	if(m_nSize < 0)
  	{
		return -1;
	}

	// Number of the points 
	N = m_nSize;   
	
	// Desired number of Clusters 
	Ms = pBnd->size_m;  //m_nClustersNumber;
	Ms = (Ms < N) ? Ms : N-1;

	/////////////////////////
    // m=1;
   	// nn = 0 * pD2->size_x + 0 for (0,0);; 
	pD2->data[0] = 0;

	m  = 1;
    n1 = pBnd->left[m];
	n2 = pBnd->right[m];
	//printf("\n*****  m=%d) %d -- %d ****\n", m, n1, n2);

	for(n = n1; n <= n2; n++) 
	{
	   //CalcVarianceFromPToP(0, n, &E2);
 	   omega  = m_pMomentZero[n]; 
  	   myu    = m_pMomentOne[n]; 
	   lambda = m_pMomentTwo[n]; 
	   pi    = m_pMomentThree[n]; 
	   nyu    = m_pMomentFour[n]; 
	   
	   //E2 = lambda - (myu*myu)/omega;
	   means = nyu/pi;
	   E2 = lambda - 2*myu*means + omega*means*means;
 
 	   nn = n * pNodes->size_x + (m-pBnd->bottom[n]);
	   pNodes->data[nn] = 0;

       nn = n * pD2->size_x + (m - pBnd->bottom[n]);
       pD2->data[nn] = E2;

	   //printf("E[0,%d] = %.12f\n", n, E2);
	}

	///////////////////////////////////////
	// Other m's.
	for(m = 2; m <= Ms; m++) 
	{   
  	    n1 = pBnd->left[m];
		n2 = pBnd->right[m];

		//printf("\n*****  m=%d) n = %d -- %d ****\n", m, n1, n2);
		new_left  = -1;

		for(n = n1; n <= n2; n++) 
		{
            i_argmin = -99;
		    val_min  =  1.0e35;

		    I1 = pBnd->left[m-1];
		    I2 = n-1; 

			if(new_left  != -1) I1 = new_left;
		
            //printf("m=%ld, n=%ld) %ld - %ld \n", m, n, I1, I2);

			omega  = m_pMomentZero[n]; 
            myu    = m_pMomentOne[n]; 
			lambda = m_pMomentTwo[n]; 
	        pi     = m_pMomentThree[n]; 
		    nyu    = m_pMomentFour[n]; 

            ////////////////////////////////////////
  		    for(i = I1; i <= I2; i++) 
			{ 
			    // Min{ D[i][m-1] + d(i,n); } 
                // (m-1) Clusters for [I1,i] points;
			    nn1 = i * pD2->size_x + (m-1 - pBnd->bottom[i]);  
                E1  = pD2->data[nn1];

			    // One Cluster for [i+1,n] points; 
			    //CalcVarianceFromPToP(i+1, n, &E2);

  	  		    if(n == i+1)
				{
	               E2 = 0.0;
				}
				else //  i = n-1;
				{ 
                   d_omega  = omega  - m_pMomentZero[i]; 
  	               d_myu    = myu    - m_pMomentOne[i];  
				   d_lambda = lambda - m_pMomentTwo[i];  

	               d_pi     = pi     - m_pMomentThree[i]; 
				   d_nyu    = nyu    - m_pMomentFour[i]; 

 			 	   //E2 = d_lambda - (d_myu*d_myu)/d_omega;
  			       means = d_nyu/d_pi;
				   E2 = d_lambda - 2*d_myu * means + d_omega *means*means;
				}
		
	            val_curr = E1 + E2;
                if(val_curr < val_min) 
				{
			            val_min  = val_curr;
                        i_argmin = i;
			   	        new_left = i;
				}//if
			}//for(i

		     nn = n * pNodes->size_x + (m-pBnd->bottom[n]);
	         pNodes->data[nn] = i_argmin;

		     nn = n * pD2->size_x + (m - pBnd->bottom[n]);
			 pD2->data[nn] = val_min;
		}//for(n

	 }//for(m

     return 0;
}
*/
/*
int CHistoData::CalcStatisticsPhi()
{   int     er= 0;
    ULONG   n;
    double  r, r2, phi, p, S0;

	for(n = 0, S0=0.0; n < m_nSize; n++)
	{
	    S0 += (double)m_pHisto[n];
	}

	// Probability
	for(n = 0; n < m_nSize; n++)
	{
       m_pProby[n] = (double)m_pHisto[n] / S0;
	}

	// Cumulants
    p   = m_pProby[0];
    r   = m_pR[0];

    r2  = r*r;
	phi = m_pX[0];

 	m_pMomentZero[0]   = r2 * p;             
	m_pMomentOne[0]    = r2 * p * phi;       
	m_pMomentTwo[0]    = r2 * p * phi * phi; 
    m_pMomentThree[0]  = p;             
	m_pMomentFour[0]   = p * phi;             

	n = 0;
	for(n = 1; n <= m_nSize-1; n++)
	{
	   p   = m_pProby[n];
	   r   = m_pR[n];

	   r2  = r*r;
	   phi = m_pX[n];

       m_pMomentZero[n]  = r2 * p             + m_pMomentZero[n-1];
       m_pMomentOne[n]   = r2 * p * phi       + m_pMomentOne[n-1] ;
       m_pMomentTwo[n]   = r2 * p * phi * phi + m_pMomentTwo[n-1] ;
	   m_pMomentThree[n] = p                  + m_pMomentThree[n-1];
	   m_pMomentFour[n]  = p * phi            + m_pMomentFour[n-1];

	   m_pClusterNumber[n] = 0;
	}//for(n


	return 0;
}
*/

/****** *************************************************************************
                  E N D 
*******************************************************************************/

