/*******************************************************************************
*  Filename: TableStructs.cpp
*
*  Arrays for the data storage
*  
*
*  Written by: 
*      Alexander Kolesnikov
*
*   Email: 
*      koles@cs.joensuu.fi
*
*   Started:
*      July 2001 
**
**  Revised 
**     27.02.2003 
**
**  Copyright (c) 2001-2002  Alexander Kolesnikov 
**
********************************************************************************/
#include "stdafx.h"

#include "TableStructs.h"

/********************************************************************
*  Create 2D Table of doubles
*
*   int size_y 
*   int size_x
*
*
********************************************************************/
DTable* CreateDTable(long int size_y, int size_x)
{   DTable  *pTable = NULL;

	if((size_x < 0) || (size_y < 0))
	{
        return NULL;
	}

  
   /* Init pTable */
    pTable = (DTable *)malloc(sizeof(DTable)); 
	if(pTable == NULL)
	{
        return NULL;
	}

    pTable->size_x = size_x;
    pTable->size_y = size_y;

	pTable->data = (double *)malloc(pTable->size_y * pTable->size_x * sizeof(double)); 

	if(pTable->data == NULL)
	{
         free(pTable);
		 pTable = NULL;
         //printf("SORRY! Not enough memory for table %d x %d! \n", 
		 //	     pTable->size_y, pTable->size_x);

         return NULL;
	}

    return pTable;
}

/******************************************************
   Init the table with double value val

*****************************************************/
int InitDTable(DTable *pTable, double val)
{
	ULONG     i, length;
    double    *pD;

	if(pTable == NULL)
	{
		return -1;
	}

	if((pTable->size_x < 0) || (pTable->size_y < 0))
	{
		return -1;
	}

    pD = (double*)pTable->data;

	if(pD == NULL)
	{
		return -1;
	}

	length = (long int)pTable->size_y * (long int)pTable->size_y;

  	for(i = 0; i < length; i++)
	{
		*pD++ = val;
	}
 
	return 0;
}


void DestroyDTable(DTable **ppTable)
{
   DTable  *pTable = NULL;
   
   if(*ppTable != NULL)
   {
      pTable = (DTable *)*ppTable;

      if(pTable->data != NULL)
	  {
	     free(pTable->data);
	     pTable->data = NULL;
	  }

 	  pTable->size_x = -1;
      pTable->size_y = -1;
 
      free(pTable);
      pTable = NULL;
  }
}


/*******************************************************************************
*
*
*
*
********************************************************************************/
int TypeDTable(DTable *pTable, int num)
{
	long int    k, L, nn, L_end;
	double  a;

	if(pTable == NULL)
	{
		return -1;
	}

    printf("-----------------------------------------------------\n");

	printf("size_x = %4ld; size_y = %ld  \n", pTable->size_x, pTable->size_y);

	if(num >= 0)
	{
  	  for(k = 0; k < pTable->size_y; k++)
	  {
	    L_end = pTable->size_x;

		printf("%4ld)  ", k);

		for(L = 0; (L < L_end) && (L < (long int) num); L++)
		{
			nn = k * pTable->size_x + L;
			a = pTable->data[nn];


			if(a < 0) 
			{
				printf("       ");
			} 
			else 
			{	
         	   if(a > 1.0e08) 
			   {
				   printf(" Inf   ");
			   }
			   else 
			   {
		
		  	     if(a >= 1.0e6)
				 {
			        printf("%4ldM ", (int)(a/1000000 + 0.5));
				 }
			     else 
				 {
			  	    if(a >= 1000)
					{
			         //   printf("%4dK ", (int)(a/1000 + 0.5));
						  printf("%8ld  ", (int)(a + 0.5));
					} 
				    else
					{
				       if(a != -1)
					   {
                           printf("%8.2f ", a);
					   }
				       else
					   {
                           printf("   * ");
					   }
					}
				 }
			   
			   } 
			}
		}
		printf("\n");
	  }
	}


	else 
	{

  	  num = -num;

 	  for(L = 0; L < pTable->size_x; L++)
	  {

		printf("%8d) ", L);
		for(k = 0; k < (long int)num; k++)
		{
			//nn = L * pTable->size_x + (pTable->size_x-y-k);
			//nn = (pTable->size_y-1-k) * pTable->size_x + L;
			nn = k * pTable->size_x + L;
			a = pTable->data[nn];

		    
			if(a < 0) 
			{
				printf("         ");
			} 
			else 
			{	
         	   if(a > 1.0e08) 
			   {
				   printf(" Inf   ");
			   }
			   else 
			   {
		
		  	     if(a >= 1.0e6)
				 {
			        printf("%4dM ", (int)(a/1000000 + 0.5));
				 }
			     else 
				 {
			  	    if(a >= 1000)
					{
			         //   printf("%4dK ", (int)(a/1000 + 0.5));
						  printf("%7d  ", (int)(a + 0.5));
					} 
				    else
					{
				       if(a != -1)
					   {
                           printf("%7d ", (int)(a + 0.5));
					   }
				       else
					   {
                           printf("   * ");
					   }
					}
				 }
			   
			   } 
			}
		}
		printf("\n");
	  }
	}

 
	return 0;
}

/*******************************************************************************
*
*
*
*
********************************************************************************/
int TypeDTable(DTable *pTable, Bounds *pBnd, int num)
{
	long int  k, L, nn, L_end;
	double  a;

	if(pTable == NULL)
	{
		return -1;
	}

    printf("-----------------------------------------------------\n");

	printf("size_x = %4d; size_y = %d  \n", pTable->size_x, pTable->size_y);

	if(num >= 0)
	{
  	  for(k = 0; k < pTable->size_y; k++)
	  {
	    L_end = pTable->size_x;

		printf("%4d)  ", k);

		for(L = 0; (L < L_end) && (L < (long int)num); L++)
		{
			nn = k * pTable->size_x + L;
			a = pTable->data[nn];


			if(a < 0) 
			{
				printf("       ");
			} 
			else 
			{	
         	   if(a > 1.0e08) 
			   {
				   printf(" Inf   ");
			   }
			   else 
			   {
		
		  	     if(a >= 1.0e6)
				 {
			        printf("%4dM ", (int)(a/1000000 + 0.5));
				 }
			     else 
				 {
			  	    if(a >= 1000)
					{
			         //   printf("%4dK ", (int)(a/1000 + 0.5));
						  printf("%8d  ", (int)(a + 0.5));
					} 
				    else
					{
				       if(a != -1)
					   {
                           printf("%8.2f ", a);
					   }
				       else
					   {
                           printf("   * ");
					   }
					}
				 }
			   
			   } 
			}
		}
		printf("\n");
	  }
	}


	else 
	{

  	  num = -num;

 	  for(L = 0; L < pTable->size_x; L++)
	  {

		printf("%8d) ", L);
		for(k = 0; k < (long int)num; k++)
		{
			//nn = L * pTable->size_x + (pTable->size_x-y-k);
			//nn = (pTable->size_y-1-k) * pTable->size_x + L;
			nn = k * pTable->size_x + L;
			a = pTable->data[nn];

		    
			if(a < 0) 
			{
				printf("         ");
			} 
			else 
			{	
         	   if(a > 1.0e08) 
			   {
				   printf(" Inf   ");
			   }
			   else 
			   {
		
		  	     if(a >= 1.0e6)
				 {
			        printf("%4dM ", (int)(a/1000000 + 0.5));
				 }
			     else 
				 {
			  	    if(a >= 1000)
					{
			         //   printf("%4dK ", (int)(a/1000 + 0.5));
						  printf("%7d  ", (int)(a + 0.5));
					} 
				    else
					{
				       if(a != -1)
					   {
                           printf("%7d ", (int)(a + 0.5));
					   }
				       else
					   {
                           printf("   * ");
					   }
					}
				 }
			   
			   } 
			}
		}
		printf("\n");
	  }
	}

 
	return 0;
}


/***************************************************************
* *  Create 2D Table of integers
*
*
*
*
*****************************************************************/
NTable* CreateNTable(long int size_y, int size_x)
{  NTable  *pTable = NULL;

	if((size_x < 0) || (size_y < 0))
	{
        return NULL;
	}
  
    /* Init pTable */
    pTable = (NTable *)malloc(sizeof(NTable)); 
	if(pTable == NULL)
	{
        return NULL;
	}

    pTable->size_x = size_x;
    pTable->size_y = size_y;

	pTable->data = (int *)malloc(pTable->size_y * pTable->size_x * sizeof(int)); 

	if(pTable->data == NULL)
	{
         free(pTable);
		 pTable = NULL;
         printf("SORRY! Not enough memory for table %d x %d! \n", 
		 	     size_y, size_x);

         return NULL;
	}

    return pTable;
}

void DestroyNTable(NTable **ppTable)
{
   NTable  *pTable = NULL;
   
   if(*ppTable != NULL)
   {
      pTable = (NTable *)*ppTable;

      if(pTable->data != NULL)
	  {
	     free(pTable->data);
	     pTable->data = NULL;
	  }

 	  pTable->size_x = -1;
      pTable->size_y = -1;
 
      free(pTable);
      pTable = NULL;
  }
}

/*******************************************************************************
*
*
*
*
********************************************************************************/
int TypeNTable(NTable *pTable, int num)
{
	long int  k, L, nn, L_end, a;

	if(pTable == NULL)
	{
		return -1;
	}

    printf("-----------------------------------------------------\n");

	if(num >= 0)
	{
  	  for(k = 0; k < pTable->size_y; k++)
	  {
	    L_end = pTable->size_x;

		printf("%4d) ", k);

		for(L = 0; (L < L_end) && (L < (long int)num); L++)
		{
			nn = k * pTable->size_x + L;
			a = pTable->data[nn];

			if((L <= (long int)num-2) && (L <= L_end-2))
			{
			   if(a  == -99999) 
			   {
                   printf("    ");
			   }
			   else {
                  if(a >= 0)  printf(" %2d ", a);
                  if(a <  0)  printf("%2d ", a);
			   }
			}

		}
		printf("\n");
	  }
    }
	else
	{

	   num = -num;

 	  for(L = 0; L < pTable->size_x; L++)
	  {

		printf("%8d) ", L);
		for(k = 0; k < (long int)num; k++)
		{

			nn = (pTable->size_y-1-k) * pTable->size_x + L;
			a = pTable->data[nn];
			if(a >= 0)  printf("%4d ", a);
            else        printf("     ");
		}
		printf("\n");
	  }
	}

 

 
	return 0;
}



int TypeFlowField(NTable *pNodes, Bounds *pBnd)
{
	long int m, n, n1, n2, m1, m2, nn;

	if((pNodes == NULL) || (pBnd == NULL))
	{
		return -1;
	}

	for(m = 1; m < pNodes->size_x; m++)
    {
	   for(n = 1; n < pNodes->size_y; n++)   
	   {
			nn = n * pNodes->size_x + (m - pBnd->bottom[n]);

			n2 = pNodes->data[nn];
			if(n2 >= 0) 
			{
			   n1 = n;

			   m1 = m;
    		   m2 = m-1;
  
  			   printf("%d %d 2\n", n1, m1);
			   printf("%d %d 2\n", n2, m2);
			}//if
		}//for(m
	}//for(n

	return 0;
}

int SetNTableToValue(NTable *pTable, int value)
{
	int  k, n, size_x, size_y;
	int *pTmp;

	if(pTable == NULL)
	{
		return -1;
	}

    size_x = pTable->size_x;
	size_y = pTable->size_y;

	if((size_x < 0) || (size_y < 0))
	{
		return -1;
	}

	if(pTable->data == NULL)
	{
		return -1;
	}

  	for(k = 0; k < size_y; k++)
	{
		pTmp = (int*)&pTable->data[k * size_x];

		for(n = 0; n < size_x; n++)
		{
		   //nn = k * pTable->size_x + n;
	       //pTable->data[nn] = value;
		   *pTmp++ = value;
		}
	}
 
	return 0;
}

int SetDTableToValue(DTable *pTable, double value)
{
	int     k, n, size_x, size_y, nn;
	double  *pTmp;

	if(pTable == NULL)
	{
		return -1;
	}

	if(pTable->data == NULL)
	{
		return -1;
	}

    size_x = pTable->size_x;
	size_y = pTable->size_y;

	if((size_x < 0) || (size_y < 0))
	{
		return -1;
	}

	nn = 0; 
  	for(k = 0; k < size_y; k++)
	{
		pTmp = (double*)&pTable->data[nn];

		for(n = 0; n < size_x; n++)
		{
		   *pTmp++ = value;
		}
		nn += size_x;
	}
 
	return 0;
}


double GetDTableValue(DTable *pTable, long int k, long int len, double *value)
{
    long int   nn; 
    double  d;

	if(pTable == NULL)
	{
       return -1;
	}

	if(pTable->data == NULL)
	{
		return -1;
	}

	if((k   < 0) || (k   > pTable->size_y) ||
	   (len < 0) || (len > pTable->size_x) )
	{
       return -1;
	}


    nn = k * pTable->size_x + len; 
	d = pTable->data[nn];

	return d;

}


int SetNTableValue(NTable *pTable, long int k, long int n, int value)
{
    long int  nn; 
    
	if(pTable == NULL)
	{
       return -1;
	}

	if(pTable->data == NULL)
	{
		return -1;
	}

	if((k < 0) || (k > pTable->size_y) ||
	   (n < 0) || (n > pTable->size_x) )
	{
       return -1;
	}

    nn = k * pTable->size_x + n; 
	pTable->data[nn] = value;

	return 0;
}


/********************************************************************
**
**  Create Bounds of the Corridor
**
**   int type    -- type of the bounding corridor  
**   int length  -- length of the corridor (number of segments Ms)
**   int width   -- Width of the corridor W
**
**   Return:
**    (Bounds* ) 
********************************************************************/
Bounds* CreateBounds(int typ, long int size_n, long int size_m, int W)
{   Bounds   *pB = NULL;
    long int  am, bn;
  
    /* Init pB  */
    if((size_n < 0) || (size_m < 0))
	{
        return NULL;
	}

    pB = (Bounds *)malloc(sizeof(Bounds)); 
	if(pB == NULL)
	{
        return NULL;
	}

    pB->type   = typ;
	pB->size_n = size_n;
	pB->size_m = size_m;
	pB->W      = W;

	// Full Search OneToOne
    if(pB->type == 1) 
	{
        am = pB->size_m + 2; // For Left&Right
		am = (am < size_n+1) ?  am : size_n+1;
		bn = pB->size_n + 2; // For Bottom & Top
	}

    // Reduced Search OneToMany
    if(pB->type == 2) 
	{
        am =  pB->size_m + 1; // For Left&Right
		am = (am < size_n+1) ?  am : size_n+1;
		bn =  pB->size_n + 2; // For Bottom & Top
	}
   
	/******************************************************/
    pB->refer = (long int *)malloc(am * sizeof(long int)); 
	if(pB->refer == NULL)
	{
         return NULL;
	}

	pB->left  = (long int *)malloc(am * sizeof(long int)); 
	if(pB->left == NULL)
	{
		 free(pB->refer);  pB->refer = NULL;
		 free(pB);		   pB = NULL;

         return NULL;
	}

    pB->right = (long int *)malloc(am * sizeof(long int)); 
	if(pB->right == NULL)
	{
    	 free(pB->refer);  pB->refer = NULL; 
		 free(pB->left);   pB->left  = NULL;
		 free(pB);		   pB = NULL;

         return NULL;
	}

    pB->top = (long int *)malloc(bn * sizeof(long int)); 
	if(pB->top == NULL)
	{
  		 free(pB->refer); pB->refer = NULL;
		 free(pB->left);  pB->left  = NULL;
		 free(pB->right); pB->right = NULL;
		 free(pB);	      pB = NULL;

         return NULL;
	}

    pB->bottom = (long int *)malloc(bn * sizeof(long int)); 
	if(pB->bottom == NULL)
	{
		 free(pB->refer); pB->refer = NULL; 
		 free(pB->left);  pB->left  = NULL;
		 free(pB->right); pB->right = NULL;
		 free(pB->top);   pB->top   = NULL;
         free(pB);	     pB = NULL;

         return NULL;
	}

    return pB;
}


void DestroyBounds(Bounds **ppB) 
{
   Bounds  *pB = NULL;
   
   if(*ppB != NULL)
   {
				  
      pB = (Bounds *)*ppB;
                
  
	  if(pB->refer != NULL) 
	  {
		  free(pB->refer);   pB->refer = NULL;
	  }

	  if(pB->right != NULL) 
	  {
		  free(pB->right); 	  pB->right = NULL;
	  }
	  
      if(pB->left != NULL) 
	  {
		  free(pB->left);     pB->left = NULL;
	  }

      if(pB->top != NULL) 
	  {
		  free(pB->top);      pB->top = NULL;
	  }
      if(pB->bottom != NULL) 
	  {
		  free(pB->bottom);    pB->bottom = NULL;
	  }

      if(pB != NULL)
	  {
		  free(pB);    pB = NULL;
	  }

   }
}




/********************************************************************
*  Create 1D array of int numbers 
*   to keep numbers of segments in MultiShape Map
*
*
*
********************************************************************/
NArray* CreateNArray(long int len)
{   NArray *pN = NULL;
    long int   k;
  
    /* Init pN  */
    if(len < 0)
	{
        return NULL;
	}

    pN = (NArray *)malloc(sizeof(NArray)); 
	if(pN == NULL)
	{
        return NULL;
	}

    pN->size = len;
	
	pN->value  = (unsigned long*)malloc( pN->size * sizeof(unsigned long)); 
	if(!pN->value)
	{
         free(pN);	 pN = NULL;
         printf("SORRY! Not enough memory for Numbers %d! \n", pN->size);

         return NULL;
	}
    for(k = 0; k < pN->size; k++) 
	{
		pN->value[k] = 0;
	}

    return pN;
}


void DestroyNArray(NArray **ppN) 
{
   NArray  *pN = NULL;
   
   if(*ppN != NULL)
   {
				  
      pN = (NArray *)*ppN;
	   
	  if(pN->value != NULL) 
	  {
		  free(pN->value); 
		  pN->value = NULL;
	  }
	  
      free(pN);
      pN = NULL;
   }
}


int TypeNArray(NArray *pB)
{
	long int n;

	if(pB == NULL)
	{
		return -1;
	}

	printf("*** NArray size: %d\n", pB->size);

    for(n = 0; n < pB->size; n++)
    {
          printf("N[%d]=%4d; ", n, pB->value[n]);
    }
	printf("\n");

	return 0;
}

/********************************************************************
*  Create 1D array of int numbers 
*   to keep numbers of segments in MultiShape Map
*
*
*
********************************************************************/
DArray* CreateDArray(long int len)
{   DArray *pN = NULL;
    long int k;
  
    /* Init pN  */
    if(len < 0)
	{
        return NULL;
	}

    pN = (DArray *)malloc(sizeof(NArray)); 
	if(pN == NULL)
	{
        return NULL;
	}

    pN->size = len;

	
    pN->value  = (double *)malloc( pN->size * sizeof(double)); 
	if(pN->value == NULL)
	{
         free(pN);
		 pN = NULL;

         return NULL;
	}
    
    for(k = 0; k < len; k++) pN->value[k] = 0;


    return pN;
}


void DestroyDArray(DArray **ppN) 
{
   DArray  *pN = NULL;
   
   if(*ppN != NULL)
   {
      pN = (DArray *)*ppN;
	   
	  if(pN->value != NULL) 
	  {
		  free(pN->value); 
		  pN->value = NULL;
	  }
	  
      free(pN);
      pN = NULL;
   }
}


int TypeDArray(DArray *pB)
{
	long int  n;

	if(pB == NULL)
	{
		return -1;
	}

	printf("*** NArray size: %d\n", pB->size);

    for(n = 0; n <= pB->size; n++)
    {
          printf("N[%d]=%.2f; ", n, pB->value[n]);
    }
	printf("\n");

	return 0;
}


/******************************************************************
 *                          E N D                                 *
 ******************************************************************/

