/*******************************************************************************
**  Filename: Table Structs.h
**  2D arrays for the data storage
**
**   Written by: Alexander Kolesnikov
**   Email:      koles@cs.joensuu.fi
**
**  Copyright (c) 2001-2002  Alexander Kolesnikov 
**
********************************************************************************/


#ifndef TABLESTRUCTS_H
#define TABLESTRUCTS_H



#include <math.h>
#include <stdlib.h>
#include <stdio.h>

#include "VectSTD.h"

typedef unsigned long int ULONG;


/* double table */
DTable* CreateDTable(long int size_x, int  size_y);
void    DestroyDTable(DTable **ppTable);
int     InitDTable(DTable *pTable, double  value);

int     TypeDTable(DTable *pTable, long int  size);
int     TypeDTable(DTable *pTable, Bounds *pBnd, long int size);

double  GetTableValue(DTable *pTable, int k, int len, double *value);

int     SetDTableToValue(DTable *pTable, double value);
int     SetNTableToValue(NTable *pTable, int value);

/* Integer table  */
NTable* CreateNTable(long int  size_x, int size_y);
void    DestroyNTable(NTable **ppTable);
int     TypeNTable(NTable *pTable, long int  size);
int     TypeFlowField(NTable *pTable, Bounds *pBnd);

/* 1D array of integers */
NArray* CreateNArray(long int  len);
void    DestroyNArray(NArray **ppN);
int     TypeNArray(NArray *pB);

/* 1D array of doubles */
DArray* CreateDArray(long int  len);
void    DestroyDArray(DArray **ppN);
int     TypeDArray(DArray *pB);


/* Bounding corridor in State Space */
Bounds* CreateBounds(int type, long int size_n, long int size_m, int width);
void    DestroyBounds(Bounds **ppB); 
int     TypeBounds(Bounds *pB);

#endif /* TABLESTRUCTS_H */
