/******************************************************************
**  Standard types  for vector applications
**
** Author
**    Alexander Kolesnikov <koles@cs.joensuu.fi>
**
** Created
**
**    05-June-2001
**
** Modified:
**    21-Febr-2002 
** 
**  Copyright (c) 2001-2002  Alexander Kolesnikov 
**
******************************************************************/


#ifndef VECTSTD_H
#define VECTSTD_H

typedef unsigned long int NODE_TYPE;

///////////////////////////////////////////////////////////////
/* Point */
struct DPoint { 
    double  x, y; 
}; 


/* Line */
struct DLine { 
    double  x1, y1; 
    double  x2, y2; 
}; 

/* Rectangulare shape */
struct DArea
{
    double  x1, y1;
	double  x2, y2;

};


/* Rectangulare shape */
struct BBox
{
    double  x_min, y_min;
	double  x_max, y_max;
};



/* Polyline structure */
struct DPolyline { 
	NODE_TYPE      ID;
    long int            size;    /* Number of points        */
    double*        x;       /* X-coordinates           */ 
	double*        y;       /* Y-coordinates           */ 

    NODE_TYPE     *node;    /* index-vector of approximation points */

	// 02.09.2003
	char          *admissible;    /* Admissible approxiamtion points  */

	// For 1-path approximation 
    int    segms_num;
	int    *segm_start;    /* number of the approxiamting segment start point  */ 
    int    *segm_length;   /* number of "zero" nodes   */ 

	// Multi-path for PostOptimization and the Strong Min-# problem
    int       paths_number;    // Number of Paths: max - min + 1;
    int       segms_num_min;   // Min number of segments 
	int       segms_num_max;   // Max number of segments
    double    *E2;        // Array of approx. errors 
 
	// Bounding Box
    BBox   bbox;    
}; 


/* 3D Polyline structure */

// 2D array of doubles 
struct DTable {
    long int     size_x;
	int     size_y;
	double  *data;
};

// 2D array of integers 
struct NTable {
    long int     size_x;
	int     size_y;
	int     *data;
};


// 3D Cumulants for coordinates to calc approx error L_2 
struct Cumulants3D {
    long int   size;

	double     *x1;
	double     *y1;
    double     *z1;

	double     *x2;
    double     *y2;
	double     *z2;

	double     *xy;
	double     *xz;
	double     *yz;
};

// Cumulants for coordinates to calc approx error L_2
struct Cumulants {
    long int   size;
	double     *x1;
	double     *x2;

	double     *y1;
	double     *y2;

	double     *xy;
};


// Struct for Bounding Corridor in State Space 
struct Bounds {
	int        type;   // 1 = One2One;  2 = One2Many;
    long int   size_n;
	long int   size_m;
	int        W;

    long int    *refer;
	long int    *left;
	long int    *right;

	long int    *top;
	long int    *bottom;

};


struct NArray { 
    long int   size;
	unsigned long   *value;
};

struct DArray { 
    long int     size;
	double  *value;
};



#endif // VECTSTD_H
