/*-------------------------------------------------------------------*/
/* CC genrator     Alexander Akimov	                     */
/*                                                                   */
/* Compression and decompression of vector maps		                 */
/*                                                                   */
/*-------------------------------------------------------------------*/


#include <math.h>

#define ProgName      "GCTFilter"
#define VersionNumber "V 1.0"
#define LastUpdated   "24.03.2006"
#define Copyright     "(C) 2007 A. Podlasov A. Akimov"
#define FACTFILE      "main.fac"

extern "C"
{
	#include <stdio.h>
	#include <string.h>
	#include <time.h>
	#include <stdlib.h>
	#include "parametr.c"
	#include "interfc.h"
}

#include "owntypes.h"
#include "ContextTree/MyCT.h"

//#define PGM 1

void PrintInfo(void)
{
  // 5.pbm  5.ct 5.out 6.pbm 7.pbm 8.pbm
  PrintMessage("\r\n%s  %s  (last modified %s)\r\n",ProgName,VersionNumber,LastUpdated);
  PrintMessage("%s\r\n",Copyright);
  PrintMessage("\r\nProgram usage syntax:\r\n");
  PrintMessage("%s {%coption} InputFile OutFile \r\n",ProgName, OPTION_SYMBOL);
  PrintMessage("Options:\r\n");
  PrintOptions();
  PrintMessage("\r\n%s     %s\r\n",ProgName,VersionNumber);
}

struct ColorLutItem
{
	RGBPIXEL Color;
	int index;
	ULONG occurence;
	ColorLutItem *pNextItem;
};


int InitColorLutAndGrayImage(IMAGE *pColorImage, IMAGE *pGrayImage, ColorLutItem* &pList)
{
	ULONG x,y;
	int CurrentIndex = 0;

	for(y=1; y<=pColorImage->ImageSizeY; y++)
	{
		for(x=1; x<=pColorImage->ImageSizeX; x++)
		{
			RGBPIXEL pixel = GetRGBPixel(pColorImage, x, y);

			ColorLutItem *pcur = 0;

			for(pcur=pList; pcur && pcur->Color!=pixel; pcur = pcur->pNextItem);
			
			if(pcur)
			{
			   pcur->occurence++;
			   PutGrayPixel(pGrayImage,x,y,pcur->index);
			}
			else
			{
				ColorLutItem *pnew = new ColorLutItem;
				pnew->Color = pixel;
				pnew->index = CurrentIndex;
				pnew->occurence=1;
				pnew->pNextItem=pList;
				pList=pnew;
				CurrentIndex++;

				PutGrayPixel(pGrayImage,x,y,pnew->index);
			}
		}
	}
//	ImageRewind(pColorImage);
//	ImageRewind(pGrayImage);
	return CurrentIndex;
}

int GrayToColorImage(IMAGE *pGrayImage, IMAGE *pColorImage,  ColorLutItem* pLutArray)
{
	ULONG x,y;
	int CurrentIndex = 0;

	for(y=1; y<=pColorImage->ImageSizeY; y++)
	{
		for(x=1; x<=pColorImage->ImageSizeX; x++)
		{
			RGBPIXEL pixel = GetGrayPixel(pGrayImage, x, y);
			PutRGBPixel(pColorImage,x,y,pLutArray[(int)pixel].Color);
		}
	}
	return CurrentIndex;
}

void DeinitColorLut(ColorLutItem* pList)
{
	ColorLutItem *pcur, *ptodie;
	pcur=pList; 
	for(; pcur!=NULL; )
	{
		ptodie = pcur;
		pcur = pcur->pNextItem;
		delete ptodie;
	}
}

/*===========================  M A I N  =============================*/
void main(int argc, char* argv[])
{
	char            InputName[127] = {'\0'};
	char            OutputName[127]= {'\0'};
	FILETYPE        Extention;

	ParameterInfo paraminfo[2] = {{NULL, "",    0, INFILE },
								{NULL, "", 0, OUTFILE}};

	paraminfo[0].FileName = InputName;
	paraminfo[1].FileName = OutputName;

	ParseParameters(argc, argv, 2, paraminfo);

	if (CheckExtention(InputName, FormatNamePGM))
	  {
		Extention = PGM;
	  }
	  else
	  {
		  if (CheckExtention(InputName, FormatNamePPM))
		  {
			  Extention = PPM;
		  }
		  else
		  {
			  ErrorMessage("Input FileType is not supported.\n");
			  Extention = UNKNOWN;
			  exit(0);
		  }
	  }

	  ColorLutItem *pColorLut=0;
	  int ColorsTotal=0;
	  ColorLutItem *ColorLutArray=0;

	  IMAGE InputImage;

	  if(Extention==PPM)
	  {
		  IMAGE ColorImage;
		  PrintMessage("COLOR MODE\n");

		  // Make color Lut and convert image to PGM according to lut
		  ImageInit(InputName, &ColorImage, INPUT, 1, BUFACCESS, RGBMODE, DISK, 10);

		  SetImageSize(&InputImage,ColorImage.ImageSizeX,ColorImage.ImageSizeY);
		  ImageInit(InputName, &InputImage, OUTPUT, 1, WHOLEIMAGE, GRAYMODE, MEMORY, 256);

		  ColorsTotal = InitColorLutAndGrayImage(&ColorImage, &InputImage, pColorLut);
//		  DumpImage(&InputImage,"dump.pgm");
		  if(Value(FilterType)==DO_GRAY)
		  {
			  DumpImage(&InputImage,OutputName);
			  exit(0);
		  }

		  ColorLutArray = new ColorLutItem[ColorsTotal];
		  ColorLutItem *ptemp;
		  for(ptemp=pColorLut; ptemp; ptemp=ptemp->pNextItem)
		  {
			  ColorLutArray[ptemp->index].Color = ptemp->Color;
			  ColorLutArray[ptemp->index].occurence = ptemp->occurence;
			  ColorLutArray[ptemp->index].pNextItem = 0;
		  }

		  DeinitColorLut(pColorLut);
		  ImageDone(&ColorImage);
	  }
	  else
	  {
		  PrintMessage("GRAY MODE\n");
		  ImageInit(InputName, &InputImage, INPUT, 1, BUFACCESS, GRAYMODE, DISK, 10);
	  }

	  IMAGE OutputImage;
	  SetImageSize(&OutputImage,InputImage.ImageSizeX,InputImage.ImageSizeY);

	  if(Extention==PPM)
		  ImageInit(OutputName, &OutputImage, OUTPUT, 1, WHOLEIMAGE, GRAYMODE, MEMORY, 256);
	  else
		  ImageInit(OutputName, &OutputImage, OUTPUT, 1, WHOLEIMAGE, GRAYMODE, DISK, 256);

	if(Value(QuietLevel) >= NORMAL)
	{
		PrintMessage("Input file: %s\n",InputName);
		PrintMessage("Output file: %s\n",OutputName);
		PrintMessage("Context size: %i\n",Value(CTSize));
		PrintMessage("Noise level: %i%%\n",Value(NoiseLevel));
		PrintMessage("Pruning treshold level: %i pixels\n",Value(PruningTreshold));
		PrintMessage("Tresholding type: %s\n", NameOfValue(TresholdType,Value(TresholdType)));
		PrintMessage("Filtering type: %s\n", NameOfValue(FilterType,Value(FilterType)));
	}

	char				OutputPath[256] = { '\0' };
	int ct_size;

	/* Log file, soderzhit vsyakuyu 
	sluzhebnuyu informaciyu o resultatah 
	szhatiya*/

	FILE *file;
	file = fopen("zresults.txt","a");
	fprintf(file, "\n %s -> %s", InputName, OutputName);
	fclose(file);

	ct_size = Value(CTSize);

	FilterPGMImage(&InputImage, &OutputImage, ct_size, (double)Value(NoiseLevel)/100.0, Value(TresholdType), Value(FilterType), Value(PruningTreshold), "zzzMyPPPM.zz" );

	if(Extention==PPM)
	{
		IMAGE ColorImage;
		SetImageSize(&ColorImage,InputImage.ImageSizeX,InputImage.ImageSizeY);
		ImageInit(OutputName, &ColorImage, OUTPUT, 1, WHOLEIMAGE, RGBMODE, DISK, 256);

//		DumpImage(&InputImage,"in.pgm");
//		DumpImage(&OutputImage,"out.pgm");
		GrayToColorImage(&OutputImage,&ColorImage,ColorLutArray);
//		DumpImage(&ColorImage,OutputName);
		ImageDone(&ColorImage);
	}

//	double t1 = clock();
//	int NumberOfClusters = 1;


	ImageDone(&InputImage);
	ImageDone(&OutputImage);

	if(ColorLutArray)
		delete [] ColorLutArray;
}
